// $Id: Capture.c 25617 2013-09-19 11:18:32Z mh $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * This Capture example shows how to capture data using the LibPCAP and NTCI
 * zero copy interfaces.
 */

#ifdef WIN32
#include <WINSOCK2.H>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <process.h>
#include "platformwin.h"
#else
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <pthread.h>

/* Unix-specific includes */
#include <termios.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <sched.h>
#endif

#include "../common/device.h"
#include "../common/cmdparse.h"

#define ULL unsigned long long

/* The NTCommonInterface.h header provides the interface definitions
 * for the Napatech NTCI library */
#include <NTCommonInterface.h>

/* These headers are needed when capturing traffic */
#include <packetfeedstructure.h>
#include <packetdescriptor.h>
#include <NTCI_packetfeed.h>
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>
#include <NTCI_packetclassification.h>
#include <packetclassificationstructure.h>

static char *cmd_file=NULL;
static char *cmd_if=NULL;
static char *cmd_pcap=NULL;
static char *cmd_ntpl=NULL;
static char *cmd_time=NULL;
static char *cmd_size=NULL;
static char *cmd_pcapnano=NULL;
static char *cmd_walk_packets=NULL;

#ifdef WITH_PCAP
/* Lib Pcap header */
#include <pcap.h>
#endif

#define PCAP_MAGIC      0xa1b2c3d4
#define PCAP_SWAPPED_MAGIC    0xd4c3b2a1
#define PCAP_MODIFIED_MAGIC   0xa1b2cd34
#define PCAP_SWAPPED_MODIFIED_MAGIC 0x34cdb2a1
#define PCAP_NSEC_MAGIC     0xa1b23c4d
#define PCAP_SWAPPED_NSEC_MAGIC   0x4d3cb2a1

/*
  The following has been taken from:
  http://wiki.wireshark.org/Development/LibpcapFileFormat
*/
typedef struct pcap_hdr_s {
  uint32_t magic_number;   /* magic number                              = 0xa1b2c3d4 */
  uint16_t version_major;  /* major version number                      =  2         */
  uint16_t version_minor;  /* minor version number                      =  4         */
  int32_t  thiszone;       /* GMT to local correction                   =  0         */
  uint32_t sigfigs;        /* accuracy of timestamps                    =  0         */
  uint32_t snaplen;        /* max length of captured packets, in octets = 10k        */
  uint32_t network;        /* data link type                            =  1         */
} pcap_hdr_t;

#ifdef WITH_PCAP
static pcap_hdr_t pcapGlobalHdr =
{
  PCAP_MAGIC,
  2,
  4,
  0,
  0,
  10000,
  1
};

static pcap_hdr_t pcapGlobalHdrNano =
{
  PCAP_NSEC_MAGIC,
  2,
  4,
  0,
  0,
  10000,
  1
};

#endif

#ifdef WIN32
#define FILE_FLAGS GENERIC_WRITE
#define FILE_RIGHTS CREATE_ALWAYS

#define LIBPCAP "winpcap"
#else
#ifdef O_NOATIME
#define FILE_FLAGS O_WRONLY | O_CREAT | O_TRUNC | O_NOATIME
#else
#define FILE_FLAGS O_WRONLY | O_CREAT | O_TRUNC
#endif
#define FILE_RIGHTS 0666
#define fileopen(a, b, c) open(a, b, c)
#define filewrite(a, b, c) write(a, b, c)
#define fileclose(a) close(a)

#define NT_HANDLE     int
#define NT_OPEN_ERROR -1

#define LIBPCAP "libpcap"
#endif

/* Possible modes. */
enum { CAPTURE_TO_DISK, CAPTURE_AND_DISCARD };

/* Size of capture buffer to use.  Once this amount of traffic has been captured, the
   feed will stop. */
/* 10GB */
#define CAPTURE_BUFFER_SIZE 0x280000000LL /* bytes */
/* The time the program will run until it stops automatically. */
#define CAPTURE_TIME_TOTAL  60 /* seconds. */

/* Global data variables. */
bool_t applicationRunning=FALSE; /* Flag to signal that application is running or terminating. */

static double dataStored=0;   /* Global storage counter. */
static uint64_t packets=0;      /* Global packet counter. */
#ifdef WITH_PCAP
static pcap_t *pcapHandle=NULL; /* PCAP session handle */
#endif

/* Static function forward decleration. */
static void StopCapture(void);
static void StoreData(char *file, uint8_t *data, uint32_t length, uint32_t numPackets);

/* Function to be called when user is pressing CTRL-C */
#ifdef WIN32
BOOL WINAPI StopApplication(DWORD sig)
#else
static void StopApplication(int sig)
#endif
{
  if(applicationRunning) {
    printf("Stopping application. Received signal %d\n", sig);
    applicationRunning=FALSE;
    StopCapture();
    StoreData(NULL,0,0,0); /* Close the file. */
  }
#ifdef WIN32
  return TRUE;
#endif
}

typedef struct
{
  int captureType;
  double captureSize;
  uint32_t captureTime;
  uint32_t secsRunning;
  double lastDataSet;
}ThreadData_t;

ThreadData_t threadData;

#ifndef WIN32
static FILE *f;
static int clearScreenInitialized=0;
static char cls [FILENAME_MAX];
void ClearScreen(void)
{
  if(clearScreenInitialized==0) {
    /* Get POSIX clear screen control sequence. */
    f = popen ("tput clear", "r");
    if(fgets (cls, FILENAME_MAX, f)==(char*)EOF) {
      // Avoid compile warning
    }
    pclose (f);
    clearScreenInitialized=1;
  }
  fputs (cls, stdout);
}
#endif

/*
  Monitor thread. This thread will monitor the amount of data being written to the
  storage and calculate a throughput.
  It will also auto-stop the application if CAPTURE_TIME_TOTAL is reached.
*/
THREAD_API MonitorThread(void *data)
{
  ThreadData_t *pData = (ThreadData_t*)data;

  while(applicationRunning) {
    sleep(1); /* Wait one second. */
    if (!applicationRunning) {
      THREAD_RETURN_OK;
    }

    pData->secsRunning++;

    if((threadData.captureTime) && (threadData.secsRunning >= threadData.captureTime)) {
      StopApplication(0);
      break;
    }
#if 1
    ClearScreen();
#else
    system("clear");
#endif

    if(pData->captureType == CAPTURE_TO_DISK) {
      if(pData->captureTime) {
        printf("Time left      :             %02d sec.\n",
      pData->captureTime-pData->secsRunning);
      } else {
        printf("Time running   :             %02d sec.\n",
      pData->secsRunning);
      }

      if(pData->captureSize) {
        printf("Space left     :     %10.02f MB\n",
               (pData->captureSize-dataStored)/1024/1024);
      }
      printf("--------------------------------------\n");
      printf("Stored         :        %10.02f MB\n", dataStored==0?0:dataStored/1024/1024);
      printf("Throughput     : Total: %10.02f Mbps\n"
             "                 1 sec: %10.02f Mbps\n"
             "Packets        :  %13lld\n",
             dataStored==0?0:(((dataStored*8)/pData->secsRunning)/1000000LL),
             dataStored==0?0:(((dataStored-pData->lastDataSet)*8)/1000000LL),
             (unsigned long long) packets);
    } else {
      if(pData->captureTime) {
        printf("Time left      :             %02d sec.\n",
               pData->captureTime-pData->secsRunning);
      } else {
        printf("Time running   :               %3d sec.\n",
               pData->secsRunning);
      }
      printf("--------------------------------------\n");
      printf("Data received  :        %10.02f MB\n", dataStored==0?0:(dataStored/1024/1024));
      printf("Throughput     : Total: %10.02f Mbps\n"
             "                 1 sec: %10.02f Mbps\n"
             "Packets        :  %13lld\n",
             dataStored==0?0:(((dataStored*8)/pData->secsRunning)/1000000LL),
             dataStored==0?0:(((dataStored-pData->lastDataSet)*8)/1000000LL),
             (unsigned long long) packets);

    }
    pData->lastDataSet = dataStored;
  }

  THREAD_RETURN_OK;
}


static void StopCapture(void)
{
#ifdef WITH_PCAP
  if(pcapHandle != NULL) {
    pcap_breakloop(pcapHandle);
  }
#endif
}


/* Function to be called whenever data is to be stored on disk. */
static void StoreData(char *file, uint8_t *data, uint32_t length, uint32_t numPackets)
{
  static NT_HANDLE fHandle=NT_OPEN_ERROR;
  uint32_t writeLength=length;
  uint32_t pkts=numPackets;
  if((threadData.captureSize) && ((dataStored+length) > threadData.captureSize)) {
    // Detect the amount of packet data to copy
    uint32_t offset=0;
    pkts=0;
    while((offset+((PacketDescriptorNtSeriesType1_t*)(data+offset))->storedLength) <= (threadData.captureSize-dataStored)) {
      pkts++;
      offset+=((PacketDescriptorNtSeriesType1_t*)(data+offset))->storedLength;
    }
    writeLength=offset;
  }

  dataStored+=writeLength;

  packets+=pkts;

  if(applicationRunning) {
    if(*file != 0) {
      if(fHandle==NT_OPEN_ERROR) {
        if((fHandle = fileopen(file, FILE_FLAGS, FILE_RIGHTS))==NT_OPEN_ERROR) {
          fprintf(stderr, "Failed to open %s for writing.\n", file);
          StopApplication(0);
          return;
        }
#ifdef WITH_PCAP
        if(pcapHandle) {
          if(cmd_pcapnano) {
            /* Write pcap global header. */
            if(filewrite(fHandle, &pcapGlobalHdrNano, sizeof(pcapGlobalHdrNano))!=sizeof(pcapGlobalHdrNano)) {
              // Avoid compile warning
            }
          } else {
            /* Write pcap global header. */
            if(filewrite(fHandle, &pcapGlobalHdr, sizeof(pcapGlobalHdr))!=sizeof(pcapGlobalHdr)) {
              // Avoid compile warning
            }
          }
        }
#endif
      }
      if(filewrite(fHandle, data, writeLength)!=(ssize_t)writeLength) {
        // Avoid compile warning
      }
    }
  } else {
    if(fHandle!=NT_OPEN_ERROR) {
      fileclose(fHandle);
      fHandle=NT_OPEN_ERROR;
    }
  }
  if((threadData.captureSize) && ((dataStored+length) > threadData.captureSize)) {
    // Did we get pass the boundary this time?
    StopApplication(0);
  }
}

typedef struct {
  struct {
    uint32_t tv_sec;
    uint32_t tv_usec;
  }ts;
  uint32_t StoredLength;
  uint32_t WireLength;
} intlpcap_t;

#ifdef WITH_PCAP
/* LIBPCAP capture. */
uint8_t pktBuf[32*1024];
static void LibPcapHandler( u_char *file,
                            const struct pcap_pkthdr *h,
                            const u_char *p)
{
  intlpcap_t hdr;
  /* Store data to disk. */
  if(applicationRunning) {
    /* Libpcap returns the header and packet in two
       separate buffers. Write them both. */
    /* Store pcap in 32bit format. */
    hdr.ts.tv_sec  = (uint32_t)h->ts.tv_sec;
    hdr.ts.tv_usec = (uint32_t)h->ts.tv_usec;
    hdr.StoredLength = h->caplen;
    hdr.WireLength   = h->len;
    memcpy(pktBuf, &hdr, sizeof(intlpcap_t));
    memcpy(&pktBuf[sizeof(intlpcap_t)], p, h->caplen);
    StoreData((char*)file, (uint8_t*)&pktBuf, sizeof(intlpcap_t)+h->caplen, 1);
  }
}
#endif

#ifdef WITH_PCAP
int LibPcapRead(char *dev, char *file)
{
  char errbuf[PCAP_ERRBUF_SIZE];  /* Error string */

  /* Open the session in promiscuous mode */
  if((pcapHandle = pcap_open_live(dev, BUFSIZ, 1, 0, errbuf)) == NULL) {
    fprintf(stderr, "Couldn't open device %s: %s\n", dev, errbuf);
    return -1;
  }
  /* This function will hang until breaked. */

  printf("Starting to capture using " LIBPCAP ":\n");
  if (pcap_loop(pcapHandle, -1, LibPcapHandler, (u_char*)file) == -1) {   //<-- Note: -2 and 0 are OK return values
    fprintf(stderr, "pcap_loop returned -1\n");
    return -1;
  }
  pcap_close(pcapHandle);
  pcapHandle = NULL;

  return 0;
}
#endif

/* NTCI capture. */
int NtciRead(int device, char *file)
{
  handle_t cardHandle=NULL;
  uint32_t packetFeedHandle=0;
  uint32_t result;
  FeedConfigSampledType2_t feedConfig;
  BufferInformationSectionType1_t bufferInformation;
  FeedBufferInformationType1_t bufferInfo;
  BufferRequestSampledType2_t request;
  CardHardwareCapabilitiesType1_t cardCapabilities;
  uint32_t channelsBitmask=0;
  uint32_t numFeeds;

  memset(&feedConfig, 0, sizeof(feedConfig));

  request.common.waitTimeoutMS=GET_NEXT_BUFFER_BLOCKING;

  /* Try to open the correct card. */
  printf("Trying to open NTCI based adapter #%d\n", device);
  cardHandle=NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, device);

  if(cardHandle==NULL) {
    fprintf(stderr, "Failed to open adapter %d.\n", device);
    StopApplication(0);
    return -1;
  }

  /* Check if we have feeds running already. */
  if(NTCI_GetPacketFeedCount(cardHandle, &numFeeds) == NTCI_ERRCODE_SUCCESS) {
    while(numFeeds) {
      FeedDescriptionType1_t description;
      result = NTCI_GetPacketFeedDescription(cardHandle,
                                             numFeeds-1,
                                             STID_FEED_DESCRIPTION_TYPE1,
                                             &description);

      if (result == NTCI_ERRCODE_INVALID_PARAMETER) {
        /* A slot has been allocated for this feed, but there is no
         * feed created. Therefore ignore this feed */
        numFeeds--;
        continue;
      }

      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, ">>>Error: Failed getting description for packet feed #%i\n", numFeeds-1);
        NTCI_CloseCard(cardHandle);
        cardHandle = NULL;
        StopApplication(0);
        return -1;
      }

      fprintf(stderr, ">>> Error: Cannot use this adapter. Feeds are already running. " LIBPCAP " perhaps?\n");
      NTCI_CloseCard(cardHandle);
      cardHandle = NULL;
      StopApplication(0);
      return -1;
    }
  }

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_HARDWARE,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_CAPABILITIES_TYPE1,
                                    &cardCapabilities)) != NTCI_ERRCODE_SUCCESS) {

    fprintf(stderr, "Error finding adapter capabilities: %i\n", result);
    NTCI_CloseCard(cardHandle);
    StopApplication(0);
  }
  channelsBitmask = cardCapabilities.totalChannelsBitmask;

  /* Set up the packet feed engine for a single feed */
  if((result = NTCI_SetupPacketFeedEngine(cardHandle, 1, NULL)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Failed to initialise packet feed engine (%i)\n",
            result);
    NTCI_CloseCard(cardHandle);
    StopApplication(0);
    return -1;
  }

  if (cmd_ntpl) {
    channelsBitmask = 0;
  }

  // Configure feed
  feedConfig.timestampType  = TIMESTAMP_NATIVE_UNIX;  //<-- Return NT hardware native timestamping
  feedConfig.sliceLength    = 0;                      //<-- Don't cap input buffer

  // Create the packet feed
  if((result = NTCI_CreatePacketFeed(cardHandle,                                      //<-- hCard. Handle to open adapter.
                                     0,                                               //<-- Index of feed to create. Zero based.
                                     channelsBitmask,                                 //<-- Channel bitmask, which channels/ports are included in this feed. We use NTPL below if this is zero.
                                     0,                                               //<-- Reserved. Just set to zero.
                                     0,                                               //<-- Reserved. Just set to zero.
                                     STID_FEED_CONFIG_SAMPLED_TYPE2,                  //<-- Type of below feedConfig structure.
                                     &feedConfig,                                     //<-- Pointer to a stid of above format
                                     &packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {   //<-- Out handle. This we need to fetch data from the feed.
    fprintf(stderr, "Failed to create a packet feed (%i)\n", result);
    NTCI_CloseCard(cardHandle);
    StopApplication(0);
    return -1;
  }

  if(channelsBitmask==0) {
    PassFilter_t command;
    command.WriteHW = TRUE;
    sprintf(command.achFilterString, "Capture[Priority=0;Feed=0]=ALL");

    if(NTCI_PacketClassification(cardHandle, STID_PASS_FILTER,
                                 &command) != NTCI_ERRCODE_SUCCESS) {
      FilterError_t error;
      (void)NTCI_PacketClassification(cardHandle, STID_GET_FILTER_ERROR,
                                      &error);

      fprintf(stderr, "%s\n",error.achFilterError1);
      fprintf(stderr, "%s\n",error.achFilterError2);
      fprintf(stderr, "%s\n",error.achFilterError3);
      NTCI_CloseCard(cardHandle);
      StopApplication(0);
      return -1;
    }
  }

  if((result = NTCI_StartPacketFeed(cardHandle, packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error while starting packet feed (%i)\n",
            result);
    NTCI_CloseCard(cardHandle);
    StopApplication(0);
    return -1;
  }

  /* find the location of the feed buffer where new
   * data will be stored */
  if((result = NTCI_GetPacketFeedBuffers(cardHandle,
                                         packetFeedHandle,
                                         STID_FEED_BUFFER_INFORMATION_TYPE1,
                                         &bufferInfo)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Unable to retrieve the location of the feed buffer.\n");
    NTCI_CloseCard(cardHandle);
    StopApplication(0);
    return -1;
  }

  printf("Starting to capture using NTCI:\n");
  while(applicationRunning) {
    uint i;
    result = NTCI_GetNextBuffer(cardHandle,
                                packetFeedHandle,
                                STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                &request,
                                STID_BUFFER_INFORMATION_SECTION_TYPE1,
                                &bufferInformation);

    if(result == NTCI_ERRCODE_SUCCESS) {
      if(cmd_walk_packets) {
        int offset = 0;
        bufferInformation.numBytesAvailable=0;
        for(i=0; i<bufferInformation.numDescriptorsAvailable; i++) {
          offset += ((PacketDescriptorNtSeriesType1_t*)((uint8_t*)bufferInfo.bufferBaseAddress+bufferInformation.sectionOffset+offset))->storedLength;
        }
        bufferInformation.numBytesAvailable = offset;
      }

      StoreData(file,
                ((uint8_t*)bufferInfo.bufferBaseAddress)+bufferInformation.sectionOffset,
                bufferInformation.numBytesAvailable,
                bufferInformation.numDescriptorsAvailable);
      
      NTCI_ReleaseBuffer(cardHandle,
                         packetFeedHandle,
                         STID_BUFFER_INFORMATION_SECTION_TYPE1,
                         &bufferInformation);
    } else if(result != NTCI_STATUSCODE_NO_SEGMENT_AVAILABLE) {
      fprintf(stderr, "Failed to get buffers. Error code %d\n", result);
      StopApplication(0);
      break;
    }
  }

  if((result = NTCI_StopPacketFeed(cardHandle, packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Failed to stop packet feed! (%i)\n", result);
  }

  if((result = NTCI_DestroyPacketFeed(cardHandle, packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Failed to destroy packet feed! (%i)\n", result);
  }

  if(channelsBitmask==0) {
    PassFilter_t command;
    command.WriteHW = TRUE;
    sprintf(command.achFilterString, "DeleteFilter=All");

    if (NTCI_PacketClassification(cardHandle, STID_PASS_FILTER,
                                  &command) != NTCI_ERRCODE_SUCCESS) {
      FilterError_t error;
      (void)NTCI_PacketClassification(cardHandle, STID_GET_FILTER_ERROR,
                                      &error);

      fprintf(stderr, "%s\n",error.achFilterError1);
      fprintf(stderr, "%s\n",error.achFilterError2);
      fprintf(stderr, "%s\n",error.achFilterError3);
      NTCI_CloseCard(cardHandle);
      StopApplication(0);
      return -1;
    }
  }

  NTCI_CloseCard(cardHandle);
  return 0;
}

static CommandLineOptions_t app_options[] = {
  {"f", "Capture file.", "file name", OPTION_OPTIONAL, ARG_STR, &cmd_file},
  {"i", "Adapter to use. (Default is NTCI adapter 0.)",  "ntxc?", OPTION_OPTIONAL, ARG_STR, &cmd_if},
#ifndef WITH_PCAP
  {"pcap", "PCAP is disabled in this version.", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_pcap},
#else
  {"pcap", "Use PCAP instead of NTCI. (Default is NTCI.)", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_pcap},
  {"nano", "Store packets in a PCAP_NANOTIME file type sec:nsec. (Default is sec:usec.)", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_pcapnano},
#endif
  {"ntpl", "Use NTPL to set up capture filter.\n\t       (Default is to use channelBitMask.)", NULL, OPTION_BOOLEAN, ARG_NUM, &cmd_ntpl},
  {"t", "Time (in seconds) to run capture. (Default is 60 seconds.)", "0..?", OPTION_OPTIONAL, ARG_NUM, &cmd_time},
  {"s", "Number of bytes to capture. (Default is 10 GB.)",  "0..?", OPTION_OPTIONAL, ARG_NUM, &cmd_size},
  {"w", "Walk packets within a segment before storing them.", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_walk_packets},
  {NULL, NULL, NULL, 0, 0, NULL}
};

static char app_description[] = {
  "The program captures data to disk for up to 60 seconds or until it\n"
  "has captured 10 GB of data.\n"
  "To disable either the 60 seconds or the 10 GB limits, set the limits to zero.\n"
  "If you leave out the target capture file, the program calculates the\n"
  "throughput, but it does not store any data.\n" };

int main(int argc, char *argv[])
{
  char captureFile[255];
  int device;
  int mode=NTCI_CAPTURE;
  char *errstr;
  pthread_t monitorThreadHandle;
#ifndef WIN32
  struct sigaction new;
#endif


  memset(captureFile, 0x00, 255);
  errstr=NULL;
  threadData.captureType = CAPTURE_AND_DISCARD;
  threadData.captureSize = CAPTURE_BUFFER_SIZE;
  threadData.captureTime = CAPTURE_TIME_TOTAL;
  threadData.secsRunning = 0;

  /* Parse the command line and initialize the needed variables. */
  if(InitApplication(&argc, argv, app_options, app_description) == -1) {
    exit(-1);
  }

  applicationRunning = TRUE;
  /* Call this function when CTRL-C is called. */
#ifdef WIN32
  SetConsoleCtrlHandler((PHANDLER_ROUTINE)StopApplication, TRUE);
#else
  memset(&new, 0, sizeof(new));
  new.sa_handler = StopApplication;
  if (sigaction(SIGINT, &new, NULL) < 0) {
    fprintf(stderr, "Failed to register sigaction.\n");
    exit(-1);
  }
#endif

  if(cmd_file) {
    NT_HANDLE fHandle=NT_OPEN_ERROR;
    snprintf(captureFile, 255, "%s", cmd_file);
    /* Trucate the file. */
    printf("Checking for existing file and deleting it if it exists.\n");
    if((fHandle = fileopen(captureFile, FILE_FLAGS, FILE_RIGHTS))==NT_OPEN_ERROR) {
      fprintf(stderr, "Cannot create file %s.\n", captureFile);
      return -1;
    }
    fileclose(fHandle);
    threadData.captureType=CAPTURE_TO_DISK;
  }
  if(cmd_pcap) {
#ifndef WITH_PCAP
    printf("PCAP support was not included at compile time. (Argument \"with-pcap\" was not specified during installation).\n");
    printf("Please re-install the Applications package and specify \"with-pcap\" as argument to install script to enable PCAP support.\n");
    return -1;
#else
    mode = PCAP_CAPTURE;
#endif
  }
  if (cmd_ntpl) {
    if (cmd_pcap) {
      printf("NTPL and PCAP can not be specified at the same time\n");
      return -1;
    }
  }
  if(cmd_pcapnano) {
    if(cmd_pcap==NULL) {
      printf("Cannot store filetype in nanotime format unless pcap is enabled.\n");
      return -1;
    }
  }
  if(cmd_time) {
    threadData.captureTime = strtol(cmd_time, NULL, 0);
  }
  if(cmd_size) {
    threadData.captureSize = strtol(cmd_size, NULL, 0);
  }

  /* try / catch */
  do {
    if(cmd_if) {
      if(!cmd_pcap) {
        device = ParseDeviceName(cmd_if);
      }else{
        device = 0;
      }
     } else {
      cmd_if = malloc(DEVICE_NAME_SIZE);
      cmd_if[0]=0x0;
      device = GetDevice(cmd_if, mode, FALSE);
    }
    if (device<0) {
      printf("No devices found. Is the driver loaded?\n");
      return -1;
    }

    /* Create the monitor thread. */
    if(pthread_create(&monitorThreadHandle, NULL,
                      MonitorThread, (void*)&threadData) != 0) {
      errstr = "Failed to create monitor thread";
      break;
    }

#ifdef WITH_PCAP
    if(mode==PCAP_CAPTURE) {
      if (LibPcapRead(cmd_if, captureFile)) {
        errstr = "LibPcapRead(..) returned error";
        break;
      }
    } else
#endif
    {
      if (NtciRead(device, captureFile)) {
        errstr = "NtciRead(..) returned error";
        break;
      }
    }

    applicationRunning=FALSE;

    /* Destroy the monitor thread. */
    if (pthread_join(monitorThreadHandle, NULL)) {
      errstr = "pthrad_join(..) returned error";
      break;
    }

    if(threadData.captureType == CAPTURE_TO_DISK) {
      if(threadData.captureTime) {
        printf("Time left      :             %02d sec.\n",
          threadData.captureTime-threadData.secsRunning);
      } else {
        printf("Time running   :             %02d sec.\n",
               threadData.secsRunning);
      }
      if(threadData.captureSize) {
        printf("Space left     :     %10.02f MB\n",
               (threadData.captureSize-dataStored)/1024/1024);
      }
      printf("--------------------------------------\n");
      printf("Stored         :        %10.02f MB\n", dataStored/1024/1024);
      printf("Throughput     : Total: %10.02f Mbps\n"
             "                 1 sec: %10.02f Mbps\n"
             "Packets        :  %13lld\n",
             ((dataStored*8)/threadData.secsRunning)/1000000LL,
             ((dataStored-threadData.lastDataSet)*8)/1000000LL,
             (unsigned long long) packets);
    } else {
      printf("Time running   :               %3d sec.\n",
             threadData.secsRunning);
      printf("--------------------------------------\n");
      printf("Data received  :        %10.02f MB\n", dataStored/1024/1024);
      printf("Throughput     : Total: %10.02f Mbps\n"
             "Packets        :  %13lld\n",
             ((dataStored*8)/threadData.secsRunning)/1000000LL,
             (unsigned long long) packets);

    }
  }while(0);

  if (errstr) {
    printf("Error: %s \n", errstr);
    return -1;
  }

  return 0;
}
