
// $Id: Replay.c 25419 2013-09-10 13:03:28Z am $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * Replay example that shows how to transmit using the NTCI interface.
 */

#ifdef WIN32
#include <winsock2.H>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <process.h>
#include "platformwin.h"
#else
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#if ! _NT_OS_TYPE==_NT_OS_TYPE_FREEBSD
#include <sys/mount.h>
#endif

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <pthread.h>

/* Unix-specific includes */
#include <termios.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <sched.h>
#include <sys/mman.h>
#endif

#include "../common/device.h"
#include "../common/cmdparse.h"

#ifdef WITH_PCAP
/* Lib Pcap header */
#include <pcap.h>
#endif

#define	PCAP_MAGIC			0xa1b2c3d4
#define	PCAP_SWAPPED_MAGIC		0xd4c3b2a1
#define	PCAP_MODIFIED_MAGIC		0xa1b2cd34
#define	PCAP_SWAPPED_MODIFIED_MAGIC	0x34cdb2a1
#define	PCAP_NSEC_MAGIC			0xa1b23c4d
#define	PCAP_SWAPPED_NSEC_MAGIC		0x4d3cb2a1

#define ULL unsigned long long

#ifdef WIN32
#define FILE_FLAGS GENERIC_READ
#define FILE_RIGHTS OPEN_EXISTING
#define LIBPCAP_NAME "winPCAP"
#else
#define LIBPCAP_NAME "libPCAP"
#ifdef O_NOATIME
#define FILE_FLAGS O_RDONLY | O_NOATIME
#else
#define FILE_FLAGS O_RDONLY
#endif
#define FILE_RIGHTS 0666
#define fileopen(a, b, c) open(a, b, c)
#define fileclose(a) close(a)
#define fileread(a, b, c) read(a, b, c)
#define filelseek(a, b, c) lseek(a, b, c)

#define NT_HANDLE     int
#define NT_OPEN_ERROR -1
#endif

static int swapped=0;
static int pcap_detected=0;
/* The NTCommonInterface.h header provides the interface definitions
 * for the Napatech NTCI library */
#include <NTCommonInterface.h>

/* These headers are needed when capturing traffic */
#include <packetfeedstructure.h>
#include <packetdescriptor.h>
#include <NTCI_packetfeed.h>
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>
#include <NTCI_cardconfiguration.h>
#include <cardconfigurationstructure.h>
#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>

static int pktTsLast=1; // Default is end of frame timestamp
uint32_t pcap_magic_number=0;

struct nttimeval
{
  uint32_t tv_sec;
  uint32_t tv_usec;
};

/* Global data variables. */

/* Flag that signals whether the application is running or closing down. */
volatile bool_t applicationRunning=FALSE;
/* ongoingTxFlush signals whether Replay is in a state where it is flushing
   the TX ports on the adapter. */
bool_t ongoingTxFlush = FALSE;

#ifdef WITH_PCAP
static pcap_t *pcapHandle=NULL; /* PCAP session handle */
#endif

uint64_t                         PacketsWithTooShortIFG=0;
uint64_t                         PacketsTxSliced=0, BytesTxSliced=0;
uint64_t                         PacketsTxHardSliced=0, BytesTxHardSliced=0;
uint64_t                         PacketsTxCount=0, BytesTxCount=0, TxDropped=0;

int mode=NTCI_CAPTURE;
/* Assume it is not necessary to override CRC.  Override only occurs
   if the file in is PCAP format and the transmission uses NTCI mode
   (not win/libPCAP). */
static bool_t txCrcOverridden = FALSE;
FeedBufferInformationType1_t     bufferInfo;
BufferRequestSampledType2_t      request;
BufferInformationSectionType3_t  bufferSectionTx;
FeedHandle                       PakFeedHandle;
handle_t                         cardHandle=NULL;
uint32_t                         txSegmentSize=0;
PacketDescriptorNtSeriesType1_t *pCurDescriptor=NULL;
uint32_t                         channelsBitmask=0;
static char *cmd_file=NULL;
static char *cmd_if=NULL;
static char *cmd_loops=NULL;
static char *cmd_channels=NULL;
static char *cmd_rate=NULL;
static char *cmd_pcap=NULL;
static char *cmd_raw=NULL;
static char *cmd_ts = NULL;
static char *cmd_gsync = NULL;
static char *cmd_delay = NULL;
static char *cmd_mod_fac = NULL;
static char *cmd_start_time = NULL;
static char *cmd_ntpcap = NULL;

static CommandLineOptions_t app_options[] = {
  {"f", "The capture file to replay.", "file name", OPTION_REQUIRED, ARG_STR, &cmd_file},
  {"ts", "Find the first time stamp (in 10-ns ticks) in the capture file.", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_ts},
  {"g", "Enable global synchronization mode.",
   "earliest 10-ns ticks time stamp to synchronize with",
   OPTION_OPTIONAL, ARG_NUM, &cmd_gsync},
  {"d", "Transmission delay in milliseconds.",
   "delay in milliseconds",
   OPTION_OPTIONAL, ARG_NUM, &cmd_delay},
  {"m", "Transmission begin time stamp modulus factor (default 15 secs.).",
   "modulus factor in seconds",
   OPTION_OPTIONAL, ARG_NUM, &cmd_mod_fac},
  {"s", "Absolute transmission begin time stamp.",
   "YYYY/MM/DD-HH:MM:SS",
   OPTION_OPTIONAL, ARG_STR, &cmd_start_time},
  {"i","Adapter to use.",  "ntxc?", OPTION_OPTIONAL, ARG_STR, &cmd_if},
#ifndef WITH_PCAP
  {"pcap", LIBPCAP_NAME " is disabled in this version.", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_pcap},
#else
  {"pcap", "Use " LIBPCAP_NAME " instead of NTCI. (Default is NTCI.)", NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_pcap},
#endif
  {"l", "Number of times to replay the capture file.", "0=infinite until CTRL-C", OPTION_OPTIONAL, ARG_NUM, &cmd_loops},
  {"c", "Channels bitmask. Each packet is sent onto the channels\n\t       that the bitmask indicates.", "0x1..0xFF", OPTION_OPTIONAL, ARG_NUM, &cmd_channels},
  {"r", "Force packets to be sent with the given rate in Mbps.", "0=txNow,1..10000", OPTION_OPTIONAL, ARG_NUM, &cmd_rate},
  {"org", "Send cap file without manipulating the standard header\n"
   "               (original header).",
   NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_raw},
  {"ntpcap",
   "Denotes that the capture file is in Napatech PCAP format, which\n"
   "               "
   "includes a checksum (CRC) for each packet, whereas a capture file\n"
   "               "
   "created with standard " LIBPCAP_NAME " does not.  Leave out this option\n"
   "               "
   "when replaying a capture file created with standard " LIBPCAP_NAME ", in\n"
   "               "
   "which case Replay sets aside four additional bytes for the CRC.",
   NULL, OPTION_BOOLEAN, ARG_NONE, &cmd_ntpcap},
  {NULL, NULL, NULL, 0, 0, NULL}
};

static char app_description[] = {
  "The \"Replay\" application replays a capture file on the specified adapter.\n"
  "The packets are sent onto the channels on which they were received. This may\n"
  "cause packets not to be sent if they were received on channels that do not\n"
  "exist on the adapter being used for replaying, for example, an NT4E capture\n"
  "file replayed on an NT20E adapter.\n"
  "The \"-r\" option applies to all channels and fails if the rate exceeds the\n"
  "negotiated speed for the channel.\n"
  "When using the \"-r\" and \"-c\" options simultaneously, the combined\n"
  "throughput cannot exceed the PCI bandwidth and results in lower rates than\n"
  "configured.\n"
  "NOTES:\n"
  "  * The NT adapters replay the traffic at the same speed at which it was\n"
  "    received unless the \"-r\" is used,\n"
  "    whereas replay using " LIBPCAP_NAME " replays as fast as possible.\n"
  "  * The only supported capture timestamp is NATIVE unless the capture\n"
  "    file is PCAP compatible. \"Replay\" cannot detect timestamp formats,\n"
  "    and if the file provided does not replay as expected, the reason can\n"
  "    be that the timestamps are not in native format. In this case, use\n"
  "    the \"-r\" option to set the replay rate.\n"
  "  * \"Replay\" uses the adapter to control IFG unless the IFG exceeds one\n"
  "    second. In this case the driver controls the IFG by performing\n"
  "    software wait loops and tries to send the packets with the correct\n"
  "    IFG. If the IFG is greater than two seconds, \"Replay\" enforces a\n"
  "    fixed two second IFG and prints a relevant message on the screen.\n"
  "  * The application does not control the adapter IFG, and any IFG\n"
  "    parameters set with NTPL impacts the transmission rate.\n"
  "  * Replay of a PCAP capture file is supported but not recommended. A PCAP\n"
  "    file does not contain any port relations, and replay of this type of\n"
  "    file only uses port zero. This creates a problem when the PCAP\n"
  "    capture file contains packets from a merged feed because it means\n"
  "    that packets are possibly sent at a rate that is too high and\n"
  "    unsupported by the port. \"Replay\" detects this situation and\n"
  "    sends the packets at line rate and increments the\n"
  "    \"PacketWithTooShortIFG\" counter.\n"
  "  * Replay of a capture file at a too high and unsupported rate occurs at\n"
  "    the adapter's line-rate, and the application increments the\n"
  "    \"PacketWithTooShortIFG\" counter for packets that should be sent at\n"
  "    a higher rate.\n"
  "  * Global synchronization mode (GSYNC) enables synchronized replay on\n"
  "    one or more adapters using one or more servers.\n"
  "  * When using GSYNC with multiple adapters, the adapters must be time\n"
  "    synchronized.\n"
  "  * The argument \"-g\" takes a 10-ns ticks time stamp; the value to use\n"
  "    can be determined by invoking Replay with the \"-ts\" option on each\n"
  "    of the capture files that are about to be replayed.  The value to be\n"
  "    used with all invocations of Replay is the same, namely the smallest\n"
  "    10-ns ticks time stamp in all of the capture files.\n"
  "  * The adapter transmits hard sliced packets with CRC errors unless\n"
  "    (1) the capture is in PCAP format and (2) the packets are transmitted\n"
  "    with the NTCI interface (not " LIBPCAP_NAME "), in which case\n"
  "    the adapter overrides the erroneous CRC with a valid CRC.\n"
  "  * The adapter does not transmit soft sliced packets.\n"
};

/* Global variables */
static uint32_t gsyncMode = 0U;  /* Set to one in GSYNC mode (-g) option */
static uint64_t gsyncOffset = 0U; /* The offset on the adapter in gsync mode */

/* Forward Declarations */
static uint64_t GetFirstTimeStamp(const char *fileName);
static uint64_t CalcOffset(uint64_t ts_start, uint32_t delay,
			   uint64_t ts1, uint64_t ts2);
static uint64_t CalcStartTS(uint64_t ts, uint32_t mod_fac);
static void PrintTimeStamp(uint64_t ts);
static int GetAdapterTS(handle_t h_card, uint64_t *ts);
static int ParseStartTime(const char *dts, uint64_t *ts);
static int CheckLinkStatusAndSpeed(
  const CardHardwareCapabilitiesType2_t *cardCapabilities,
  int32_t speed, uint32_t channelsBitmask);

/*
 * The function called when user presses CTRL-C
 */
#ifdef WIN32
  BOOL WINAPI StopApplication(DWORD sig)
#else
  static void StopApplication(int sig)
#endif
{
  if (applicationRunning) {
    applicationRunning = FALSE;
    printf("Stopping application: (%d)%40s\n", sig, "");
    fflush(stdout);
  }

#ifdef WIN32
  return TRUE;
#endif
}

/* Global variables, part two */
#define MAX_NUM_CHANNELS 4
// previous channel timestamp.
static uint64_t prevChanTime[MAX_NUM_CHANNELS]={0,0,0,0};
// chanTime array contains the timestamp for the next packet
static uint64_t chanTime[MAX_NUM_CHANNELS]={0x1000,0x1000,0x1000,0x1000}; // Start with 0x1000 offset. Just so we don't use 0
static uint32_t ifSpeed[MAX_NUM_CHANNELS]={0,0,0,0}; // The interface speed.
static uint8_t fpgaV=0;
static uint32_t relativeTxTiming = 1U;
static uint64_t gsyncSyncTS = 0U; /* The -g option value converted to a number */
static uint64_t gsyncFirstTS = 0U;  /* The first time stamp in the capture file */
static uint64_t gsyncBeginTS = 0U;  /* Synchronization time stamp */
static uint32_t gsyncDelay = 0U;  /* Transmission delay used on first replay */
static uint64_t gsyncAdapterTS = 0U; /* The adapter time stamp */
static uint32_t gsyncModFactor = 15U; /* Default modulus factor */
static uint64_t gsyncMaxTS = 0U; /* Maximum (last) TS used in gsync mode */
static uint64_t gsyncStartTimeOption = 0U; /* Option -s in secs since epoch */
 /* Used for multiple replay iterations of capture file */
static uint64_t gsyncAdditionalTsOffset[MAX_NUM_CHANNELS + 1] =
{ 0U, 0U, 0U, 0U, 0U };
/* Time stamp of first packet of an iteration of a capture file */
static uint64_t gsyncIterTsBegin[MAX_NUM_CHANNELS + 1] =
{ 0U, 0U, 0U, 0U, 0U };
/* Time stamp of the last packet of an iteration of a capture file */
static uint64_t gsyncIterTsEnd[MAX_NUM_CHANNELS + 1] =
{ 0U, 0U, 0U, 0U, 0U };


/*
 * Transmit a packet without changing the descriptor
 */
static int TxPacketRaw(const uint8_t *packet, uint8_t firstPacket)
{
  int skipLen=0;
  uint32_t storedLength, cpyLength;
#if 0
  uint32_t wireLength;
#endif
  uint32_t result;
  uint64_t *pSrc=NULL, *pDest=NULL;
  uint32_t descriptor_swapped[4];
  PacketDescriptorNtSeriesType1_t *pDescriptor = (PacketDescriptorNtSeriesType1_t*)packet;
  if(swapped) {
    descriptor_swapped[0]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+0));
    descriptor_swapped[1]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+4));
    descriptor_swapped[2]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+8));
    descriptor_swapped[3] = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+12));
    pDescriptor=(PacketDescriptorNtSeriesType1_t *)descriptor_swapped;
  }
#if 0
  wireLength   = pDescriptor->wireLength;
#endif
  storedLength = pDescriptor->storedLength;
  cpyLength    = pDescriptor->wireLength;

  if((pDescriptor->hardSlice==1) &&
     (pDescriptor->frameSliced==1)) {
    // Sliced packet. Use stored length as length.
    cpyLength=(storedLength-
               sizeof(PacketDescriptorNtSeriesType1_t)-(pDescriptor->extensionLength<<3));
    PacketsTxHardSliced++;
    BytesTxHardSliced+=cpyLength;
  } else if(pDescriptor->frameSliced==1) {
    cpyLength=(storedLength-
               sizeof(PacketDescriptorNtSeriesType1_t)-(pDescriptor->extensionLength<<3));
    PacketsTxSliced++;
    BytesTxSliced+=cpyLength;
  }

  if(pCurDescriptor != NULL) {
    int releaseSegment=0;

    // Check whether this is the first packet in the file
    if((bufferSectionTx.numBytesAvailable+storedLength) > txSegmentSize) {
      releaseSegment=1;
    }

    if(releaseSegment) {
      /* Setting the transmitFlag to one and releasing the buffer causes
         the transmit to occur. */
      bufferSectionTx.transmitFlag = 1;
      NTCI_ReleaseBuffer(cardHandle,
                         PakFeedHandle,
                         STID_BUFFER_INFORMATION_SECTION_TYPE3,
                         &bufferSectionTx);
      pCurDescriptor = NULL;
    }
  }

  // Check if we already have a segment to use
  if(pCurDescriptor == NULL) {
    // Force wait for new segment
    request.common.waitTimeoutMS = GET_NEXT_BUFFER_BLOCKING;

    result = NTCI_GetNextBuffer(cardHandle,
                                PakFeedHandle,
                                STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                &request,
                                STID_BUFFER_INFORMATION_SECTION_TYPE3,
                                &bufferSectionTx);

    if(result == NTCI_ERRCODE_SUCCESS) {
      /* Insert packet into the transmit-only buffer */
      pCurDescriptor = (PacketDescriptorNtSeriesType1_t *)((char *)bufferInfo.bufferBaseAddress + bufferSectionTx.sectionOffset);
      txSegmentSize = bufferSectionTx.numBytesAvailable;
      // Segment now contains no data.
      bufferSectionTx.numBytesAvailable = 0;
    } else if(result == NTCI_STATUSCODE_NO_SEGMENT_AVAILABLE) {
      // We cannot send. Return 0 bytes sent.
      return 0;
    } else {
      return -1;
    }
  }

  pDest = (uint64_t*)pCurDescriptor;
  pSrc  = (uint64_t*)(packet);

  // Copy the standard header
  *pDest=*((uint64_t*)pDescriptor+0);
  *(pDest+1)=*((uint64_t*)pDescriptor+1);

  cpyLength=storedLength-16; // Don't copy the std descr, this is already done via pDest assignments

  /* Initialize bits pertaining to absolute TX timing mode */
  if (firstPacket) {
    ((PacketDescriptorNtSeriesType1_t *) pDest)->txSetClock = 1;
    ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
  }

  // Copy payload
  pSrc += (2 + (skipLen >> 3));
  pDest += 2;
  (void) memcpy(pDest, pSrc, cpyLength);


  /* Check the packet length is dividable by eight, it is an error
     (undefined) to hand over a frame with an illegal size, and it
     typically causes the adapter to discard the packet. */
  if (pCurDescriptor->storedLength % 8U) {
    fprintf(stderr,
            ">>> Error: The capture file contains a packet with size = %u,\n"
            "           which is not dividable by eight.\n",
            pCurDescriptor->storedLength);
    /* A negative return value causes Replay to exit */
    return -1;
  }

  pCurDescriptor=(PacketDescriptorNtSeriesType1_t*)((uint8_t*)pCurDescriptor+storedLength);
  bufferSectionTx.numBytesAvailable+=storedLength;

  PacketsTxCount++;
  BytesTxCount += cpyLength;

  return cpyLength;
}


/*
 * Transmit a packet.
 */
static int TxPacket(const uint8_t *packet, int32_t speed, int8_t channel, uint8_t firstPacket, uint8_t firstIteration)
{
  /* speed == -1: use packet ts; speed == 0: line rate; speed > 0: the rate */
  /* channel == -1: use channel from captured packet descriptor;
     channel >= 0: use that channel. */
  uint32_t storedLength, wireLength, cpyLength;
  uint64_t curTs; // Variable to "point" to the correct chanTime;
  uint32_t descriptor_swapped[4];
  PacketDescriptorNtSeriesType1_t *pDescriptor = (PacketDescriptorNtSeriesType1_t*)packet;
  if(swapped) {
    descriptor_swapped[0]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+0));
    descriptor_swapped[1]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+4));
    descriptor_swapped[2]  = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+8));
    descriptor_swapped[3] = ntohl(*(uint32_t*)((uint8_t*)pDescriptor+12));
    pDescriptor=(PacketDescriptorNtSeriesType1_t *)descriptor_swapped;
  }
  wireLength   = pDescriptor->wireLength;
  storedLength = pDescriptor->storedLength;

  cpyLength    = wireLength;

  /* Check channel because it is used as an array index */
  if (channel < -1 || channel >= MAX_NUM_CHANNELS)
    return -1;

  if(mode == NTCI_CAPTURE) {
    uint32_t result;
    uint64_t *pSrc=NULL, *pDest=NULL;
    int skipLen=0;
    int sliced=0;

    /* Check for libpcap. Convert to PacketDescriptorNtSeriesType1_t. */
    if(pDescriptor->descriptorType==PCAP_DESCRIPTOR) {
      struct nttimeval *pcap_time = (struct nttimeval *)(pDescriptor);
      if(channel==-1) {
        channel=0; // PCAP default only send on port 0
      }
      // Convert the timestamp.
      if((pcap_magic_number == PCAP_NSEC_MAGIC) || (pcap_magic_number == PCAP_SWAPPED_NSEC_MAGIC)) {
        // PCAP nano
        curTs=((uint64_t)pcap_time->tv_sec*1000000000)+pcap_time->tv_usec;
        curTs/=10;  // 10ns tick
      } else {
        curTs = ((uint64_t)pcap_time->tv_sec * 1000000 +
                 pcap_time->tv_usec) * 100; // 10ns ticks
      }

      if (wireLength > storedLength) {
	sliced=1;
        PacketsTxSliced++;
        wireLength=storedLength;
      }

      /* PCAP files created with libpcap do not contain CRC, whereas PCAP files
         created by Napatech adapters (using the capture application) do. */
      if (!cmd_ntpcap) {
        wireLength += 4; // Remember to set aside space for crc
      }
      cpyLength=wireLength;
      storedLength=((wireLength+sizeof(PacketDescriptorNtSeriesType1_t)+7)&~7); /* 8 bytes alignment. */
    } else {
      if(channel==-1) {
        channel=pDescriptor->channel;
      }
      curTs=*((uint64_t*)pDescriptor);
      if((pDescriptor->hardSlice==1) &&
         (pDescriptor->frameSliced==1)) {
        // Sliced packet. Use stored length as length.
        cpyLength=(storedLength-
                   sizeof(PacketDescriptorNtSeriesType1_t)-(pDescriptor->extensionLength<<3));
        PacketsTxHardSliced++;
        BytesTxHardSliced+=cpyLength;
      } else if(pDescriptor->frameSliced==1) {
        cpyLength=(storedLength-
                   sizeof(PacketDescriptorNtSeriesType1_t)-(pDescriptor->extensionLength<<3));
	      sliced=1;
        PacketsTxSliced++;
        BytesTxSliced+=cpyLength;
      }
    }

    /**
     * Check if we need to introduce a delay.
     */
    if(speed>0) {
      /**
       * Calculate timestamp jump value
       * One byte takes 8/speed sec to transmit
       */
      uint64_t trans_time;
      /* +20 because of default 12 IPG and 8 PREEAMPLE. 10ns Ticks */
      trans_time = (((((uint64_t) wireLength + 20) * 8 * 1000000)) / speed);
      /* Round up to nearest 10 */
      trans_time = ((trans_time + 9) / 10) * 10;
      if (pktTsLast) {
        chanTime[channel] += trans_time;
        curTs = chanTime[channel] / 10;
      } else {
        /* Next packet time should be: */
        curTs = chanTime[channel] / 10;
        chanTime[channel] += trans_time;
      }
    }

    /* Adjust curTs in global sync mode */
    if (gsyncMode) {
      uint64_t trans_time;
      static int print_first_ts = 1;

      if (firstPacket && firstIteration) {
	gsyncIterTsBegin[channel + 1] = curTs;
	gsyncIterTsEnd[channel + 1] = curTs;
      }

      if (firstPacket) {
	const uint64_t round_f = 1000000000 / 1000 / 10; /* 10 ns-ticks */
	uint64_t increment =
	  gsyncIterTsEnd[channel + 1] - gsyncIterTsBegin[channel + 1];
        /* Round up to nearest millisecond to ensure synchronization
           when using multiple replay applications on links with different
           speeds. */
	increment = ((increment + (round_f - 1)) / round_f) * round_f;
	gsyncAdditionalTsOffset[channel + 1] += increment;
	gsyncIterTsBegin[channel + 1] = gsyncIterTsEnd[channel + 1];
      }

      curTs += gsyncAdditionalTsOffset[channel + 1];

      if (curTs > gsyncMaxTS)
        gsyncMaxTS = curTs;

      /* Calculate number of 10ns-ticks to transmit packet */
      trans_time = (((((uint64_t) wireLength + 20) * 8 * 100)) / ifSpeed[channel]);
      /* Round up to nearest 10 */
      trans_time = ((trans_time + 9) / 10) * 10;
      gsyncIterTsEnd[channel + 1] = curTs + trans_time;

      if (firstPacket && print_first_ts) {
        print_first_ts = 0;
        printf("Transmission begins on ");
        PrintTimeStamp(curTs + gsyncOffset);
        printf(" (%llu nsec)\n", (ULL) ((curTs + gsyncOffset) * 10U));
      }
    }

    if(pCurDescriptor != NULL) {
      int releaseSegment=0;
      int waitInSw=0;

      // Check whether this is the first packet in the file
      if((firstPacket==0) && (speed == -1)) {
        /**
         * If the delay between two packets is larger than 10 msec, release
	 * the segment to become more responsive to CTRL-C.
         */
        if(prevChanTime[channel]!=0) {
          if(prevChanTime[channel] < curTs) {
            if((curTs-prevChanTime[channel])>1000000) {
              releaseSegment=1;
            }
            /**
             * If the delay is larger than 1 sec., wait in software to become
	     * more responsive to CTRL-C.
             */
            if((curTs-prevChanTime[channel])>100000000) {
              waitInSw=1;
              releaseSegment=1;
            }
          }
        }
      }
      if((bufferSectionTx.numBytesAvailable+storedLength) > txSegmentSize) {
        releaseSegment=1;
      }

      if(releaseSegment) {
	/* Setting the transmitFlag to one and releasing the buffer causes
	   the transmit to occur. */
        bufferSectionTx.transmitFlag = 1;
        NTCI_ReleaseBuffer(cardHandle,
                           PakFeedHandle,
                           STID_BUFFER_INFORMATION_SECTION_TYPE3,
                           &bufferSectionTx);
        pCurDescriptor = NULL;
      }

      if(waitInSw) {
        struct timeval tv;
        tv.tv_sec = (curTs-prevChanTime[channel])/100000000;
        tv.tv_usec= (((curTs-prevChanTime[channel])%100000000)/100);
        if(tv.tv_sec > 2) {
          printf("Pause between packets is large: %ld.%ld\n", (unsigned long)tv.tv_sec,(unsigned long)tv.tv_usec);
          printf("Force to only wait 2 sec.\n");
          tv.tv_sec=2;
          tv.tv_usec=0;
        }
#ifdef WIN32
        Sleep((tv.tv_sec*1000)+(tv.tv_usec/1000));
#else
        select(0, NULL, NULL, NULL, &tv);
#endif
        prevChanTime[channel]=0;
      }
    }

    // Check if we already have a segment to use
    if(pCurDescriptor == NULL) {
      // Force wait for new segment
      request.common.waitTimeoutMS = GET_NEXT_BUFFER_BLOCKING;

      /* The call to NTCI_GetNextBuffer constitutes a significant waiting
         point during replay.  When the CTRL-C handler runs, this handler
         flushes all transmission ports, which causes new segments to be
         available, which in turns causes NTCI_GetNextBuffer to return
         (a segment). */
      result = NTCI_GetNextBuffer(cardHandle,
                                  PakFeedHandle,
                                  STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                  &request,
                                  STID_BUFFER_INFORMATION_SECTION_TYPE3,
                                  &bufferSectionTx);

      if(result == NTCI_ERRCODE_SUCCESS) {
        /* Insert packet into the transmit-only buffer */
        pCurDescriptor = (PacketDescriptorNtSeriesType1_t *)((char *)bufferInfo.bufferBaseAddress + bufferSectionTx.sectionOffset);
        txSegmentSize = bufferSectionTx.numBytesAvailable;
        // Segment now contains no data.
        bufferSectionTx.numBytesAvailable = 0;
      } else if(result == NTCI_STATUSCODE_NO_SEGMENT_AVAILABLE) {
        // We cannot send. Return 0 bytes sent.
        return 0;
      } else {
        return -1;
      }
    }

    pDest = (uint64_t*)pCurDescriptor;
    pSrc  = (uint64_t*)(packet);

    // Copy the standard header
    *pDest=*((uint64_t*)pDescriptor+0);
    *(pDest+1)=*((uint64_t*)pDescriptor+1);

    if(sliced) {
      ((PacketDescriptorNtSeriesType1_t*)pDest)->frameSliced=1;
    }

    ((PacketDescriptorNtSeriesType1_t*)pDest)->txIgnore=0;
    // Set TX now to 0 to ensure packets are sent with the RX rate. (Might be set otherwise further down)
    ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow=0;

    if(pDescriptor->extensionFormat ||
       pDescriptor->extensionLength) {
      if(((fpgaV >= 40) &&
          (pDescriptor->extensionFormat < 5)) ||
         ((fpgaV < 40) &&
          (pDescriptor->extensionFormat!=0x05)) ||
         ((pDescriptor->extensionLength!=2))) {
        /* TX has limited extended header support. Convert to the standard descriptor. */
        skipLen=(((PacketDescriptorNtSeriesType1_t*)pSrc)->extensionLength<<3);
        storedLength=(storedLength-skipLen);
        ((PacketDescriptorNtSeriesType1_t*)pDest)->descriptorType=STANDARD_DESCRIPTOR;
        ((PacketDescriptorNtSeriesType1_t*)pDest)->extensionLength=0;
        ((PacketDescriptorNtSeriesType1_t*)pDest)->extensionFormat=0;
        ((PacketDescriptorNtSeriesType1_t*)pDest)->storedLength = storedLength;
      }
    }

    /**
     * When in relative timing mode, NT series cards do not detect jumps
     * backwards in time as min IPG, so force the adapter to send the first
     * frame ASAP, and the rest of the frames will be sent with correct IPG.
     * When in absolute timing  mode, the adapter will detect the backward jump.
     *
     * Do not use txNow if rate is to be overridden.
     */
    if (prevChanTime[channel] >= curTs && speed == -1) {
      if (!gsyncMode)
	((PacketDescriptorNtSeriesType1_t *) pDest)->txNow = 1;
    }

    /**
     * The very first frame, or the first frame after a delay introduced by
     * this application, must be sent ASAP.
     */
    if(!prevChanTime[channel]) {
      if (!gsyncMode)
	((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
    }
    /**
     * If speed is set to 0 it is equal to txNow.
     */
    if(speed==0) {
      /* speed is never zero in gsyncMode, but we check anyway */
      if (!gsyncMode)
	((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
    }

    if(pDescriptor->descriptorType==PCAP_DESCRIPTOR) {
      /* Convert to STD header. */
      ((PacketDescriptorNtSeriesType1_t*)pDest)->wireLength   = wireLength;
      ((PacketDescriptorNtSeriesType1_t*)pDest)->storedLength = storedLength;
      ((PacketDescriptorNtSeriesType1_t*)pDest)->descriptorType=STANDARD_DESCRIPTOR;
      ((PacketDescriptorNtSeriesType1_t*)pDest)->txCrcOverride=1;
      txCrcOverridden = TRUE;
    } else {
      cpyLength=storedLength-16; // Don't copy the std descr, this is already done via pDest assignments
    }

    // Save the timestamp.
    *((uint64_t*)(pDest))=curTs;

    if(cmd_channels) {
      /* Overwrite channel. */
      ((PacketDescriptorNtSeriesType1_t*)pDest)->txChannel = channel;
    } else {
      /* Transmit on the same channel we received the frame */
      ((PacketDescriptorNtSeriesType1_t*)pDest)->txChannel = pDescriptor->channel;
    }

    /**
     * Detect whether packets are being sent with too short IFG.
     * This can occur if the user transmits a merged PCAP file, or
     * if the user sends an NT20 capture file on an NT4 port, or more generally,
     * if the user sends a capture file at a rate that exceeds what the
     * adapter supports.  Replay will attempt to detect this case and issue a
     * txNow and increment an error counter.
     */
    if((curTs-prevChanTime[channel]) != 0) {
      if((((((uint64_t)wireLength+20)*8*100))/(curTs-prevChanTime[channel])) > ifSpeed[channel]) {
        PacketsWithTooShortIFG++;
        if(fpgaV < 40) {
          ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
        }
      }
    } else {
      PacketsWithTooShortIFG++;
      if(fpgaV < 40) {
        // FPGA 40+ can handle that packets are sent too close together.
        ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
      }
    }

    /* Initialize bits pertaining to absolute timing mode */
    ((PacketDescriptorNtSeriesType1_t *) pDest)->txSetClock = 0;
    if (gsyncMode) {
      if (firstPacket && firstIteration) {
	/* Set up GSYNC offset register */
	TimeSynchronizedTxConfigurationType1_t card_cfg;
	if (channelsBitmask == 0)
	  card_cfg.timeSynchronizedTxPortEnableMask = ~channelsBitmask;
	else
	  card_cfg.timeSynchronizedTxPortEnableMask = channelsBitmask;
        /* The adapter implements the following logic in GSYNC mode.
           Let Tp denote the time stamp of a packet, t the current (running)
           time, and offset the GSYNC offset in the adapter.

             Tp  > t + offset   : wait, do not send the packet yet
             Tp <= t + offset   : send the packet

           Replay calculates gsyncOffset a little differently, namely as

             t - Tp = gsyncOffset       and hence
             Tp = -gsyncOffset + t

           If we insert this last equation into the right-hand side of the
           expression used in the adapter (Tp = t + offset), we get

             Tp = t + offset = -gsyncOffset + t

           and therefore

             offset = -gsyncOffset

           Replay therefore uses the negated value of gsyncOffset as the
           offset value on the adapter.
        */
        /* Use the two-complement negated value of gsyncOffset as offset. */
	card_cfg.timeSynchronizedTxOffset = (~gsyncOffset) + 1;
	card_cfg.timeSynchronizedTxOffsetUpdate = 1;
	result = NTCI_SetCardConfigurationDetails(
	  cardHandle,
	  channelsBitmask,
	  STID_TIME_SYNCHRONIZED_TX_CONFIGURATION_TYPE1,
	  &card_cfg);
	if (result != NTCI_ERRCODE_SUCCESS) {
	  fprintf(stderr, "Failed to enable GSYNC and set offset to %llu.\n",
		  (unsigned long long) card_cfg.timeSynchronizedTxOffset);
	  return -1;
	}
      }
    } else {
      if (!relativeTxTiming && firstPacket && firstIteration) {
	/* Disable GSYNC on all ports */
	TimeSynchronizedTxConfigurationType1_t card_cfg;
	/* Set the enable mask to 0x0 to disable GSYNC on all ports */
	card_cfg.timeSynchronizedTxPortEnableMask = 0x0;
	card_cfg.timeSynchronizedTxOffset = 0;
	card_cfg.timeSynchronizedTxOffsetUpdate = 1;
	result = NTCI_SetCardConfigurationDetails(
	  cardHandle, 0x0,
	  STID_TIME_SYNCHRONIZED_TX_CONFIGURATION_TYPE1, &card_cfg);
	if (result != NTCI_ERRCODE_SUCCESS) {
	  fprintf(stderr, "Failed to disable GSYNC mode.\n");
	  return -1;
	}
      }

      if (!relativeTxTiming && firstPacket) {
	((PacketDescriptorNtSeriesType1_t *) pDest)->txSetClock = 1;
	if (firstIteration) {
	  /* must use txdelay */
	  ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 0;
	}
	else {
	  ((PacketDescriptorNtSeriesType1_t*)pDest)->txNow = 1;
	}
      }
    }

    // Store timestamp;
    prevChanTime[channel]=curTs;

    // Copy payload
    pSrc += (2 + (skipLen >> 3));
    pDest += 2;
    (void) memcpy(pDest, pSrc, cpyLength);


    /* Check the packet length is dividable by eight, it is an error
       (undefined) to hand over a frame with an illegal size, and it
       typically causes the adapter to discard the packet. */
    if (pCurDescriptor->storedLength % 8U) {
      fprintf(stderr,
              ">>> Error: The capture file contains a packet with size = %u,\n"
              "           which is not dividable by eight.\n",
              pCurDescriptor->storedLength);
      /* A negative return value causes Replay to exit */
      return -1;
    }

    pCurDescriptor=(PacketDescriptorNtSeriesType1_t*)((uint8_t*)pCurDescriptor+storedLength);
    bufferSectionTx.numBytesAvailable+=storedLength;

    PacketsTxCount++;
    BytesTxCount += cpyLength;

    return cpyLength;
  } else {
    /* PCAP mode */
#ifdef WITH_PCAP
    const uint8_t *pkt;
    int ret;
    // Convert header to pcap.
    if(pDescriptor->descriptorType==STANDARD_DESCRIPTOR) {
      if((pDescriptor->hardSlice==1) ||
         (pDescriptor->frameSliced==1)) {
        // Sliced packet. Use stored length as length.
        PacketsTxHardSliced++;
        BytesTxHardSliced+=cpyLength;
        cpyLength=(storedLength-
                   sizeof(PacketDescriptorNtSeriesType1_t)-(pDescriptor->extensionLength<<3));
      } else if(pDescriptor->frameSliced==1) {
        PacketsTxSliced++;
        BytesTxSliced+=cpyLength;
      } else {
        cpyLength = (pDescriptor->wireLength-4);
      }
      pkt = (packet+sizeof(PacketDescriptorNtSeriesType1_t)+
             (pDescriptor->extensionLength<<3));
    } else {
      pkt = (packet+sizeof(PacketDescriptorNtSeriesType1_t));
    }
    ret=pcap_sendpacket(pcapHandle, pkt, cpyLength);

    if(ret==0) {
      PacketsTxCount++;
      BytesTxCount += cpyLength;
      return cpyLength;
    } else {
      return -1;
    }
#else
    fprintf(stderr, "PCAP not compiled in.\n");
    return -1;
#endif
  }
}


/*
 * Transmit function. Sends out the contents of input file "fileName"
 * without changing anything in the standart header
 */
static void TxCaptureRaw(IN char                            *dev,
                         IN char                            *fileName)

{
  uint32_t                         result=0;
  FeedConfigSampledType5_t         feedConfig;
  FeedStatusSampledType3_t         feedStatusTx;
  NT_HANDLE                        fHandle = NT_OPEN_ERROR;
  uint32_t                         offset=0;
  uint8_t                          chBitmask=0;
  uint64_t                         numBytesAvailable=0;
  uint32_t pktInFile=0;
  PacketDescriptorNtSeriesType1_t *pDescriptor=NULL;
  CardHardwareCapabilitiesType2_t cardCapabilities;
  CardHardwareCapabilitiesType4_t cap4;
  CardPacketFeedCapabilitiesType2_t cardPacketFeed;
  CardHardwareVersionCapabilitiesType3_t data3;
  int adapterNumber;
  uint64_t fpgaId=0;

#ifdef WIN32
  HANDLE hMemoryMappedFile = INVALID_HANDLE_VALUE;
#endif

  memset(&feedConfig, 0, sizeof(feedConfig));

  /* Open the correct card. */
  // parse adapter card string
  adapterNumber = ParseDeviceName(dev);
  // check parse result - issue error if needed
  if (adapterNumber <= -1) {    /* stop transmitting packets */
    NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
    NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);

    return;
  }

  printf("Trying to open NTCI based card #%d\n", adapterNumber);
  if(strstr(dev, "ntxc") != NULL) {
    cardHandle=NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, adapterNumber);
  }
  if(cardHandle==NULL) {
    fprintf(stderr, "Failed to open card: '%s'\n", dev);
    return;
  }
  printf("Step1\n");

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_VERSION,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                    &data3)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Failed to get capabilities\n");
    NTCI_CloseCard(cardHandle);
    return;
  }
  if(data3.AdapterCard.dwInfoType==1) {
    fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==2) {
    fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==3) {
    fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
  }
  fpgaV=((fpgaId&0x0000FF00) >> 8);

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_HARDWARE,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_CAPABILITIES_TYPE4,
                                    &cap4)) == NTCI_ERRCODE_SUCCESS) {
    if(cap4.pktTsMethod == PACKET_TIMESTAMP_FIRST) {
      pktTsLast = 0;
    }
    relativeTxTiming = cap4.relativeTxTiming;
  }
  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_HARDWARE,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_CAPABILITIES_TYPE2,
                                    &cardCapabilities)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error finding card capabilities: %i\n", result);
    NTCI_CloseCard(cardHandle);
    return;
  } else {
    if(cardCapabilities.txChannelsSupportedBitmask==0) {
      fprintf(stderr, "Adapter doesn't support TX.\n");
      NTCI_CloseCard(cardHandle);
      return;
    }
  }
  if(NTCI_GetCapabilities(cardHandle, CAPID_PACKET_FEED, NOT_SUB_CAPABILITIES,
                          STID_CARD_PACKET_FEED_CAPABILITIES_TYPE2,
                          &cardPacketFeed) == NTCI_ERRCODE_SUCCESS) {
    /* Never in PCAP mode in TxCaptureRaw, ensure txBoundToPort is unset. */
    if(cardPacketFeed.type1.txBoundToPort) {
      fprintf(stderr, "Adapter is running with PerPortTx. Please disable this when loading the driver.\n");
      NTCI_CloseCard(cardHandle);
      return;
    }
  }

  BytesTxCount = 0;
  PacketsTxCount = 0;

  /* Set up the packet feed engine for a single feed */
  result = NTCI_SetupPacketFeedEngine(cardHandle, 1, NULL);
  if (result != NTCI_ERRCODE_SUCCESS)
  {
    fprintf(stderr, "Failed to initialise packet feed engine (%i)\n",
            result);
    fprintf(stderr, "Make sure that no existing packet feeds are still running/created - use KillFeeds example\n");
    NTCI_CloseCard(cardHandle);
    return;
  }

  if (CheckLinkStatusAndSpeed(&cardCapabilities, 0, 0) < 0) {
    (void) NTCI_CloseCard(cardHandle);
    return;
  }

  // Configure a packet feed
  feedConfig.timestampType        = TIMESTAMP_NATIVE_UNIX;  //<-- Return UNIX timestamping
  feedConfig.sliceLength          = 0;                      //<-- Don't cap input buffer
  feedConfig.descriptorType       = STANDARD_DESCRIPTOR;    //<-- Prepend each packet with a standard descriptor
  feedConfig.maxLatency           = 1000;                   //<-- Timeout in usecs. After 1ms cause NTCI_GetNextBuffer() to return if called blocking and not enough packets has been received to fill a segment
  feedConfig.numTransmitSections  = 16;                     //<-- Number of sections for transmit
  //    This divides up the transmit buffer into 16 segments,
  //    these can be used simultaneously, i.e. you could be filling
  //    one section whilst transmitting from another, these segments
  //    will be queued in the driver/hardware when ReleaseBuffer is
  //    called

  // Create the packet feed
  result = NTCI_CreatePacketFeed(cardHandle,                                      //<-- hCard. Handle to open adapter.
                                 0,                                               //<-- Index of feed to create. Zero based.
                                 0,                                               //<-- Channel bitmask. 0 because we don't want to have a filter created.
                                 0,                                               //<-- Reserved. Just set to zero.
                                 0,                                               //<-- Reserved. Just set to zero.
                                 STID_FEED_CONFIG_SAMPLED_TYPE5,                  //<-- Type of below feedConfig structure.
                                 &feedConfig,                                     //<-- Pointer to a stid of above format
                                 &PakFeedHandle);                                 //<-- Out handle. This we need to fetch data from the feed.

  if (result != NTCI_ERRCODE_SUCCESS) {
    char buf[128];
    fprintf(stderr, "Failed to create a packet feed (%i)\n", result);
    NTCI_ExplainLastManagerError(buf, sizeof(buf));
    puts(buf);
    NTCI_CloseCard(cardHandle);
    return;
  }

  /* start the packet feed */
  result = NTCI_StartPacketFeed(cardHandle, PakFeedHandle);

  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error while starting packet feed (%i)\n",
            result);
    (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
    (void) NTCI_CloseCard(cardHandle);
    return;
  }

  /* find the location of the feed buffer where new
   * data will be stored */
  result = NTCI_GetPacketFeedBuffers(cardHandle,
                                     PakFeedHandle,
                                     STID_FEED_BUFFER_INFORMATION_TYPE1,
                                     &bufferInfo);

  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Unable to retrieve the location of the feed "
            "buffer.\n");
    (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
    (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
    (void) NTCI_CloseCard(cardHandle);
    return;
  }

  /* Open the input file */
  if((fHandle = fileopen(fileName, FILE_FLAGS, FILE_RIGHTS)) != NT_OPEN_ERROR) {
    uint64_t cookie=0;
    uint8_t *pCapture = NULL;
    if(fileread(fHandle, &cookie, sizeof(cookie))!=sizeof(cookie)) {
      fprintf(stderr, "File invalid size.\n");
      (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_CloseCard(cardHandle);
      return;
    }
    if(numBytesAvailable == 0) {
      /* Set pointer back to the start. */
      filelseek(fHandle, 0, SEEK_SET);
      numBytesAvailable = filelseek(fHandle,0, SEEK_END);
    }

    printf("Replay file \"%s\" starts at offset: %d\n", fileName, (uint32_t)offset);

    printf("Feed started. Transmitting packets ..\n");
    filelseek(fHandle,0, SEEK_SET);

#ifdef WIN32
    hMemoryMappedFile = CreateFileMapping(fHandle, NULL, PAGE_READONLY, 0, numBytesAvailable, NULL);
    if (hMemoryMappedFile == NULL) {
      pCapture = MAP_FAILED;
    }
    else {
      pCapture = MapViewOfFile(hMemoryMappedFile, FILE_MAP_READ, 0, 0, numBytesAvailable);
    }
#else
    pCapture = mmap(NULL, numBytesAvailable, PROT_READ,
                    MAP_PRIVATE, fHandle, 0);
#endif
    if(pCapture != MAP_FAILED) {
      /* This is the send data main loop
       * We stay in this loop until the number of requested iterations has been reached
       */
//      while(applicationRunning) {
        uint8_t firstPacket;

        /*
         * This is the send data inner loop
         * We stay in this loop until we have reached the end of the input file
         */
        pDescriptor = ((PacketDescriptorNtSeriesType1_t*)(pCapture+offset));
        pktInFile = 0;
        firstPacket=1; // Indicate this it the first packet in a file
        while(applicationRunning && ((uint8_t*)pDescriptor < (pCapture+numBytesAvailable))) {
          uint32_t storedLength;
          if(swapped) {
            storedLength = ntohl(*(((uint32_t*)pDescriptor)+2));
          } else {
            storedLength = pDescriptor->storedLength;
          }
          // Sanity checks
          if((storedLength > sizeof(PacketDescriptorNtSeriesType1_t)) &&
             ((uint8_t*)pDescriptor+storedLength <= (pCapture+numBytesAvailable))) {
            int ret=0;
            // Transmit the packet
#ifdef WIN32
            while ((ret=TxPacketRaw((uint8_t*)pDescriptor, firstPacket)) == 0)
            {
              if (applicationRunning == FALSE)
                break;
            }
            if (ret < 0)
            {
              fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
              if(ret==-2) {
                TxDropped++;
              } else {
                applicationRunning = FALSE;
                break;
              }
            }
#else
            if((ret=TxPacketRaw((uint8_t*)pDescriptor, firstPacket)) <= 0) {
              fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
              if(ret==-2) {
                TxDropped++;
              } else {
                applicationRunning = FALSE;
                break;
              }
            }
#endif
            pktInFile++;
          } else {
            fprintf(stderr, "Error detected in file. Packet #%d\n", pktInFile);
            applicationRunning = FALSE;
            break;
          }
          if(pcap_detected) {
            pDescriptor = (PacketDescriptorNtSeriesType1_t*)((uint8_t*)pDescriptor+storedLength+16);
          } else {
            pDescriptor = (PacketDescriptorNtSeriesType1_t*)((uint8_t*)pDescriptor+(pDescriptor->descriptorType==PCAP_DESCRIPTOR?storedLength+16:storedLength));
          }
          fprintf(stderr, "pktInFile:%d\n", pktInFile);
          firstPacket=0;
        } /* end of send data inner loop */
//      } /* End of send data main loop */

      if(mode==NTCI_CAPTURE) {
        if(pCurDescriptor) {
          bufferSectionTx.transmitFlag = 1;

          NTCI_ReleaseBuffer(cardHandle,
                             PakFeedHandle,
                             STID_BUFFER_INFORMATION_SECTION_TYPE3,
                             &bufferSectionTx);
          pCurDescriptor = NULL;
          sleep(1);
        }
        printf("Waiting for transmit to complete...\n");
        for (;;) {
          result = NTCI_GetPacketFeedStatus(cardHandle,
                                            PakFeedHandle,
                                            STID_FEED_STATUS_SAMPLED_TYPE3,
                                            &feedStatusTx);
          if (result == NTCI_ERRCODE_SUCCESS) {
            if (feedStatusTx.numBytesPendingTransmit) {
              printf("Still %d bytes to be sent.\r", feedStatusTx.numBytesPendingTransmit);
              fflush(stdout);
              sleep(1);
            } else {
              printf("%80s\rDONE\n", "");
              break;
            }
            if (applicationRunning == FALSE) {
              break;
            }
          } else {
            printf("FAILED to get packet feed status.  Error code: %d\n",
                   result);
            break;
          }
        }
      }
#ifdef WIN32
      UnmapViewOfFile(pCapture);
      CloseHandle(hMemoryMappedFile);
#else
      munmap(pCapture, numBytesAvailable);
#endif
    }

    //Capture card  all ports must be flushed
    chBitmask = cardCapabilities.txChannelsSupportedBitmask;
    NTCI_FlushTxPorts(cardHandle, chBitmask, FALSE);

    /* stop transmitting packets */
    NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
    NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);

    PakFeedHandle = (FeedHandle)0;

    if (result != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error while stopping packet feed (%i)\n",
              result);
    }
    printf("Packet feed stopped. \n");
    printf("Total packets Transmitted: %lld\n", (unsigned long long)PacketsTxCount);
    printf("Total bytes Transmitted: %lld\n", (unsigned long long)BytesTxCount);
    printf("Total packets (soft sliced): %lld\n", (unsigned long long)PacketsTxSliced);
    printf("Total bytes (soft sliced): %lld\n", (unsigned long long)BytesTxSliced);
    printf("Total packets (hard sliced): %lld\n", (unsigned long long)PacketsTxHardSliced);
    printf("Total bytes (hard sliced): %lld\n", (unsigned long long)BytesTxHardSliced);
    printf("Total packets with too short IFG: %lld\n", (unsigned long long)PacketsWithTooShortIFG);
    fileclose(fHandle);
    if ((PacketsTxHardSliced > 0 && (!txCrcOverridden)) ||
        PacketsTxSliced > 0) {
      printf("Note:\n");
      if (PacketsTxHardSliced > 0 && (!txCrcOverridden))
        printf("HardSliced packets are transmitted with CRC errors.\n");
      if (PacketsTxSliced > 0)
        printf("Soft sliced packets are not sent by the adapter.\n");
    }
  }

  NTCI_CloseCard(cardHandle);
  return;
}


/*
 * Transmit function. Sends out the contents of input file "fileName" for
 * "iterationsRequested" number of times.
 */
static void TxCapture(IN char                            *dev,
                      IN char                            *fileName,
                      IN uint32_t                        channelsBitmask,
                      IN uint32_t                        iterationsRequested,
                      IN int32_t                         speed,
                      IN uint32_t                        mode)
{
  uint32_t                         result=0;
  FeedConfigSampledType5_t         feedConfig;
  BufferInformationSectionType1_t  secBufInfo;
  FeedStatusSampledType3_t         feedStatusTx;
  NT_HANDLE                        fHandle = NT_OPEN_ERROR;
  uint32_t                         iterationsPerformed=0, offset=0;
  uint8_t                          chBitmask=0, channel=0;
  uint64_t                         numBytesAvailable=0;
  uint32_t pktInFile=0;
  PacketDescriptorNtSeriesType1_t *pDescriptor=NULL;
  CardHardwareCapabilitiesType2_t cardCapabilities;
  CardHardwareCapabilitiesType4_t cap4;
  CardPacketFeedCapabilitiesType2_t cardPacketFeed;
  uint8_t firstIteration;

#ifdef WIN32
  HANDLE hMemoryMappedFile = INVALID_HANDLE_VALUE;
#endif

  memset(&feedConfig, 0, sizeof(feedConfig));

  /* Open the correct card. */
  if(mode==NTCI_CAPTURE) {
    CardHardwareVersionCapabilitiesType3_t data3;
    int adapterNumber;
    uint64_t fpgaId=0;
    // parse adapter card string
    adapterNumber = ParseDeviceName(dev);
    // check parse result - issue error if needed
    if (adapterNumber <= -1) {
      return;
    }

    printf("Trying to open NTCI based card #%d\n", adapterNumber);
    if(strstr(dev, "ntxc") != NULL) {
      cardHandle=NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, adapterNumber);
    }
    if(cardHandle==NULL) {
      fprintf(stderr, "Failed to open card: '%s'\n", dev);
      return;
    }

    if((result = NTCI_GetCapabilities(cardHandle,
                                      CAPID_VERSION,
                                      NOT_SUB_CAPABILITIES,
                                      STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                      &data3)) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Failed to get capabilities\n");
      NTCI_CloseCard(cardHandle);
      return;
    }
    if(data3.AdapterCard.dwInfoType==1) {
      fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
    }else if(data3.AdapterCard.dwInfoType==2) {
      fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
    }else if(data3.AdapterCard.dwInfoType==3) {
      fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
    }
    fpgaV=((fpgaId&0x0000FF00) >> 8);

    if((result = NTCI_GetCapabilities(cardHandle,
                                      CAPID_HARDWARE,
                                      NOT_SUB_CAPABILITIES,
                                      STID_CARD_HARDWARE_CAPABILITIES_TYPE4,
                                      &cap4)) == NTCI_ERRCODE_SUCCESS) {
       if(cap4.pktTsMethod == PACKET_TIMESTAMP_FIRST) {
        pktTsLast = 0;
      }
      relativeTxTiming = cap4.relativeTxTiming;
    }
    /* Check whether driver supports gsync mode */
    if (gsyncMode && relativeTxTiming) {
      fprintf(stderr,
	      "Global sync mode is not supported in relative timing mode.\n"
	      "Please enable absolute timing mode by setting RelativeTxTiming=0"
	      "\nwhen loading the driver.\n");
      NTCI_CloseCard(cardHandle);
      return;
    }
    if((result = NTCI_GetCapabilities(cardHandle,
                                      CAPID_HARDWARE,
                                      NOT_SUB_CAPABILITIES,
                                      STID_CARD_HARDWARE_CAPABILITIES_TYPE2,
                                      &cardCapabilities)) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error finding card capabilities: %i\n", result);
      NTCI_CloseCard(cardHandle);
      return;
    } else {
      if(cardCapabilities.txChannelsSupportedBitmask==0) {
        fprintf(stderr, "Adapter doesn't support TX.\n");
        NTCI_CloseCard(cardHandle);
        return;
      }
    }
    if(NTCI_GetCapabilities(cardHandle, CAPID_PACKET_FEED, NOT_SUB_CAPABILITIES,
                            STID_CARD_PACKET_FEED_CAPABILITIES_TYPE2,
                            &cardPacketFeed) == NTCI_ERRCODE_SUCCESS) {
      if (cmd_pcap) {
        /* Check txBoundToPort (PcapTx=1 or PerPortTx=1) in PCAP mode */
        if (! cardPacketFeed.type1.txBoundToPort) {
          fprintf(stderr,
                  "Adapter must be configured with PerPortTx=1 in PCAP mode. "
                  "Please enable this when loading the driver.\n");
          NTCI_CloseCard(cardHandle);
          return;
        }
      } else {
        if (cardPacketFeed.type1.txBoundToPort) {
          fprintf(stderr,
                  "Adapter is running with PerPortTx=1. Please disable "
                  "this when loading the driver.\n");
          NTCI_CloseCard(cardHandle);
          return;
        }
      }
    }

    BytesTxCount = 0;
    PacketsTxCount = 0;

    /* Set up the packet feed engine for a single feed */
    result = NTCI_SetupPacketFeedEngine(cardHandle, 1, NULL);
    if (result != NTCI_ERRCODE_SUCCESS)
    {
      fprintf(stderr, "Failed to initialise packet feed engine (%i)\n",
              result);
      fprintf(stderr, "Make sure that no existing packet feeds are still running/created - use KillFeeds example\n");
        NTCI_CloseCard(cardHandle);
      return;
    }

    /* Configure packet feed */

    // check channel-bitmask
    if(channelsBitmask) {
      if ( (channelsBitmask & cardCapabilities.type1.totalChannelsBitmask) != channelsBitmask) {
        fprintf(stderr, ">>> Error: Channelbitmask (0x%02x) references channels outside the available adapter channels (0x%02x).\n",
                channelsBitmask, cardCapabilities.type1.totalChannelsBitmask);
        NTCI_CloseCard(cardHandle);
        exit(0);
      }
    }

    if (CheckLinkStatusAndSpeed(&cardCapabilities, speed, channelsBitmask) < 0) {
      (void) NTCI_CloseCard(cardHandle);
      exit(0);
    }

    // Configure a packet feed
    feedConfig.timestampType        = TIMESTAMP_NATIVE_UNIX;  //<-- Return UNIX timestamping
    feedConfig.sliceLength          = 0;                      //<-- Don't cap input buffer
    feedConfig.descriptorType       = STANDARD_DESCRIPTOR;    //<-- Prepend each packet with a standard descriptor
    feedConfig.maxLatency           = 1000;                   //<-- Timeout in usecs. After 1ms cause NTCI_GetNextBuffer() to return if called blocking and not enough packets has been received to fill a segment
    feedConfig.numTransmitSections  = 16;                     //<-- Number of sections for transmit
                                                              //    This divides up the transmit buffer into 16 segments,
                                                              //    these can be used simultaneously, i.e. you could be filling
                                                              //    one section whilst transmitting from another, these segments
                                                              //    will be queued in the driver/hardware when ReleaseBuffer is
                                                              //    called

    // Create the packet feed
    result = NTCI_CreatePacketFeed(cardHandle,                                      //<-- hCard. Handle to open adapter.
                                   0,                                               //<-- Index of feed to create. Zero based.
                                   0,                                               //<-- Channel bitmask. 0 because we don't want to have a filter created.
                                   0,                                               //<-- Reserved. Just set to zero.
                                   0,                                               //<-- Reserved. Just set to zero.
                                   STID_FEED_CONFIG_SAMPLED_TYPE5,                  //<-- Type of below feedConfig structure.
                                   &feedConfig,                                     //<-- Pointer to a stid of above format
                                   &PakFeedHandle);                                 //<-- Out handle. This we need to fetch data from the feed.

    if (result != NTCI_ERRCODE_SUCCESS) {
      char buf[128];
      fprintf(stderr, "Failed to create a packet feed (%i)\n", result);
      NTCI_ExplainLastManagerError(buf, sizeof(buf));
      puts(buf);
      NTCI_CloseCard(cardHandle);
      return;
    }

    if (gsyncMode) {
      /* Check we can get the adapter time stamp and that the format
	 is native UNIX. */
      result = GetAdapterTS(cardHandle, &gsyncAdapterTS);
      if (result != NTCI_ERRCODE_SUCCESS) {
	fprintf(stderr, ">>> Error: Problem with the adapter's time stamp.\n");
        (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
	(void) NTCI_CloseCard(cardHandle);
	return;
      }
      if (cmd_start_time) {
	/* Validate start time is not in the past */
	if (gsyncStartTimeOption <= gsyncAdapterTS) {
          fprintf(stderr, ">>> Error: Start time is in the past.\n");
          (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
          (void) NTCI_CloseCard(cardHandle);
          return;
        }
        gsyncBeginTS = gsyncStartTimeOption;
      } else {
	gsyncBeginTS = CalcStartTS(gsyncAdapterTS, gsyncModFactor);
      }
      gsyncOffset =
	CalcOffset(gsyncBeginTS, gsyncDelay, gsyncFirstTS, gsyncSyncTS);
    }

    /* start the packet feed */
    result = NTCI_StartPacketFeed(cardHandle, PakFeedHandle);

    if (result != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error while starting packet feed (%i)\n",
              result);
      (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_CloseCard(cardHandle);
      return;
    }

    /* find the location of the feed buffer where new
     * data will be stored */
    result = NTCI_GetPacketFeedBuffers(cardHandle,
                                       PakFeedHandle,
                                       STID_FEED_BUFFER_INFORMATION_TYPE1,
                                       &bufferInfo);

    if (result != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Unable to retrieve the location of the feed "
              "buffer.\n");
      (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_CloseCard(cardHandle);
      return;
    }
  } else {
#ifdef WITH_PCAP
    // PCAP.
    char errbuf[PCAP_ERRBUF_SIZE];	/* Error string */
    if((pcapHandle = pcap_open_live(dev, BUFSIZ, 1, 0, errbuf)) == NULL) {
      fprintf(stderr, "Couldn't open device %s: %s\n", dev, errbuf);
      return;
    }
#endif
  }

  /* Open the input file */
  if((fHandle = fileopen(fileName, FILE_FLAGS, FILE_RIGHTS)) != NT_OPEN_ERROR) {
    uint64_t cookie=0;
    uint8_t *pCapture = NULL;
    if(fileread(fHandle, &cookie, sizeof(cookie))!=sizeof(cookie)) {
      fprintf(stderr, "File invalid size.\n");
      (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_CloseCard(cardHandle);
      return;
    }
    if(cookie==0xFEDEABBA) {
      cookie = 0;
      if(fileread(fHandle, &cookie, sizeof(cookie))!=sizeof(cookie)) {
        (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
        (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
        (void) NTCI_CloseCard(cardHandle);
        fprintf(stderr, "File invalid size.\n");
        return;
      }
      if(cookie==0xFEDEABBA) {
        /* We have a segment dump. Not a capture. */
        if(fileread(fHandle, &secBufInfo, sizeof(BufferInformationSectionType1_t))!=sizeof(BufferInformationSectionType1_t)) {
          fprintf(stderr, "File invalid size.\n");
          (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
          (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
          (void) NTCI_CloseCard(cardHandle);
          return;
        }
        numBytesAvailable=secBufInfo.numBytesAvailable;
        offset = sizeof(uint64_t)*2+sizeof(BufferInformationSectionType1_t);
      } else {
        offset = 0;
      }
    } else {
      /* Check for PCAP global header. */
      filelseek(fHandle, 0, SEEK_SET);
      if(fileread(fHandle, &pcap_magic_number, sizeof(pcap_magic_number)) != sizeof(pcap_magic_number)) {
        fprintf(stderr, "File invalid size.\n");
        (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
        (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
        (void) NTCI_CloseCard(cardHandle);
        return;
      }
      if((pcap_magic_number==PCAP_SWAPPED_MAGIC) || (pcap_magic_number==PCAP_SWAPPED_MODIFIED_MAGIC) || (pcap_magic_number==PCAP_SWAPPED_NSEC_MAGIC)) {
        /* We have a pcap global header. */
        offset = 24;
        swapped=1;
        pcap_detected=1;
      } else if((pcap_magic_number==PCAP_MAGIC) || (pcap_magic_number==PCAP_MODIFIED_MAGIC) || (pcap_magic_number==PCAP_NSEC_MAGIC)) {
        /* We have a pcap global header. */
        offset = 24;
        pcap_detected=1;
      }
    }
    if(numBytesAvailable == 0) {
      /* Set pointer back to the start. */
      filelseek(fHandle, 0, SEEK_SET);
      numBytesAvailable = filelseek(fHandle,0, SEEK_END);
    }

    if (cmd_ntpcap && (!pcap_detected)) {
      fprintf(stderr,
              "Error: The -ntpcap option cannot be used because\n"
              "       the capture file is not in PCAP format.\n");
      (void) NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
      (void) NTCI_CloseCard(cardHandle);
      return;
    }

    printf("Replay file \"%s\" starts at offset: %d\n", fileName, (uint32_t)offset);
    if(swapped) {
      printf("File is in swapped format\n");
    }
    printf("Feed started. Transmitting packets ..\n");
    filelseek(fHandle,0, SEEK_SET);

#ifdef WIN32
    hMemoryMappedFile = CreateFileMapping(fHandle, NULL, PAGE_READONLY, 0, numBytesAvailable, NULL);
    if (hMemoryMappedFile == NULL) {
      pCapture = MAP_FAILED;
    }
    else {
      pCapture = MapViewOfFile(hMemoryMappedFile, FILE_MAP_READ, 0, 0, numBytesAvailable);
    }
#else
    pCapture = mmap(NULL, numBytesAvailable, PROT_READ,
                    MAP_PRIVATE, fHandle, 0);
#endif

    if(pCapture != MAP_FAILED) {
      /* This is the send data main loop.
       * Loop for the number of iterations or until the user presses ctrl-c.
       */
      iterationsPerformed=0;
      firstIteration=1;
      while (applicationRunning) {
        uint8_t firstPacket;
        if(iterationsRequested) {
          if(iterationsPerformed++ >= iterationsRequested) {
            break;
          }
        }
        /*
         * This is the send data inner loop
         * We stay in this loop until we have reached the end of the input file
         */
        pDescriptor = ((PacketDescriptorNtSeriesType1_t*)(pCapture+offset));
        pktInFile = 0;
        firstPacket=1; // Indicate this it the first packet in a file
        while(applicationRunning && ((uint8_t*)pDescriptor < (pCapture+numBytesAvailable))) {
          // Sanity checks
          uint16_t storedLength;
          if(swapped) {
            storedLength = ntohl(*(((uint32_t*)pDescriptor)+2));
          } else {
            storedLength = pDescriptor->storedLength;
          }
          if((storedLength > sizeof(PacketDescriptorNtSeriesType1_t)) &&
             ((uint8_t*)pDescriptor+storedLength <= (pCapture+numBytesAvailable))) {
            int ret=0;

            if(mode==NTCI_CAPTURE) {
              if(cmd_channels) {
                const uint32_t totalChannelsBitmask =
                  cardCapabilities.type1.totalChannelsBitmask;
                /* Replicate the packet onto all channels in channelsBitmask. */
                for(channel=0; ; channel++) {

                  /* Check channel is available on the adapter */
                  if (!((totalChannelsBitmask >> channel) & 0x1))
                    break;

                  if(channelsBitmask&(1<<channel)) {
                    // Transmit the packet
#ifdef WIN32
                    while ((ret=TxPacket((uint8_t*)pDescriptor, speed, channel, firstPacket, firstIteration)) == 0)
                    {
                      if (applicationRunning == FALSE)
                        break;
                    }
                    if (ret < 0)
                    {
                      fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                      if(ret==-2) {
                        TxDropped++;
                      } else {
                        applicationRunning = FALSE;
                        break;
                      }
                    }
#else
		    if (gsyncMode)
		      do {
			/* TxPacket may return zero in GSYNC mode, which
			   indicates that no segments are available because
			   the adapter withholds the packet(s) until the right
			   moment. */
			ret = TxPacket((uint8_t *) pDescriptor, speed, channel,
				       firstPacket, firstIteration);
		      } while (ret == 0);
		    else
		      ret=TxPacket((uint8_t*)pDescriptor, speed, channel,
				   firstPacket, firstIteration);
                    if (ret <= 0) {
                      fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                      if(ret==-2) {
                        TxDropped++;
                      } else {
                        applicationRunning = FALSE;
                      }
                      break;
                    }
#endif
                  }
                }
              } else {
                // Transmit the packet
#ifdef WIN32
                while ((ret=TxPacket((uint8_t*)pDescriptor, speed, -1, firstPacket, firstIteration)) == 0)
                {
                  if (applicationRunning == FALSE)
                    break;
                }
                if (ret < 0)
                {
                  fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                  if(ret==-2) {
                    TxDropped++;
                  } else {
                    applicationRunning = FALSE;
                    break;
                  }
                }
#else
		if (gsyncMode)
		  do {
		    /* TxPacket may return zero in GSYNC mode, which
		       indicates that no segments are available because
		       the adapter withholds the packet(s) until the right
		       moment. */
		    ret = TxPacket((uint8_t *) pDescriptor, speed, -1,
				   firstPacket, firstIteration);
		  } while (ret == 0);
		else
		  ret = TxPacket((uint8_t *) pDescriptor, speed, -1,
				 firstPacket, firstIteration);
                if (ret <= 0) {
                  fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                  if(ret==-2) {
                    TxDropped++;
                  } else {
                    applicationRunning = FALSE;
                    break;
                  }
                }
#endif
              }
            } else {
#ifdef WIN32
              // Transmit the packet
              while ((ret=TxPacket((uint8_t*)pDescriptor, speed, 0, firstPacket, firstIteration)) == 0)
              {
                if (applicationRunning == FALSE)
                  break;
              }
              if (ret < 0)
              {
                fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                if(ret==-2) {
                  TxDropped++;
                } else {
                  applicationRunning = FALSE;
                  break;
                }
              }
#else
              if((ret=TxPacket((uint8_t*)pDescriptor, speed, 0, firstPacket, firstIteration)) <= 0) {
                fprintf(stderr, "Failed to send packet #%d\n", pktInFile);
                if(ret==-2) {
                  TxDropped++;
                } else {
                  applicationRunning = FALSE;
                }
                break;
              }
#endif
            }
            pktInFile++;
          } else {
            fprintf(stderr, "Error detected in file. Packet #%d\n", pktInFile);
            applicationRunning = FALSE;
            break;
          }
          if(pcap_detected) {
            pDescriptor = (PacketDescriptorNtSeriesType1_t*)((uint8_t*)pDescriptor+storedLength+16);
          } else {
            pDescriptor = (PacketDescriptorNtSeriesType1_t*)((uint8_t*)pDescriptor+(pDescriptor->descriptorType==PCAP_DESCRIPTOR?storedLength+16:storedLength));
          }
          firstPacket=0;
        } /* end of send data inner loop */
	firstIteration=0;
      } /* End of send data main loop */

      if(mode==NTCI_CAPTURE) {
        if(pCurDescriptor) {
          bufferSectionTx.transmitFlag = 1;

          if (NTCI_ReleaseBuffer(
                cardHandle, PakFeedHandle,STID_BUFFER_INFORMATION_SECTION_TYPE3,
                &bufferSectionTx) != NTCI_ERRCODE_SUCCESS) {
            fprintf(stderr, "%s: Failed to release buffer.\n", __func__);
          }
          pCurDescriptor = NULL;
        }

        printf("Waiting for transmit to complete...\n");
        for (;;) {
          result = NTCI_GetPacketFeedStatus(cardHandle,
                                            PakFeedHandle,
                                            STID_FEED_STATUS_SAMPLED_TYPE3,
                                            &feedStatusTx);
          if (result == NTCI_ERRCODE_SUCCESS) {
            if (feedStatusTx.numBytesPendingTransmit) {
              /* If the user pressed ctrl-c and there are bytes
                 waiting to be sent, flush all TX ports.  Flushing must occur
                 *after* the last call to NTCI_ReleaseBuffer. */
              if (!applicationRunning && !ongoingTxFlush) {
                ongoingTxFlush = TRUE;
                /* Replay does not support per port flushing, flush all ports */
                chBitmask = cardCapabilities.txChannelsSupportedBitmask;
                if (NTCI_FlushTxPorts(cardHandle, chBitmask, TRUE) !=
                    NTCI_ERRCODE_SUCCESS)
                  fprintf(stderr, "%s: NTCI_FlushTxPorts(TRUE) failed\n", __func__);
              }
              printf("Still %d bytes to be sent.          \r",
                     feedStatusTx.numBytesPendingTransmit);
              fflush(stdout);
              sleep(1);
            } else {
              printf("%80s\rDONE - no more pending packets.\n", "");
              break;
            }
#if 0
            if (applicationRunning == FALSE) {
              break;
            }
#endif
          } else {
            printf("FAILED to get packet feed status.  Error code: %d\n",
                   result);
            break;
          }
        }
      }
#ifdef WIN32
      UnmapViewOfFile(pCapture);
      CloseHandle(hMemoryMappedFile);
#else
      munmap(pCapture, numBytesAvailable);
#endif
    }

    /* Handle GSYNC mode specific close-down.
       When in GSYNC mode, even though the transmission is completed, there
       may be packets that have not yet been sent onto the wire.
    */
    if (gsyncMode) {
      TimeSynchronizedTxConfigurationType1_t card_cfg;
      int clean_up_console = 0;
      const uint64_t last_ts = gsyncMaxTS + gsyncOffset;

      while (applicationRunning) {
        char msg[128];
        const int console_width = 80;
        uint64_t adap_ts;

        if (GetAdapterTS(cardHandle, &adap_ts) != NTCI_ERRCODE_SUCCESS) {
          fprintf(stderr, ">>> Error: Failed to read adapter's time stamp.\n");
          break;
        }

        if (adap_ts > last_ts)
          break;

        (void) snprintf(msg, console_width,
                        "GSYNC mode: awaiting completion in %llu seconds",
                        (ULL) (((last_ts - adap_ts) / 100000000) + 1));
        printf("%-80s\r", msg);
        fflush(stdout);
        clean_up_console = 1;
        sleep(1);
      }
      if (clean_up_console) {
        printf("%80s\r", "");
        fflush(stdout);
      }

      if (!applicationRunning && !ongoingTxFlush) {
        /* Handle ctrl-c from user: Flush all TX ports */
        chBitmask = cardCapabilities.txChannelsSupportedBitmask;
        if (NTCI_FlushTxPorts(cardHandle, chBitmask, TRUE)
            == NTCI_ERRCODE_SUCCESS) {
          ongoingTxFlush = TRUE;
          sleep(2); /* Allow a little while for the flush */
        } else {
          fprintf(stderr, "%s: NTCI_FlushTxPorts(TRUE) failed\n", __func__);
        }
      }

      /* Put the adapter back in normal IFG mode so that other
         applications work as expected. */
      /* Set the enable mask to 0x0 to disable GSYNC on all ports */
      card_cfg.timeSynchronizedTxPortEnableMask = 0x0;
      card_cfg.timeSynchronizedTxOffset = 0U;
      card_cfg.timeSynchronizedTxOffsetUpdate = 1;
      result = NTCI_SetCardConfigurationDetails(
        cardHandle, 0x0,
        STID_TIME_SYNCHRONIZED_TX_CONFIGURATION_TYPE1, &card_cfg);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, "Failed to disable GSYNC and set offset to zero.\n");
      }
    }

    if(mode==NTCI_CAPTURE) {
      /* stop transmitting packets */
      result = NTCI_StopPacketFeed(cardHandle, PakFeedHandle);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, "Error: Failed to stop packet feed; error %d\n",
		result);
      } else {
	result = NTCI_DestroyPacketFeed(cardHandle, PakFeedHandle);
        if (result != NTCI_ERRCODE_SUCCESS) {
          fprintf(stderr, "Error: Failed to destroy packet feed; error %d\n",
                  result);
        }
      }

      PakFeedHandle = (FeedHandle)0;

      if (result == NTCI_ERRCODE_SUCCESS)
        printf("Packet feed stopped.\n");

      if (ongoingTxFlush) {
        if (NTCI_FlushTxPorts(cardHandle, chBitmask, FALSE) !=
            NTCI_ERRCODE_SUCCESS)
          fprintf(stderr,
                  "Error: Failed to disable flushing of Tx Segments. %x\n",
                  result);
        ongoingTxFlush = FALSE;
      }
    }
    printf("Total packets Transmitted: %lld\n", (unsigned long long)PacketsTxCount);
    printf("Total bytes Transmitted: %lld\n", (unsigned long long)BytesTxCount);
    printf("Total packets (soft sliced): %lld\n", (unsigned long long)PacketsTxSliced);
    printf("Total bytes (soft sliced): %lld\n", (unsigned long long)BytesTxSliced);
    printf("Total packets (hard sliced): %lld\n", (unsigned long long)PacketsTxHardSliced);
    printf("Total bytes (hard sliced): %lld\n", (unsigned long long)BytesTxHardSliced);
    printf("Total packets with too short IFG: %lld\n", (unsigned long long)PacketsWithTooShortIFG);
    fileclose(fHandle);
    if ((PacketsTxHardSliced > 0 && (!txCrcOverridden)) ||
        PacketsTxSliced > 0) {
      printf("Note:\n");
      if (PacketsTxHardSliced > 0 && (!txCrcOverridden))
        printf("HardSliced packets are transmitted with CRC errors.\n");
      if (PacketsTxSliced > 0)
        printf("Soft sliced packets are not sent by the adapter.\n");
    }
  }

  if(mode==NTCI_CAPTURE) {
    NTCI_CloseCard(cardHandle);
  } else {
#ifdef WITH_PCAP
    pcap_close(pcapHandle);
    pcapHandle = NULL;
#endif
  }

  return;
}

int main(int argc, char *argv[])
{
  char device[DEVICE_NAME_SIZE];
  uint32_t loopCount=1;
  int32_t rate=0;
  int32_t result;

  memset(device, 0x00, DEVICE_NAME_SIZE);

  if(InitApplication(&argc, argv, app_options, app_description) == -1) {
    exit(-1);
  }

  applicationRunning = TRUE;
  /* Call this function when CTRL-C is called. */
#ifdef WIN32
  SetConsoleCtrlHandler((PHANDLER_ROUTINE)StopApplication, TRUE);
#else
  signal(SIGINT, StopApplication);
#endif

  /* Parse the commandline. */
  if (cmd_ts) {
    uint64_t firstTS = 0;
    /* InitApplication has already checked that -f option is present, but
       check anyway to be on the safe side. */
    if (!cmd_file) {
      fprintf(stderr,
	      "The -ts option requires that a file name be specified "
	      "with the -f option\n.");
      return -2;
    }
    firstTS = GetFirstTimeStamp(cmd_file);
    printf("%llu\n", (unsigned long long) firstTS);
    return (firstTS == 0);
  }

  if (cmd_gsync) {
    gsyncMode = 1;
    gsyncSyncTS = strtoull(cmd_gsync, NULL, 0);
    gsyncDelay = (cmd_delay ? strtol(cmd_delay, NULL, 0) * (1000000/10) : 0);
    if (cmd_mod_fac && cmd_start_time) {
      fprintf(stderr,
	      ">>> Error: Options -m and -s "
	      "cannot be used at the same time.\n");
      return -1;
    }
    gsyncModFactor =
      (cmd_mod_fac ? strtoul(cmd_mod_fac, NULL, 0) : gsyncModFactor);
    if (gsyncModFactor == 0) {
      fprintf(stderr, "Option -m must be a positive value.\n");
      return -1;
    }
    if (cmd_start_time) {
      uint64_t begin_ts = 0U;
      if (ParseStartTime(cmd_start_time, &begin_ts) < 0) {
	fprintf(stderr, ">>> Error: Failed to parse start time '%s'\n",
		cmd_start_time);
	return -1;
      }
      /* convert to 10-nsec ticks */
      gsyncStartTimeOption = begin_ts * 100000000;
    }
    /* Check user has not used incompatible options. */
    if (cmd_raw || cmd_rate || cmd_pcap) {
      fprintf(stderr,
	      ">>> Error: Option -g cannot be used together with "
	      "-org, -r, or -pcap options.\n");
      return -1;
    }
    /* InitApplication has already checked that -f option is present, but
       check anyway to be on the safe side. */
    if (!cmd_file) {
      fprintf(stderr,
	      "The -g option requires that a file name be specified "
	      "with the -f option\n.");
      return -2;
    }
    gsyncFirstTS = GetFirstTimeStamp(cmd_file);
  }

  if (!cmd_gsync && !cmd_raw) {
    /* Check for unsupported options in plain replay mode */
    if (cmd_delay || cmd_mod_fac || cmd_start_time) {
      fprintf(stderr, ">>> Error: Plain replay mode does not support "
              "-d, -m, and -s options.\n");
      return -1;
    }
  }

  if(cmd_file) {
    NT_HANDLE fHandle=NT_OPEN_ERROR;
    /* Open the captured data file */
    printf("Validating the presence of capture data file\n");
    if((fHandle = fileopen(cmd_file, FILE_FLAGS, FILE_RIGHTS))==NT_OPEN_ERROR) {
#ifdef WIN32
      // Retrieve the system error message for the last-error code
      char msgBuf[500];
      FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                    NULL,
                    GetLastError(),
                    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                    (LPTSTR)msgBuf,
                    500,
                    NULL );
      fprintf(stderr, "Cannot open file \"%s\" - %s\n", cmd_file, msgBuf);
#else
      fprintf(stderr, "Cannot open file %s.\n", cmd_file);
#endif
      return -1;
    } else {
      printf("File present OK\n");
    }
    fileclose(fHandle);
  }
  if(cmd_if) {
    snprintf(device, DEVICE_NAME_SIZE-1, "%s", cmd_if);
  }

  if (cmd_raw) {
    if (cmd_loops || cmd_rate || cmd_pcap || cmd_channels || cmd_gsync ||
      cmd_mod_fac || cmd_start_time || cmd_delay) {
      fprintf(stderr, ">>> Error: -org option cannot be used together with -c, -d, -g, -r, -l, -m, -pcap, or -s options\n");
      return 0;
    }
    fprintf(stdout, ">>> org mode enabled!!! Adapter: %s.\n", cmd_if);
    TxCaptureRaw(device, cmd_file);
    return 0;
  }

  /* Global variable channelsBitmask, used by CTRL-C handler and TxPacket */
  channelsBitmask = cmd_channels?strtol(cmd_channels, NULL, 0):0;
  if (cmd_channels && (channelsBitmask == 0)) {
    fprintf(stderr, ">>> Error: Invalid channelbitmask.\n");
    return -1;
  }
  loopCount       = cmd_loops?strtol(cmd_loops, NULL, 0):1;
  rate            = cmd_rate?strtol(cmd_rate, NULL, 0)*1000:-1;

  if(cmd_pcap) {
#ifndef WITH_PCAP
    printf("PCAP support was not included at compile time. (Argument \"with-pcap\" was not specified during installation).\n");
    printf("Please re-install the Applications package and specify \"with-pcap\" as argument to install script to enable PCAP support.\n");
    return -1;
#else
    mode = PCAP_CAPTURE;
    /* Check for illegal command-line options */
    if (cmd_channels || cmd_delay || cmd_gsync || cmd_mod_fac ||
        cmd_raw || cmd_rate || cmd_start_time) {
      fprintf(stderr,
              ">>> Error: -pcap option cannot be used with -c, -d, -g, -m,\n"
              "           -org, -r, or -s options\n");
      return 0;
    }
#endif
  }

  if(*device==0) {
    if((result=GetDevice(device, mode, FALSE)) < 0) {
      if(result==-2) {
        exit(0);
      }
      fprintf(stderr, "GetDevice(..) returned error");
      exit(0);
    }
  }
  if(mode==NTCI_CAPTURE) {
    if((!strncmp(device, "ntxc", 4)) || (!strncmp(device, "xyxs", 4))) {
      TxCapture(device, cmd_file, channelsBitmask, loopCount, rate, mode);
    } else {
      fprintf(stderr, "Please only select Napatech devices.\n");
    }
  } else {
    TxCapture(device, cmd_file, channelsBitmask, loopCount, rate, mode);
  }

  applicationRunning=FALSE;

  return 0;
}

static uint64_t GetFirstTimeStamp(const char *fileName)
{
  /* Retrieve the first absolute time stamp of the capture file.
     Note that the function cannot always detect whether the supplied file is
     in fact a capture file, and if not, the function returns a (silly) time
     stamp that does not make sense.
     The function returns zero in case of errors. */

  NT_HANDLE fHandle;
  const uint64_t FILE_ERROR = 0;
  uint64_t cookie = 0;
  uint64_t firstTS = 0;
  uint32_t pcap_magic_number = 0;
  BufferInformationSectionType1_t  secBufInfo;
  uint32_t offset = 0;
  ssize_t read_rv = 0;
  uint32_t pkt_desc[4];
  PacketDescriptorNtSeriesType1_t *pPktDesc;
  int swapped = 0;

  fHandle = fileopen(fileName, FILE_FLAGS, FILE_RIGHTS);
  if (fHandle == NT_OPEN_ERROR)
    return FILE_ERROR;

  if (fileread(fHandle, &cookie, sizeof(cookie)) != sizeof(cookie)) {
    fprintf(stderr, "File %s has an invalid file size.\n", fileName);
    (void) fileclose(fHandle);
    return FILE_ERROR;
  }

  if (cookie == 0xFEDEABBA) {
    cookie = 0;
    if (fileread(fHandle, &cookie, sizeof(cookie)) != sizeof(cookie)) {
      fprintf(stderr, "File %s has an invalid size.\n", fileName);
      (void) fileclose(fHandle);
      return FILE_ERROR;
    }
    if (cookie == 0xFEDEABBA) {
      /* This is a segment dump, not a capture. */
      if (fileread(fHandle, &secBufInfo,
		   sizeof(BufferInformationSectionType1_t)) !=
	  sizeof(BufferInformationSectionType1_t)) {
	fprintf(stderr, "File %s has an invalid size.\n", fileName);
        (void)fileclose(fHandle);
	return FILE_ERROR;
      }
      offset = sizeof(uint64_t) * 2 + sizeof(BufferInformationSectionType1_t);
    }
  } else {
    /* Check for PCAP global header. */
    filelseek(fHandle, 0, SEEK_SET);
    if (fileread(fHandle, &pcap_magic_number, sizeof(pcap_magic_number)) !=
	sizeof(pcap_magic_number)) {
      fprintf(stderr, "File %s has invalid size.\n", fileName);
      (void)fileclose(fHandle);
      return FILE_ERROR;
    }

    if((pcap_magic_number==PCAP_SWAPPED_MAGIC) || (pcap_magic_number==PCAP_SWAPPED_MODIFIED_MAGIC) || (pcap_magic_number==PCAP_SWAPPED_NSEC_MAGIC)) {
      /* We have a pcap global header. */
      offset = 24;
      swapped=1;
    } else if((pcap_magic_number==PCAP_MAGIC) || (pcap_magic_number==PCAP_MODIFIED_MAGIC) || (pcap_magic_number==PCAP_NSEC_MAGIC)) {
      /* We have a pcap global header. */
      offset = 24;
    }
  }

  (void) filelseek(fHandle, (off_t) offset, SEEK_SET);
  read_rv = fileread(fHandle, &pkt_desc,
		     sizeof(PacketDescriptorNtSeriesType1_t));
  if (read_rv != sizeof(PacketDescriptorNtSeriesType1_t)) {
    fprintf(stderr, "Cannot read first packet descriptor from %s.\n", fileName);
    (void)fileclose(fHandle);
    return FILE_ERROR;
  }
  (void) fileclose(fHandle);
  if(swapped) {
    pkt_desc[0]=ntohl(pkt_desc[0]);
    pkt_desc[1]=ntohl(pkt_desc[1]);
    pkt_desc[2]=ntohl(pkt_desc[2]);
    pkt_desc[3]=ntohl(pkt_desc[3]);
  }
  pPktDesc = (PacketDescriptorNtSeriesType1_t *)pkt_desc;
  if (pPktDesc->descriptorType == PCAP_DESCRIPTOR) {
    struct nttimeval pcap_time;
    memcpy(&pcap_time, &pkt_desc, sizeof(struct nttimeval));
    /* Convert the timestamp */
    if (pcap_magic_number == PCAP_NSEC_MAGIC ||
        pcap_magic_number == PCAP_SWAPPED_NSEC_MAGIC) {
      /* PCAP nano */
      firstTS = ((uint64_t) pcap_time.tv_sec * 1000000000) + pcap_time.tv_usec;
      firstTS /= 10;  // 10ns tick
    } else {
      /* 10ns ticks */
      firstTS = ((uint64_t)pcap_time.tv_sec * 1000000 + pcap_time.tv_usec) * 100;
    }
  } else {
    firstTS = pPktDesc->timestamp;
  }

  return firstTS;
}

static uint64_t
CalcOffset(uint64_t ts_start, uint32_t delay, uint64_t ts1, uint64_t ts2)
{
  /* CalcOffset calculates the global offset to set on an adapter in global
     sync mode.  The clocks of the adapters used must be synchronized.
     All time stamps supplieds as parameters must be in 10-ns ticks.
     The offset is calculated such that transmission starts on ts_start + delay.
  */

  uint64_t offset = 0U;  /* offset value */
  uint64_t ts_min = 0U;  /* minimum time stamp of ts1 and ts2 */
  uint64_t ts_f0 = 0U;   /* time stamp of first packet */

  ts_min = ts1 <= ts2 ? ts1 : ts2;
  ts_f0 = ts_start + delay;
  offset = ts_f0 - ts_min;

  return offset;
}

static uint64_t CalcStartTS(uint64_t ts, uint32_t mod_fac)
{
  uint64_t tmp_ts;
  const uint64_t ticks_per_sec = 100000000U; /* 10-nanosecs per second */

  /* Round adapter time up to nearest second */
  tmp_ts = (ts + ticks_per_sec - 1) / ticks_per_sec;
  /* Add one additional second to ensure ample time for set-up */
  tmp_ts += 1U;
  /* Find nearest time that satisfies tmp_ts % mod_fac == 0 */
  tmp_ts = ((tmp_ts + (mod_fac - 1)) / mod_fac) * mod_fac;

  return tmp_ts * ticks_per_sec; /* convert secs -> 10-ns ticks */
}

static void PrintTimeStamp(uint64_t ts)
{
  time_t tmp_ts;
  struct tm *loc_time;
  char ts_buf[256];

  tmp_ts = (time_t) (ts / 100000000);
  loc_time = localtime(&tmp_ts);
  sprintf(ts_buf, "%04d/%02d/%02d-%02d:%02d:%02d.%09lu",
	  loc_time->tm_year+1900, loc_time->tm_mon+1, loc_time->tm_mday,
	  loc_time->tm_hour, loc_time->tm_min, loc_time->tm_sec,
	  (unsigned long) (ts % 100000000 * 10));
  printf("%s", ts_buf);
}

static int GetAdapterTS(handle_t h_card, uint64_t *ts)
{
  TimestampConfigurationType1_t conf;
  uint32_t result;

  if (!ts)
    return NTCI_ERRCODE_NULL_POINTER;

  if (h_card == 0)
    return NTCI_ERRCODE_INTERNAL_ERROR;

  result = NTCI_GetTimestampConfiguration(h_card,
					  0,
					  STID_TIMESTAMP_CONFIGURATION_TYPE1,
					  &conf);

  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Unable to read time stamp configuration: %i\n", result);
    return result;
  }

  if (conf.timestampType != TIMESTAMP_NATIVE_UNIX) {
    fprintf(stderr, "Time stamp on the adapter must be in NATIVE_UNIX format"
	    " for global sync mode.\n");
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  *ts = conf.timestampValue;

  return NTCI_ERRCODE_SUCCESS;
}

static int ParseStartTime(const char *dts, uint64_t *ts)
{
  char *rem_d;
  struct tm tm;
  time_t t;
  struct tm *loc_time;
  time_t now;

  /* Reset tm structure because strptime is not required to initialize
     all members variables. */
  memset(&tm, 0, sizeof(tm));
  /* Did user specify date+time in ISO format? */
  rem_d = strptime(dts, "%Y/%m/%d-%H:%M:%S", &tm);
  if (!rem_d) {
    /* Did user specify date and time in locale's representation? */
    rem_d = strptime(dts, "%x %X", &tm);
    if (!rem_d) {
      memset(&tm, 0, sizeof(tm));
      /* Did user specify (just) time in locale's representation? */
      rem_d = strptime(dts, "%X", &tm);
      if (!rem_d) {
        memset(&tm, 0, sizeof(tm));
        /* Did user specify only hour and minute? */
        rem_d = strptime(dts, "%R", &tm);

        if (!rem_d)
          /* Failed to parse date/time */
          return -1;
      }

      /* User has only specified time, get yy/mm/dd from host, not adapter;
         assume date on host and adapter are the same. */
      now = time(NULL);
      loc_time = localtime(&now);
      tm.tm_year = loc_time->tm_year;
      tm.tm_mon = loc_time->tm_mon;
      tm.tm_mday = loc_time->tm_mday;
    }
  }


  tm.tm_isdst = -1; /* Let mktime decide about DST */
  errno = 0;
  t = mktime(&tm);

  if (t == (time_t) -1)
    return -1;

  *ts = t;
  return 0;
}

static int
CheckLinkStatusAndSpeed(
  const CardHardwareCapabilitiesType2_t *cardCapabilities,
  int32_t speed, uint32_t channelsBitmask)
{
  /* This function checks whether links are up/down and whether link speed
     is high enough to fulfill the user's request.
     The function returns zero on success, and a negative value in case
     of errors.
  */
  uint8_t numChannels = 0U;
  ChannelStatusType1_t chanStatus;
  const uint32_t totalChannelsBitmask =
    cardCapabilities->type1.totalChannelsBitmask;

  for (numChannels = 1; ; numChannels++) {

    if (!((totalChannelsBitmask >> (numChannels - 1)) & 0x1))
      break;

    if (channelsBitmask == 0 || (channelsBitmask & (1 << (numChannels-1))) ||
        gsyncMode) {
      if (NTCI_GetChannelStatus(cardHandle,
                                1 << (numChannels - 1),
                                STID_CHANNEL_STATUS_TYPE1,
                                &chanStatus) != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, "Failed to get channel %d status.\n", numChannels - 1);
        return -1;
      }

      if (chanStatus.linkUp) {
        /* Get current link speed for the current channel */
        switch(chanStatus.actualSpeed) {
        case LINK_SPEED_10M:  ifSpeed[numChannels-1] =    10; break;
        case LINK_SPEED_100M: ifSpeed[numChannels-1] =   100; break;
        case LINK_SPEED_1G:   ifSpeed[numChannels-1] =  1000; break;
        case LINK_SPEED_10G:  ifSpeed[numChannels-1] = 10000; break;
        case LINK_SPEED_40G:  ifSpeed[numChannels-1] = 40000; break;
        default:
          fprintf(stderr, "Unsupported speed detected on link %d.\n",
                  numChannels - 1);
          return -2;
        }
      } else {
        /* Set link speed to the maximum value of the adapter */
        if (chanStatus.availableSpeedsBitmask) {
          if (chanStatus.availableSpeedsBitmask & SPEED_100GB) {
            ifSpeed[numChannels - 1] = 100000;
          } else if(chanStatus.availableSpeedsBitmask & SPEED_40GB) {
            ifSpeed[numChannels - 1] = 40000;
          } else if(chanStatus.availableSpeedsBitmask & SPEED_10GB) {
            ifSpeed[numChannels - 1] = 10000;
          } else if(chanStatus.availableSpeedsBitmask & SPEED_1GB) {
            ifSpeed[numChannels - 1] = 1000;
          } else if(chanStatus.availableSpeedsBitmask & SPEED_100MB) {
            ifSpeed[numChannels - 1] = 100;
          } else if(chanStatus.availableSpeedsBitmask & SPEED_10MB) {
            ifSpeed[numChannels - 1] = 10;
          }
        } else {
          fprintf(stderr,
                  ">>> Error: Link %d is down and the application is unable to"
                  " detect available speeds.\n", numChannels - 1);
          return -3;
        }
      }

      /* Check channel state/speed if the user has not specified any bitmask
         or the user has specifically asked for this channel. */
      if (channelsBitmask == 0 ||
          ((channelsBitmask >> (numChannels - 1)) & 0x1)) {
        if (! chanStatus.linkUp) {
          if (channelsBitmask == 0) {
            fprintf(stdout,
                    ">>> NOTE: Link %d is down. The application continues "
                    "anyway and\n"
                    ">>>       assumes a link speed of %dMbps.\n",
                    numChannels - 1, ifSpeed[numChannels - 1]);
            fprintf(stdout,
                    ">>>       Packets are not sent until link "
                    "is established.\n");
          } else {
            fprintf(stderr,
                    ">>> Error: Link %d is down and cannot be used "
                    "for replay.\n", numChannels - 1);
            return -4;
          }
        }

	if (speed > 0) {
	  /* Check link is at least as fast as requested by the user */
	  if (ifSpeed[numChannels - 1] < (uint) speed / 1000) {
	    fprintf(stderr, "Invalid speed. Channel %d is running %d so the "
		    "requested %dMbps is not possible.\n",
		    numChannels - 1, ifSpeed[numChannels - 1], speed / 1000U);
	    return -5;
	  }
	}
      }
    }
  }
  return 0;
}
