// $Id: device.c 25292 2013-09-05 13:04:54Z hhd $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * GetDevice checks that the provided device and its usage mode are allowed.
 */

#include <ostype.h>

#ifdef WIN32
#include <WINSOCK2.H>
#include <fcntl.h>
#include <io.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#endif
// capabilities
#include <NTCommonInterface.h>
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>
// vpd
#include <NTCI_vpdaccess.h>
#include <vpdstructure.h>
//
#ifdef WITH_PCAP
#include <pcap.h>
#endif
#include "device.h"

/* External variable. */
extern bool_t applicationRunning;

/* This function will check if the device is
   OK and what type it is (PCAP or CAPTURE). */

static int
_getchoice(char max)
{
  int choice;

  while(applicationRunning) {
    /* input char sanity loop */
#ifndef WIN32
    if(system("stty cbreak")!=0) {
      // Avoid compile warning
    }    
    if(system("stty -echo")!=0) {
      // Avoid compile warning      
    }
#endif
    choice = getchar();
#ifndef WIN32
    if(system("stty echo")!=0) {
      // Avoid compile warning
    }
    if(system("stty -cbreak")!=0) {
      // Avoid compile warning
    }
#endif

    if (choice != '\n' && iscntrl(choice))
      return -1;

    printf("%c\n", choice);

    if((choice == 'q') || (choice == 'Q')) {
      return -1;
    }
    
    choice -= '0';

    if (choice < 0 || choice > 9) {
      printf("invalid conversion. Please enter only numbers\nChoice: ");
      continue;
    }

    if (choice < 0 || choice > max) {
      printf("Out of range. Please enter a choice between 0 and %d\nChoice: ", max);
      continue;
    }
    return choice;
  } //while applicationRunning
  return -1;
}

#ifdef WITH_PCAP
int GetDevice(char *device, int mode, int showOnly)
#else
int GetDevice(char *device, NOUSE int mode, int showOnly)
#endif
{
  int           devices=0;
  int           choice=0;
  int           i=0;
#ifdef WITH_PCAP
  char          ebuf[PCAP_ERRBUF_SIZE];
#endif
  handle_t      cardHandle=NULL;
#ifdef WITH_PCAP
	pcap_if_t     *devs=NULL;
	pcap_if_t     *cur=NULL;
#endif  
  int           devMap[17];
  char          menu[20][80]; // 20 lines each 80 chars
  int           line=0;
  char* vpdString = NULL; // RMT#2646

  memset(menu, 0x00, sizeof(menu));
  /* Only prompt for interface if none is specified. */
  if(*device == 0x00) {
    if(showOnly==FALSE) {
      printf("Interface not specified.\nInterface list:\n");
    }
#ifdef WITH_PCAP
    if(mode == PCAP_CAPTURE) {
      /* If pcap mode is used ask pcap for the list: */
      if(pcap_findalldevs(&devs, ebuf) <0) {
        fprintf(stderr, ">>> Error: %s\n", ebuf);
      } else {
        /* Present the list. */
        sprintf(menu[line++], "PCAP devices:\n");
        sprintf(menu[line++], "---------------------------------------\n");
        i=0;
        cur = devs;
        if (!cur) {
          printf("No PCAP devices found\n");
          return -1;
        }
        for(;;) {          
          if(cur->description) {
#ifdef WIN32
            if (cur->name[0] == 'n' && cur->name[1] == 't' &&
                cur->name[2] == 'x' && cur->name[3] == 'c') {
              sprintf(menu[line++], "%d: %s (%s)\n", i, cur->description, cur->name);
            }
            else {
              sprintf(menu[line++], "%d: %s\n", i, cur->description);
            }
#else
            sprintf(menu[line++], "%d: %s (%s)\n", i, cur->name, cur->description);
#endif // END: WIN32
          } else {
            sprintf(menu[line++], "%d: %s\n", i, cur->name);
          }
          cur = cur->next;
          if(cur) {
            i++;
            devices++;
          } else {
            cur = devs;
            break;
          }
        }
        sprintf(menu[line++], "---------------------------------------\n");
        if(showOnly==FALSE) {
          sprintf(menu[line++], "Q: Quit application\n");
          
          if(devices>1) {
            for(i=0; i<line; i++) {
              printf("%s", menu[i]);
            }
            /* Get user input */
            printf("Please enter PCAP interface # to use: ");
            
            
            if((choice = _getchoice(devices)) < 0)
              return -2;
            
            for(cur = devs; choice--; cur = cur->next);
          }
          
          strncpy(device, cur->name, DEVICE_NAME_SIZE);
          
          printf("Chosen interface: %s\n", device);
        }
        return 0;
      }
    } else 
#endif // END: WITH_PCAP
    {
      /* NTCI capture. List devices. */
      sprintf(menu[line++], "NTCI devices:\n");
      sprintf(menu[line++], "---------------------------------------\n");
      devices=0;
      for(i=0;i<16;i++) {
        if((cardHandle = NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, i)) != NULL) {
          /* We only show devices with packetfeed support. */
          CardServicesCapabilitiesType1_t services;
          if(NTCI_GetCapabilities(cardHandle, CAPID_SERVICES, NOT_SUB_CAPABILITIES,
                                  STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                                  &services) != NTCI_ERRCODE_SUCCESS) {
            NTCI_CloseCard(cardHandle);
            continue;
          }
          if(!services.packetFeedSupported) {
            /* Skip card it has no packetfeed support. */
            NTCI_CloseCard(cardHandle);
            continue;
          }
          //
          // RMT#2646: Add adapter cardname/-description to adapter selection list (RMT#2646)
          //
          devMap[devices] = i;
          vpdString = GetVPDStringAdapterName(cardHandle);
          if ( vpdString ) {
            {
              CardHardwareVersionCapabilitiesType3_t data3;
              uint64_t fpgaId=0;

              if((NTCI_GetCapabilities(cardHandle,
                                       CAPID_VERSION,
                                       NOT_SUB_CAPABILITIES,
                                       STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                       &data3)) != NTCI_ERRCODE_SUCCESS) {
                NTCI_CloseCard(cardHandle);
                continue;
              }
              if(data3.AdapterCard.dwInfoType==1) {
                fpgaId = (data3.AdapterCard.u.InfoType1.qwFpgaId&0xFFFF0000)>>16;
              }else if(data3.AdapterCard.dwInfoType==2) {
                fpgaId = (data3.AdapterCard.u.InfoType2.qwFpgaId&0xFFFF0000)>>16;
              }else if(data3.AdapterCard.dwInfoType==3) {
                fpgaId = (data3.AdapterCard.u.InfoType3.qwFpgaId&0xFFFF0000)>>16;
              }
              if(fpgaId == 9220) {
                sprintf(menu[line++], "%d: NT20E2 %s\n", devices, vpdString);
              } else if(fpgaId == 9221) {
                sprintf(menu[line++], "%d: NT40E2_4 %s\n", devices, vpdString);
              } else if(fpgaId == 9222) {
                sprintf(menu[line++], "%d: NT40E2_1 %s\n", devices, vpdString);
              } else {
                sprintf(menu[line++], "%d: %s%d: %s\n", devices, "ntxc", i, vpdString);
              }
            }
            free(vpdString);
            vpdString = NULL;
          } else  {
            sprintf(menu[line++], "%d: %s%d: %s\n", devices, "ntxc", i, "NA: no adapter description found");
          }
          devices++;
        } else {
          break;
        }
        NTCI_CloseCard(cardHandle);
      }
      sprintf(menu[line++], "---------------------------------------\n");
      if(showOnly==FALSE) {
        sprintf(menu[line++], "Q: Quit application\n");
        if(devices) {
          if(devices > 1) {
            for(i=0; i<line; i++) {
              printf("%s", menu[i]);
            }
            printf("Please enter NTCI interface # to use: ");
            if ((choice = _getchoice(devices-1)) < 0)
              return -2; 
          } else {
            choice = 0;
          }
          snprintf(device, DEVICE_NAME_SIZE, "ntxc%d", devMap[choice]);          
          printf("Chosen interface: %s\n", device);
          return choice;
        }
      } else {
        for(i=0; i<line; i++) {
          printf("%s", menu[i]);
        }
        return 0;
      }
      return -1;
    }
  }
  return -1;
}

//
//
//
int ParseDeviceName(char* deviceName)
{
  int adapterNumber;
  char* strEndPtr;

  if (strncmp(deviceName, "ntxc", 4) == 0) {
    // Make sure that our adapter number is beginning with a digit since strtol() also eats +/-prefixed numbers - and we do not...
    if (deviceName[4]>='0' && deviceName[4]<='9') {
      /* Reset errno to discard any previous errors before calling strtol;
         this is necessary because strtol may return zero as a valid value
         but also to signal a parsing error, thus setting errno to zero enables
         the code to distinguish between a legal zero return value and an
         error condition (errno == EINVAL || errno == ERANGE).
         The C standard does not permit any library function to set errno
         to zero. */
      errno = 0;
      adapterNumber = strtol(&deviceName[4], &strEndPtr, 10);
      if ( strEndPtr != NULL && *strEndPtr != '\0') adapterNumber = -1; // we did not parse the whole string - that's an error...
      if ( adapterNumber == 0 && (errno == EINVAL || errno == ERANGE) ) adapterNumber = -1; // invalid number or out of range
    } else {
      adapterNumber = -1; // error
    }
  } else {
    adapterNumber = -1; // error
  }

  // check the adapter number if negative - issue parse error
  if (adapterNumber <= -1) {
    fprintf(stderr, "Failed to parse adapter name: '%s'\n", deviceName);
  }
  return adapterNumber;
}

//
//
//
char* GetVPDStringAdapterName(handle_t cardHandle)
{
  //  int vpdVersion;
  int vpdResult;
  //  int vpdStringMaxLen;
  uint8_t* vpdString = NULL;
  char* dupPtr = NULL;
  VPDImageType4_t vpdData4;
  VPDImageType5_t vpdData5;

  //
  // RMT#2646: Add adapter cardname/-description to adapter selection list (RMT#2646)
  //
  vpdResult = NTCI_ReadVitalProductData(cardHandle, 0, STID_VPD_IMAGE_TYPE5, &vpdData5);
  if ( vpdResult == NTCI_ERRCODE_SUCCESS) {
    //    vpdVersion = 5;
    vpdString = vpdData5.mainBoardType;
    //    vpdStringMaxLen = sizeof(vpdData5.mainBoardType);
    //
  } else {
    vpdResult = NTCI_ReadVitalProductData(cardHandle, 0, STID_VPD_IMAGE_TYPE4, &vpdData4);
    if (vpdResult == NTCI_ERRCODE_SUCCESS) {
      //      vpdVersion = 4;
      vpdString = vpdData4.productName;
      //      vpdStringMaxLen = sizeof(vpdData4.productName);
      //
    } else {
      //      vpdVersion = -1; // Error
      vpdString = NULL;
      //      vpdStringMaxLen = -1;
      //
    } // vpdData4
  } // vpdData5
  //
  if ( vpdString ) {
    dupPtr = strdup((char*)vpdString);
  } else  {
    dupPtr = NULL;
  }
  return dupPtr;
}
//
//
//
