// $Id: timestamp.c 25292 2013-09-05 13:04:54Z hhd $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * Timestamp convert functions.
 */

/* The NTCommonInterface.h header provides the interface definitions
 * for the Napatech NTCI library */
#include <NTCommonInterface.h>
#include "timestamp.h"

#ifdef WIN32
#include <Windows.h>
#include <stdio.h>
#else
#include <stdio.h>
#include <sys/time.h>
#endif

#define DIV64(_dividend, _divisor, _quotient_type, _quotient, _remainder) do { \
	uint64_t __quot    = (uint64_t) _dividend; \
	uint32_t __mod     = _dividend % _divisor; \
        __quot /= _divisor; \
	_quotient          = (_quotient_type) __quot; \
	_remainder         = (_quotient_type) __mod; \
} while (0);


#define CONST64(__val) __val##ULL

/* Offset of Unix epoch (1 Jan 1970) from Windows epoch (1 Jan 1601) in
 * 100ns (Windows system tick time) units */
#define UNIX_EPOCH_OFFSET_MAGIC CONST64(116444736000000000)

struct nttimeval
{
  uint32_t tv_sec;
  uint32_t tv_usec;
};

static void ConvertNdisToPcap( uint64_t* ptime )
{
	uint64_t tmp = *ptime;
	NOUSE uint64_t tmp2;
	struct nttimeval *punix_time = (struct nttimeval *)ptime;

	/* Convert from 100ns ticks from 01/01/1601 to ticks from 01/01/1970 */
	tmp -= UNIX_EPOCH_OFFSET_MAGIC;

	/* Convert u64 100ns ticks to u64 microseconds */
	DIV64(tmp, CONST64(10), uint64_t, tmp, tmp2);

	/* Convert from u64 microseconds to u32 seconds and u32 microseconds
	 * (places directly back into time pointer) */
	DIV64(tmp, CONST64(1000000), uint32_t, punix_time->tv_sec, punix_time->tv_usec);
}


static void ConvertNdisToNanoT( uint64_t* ptime )
{
	uint64_t tmp = *ptime;
	struct nttimeval *punix_time = (struct nttimeval *)ptime;

	/* Convert from 100ns ticks from 01/01/1601 to ticks from 01/01/1970 */
	tmp -= UNIX_EPOCH_OFFSET_MAGIC;

	/* Convert u64 100ns ticks to u64 nanoseconds */
	tmp *= CONST64(100);

	/* Convert from u64 microseconds to u32 seconds and u32 nanoseconds
	 * (places directly back into time pointer) */
	DIV64(tmp, CONST64(1000000000), uint32_t, punix_time->tv_sec, punix_time->tv_usec);
}


static void ConvertPcapToNdis( uint64_t* ptime )
{
	uint64_t tmp;
	struct nttimeval *punix_time = (struct nttimeval *)ptime;

	/* Convert from u32 seconds and u32 microseconds to u64 microseconds */
	tmp = (punix_time->tv_sec * CONST64(1000000)) + punix_time->tv_usec;

	/* Convert u64 microseconds to u64 100ns ticks */
	tmp *= CONST64(10);

	/* Convert from 100ns ticks from 01/01/1970 to ticks 01/01/1601 */
	tmp += UNIX_EPOCH_OFFSET_MAGIC;

	*ptime = tmp;
}


void ConvertSystemTickTime( uint64_t* ptime,
                            TimestampType_t fromTimestampType,
                            TimestampType_t toTimestampType )
{
	bool_t bStatus = TRUE;

	if ( fromTimestampType == toTimestampType ) {
		/* No conversion is required */
		return;
	}
	else if ( (fromTimestampType == TIMESTAMP_NATIVE) ||
            (toTimestampType   == TIMESTAMP_NATIVE) ) {
		/* Native timestamp conversions are not allowed */
		bStatus = FALSE;
	}
	else {
		NOUSE uint64_t mod;

		switch ( fromTimestampType ) {
		case TIMESTAMP_NDIS:
			switch ( toTimestampType ) {
			case TIMESTAMP_PCAP:
				/* Convert from NDIS to PCAP */
				ConvertNdisToPcap(ptime);
				break;
			case TIMESTAMP_NATIVE_NDIS:
				/* Convert from NDIS 100ns res to native 10ns res */
				*ptime = *ptime * CONST64(10);
				break;
			case TIMESTAMP_NATIVE_UNIX:
				/* Convert NDIS to Native Unix */
				*ptime = (*ptime - UNIX_EPOCH_OFFSET_MAGIC) * CONST64(10);
				break;
			case TIMESTAMP_PCAP_NANOTIME:
				/* Convert from NDIS to PCAP NANOTIME */
				ConvertNdisToNanoT(ptime);
				break;
			default:
				bStatus = FALSE;
			}
			break;

		case TIMESTAMP_PCAP:
			switch ( toTimestampType ) {
			case TIMESTAMP_NDIS:
				/* Convert PCAP to NDIS */
				ConvertPcapToNdis(ptime);
				break;
			case TIMESTAMP_NATIVE_NDIS:
				/* Convert PCAP to NDIS */
				ConvertPcapToNdis(ptime);
				/* Convert NDIS to NATIVE_NDIS */
				*ptime = *ptime * CONST64(10);
				break;
			case TIMESTAMP_NATIVE_UNIX:
				/* Convert PCAP to Native Unix */
				ConvertPcapToNdis(ptime);
				*ptime = (*ptime - UNIX_EPOCH_OFFSET_MAGIC) * CONST64(10);
				break;
			case TIMESTAMP_PCAP_NANOTIME:
      {
        /* Convert from PCAP to PCAP NANOTIME */
        struct nttimeval *punix_time = (struct nttimeval *)ptime;
        punix_time->tv_usec *= 1000;
      }
      break;
			default:
				bStatus = FALSE;
			}
			break;
      
		case TIMESTAMP_NATIVE_NDIS:
			switch ( toTimestampType ) {
			case TIMESTAMP_NDIS:
				/* Convert from NATIVE_NDIS to NDIS */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				break;
			case TIMESTAMP_PCAP:
				/* Convert from NATIVE_NDIS to PCAP */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				/* Convert NDIS to PCAP */
				ConvertNdisToPcap(ptime);
				break;
			case TIMESTAMP_NATIVE_UNIX:
				/* Convert NATIVE_NDIS to Native Unix */
				*ptime = *ptime - (UNIX_EPOCH_OFFSET_MAGIC * CONST64(10));
				break;
			case TIMESTAMP_PCAP_NANOTIME:
				/* Convert from NATIVE_NDIS to NDIS */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				/* Convert from NDIS to PCAP NANOTIME */
				ConvertNdisToNanoT(ptime);
				break;
			default:
				bStatus = FALSE;
			}
			break;

		case TIMESTAMP_NATIVE_UNIX:
			switch ( toTimestampType ) {
			case TIMESTAMP_NATIVE_NDIS:
				/* Convert from NATIVE_UNIX to Native NDIS */
				*ptime = *ptime + (UNIX_EPOCH_OFFSET_MAGIC * CONST64(10));
				break;
			case TIMESTAMP_NDIS:
				/* Convert from NATIVE_UNIX to NDIS */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				*ptime = *ptime + UNIX_EPOCH_OFFSET_MAGIC;
				break;
			case TIMESTAMP_PCAP:
				/* Convert from NATIVE_UNIX to NDIS */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				*ptime = *ptime + UNIX_EPOCH_OFFSET_MAGIC;
				/* Convert NDIS to PCAP */
				ConvertNdisToPcap(ptime);
				break;
			case TIMESTAMP_PCAP_NANOTIME:
				/* Convert from NATIVE_UNIX to NDIS */
				DIV64(*ptime, 10, uint64_t, *ptime, mod);
				*ptime = *ptime + UNIX_EPOCH_OFFSET_MAGIC;
				/* Convert from NDIS to PCAP NANOTIME */
				ConvertNdisToNanoT(ptime);
				break;
			default:
				bStatus = FALSE;
			}
			break;

		case TIMESTAMP_PCAP_NANOTIME:
			switch ( toTimestampType ) {
			case TIMESTAMP_NDIS:
      {
        /* Convert from PCAP NANOTIME to PCAP */
        struct nttimeval *punix_time = (struct nttimeval *)ptime;
        punix_time->tv_usec /= 1000;
        /* Convert PCAP to NDIS */
        ConvertPcapToNdis(ptime);
      }
      break;
			case TIMESTAMP_PCAP:
      {
        /* Convert from PCAP NANOTIME to PCAP */
        struct nttimeval *punix_time = (struct nttimeval *)ptime;
        punix_time->tv_usec /= 1000;
      }
      break;
			case TIMESTAMP_NATIVE_NDIS:
      {
        /* Convert from PCAP NANOTIME to PCAP */
        struct nttimeval *punix_time = (struct nttimeval *)ptime;
        punix_time->tv_usec /= 1000;
        /* Convert PCAP to NDIS */
        ConvertPcapToNdis(ptime);
        /* Convert NDIS to NDIS_NATIVE */
        *ptime = *ptime * CONST64(10);
      }
      break;
			case TIMESTAMP_NATIVE_UNIX:
      {
        /* Convert from PCAP NANOTIME to PCAP */
        struct nttimeval *punix_time = (struct nttimeval *)ptime;
        punix_time->tv_usec /= 1000;
        /* Convert PCAP to NDIS */
        ConvertPcapToNdis(ptime);
        /* Convert NDIS to NDIS_NATIVE */
        *ptime = (*ptime - UNIX_EPOCH_OFFSET_MAGIC) * CONST64(10);
      }
      break;
			default:
				bStatus = FALSE;
			}
			break;

		default:
			bStatus = FALSE;
			break;
		}
	}

	if ( ! bStatus ) {
		fprintf(stderr, "Unsupported time stamp conversion: %u to %u\n",
            fromTimestampType, toTimestampType );
		*ptime = CONST64(0);
	}

	return;
}
