// $Id: CardStatus.c 15634 2011-11-15 15:47:21Z ke $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * CardStatus.c : Contain functions to get card status.
 */

#ifdef WIN32
#include <Windows.h>
#include <stdio.h>
#include <string.h>
#else
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#endif

/* The NTCommonInterface.h header provides the interface definitions
 * for the NTCI library */
#include <NTCommonInterface.h>

#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>
#include <NTCI_cardconfiguration.h>
#include "toolslib.h"

static char WriteBuffer[2048];
static uint32_t writePos=0;
#define DUMP_TO_SCREEN(...) writePos += WriteToBuffer(&WriteBuffer[writePos], sizeof(WriteBuffer)-writePos,  __VA_ARGS__);

typedef struct
{
    uint8_t TemperatureMSB;
    uint8_t TemperatureLSB;
    uint8_t VoltageMSB;
    uint8_t VoltageLSB;
    uint8_t TxBiasCurrentMSB;
    uint8_t TxBiasCurrentLSB;
    uint8_t TxPowerMSB;
    uint8_t TxPowerLSB;
    uint8_t RxPowerMSB;
    uint8_t RxPowerLSB;
} SfpMonitoringDataType1_t;

/* Return the description of a port type */
static const char *PortTypeDescription(enum PORT_TYPES portType)
{
  switch (portType) 
  {
  case PORT_TYPE_NOT_AVAILABLE: return "Not connected";
  case PORT_TYPE_NOT_RECOGNISED: return "Not recognised";
  case PORT_TYPE_RJ45: return "RJ-45";
  case PORT_TYPE_SFP_NOT_PRESENT: return "Slot empty";
  case PORT_TYPE_SFP_SX: return "SFP SX";
  case PORT_TYPE_SFP_SX_DD: return "SFP SX (with digital diagnostic)";
  case PORT_TYPE_SFP_LX: return "SFP LX";
  case PORT_TYPE_SFP_LX_DD: return "SFP LX (with digital diagnostic)";
  case PORT_TYPE_SFP_ZX: return "SFP ZX";
  case PORT_TYPE_SFP_ZX_DD: return "SFP ZX (with digital diagnostic)";
  case PORT_TYPE_SFP_CU: return "SFP Copper 1000";
  case PORT_TYPE_SFP_CU_DD: return "SFP Copper 1000 (with digital diagnostic)";
  case PORT_TYPE_SFP_NOT_RECOGNISED: return "SFP (not recognised)";
  case PORT_TYPE_XFP: return "XFP";
  case PORT_TYPE_XPAK: return "XPAK";
  case PORT_TYPE_SFP_CU_TRI_SPEED: return "SFP Copper 10/100/1000";
  case PORT_TYPE_SFP_CU_TRI_SPEED_DD: return "SFP Copper 10/100/1000 (with digital diagnostic)";
  case PORT_TYPE_SFP_PLUS: return "SFP+";   
  case PORT_TYPE_QSFP_PLUS: return "QSFP+";
  }
  return NULL;
}

/* Return a description of a link quality level */
static const char *LinkQualityDescription(enum LINK_QUALITY quality)
{
  switch (quality) 
  {
  case LINK_QUALITY_GOOD:  return "good";
  case LINK_QUALITY_POOR:  return "poor (link up, but bad quality)";
  case LINK_QUALITY_BAD:   return "bad (link down or too weak to register)";
  case LINK_QUALITY_ERROR: return "error (link down because an error has been detected)";    
  }
  return NULL;
}

static const char *YesOrNo(uint32_t val)
{
    if (val)
      return "Yes";
    else
      return "No ";
}

/* Converts a link speed bit into a description */
static const char *LinkSpeedDescription(uint32_t speed)
{
  switch (speed)
  {
  case LINK_SPEED_10M:  return "10 Mbit";
  case LINK_SPEED_100M: return "100 Mbit";
  case LINK_SPEED_1G:   return "1 Gbit";
  case LINK_SPEED_10G:  return "10 Gbit";
  case LINK_SPEED_40G:  return "40 Gbit";
  case LINK_SPEED_100G: return "100 Gbit";        
  }
  return NULL;
}

static void printChannelStatus3(ChannelStatusType3_t chanStatus3)
{
  DUMP_TO_SCREEN("Autonegotiation capabilities\n");  
  DUMP_TO_SCREEN("  10 MBit  :   Full duplex: %s     Half duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_10M_HALFDUPLEX ),
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_10M_FULLDUPLEX ));
  DUMP_TO_SCREEN("  100 MBit :   Full duplex: %s     Half duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_100M_FULLDUPLEX ),
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_100M_HALFDUPLEX));
  DUMP_TO_SCREEN("  1 GBit   :   Full duplex: %s     Half duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_1G_FULLDUPLEX),
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_1G_HALFDUPLEX));
  DUMP_TO_SCREEN("  10 GBit  :   Full duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_10G_FULLDUPLEX  ));
  DUMP_TO_SCREEN("  40 GBit  :   Full duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_40G_FULLDUPLEX  ));
  DUMP_TO_SCREEN("  100 GBit :   Full duplex: %s\n",
                 YesOrNo(chanStatus3.advertiseCapabilitiesdBitMask & ADVERTISE_100G_FULLDUPLEX  ));
  DUMP_TO_SCREEN("\n");
  DUMP_TO_SCREEN("Manual configuration capabilities\n");  
  DUMP_TO_SCREEN("  Full duplex : %s\n",
                 YesOrNo(chanStatus3.manuelConfDuplexBitMask & MANUAL_FULL_DUPLEX));
  DUMP_TO_SCREEN("  Half duplex : %s\n",
                 YesOrNo(chanStatus3.manuelConfDuplexBitMask & MANUAL_HALF_DUPLEX));
  DUMP_TO_SCREEN("  10 MBit     : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_10M));
  DUMP_TO_SCREEN("  100 MBit    : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_100M ));    
  DUMP_TO_SCREEN("  1 GBit      : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_1G ));
  DUMP_TO_SCREEN("  10 GBit     : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_10G));    
  DUMP_TO_SCREEN("  40 GBit     : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_40G));    
  DUMP_TO_SCREEN("  100 GBit    : %s\n",
                 YesOrNo(chanStatus3.manuelConfSpeedBitMask & LINK_SPEED_100G));    
  DUMP_TO_SCREEN("  MDIMode     : Auto: %s     MDI: %s    MDIX: %s\n",
                 YesOrNo(chanStatus3.manuelMdiModeBitMask & MANUAL_MDI_MODE_AUTO),
                 YesOrNo(chanStatus3.manuelMdiModeBitMask & MANUAL_MDI_MODE_MDI ),
                 YesOrNo(chanStatus3.manuelMdiModeBitMask & MANUAL_MDI_MODE_MDIX));
  DUMP_TO_SCREEN("\n");
  DUMP_TO_SCREEN("Available Speed Bitmask\n");
  DUMP_TO_SCREEN("  10 MBit     : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_10M));
  DUMP_TO_SCREEN("  100 MBit    : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_100M ));    
  DUMP_TO_SCREEN("  1 GBit      : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_1G ));
  DUMP_TO_SCREEN("  10 GBit     : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_10G));    
  DUMP_TO_SCREEN("  40 GBit     : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_40G));    
  DUMP_TO_SCREEN("  100 GBit    : %s\n",
                 YesOrNo(chanStatus3.availableSpeedsBitmask & LINK_SPEED_100G));    
  
  DUMP_TO_SCREEN("\n");
  if (!chanStatus3.configSuccess) {
    DUMP_TO_SCREEN(">>> ERROR: LINK CONFIGURATION DOES NOT MATCH PORT CAPABILITIES\n");
    DUMP_TO_SCREEN("\n");          
  }
}


uint32_t GetChannelStatus(uint8_t card, uint32_t channelMask)
{
  handle_t cardHandle;
  uint32_t result;
  uint8_t channel, ChannelStatusType=0;
  ChannelStatusType1_t chanStatus;
  ChannelStatusType2_t chanStatus2;
  ChannelStatusType3_t chanStatus3;
  ChannelStatusType4_t chanStatus4;
  
  writePos=0;
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }
  result = NTCI_GetChannelStatus(cardHandle, 
                                 channelMask,
                                 STID_CHANNEL_STATUS_TYPE4,
                                 &chanStatus4);
  
  if(result == NTCI_ERRCODE_SUCCESS) {
    memcpy(&chanStatus, &chanStatus4, sizeof(ChannelStatusType1_t));
    ChannelStatusType = 4;
  } else {
    result = NTCI_GetChannelStatus(cardHandle, 
                                   channelMask,
                                   STID_CHANNEL_STATUS_TYPE3,
                                   &chanStatus3);

    if(result == NTCI_ERRCODE_SUCCESS) {
      memcpy(&chanStatus, &chanStatus3, sizeof(ChannelStatusType1_t));    
      ChannelStatusType = 3;

    } else {
      result = NTCI_GetChannelStatus(cardHandle, 
                                     channelMask,
                                     STID_CHANNEL_STATUS_TYPE2,
                                     &chanStatus2);
      if(result == NTCI_ERRCODE_SUCCESS) {
        memcpy(&chanStatus, &chanStatus2, sizeof(ChannelStatusType1_t));  
        ChannelStatusType = 2;
      } else {
        result = NTCI_GetChannelStatus(cardHandle, 
                                       channelMask,
                                       STID_CHANNEL_STATUS_TYPE1,
                                       &chanStatus);
        if(result == NTCI_ERRCODE_SUCCESS) {
          ChannelStatusType = 1;
        } else {
          fprintf(stderr, ">>> ERROR: Failed getting channel status for channel bitmask 0x%02X. Errorcode: 0x%02X\n", channelMask, result);
          return result;
        }
      }
    }
  }
  
  for(channel=0; channel<32; channel++) {
    if((1<<channel) & channelMask) {
      break;
    }
  }
  
  DUMP_TO_SCREEN("AdapterStatus    : Channel %d\n", channel);
  DUMP_TO_SCREEN("  Link status    : ");

  if (chanStatus.linkUp) {
    DUMP_TO_SCREEN("connected\n");
  } else {
    DUMP_TO_SCREEN("not connected\n");
  }
  
  /* If the link is up, print some information about it */
  DUMP_TO_SCREEN("  Port type      : %s\n", PortTypeDescription(chanStatus.portType));   
  DUMP_TO_SCREEN("  Autonegotiation: %s\n", chanStatus.autoNegotiationEnabled==TRUE?"enabled":"disabled");
  DUMP_TO_SCREEN("  Link quality   : %s\n",  LinkQualityDescription(chanStatus.linkQuality));
  DUMP_TO_SCREEN("  Link Speed     : %s\n", LinkSpeedDescription(chanStatus.actualSpeed));
  DUMP_TO_SCREEN("  Duplex         : %s\n", chanStatus.fullDuplex==TRUE?"Full":"Half");  
  if (ChannelStatusType == 4) {
    DUMP_TO_SCREEN("  Linkdown events: %d\n", chanStatus4.linkDownEventCounter);    
    DUMP_TO_SCREEN("  Wis Mode       : %s\n", chanStatus4.wisModeStatus==TRUE?"enabled":"disabled");
  }

    
  DUMP_TO_SCREEN("\n");
  if (ChannelStatusType == 2) {
    DUMP_TO_SCREEN("Autonegotiation capabilities\n");  
    DUMP_TO_SCREEN("  10 MBit :   Full duplex: %s     Half duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.advertise10FullDuplex),
                   YesOrNo(chanStatus2.configCapabilities.advertise10HalfDuplex));
    DUMP_TO_SCREEN("  100 MBit:   Full duplex: %s     Half duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.advertise100FullDuplex),
                   YesOrNo(chanStatus2.configCapabilities.advertise100HalfDuplex));
    DUMP_TO_SCREEN("  1 GBit  :   Full duplex: %s     Half duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.advertise1000FullDuplex),
                   YesOrNo(chanStatus2.configCapabilities.advertise1000HalfDuplex));
    DUMP_TO_SCREEN("  10 GBit :   Full duplex: %s     Half duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.advertise10000FullDuplex),
                   YesOrNo(chanStatus2.configCapabilities.advertise10000HalfDuplex));
    DUMP_TO_SCREEN("\n");
    DUMP_TO_SCREEN("Manual configuration capabilities\n");  
    DUMP_TO_SCREEN("  Full duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualFullDuplex));
    DUMP_TO_SCREEN("  Half duplex: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualHalfDuplex));
    DUMP_TO_SCREEN("  10 MBit    : %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualSpeed10));
    DUMP_TO_SCREEN("  100 MBit   : %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualSpeed100));    
    DUMP_TO_SCREEN("  1 GBit     : %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualSpeed1000));
    DUMP_TO_SCREEN("  10 GBit    : %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualSpeed10000));    
    DUMP_TO_SCREEN("  MDIMode    : Auto: %s     MDI: %s    MDIX: %s\n",
                   YesOrNo(chanStatus2.configCapabilities.manualMdiModeAuto),
                   YesOrNo(chanStatus2.configCapabilities.manualMdiModeMDI),
                   YesOrNo(chanStatus2.configCapabilities.manualMdiModeMDIX));
    DUMP_TO_SCREEN("\n");
    if (!chanStatus2.configSuccess) {
      DUMP_TO_SCREEN(">>> ERROR: LINK CONFIGURATION DOES NOT MATCH PORT CAPABILITIES\n");
      DUMP_TO_SCREEN("\n");          
    }
  }  
  if (ChannelStatusType == 3) {
    printChannelStatus3(chanStatus3);
  }
  if (ChannelStatusType == 4) {
    printChannelStatus3(chanStatus4.type3);
  }
  
  
  WriteToLog(LOG_INFO, WriteBuffer);
  return NTCI_ERRCODE_SUCCESS;
}


/* Show the link status for all available channels. */
uint32_t GetChannelLinkStatus(uint8_t card)
{
  handle_t cardHandle;
  ChannelsLinkStatusType1_t chansLinkStatus;
  uint32_t result=0;
  uint8_t channel=0;

  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }
  
  if((result = NTCI_GetChannelsLinkStatus(cardHandle, 
                                          STID_CHANNELS_LINK_STATUS_TYPE1,
                                          &chansLinkStatus)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  for(channel=0; channel<chansLinkStatus.numberOfChannelsAvailable; channel++) {
    if (chansLinkStatus.linkUpBitmask & (1<<channel)) {
      WriteToLog(LOG_INFO, "Channel %d: Link UP.\n", channel);
    } else {
      WriteToLog(LOG_INFO, "Channel %d: Link DOWN.\n", channel);
    }
  }
  return NTCI_ERRCODE_SUCCESS;    
}
  


uint32_t _GetChannelStatus(void *arglist)
{
  uint8_t *argptr = arglist;
  uint8_t card;
  uint32_t channelMask;
  card        = get8incr(&argptr);
  channelMask = get32incr(&argptr);
  return GetChannelStatus(card, channelMask);
}

uint32_t _GetChannelLinkStatus(void *arglist)
{
  uint8_t *argptr = arglist;
  uint8_t card;

  card            = get8incr(&argptr);
  return GetChannelLinkStatus(card);
}

