// $Id: FilterTest.c 13569 2011-07-05 12:50:08Z aa $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * FilterTest.c : Contain functions related to autotest of filtering
 */

#ifdef WIN32
#include <Windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#endif

/* The NTCommonInterface.h header provides the interface definitions
 * for the NTCI library */

#include <NTCommonInterface.h>

/* These headers are needed for accessing the capabilities interface: */

#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>

/* These headers are needed for the packet classification API */

#include <NTCI_packetclassification.h>
#include <packetclassificationstructure.h>

#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>


#include "toolslib.h"
#include "FilterTest.h"

#ifdef _WIN32
#define PRINTF_64BIT "%I64d"
#else
#define PRINTF_64BIT "%Ld"
#define PRINTF_64BITHEX "%016llX"
#endif

static uint32_t LastTestFrameLibFilterID;
static uint64_t LastTestFrameLibFilterIndex;
static uint64_t LastTestFrameLibFilterColor;

/* Pass filter command string to parser */

uint32_t PassFilter(uint8_t card, char *data)
{
    uint32_t result;
    uint32_t returncode;
    PassFilter_t command;
    handle_t cardHandle;

    if (strlen(data) > sizeof(command.achFilterString)) {
      WriteToLog(LOG_ERROR, "NTPL string too long\n");
      return NTCI_ERRCODE_INTERNAL_ERROR;
    }
    if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
      return result;
    }    

    /* Use the first version of the STID. */
    command.WriteHW = TRUE;
    memset(command.achFilterString, 0, sizeof(command.achFilterString));
    memcpy(command.achFilterString, data, strlen(data));
    
    /* First try the newest STID. */       
    returncode = NTCI_PacketClassification(cardHandle,
                                           STID_PASS_FILTER,
                                           &command);
      
    if(returncode != NTCI_ERRCODE_SUCCESS) {
      FilterError_t error;
        
      result = NTCI_PacketClassification(cardHandle,
                                         STID_GET_FILTER_ERROR,
                                         &error);
        
      WriteToLog(LOG_ERROR, "%s\n",error.achFilterError1);
      WriteToLog(LOG_ERROR, "%s\n",error.achFilterError2);
      WriteToLog(LOG_ERROR, "%s\n",error.achFilterError3);
    } else {
      switch (command.ReturnData.FilterReturnDataType) {
      case FILTER_RETURN_GROUPID:
        LastTestFrameLibFilterID = command.ReturnData.u.FilterIndex.FilterId;
        LastTestFrameLibFilterIndex = command.ReturnData.u.FilterIndex.u.GroupIndex;
        WriteToLog(LOG_INFO, "Returned FilterGroupID     : 0x%04X\n", LastTestFrameLibFilterID);
        break;
      case FILTER_RETURN_SEARCHID:
        LastTestFrameLibFilterID = command.ReturnData.u.FilterIndex.FilterId;
        LastTestFrameLibFilterIndex = command.ReturnData.u.FilterIndex.u.SearchIndex.Index[0];
        WriteToLog(LOG_INFO, "Returned FilterSearchID    : 0x%04X\n", LastTestFrameLibFilterID);
        WriteToLog(LOG_INFO, "Returned FilterSearchIndex : 0x%04X\n", LastTestFrameLibFilterIndex);
        break;
      case FILTER_RETURN_INFO:
        {
          WriteToLog(LOG_INFO, "FilterReturnDataType       : 0x%02X\n", command.ReturnData.FilterReturnDataType);
          WriteToLog(LOG_INFO, "SizeCount                  : 0x%02X\n", command.ReturnData.u.FilterInfo.SizeCount);
          WriteToLog(LOG_INFO, "ProtocolCount              : 0x%02X\n", command.ReturnData.u.FilterInfo.ProtocolCount);
          WriteToLog(LOG_INFO, "ErrorCount                 : 0x%02X\n", command.ReturnData.u.FilterInfo.ErrorCount);
          WriteToLog(LOG_INFO, "PatternCount               : 0x%02X\n", command.ReturnData.u.FilterInfo.PatternCount);
          WriteToLog(LOG_INFO, "DynOffsetCount             : 0x%02X\n", command.ReturnData.u.FilterInfo.DynOffsetCount);
          WriteToLog(LOG_INFO, "Group4PlusCount            : 0x%02X\n", command.ReturnData.u.FilterInfo.Group4PlusCount);
          WriteToLog(LOG_INFO, "Group8Count                : 0x%02X\n", command.ReturnData.u.FilterInfo.Group8Count);
          WriteToLog(LOG_INFO, "SliceLookupCount           : 0x%02X\n", command.ReturnData.u.FilterInfo.SliceLookupCount);
          WriteToLog(LOG_INFO, "SliceLengthCount           : 0x%02X\n", command.ReturnData.u.FilterInfo.SliceLengthCount);
          WriteToLog(LOG_INFO, "SearchCount                : 0x%02X\n", command.ReturnData.u.FilterInfo.SearchCount);                
        }
        break;
      case FILTER_RETURN_FILTERID:
        LastTestFrameLibFilterID = command.ReturnData.u.FilterIndex.FilterId;
        WriteToLog(LOG_INFO, "Returned FilterID          : 0x%04X\n", LastTestFrameLibFilterID);
        break;
      }
    }
    return returncode;
}


uint32_t GetLastFilterID(void)
{
  TestFrameLibFuncRetVal = LastTestFrameLibFilterID;
  return NTCI_ERRCODE_SUCCESS;
}


uint32_t GetLastFilterIndex(void)
{
  TestFrameLibFuncRetVal = LastTestFrameLibFilterIndex;
  return NTCI_ERRCODE_SUCCESS;
}


uint32_t GetLastFilterColor(void)
{
  TestFrameLibFuncRetVal = LastTestFrameLibFilterColor;
  return NTCI_ERRCODE_SUCCESS;
}


uint32_t _PassFilter(void *arglist)
{
  uint8_t  *argptr = arglist;
  char     *data;
  uint8_t  card;
  
  card        = get8incr(&argptr);
  getCHARPincr(argptr, data);
  return PassFilter(card, data);
}


uint32_t _GetLastFilterID(void *arglist __attribute__((unused)))
{
  return GetLastFilterID();
}

uint32_t _GetLastFilterIndex(void *arglist __attribute__((unused)))
{
  return GetLastFilterIndex();
}

uint32_t _GetLastFilterColor(void *arglist __attribute__((unused)))
{
  return GetLastFilterColor();
}
