// $Id: PacketFeedInfo.c 24627 2013-07-11 13:47:07Z am $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * PacketFeedInfo.c : This file contain functions related to packetfeeds.
 */

#ifdef WIN32
#include <windows.h>
#include <process.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <io.h>
#include <errno.h>
#include "platformwin.h"
#else
#define _GNU_SOURCE
#define _FILE_OFFSET_BITS 64
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <sys/mman.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <pthread.h>
#include <semaphore.h>
#endif

/* The NTCommonInterface.h header provides the interface definitions
 * for the Xyratex NTCI library */

#include <NTCommonInterface.h>

/* These headers are needed when capturing traffic */

#include <packetfeedstructure.h>
#include <packetdescriptor.h>
#include <NTCI_packetfeed.h>
#include <NTCI_registeraccess.h>

/* These headers are needed for accessing the capabilities interface: */
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>
#include "toolslib.h"

/* Thse headers are needed for statistics and channel status. */
#include <NTCI_statistics.h>
#include <NTCI_cardstatus.h>
#include <statisticsstructure.h>
#include <cardstatusstructure.h>
#include <cardconfigurationstructure.h>
#include <NTCI_cardconfiguration.h>

#include "PacketFeedInfo.h"


#ifdef WIN32
#define FILE_FLAGS1 GENERIC_WRITE
#define FILE_RIGHTS1 CREATE_ALWAYS
#define FILE_FLAGS2 GENERIC_READ
#define FILE_RIGHTS2 OPEN_EXISTING
#else
#define FILE_FLAGS1 O_WRONLY | O_CREAT | O_TRUNC
#define FILE_RIGHTS1 0666
#define FILE_FLAGS2 O_RDONLY
#define FILE_RIGHTS2 0666

#define fileopen open
#define filelseek lseek
#define fileclose close
#define fileread read
#define filewrite write

#define NT_HANDLE int
#endif



/* Search for feeds currently running and shut them down, before
 * starting the new feed */
void ShutdownAllPacketFeeds(handle_t cardHandle)
{
  uint32_t numFeeds;
  uint32_t i;
  uint32_t result;
  uint32_t shutdownCount;
  CardHardwareCapabilitiesType2_t cardCapabilities;

  //First enable Flushing
  NTCI_GetCapabilities(cardHandle, 
                       CAPID_HARDWARE,
                       NOT_SUB_CAPABILITIES,
                       STID_CARD_HARDWARE_CAPABILITIES_TYPE2,
                       &cardCapabilities);
  
  result = NTCI_FlushTxPorts(cardHandle, cardCapabilities.txChannelsSupportedBitmask, TRUE);
  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "ShutdownAllPacketFeeds failed (%x) enabling Tx Flushing for channelmask: %x\n",
            result, cardCapabilities.txChannelsSupportedBitmask);
  }

  result = NTCI_GetPacketFeedCount(cardHandle, &numFeeds);

  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error getting packet feed count: %i\n", result);
    exit(-1);
  }

  /* Iterate through each feed */
  shutdownCount = 0;
  /* Flush Segments */

  for (i=0; i<numFeeds; i++) {
    FeedDescriptionType2_t description;

        
    /* Get the description */
    result = NTCI_GetPacketFeedDescription(cardHandle, 
                                           i,
                                           STID_FEED_DESCRIPTION_TYPE2,
                                           &description);


    if (result == NTCI_ERRCODE_INVALID_PARAMETER) {
      /* A slot has been allocated for this feed, but there is no
       * feed created. Therefore ignore this feed */
      continue;
    }
    
    if (result != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error getting description for packet feed #%i: %i\n",
              i, result);
      exit(-1);
    }
    
    /* Stop the feed. */
    result = NTCI_StopPacketFeed(cardHandle, description.handle);
    if (result == NTCI_ERRCODE_SUCCESS) {
      /* Destroy the feed. */
      result = NTCI_DestroyPacketFeed(cardHandle, description.handle);
    }

    if (result == NTCI_ERRCODE_SUCCESS) {
      ++shutdownCount;
    }
    else 
    {
      fprintf(stderr, "Failed to shut down packet feed #%i: %i\n",
              i, result);
      exit(-1);
    }
  }
  /* Stop segment flushing  */
  result = NTCI_FlushTxPorts(cardHandle, cardCapabilities.txChannelsSupportedBitmask, FALSE);
  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "ShutdownAllPacketFeeds failed (%x) disabling Tx Flushing for channelmask: %x\n",
            result, cardCapabilities.txChannelsSupportedBitmask);
  }

  if (shutdownCount > 0)
    fprintf(stderr, "Shut down %i existing packet feeds\n", shutdownCount);
}



uint32_t PacketFeedStatus(uint8_t card, uint8_t feed, uint8_t type)
{
  uint32_t result;
  uint32_t wPos=0;
  handle_t cardHandle;
  FeedStatusSampledType4_t feedStatus;
  FeedDescriptionType1_t description;
  char writeBuffer[1024];

  if (type == 4) {
    if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
      return result;
    }
    
    result = NTCI_GetPacketFeedDescription(cardHandle, 
                                           feed,
                                           STID_FEED_DESCRIPTION_TYPE1,
                                           &description);
    
    if (result == NTCI_ERRCODE_SUCCESS) {
        result = NTCI_GetPacketFeedStatus(cardHandle, description.handle,
                                          STID_FEED_STATUS_SAMPLED_TYPE4, &feedStatus);
        
        if (result != NTCI_ERRCODE_SUCCESS) {
          WriteToLog(LOG_ERROR, "GetPacketFeedStatus: Failed (%i)\n", result);
          return result;
        }
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "Feed %d status:\n", feed);
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "   Mode: ");
        switch((FeedStatusSampledMode)feedStatus.mode) {
        case SAMPLED_FEED_STOPPED:        
          wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                                "Stopped...\n");
          break;
        case SAMPLED_FEED_STARTED_PACKETS_AVAILABLE:
          wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                                "Segments available...\n");
          break;
        case SAMPLED_FEED_STARTED_NO_PACKETS_AVAILABLE:
          wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                                "No segments available...\n");
          break;
        case SAMPLED_FEED_STARTED_PACKETS_PENDING:
          wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                                "Data pending...\n");
          break;
        }
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "   Adapter SDRAM fill level                          : %d %%\n",
                              feedStatus.adapterFillLevel);
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "   Number of segments ready via API                  : %d\n",
                              feedStatus.numSegmentsReady);
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "   Number of segments currently held by the user     : %d\n",
                              feedStatus.numSegmentsBusy);
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              "   Number of segments waiting for adapter free slots : %d\n",
                              feedStatus.numSegmentsIdle);
        (void)WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                            "   Number of segments held by the adapter            : %d\n",
                            feedStatus.numSegmentsAdapter);
        WriteToLog(LOG_INFO, writeBuffer);
        return result;
    } else {
      WriteToLog(LOG_ERROR, "No feed %d exist.\n", feed);
    }
    return 0;
  } else {
    WriteToLog(LOG_ERROR, "No feed %d exist or type %d is not 4\n", feed, type);
    return NTCI_ERRCODE_INVALID_PARAMETER;
  }
}  

uint32_t _PacketFeedStatus(void* arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card, feed, type;
  
  card  = get8incr(&argptr);
  feed  = get8incr(&argptr);
  type  = get8incr(&argptr);
  
  return PacketFeedStatus(card, feed, type);
}


void CardSDRAMUtilization(handle_t cardHandle)
{
  uint32_t numFeeds;
  uint32_t i;
  uint32_t result;
  int32_t sdramUtil[32];
  uint32_t wPos=0;
  FeedStatusSampledType4_t feedStatus;
  char writeBuffer[1024];
  int numValidFeeds=0;
  
  memset(sdramUtil, -1, sizeof(sdramUtil));
  
  result = NTCI_GetPacketFeedCount(cardHandle, &numFeeds);

  if (result != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error getting packet feed count: %i\n", result);
    return;
  }

  /* Iterate through each feed */
  for (i=0; i<numFeeds; i++) {
    FeedDescriptionType1_t description;
    
    /* Get the description */
    result = NTCI_GetPacketFeedDescription(cardHandle, 
                                           i,
                                           STID_FEED_DESCRIPTION_TYPE1,
                                           &description);
    
    if (result == NTCI_ERRCODE_INVALID_PARAMETER) {
      /* A slot has been allocated for this feed, but there is no
       * feed created. Therefore ignore this feed */
      continue;
    }
    
    if (result != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error getting description for packet feed #%i: %i\n",
              i, result);
      return;
    }

    result = NTCI_GetPacketFeedStatus(cardHandle, description.handle,
                                      STID_FEED_STATUS_SAMPLED_TYPE4, &feedStatus);
    
    if (result != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_ERROR, "GetPacketFeedStatus: Failed (%i)\n", result);
      return;
    }
    sdramUtil[i]=(int32_t)feedStatus.adapterFillLevel;
    numValidFeeds++;
  }

  if(!numValidFeeds) {
    (void)WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                        ">>> Error: No feeds running. Cannot show status.\n");
  } else {
    // Dump the results
    wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                          "Fill level in \%: ");  
    for (i=0; i<numFeeds; i++) {
      if(sdramUtil[i] != -1) {
        wPos += WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos,
                              " %3d ", sdramUtil[i]);
      }
    }
    (void)WriteToBuffer(&writeBuffer[wPos], sizeof(writeBuffer)-wPos, "\n");
  }
  WriteToLog(LOG_INFO, writeBuffer);  
}


