#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define USE_GET_ONLY_API

#define _NTAPI_EXTDESCR_ALL_

#define ERROR_3GD_NOT_LOADED -10000
#define CHAR_DEVICE_3GD  "/dev/nt3gd"

#ifdef HAVE_NAPATECH3G_API
#include <stdint.h>
#include "pcap-napatech-3g.h"
#include "pcap-int.h"

#include <nt.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <syslog.h>
#include <pthread.h>
#include <pcap-napatech-intl.h>
#include "iniparser.h"

#define EPOCH_OFFSET 1164447360000000000ULL   // offset from 1/1 1601 to 1/1 1970 in 10ns resolution


#undef NAPATECH_DEBUG
#undef NAPATECH_CALL_TRACE




static int napatech_activate(pcap_t *p);


typedef struct _nanotimeval {
  uint32_t sec;
  uint32_t nsec;
} nanotimeval;


/* For segment based read from driver we need this precise definition. 
 * This is the format from FPGA 
 */
struct ntpcap_ts_s {
  uint32_t sec;
  uint32_t usec;
};
struct ntpcap_hdr_s {
  struct ntpcap_ts_s ts;
  uint32_t caplen;
  uint32_t wirelen;
};

#define SEGMENT_IF 0
#define PACKET_IF  1


/* struct napatech_details
 *
 * This is an additional part of the pcap_t structure, which we use to
 * hold the details of the packet feeds we create.
 */
struct napatech3G_details
{
  uint32_t StreamId;
  int TxPort;
  int HBAllow;
  int TxNumaNode;
  int TxMinHostBuffer;
  int NanoSec;

  NtNetStreamRx_t   pRxStream;
  /* Segment inferface variables */
  NtNetBuf_t           segNetBuf;
  struct NtNetBuf_s    NetBuf;
  struct ntpcap_hdr_s *pPCAPHdr;
  int32_t              segLength;
  /*-----------------------------*/
  NtNetStreamTx_t pTxStream;
  NtStatStream_t  pStatStream;

  uint64_t  NativeTimeOffset;
  uint64_t  pkts;
  uint64_t  dropped;
  
  int snaplen;
  int promisc;
  int to_ms;
  int nonblocking;

  nanotimeval tlst;
  uint64_t lst_ts;
  
  /* List */
  pcap_t *pNext;
};



static void WriteToLog(int boError, char *file, const char *function, uint32_t line, const char *format, ... )
{
  va_list ap;
  char *logStr = (char*)malloc(1024);
  va_start(ap,format);

  vsprintf(logStr, format, ap);

  if(file!=NULL) {
    syslog(boError?LOG_ERR:LOG_INFO, "NT: %s %d: %s: %s", file, line, function, logStr);
  } else {
    syslog(boError?LOG_ERR:LOG_INFO, "NT: %s", logStr);
  }
  if(boError) {
    fprintf(stderr, "NT: %s %d: %s: %s", file, line, function, logStr);
  }
  free(logStr);
}

#ifdef NAPATECH_CALL_TRACE
#define CALL_TRACE()                     WriteToLog(0, __FILE__, __FUNCTION__, __LINE__,"\n")
#else
#define CALL_TRACE() 
#endif

#ifdef NAPATECH_DEBUG
#define DEBUG_PRINT(string, args...)     WriteToLog(0, __FILE__, __FUNCTION__, __LINE__, string, ## args)
#else
#define DEBUG_PRINT(string, args...)
#endif

#define ERROR_PRINT(string,  args...)    WriteToLog(1, __FILE__, __FUNCTION__, __LINE__, string, ## args)
#define ERROR_INFO(string,  args...)     WriteToLog(0, __FILE__, __FUNCTION__, __LINE__, string, ## args)



static void napatech_pcap_cleanup( pcap_t *p );


struct DefinedStreams_s 
{
  NtInfo_t SInfo;
} DefStreams;  


struct SetupStream_s
{
  char IniSecName[MAX_DEV_NAME_SIZE];
  NtNtplInfo_t NtplInfo[MAX_FILTERS_PER_INI_SEC];
  uint32_t StreamId;
  int TxPort;
  int HBAllow;  
  int TxNumaNode;
  int TxMinHostBuffer;
  int NanoSec;
  int NTInterface;
} SetupStreams[NAPATECH3G_MAX_STREAMS];

static int SetupStreamCount = 0;
static int timestampType = 0;
static int NT3GInitialized = 0;
static int platform32 = 0;


// List of all pcap instances created. Use to destroy them on "atexit"
static pthread_mutex_t LstMutex;
static pcap_t *pHead=NULL;


static uint32_t Flock=0;

static void FastlockGet(void)
{
  uint32_t prev;
  
  do 
  {
    __asm__ __volatile__("lock; cmpxchg %1,%2"
			: "=a"(prev)
			: "r"(1), "m"(Flock), "0"(0)
			: "memory");

    if (prev) usleep(100);			
  } while (prev);			
}


void inline FastlockRelease(void)
{
  Flock = 0;
}



static void atexit_handler(void)
{
 pcap_t *cur=pHead;
 
 CALL_TRACE();
 
  // Close all nt pcap instances
  while(cur) {
    // napatech_pcap_cleanup will take cur out of the list
    napatech_pcap_cleanup(cur);
    cur=pHead;
  }

  //  NT_Done() - atexit in NT_Init() is executed automatically after this call
  FastlockGet();
  NT3GInitialized = 0;
  platform32 = 0;
  pthread_mutex_destroy(&LstMutex);
  FastlockRelease();
}




static int ExtractStreamIdFrom_nt3g_Device(const char *pDevice)
{
  int len;
  
  if ((len = strlen(pDevice)) >= 5 &&
      (pDevice[0] == 'n' || pDevice[0] == 'N') &&
      (pDevice[1] == 't' || pDevice[1] == 'T') &&
      (pDevice[2] == '3') &&
      (pDevice[3] == 'g' || pDevice[3] == 'G'))
  {
   int n = 4;
    while (n < len && pDevice[n] >= '0' && pDevice[n] <= '9') n++;
    if (n == len)
      return atoi(&pDevice[4]);
  }
  
  return -1; 
}


static int StreamIdDefinedInDefStream(int streamId)
{
 int cnt;
  for (cnt = 0; cnt < DefStreams.SInfo.u.stream.data.count; cnt++)
      if (DefStreams.SInfo.u.stream.data.streamIDList[cnt] == streamId) return 1;
  return 0;
}


static int UpdateStreamInfo(void)
{
  NtInfoStream_t hStream;
  
  if (NT_InfoOpen(&hStream, "libpcap") != NT_SUCCESS) return -1;
  DefStreams.SInfo.cmd = NT_INFO_CMD_READ_STREAM;
  if (NT_InfoRead(hStream, &DefStreams.SInfo) != NT_SUCCESS) return -1;
  NT_InfoClose(hStream);

  return NT_SUCCESS;
}


static struct SetupStream_s *StreamIdDefinedInSetupStream(uint32_t streamId)
{
 int cnt;
 
  for (cnt = 0; cnt < SetupStreamCount; cnt++)
    if (streamId == SetupStreams[cnt].StreamId) return &SetupStreams[cnt];  
  return NULL;
}

static struct SetupStream_s *NameDefinedInSetupStream(const char *name)
{
 int cnt;
 
  for (cnt = 0; cnt < SetupStreamCount; cnt++)
    if (strcmp(name, SetupStreams[cnt].IniSecName) == 0) return &SetupStreams[cnt];  
  return NULL;
}



static int ReadAndSetupIniStreams(void)
{
  int cnt, i, rx, res, TxPort, value;
  uint32_t StreamId;
  dictionary *pDict = NULL;
  char *pNTPL;
  int numSec, ConfigOk;
  NtConfigStream_t hStream;
  NtInfoStream_t hInfStream;
  char key[256], *pStrVal;
  NtInfo_t SInfo;
  NtNtplInfo_t NtplInf;
  
  pDict = iniparser_load(NAPATECH3G_NTPCAP_INI_FILE_PATH);
  if (pDict)
  {
    numSec = iniparser_getnsec(pDict);
    if (numSec >= 0)
    {  
      if (NT_ConfigOpen(&hStream, "libpcap") != NT_SUCCESS)
      {
	ERROR_PRINT("ERROR Opening Config\n");
	return -1; 
      }

      if (NT_InfoOpen(&hInfStream, "libpcap") != NT_SUCCESS)
      {
        NT_ConfigClose(hStream);
	ERROR_PRINT("ERROR Opening Info\n");
	return -1; 
      }
      SInfo.cmd = NT_INFO_CMD_READ_SYSTEM;
      if (NT_InfoRead(hInfStream, &SInfo) != NT_SUCCESS)
      {
        NT_InfoClose(hInfStream);
        NT_ConfigClose(hStream);
	ERROR_PRINT("ERROR reading Info from NT_Info_Stream\n");
	return -1; 
      }
      NT_InfoClose(hInfStream);
   
 
      SetupStreamCount = 0;
      for (i=0; i<numSec;i++)
      {	
       int chk_streamid_identical;

	if (SetupStreamCount == NAPATECH3G_MAX_STREAMS) {
	  ERROR_PRINT("Too many Libpcap devices defined (max: %d)\n", NAPATECH3G_MAX_STREAMS);
          break;
	}

	strcpy(SetupStreams[SetupStreamCount].IniSecName, iniparser_getsecname(pDict, i)); 
	
        if (NameDefinedInSetupStream(SetupStreams[SetupStreamCount].IniSecName))
	{
	  DEBUG_PRINT("[%s] Name clash!!\n");
	  continue; // drop this definition - clash with earlier defined device
	}  
	
	DEBUG_PRINT("[%s] Checking/setting up\n", SetupStreams[SetupStreamCount].IniSecName);
	
        // Check NTPL expression validity. if StreamId not specified explicitly, then check for identical StreamId setup in ntpl expr.
	rx= 1;
	ConfigOk = 1;
	StreamId = (uint32_t)-1;
        SetupStreams[SetupStreamCount].StreamId = (uint32_t)-1;  // may be checked on and indicator if we shall create the stream
        SetupStreams[SetupStreamCount].TxPort = -1;
        SetupStreams[SetupStreamCount].HBAllow = -1;
        SetupStreams[SetupStreamCount].TxNumaNode = 0;
        SetupStreams[SetupStreamCount].TxMinHostBuffer = 0;
        SetupStreams[SetupStreamCount].NanoSec = 0;
        SetupStreams[SetupStreamCount].NTInterface = SEGMENT_IF;

	// Get first StreamId - if specified
        sprintf(key,"%s:StreamId", SetupStreams[SetupStreamCount].IniSecName);
        StreamId = (uint32_t)iniparser_getint(pDict, key, -1);

        chk_streamid_identical = (StreamId == (uint32_t)-1)?1:0;

	while ((rx <= MAX_FILTERS_PER_INI_SEC) && ConfigOk)
	{
	  sprintf(key,"%s:Rx%d", SetupStreams[SetupStreamCount].IniSecName, rx);
          pNTPL = iniparser_getstring(pDict, key, NULL);
	  if (pNTPL == NULL) break; // no more keys
	
          /* Validate NTPL and get StreamId */
	  res = NT_NTPL(hStream, pNTPL, &NtplInf, 1);
	  if (res != NT_SUCCESS)
	  { 
            ERROR_INFO("[%s] NTPL error : %s\n",SetupStreams[SetupStreamCount].IniSecName, NtplInf.u.errorData.errBuffer[0]);
	    ConfigOk = 0;
	    break;
	  }  

	  if (chk_streamid_identical) {
	    if (rx == 1) {
	      StreamId = NtplInf.streamId;
	    } else
            if (StreamId == (uint32_t)-1 || StreamId != NtplInf.streamId) {
  	      ERROR_PRINT("[%s] StreamId mismatch or not unique\n", SetupStreams[SetupStreamCount].IniSecName);
	      ConfigOk = 0;
	      break;
	    } 
	  }
  
	  rx++;
	}
	
	if (!ConfigOk) 
	  continue;

        if (StreamId != (uint32_t)-1)
	  DEBUG_PRINT("[%s] StreamId %d identified\n", SetupStreams[SetupStreamCount].IniSecName, StreamId);

	/* Get hostbuffer allowance value */
        sprintf(key,"%s:HBAllowance", SetupStreams[SetupStreamCount].IniSecName);
        value = (uint32_t)iniparser_getint(pDict, key, -1);
        if (value !=-1 && (value < 1 || value > 100)) {
	  ERROR_PRINT("[%s] Invalid hostbuffer allowance value specified (%i) range is 1..100\n", SetupStreams[SetupStreamCount].IniSecName, value);
          value=-1;
	}
        SetupStreams[SetupStreamCount].HBAllow = value;
        if (value!=-1) {
          DEBUG_PRINT("[%s] Hostbuffer allowance value set to %d%%\n", SetupStreams[SetupStreamCount].IniSecName, value);
	}

	// Get Tx numa node
        sprintf(key,"%s:TxNumaNode", SetupStreams[SetupStreamCount].IniSecName);
        value = (uint32_t)iniparser_getint(pDict, key, -10000);
        if (value != -10000) {
	  if (value < 0) {
	    ERROR_PRINT("[%s] Invalid Tx numa node (%i)\n", SetupStreams[SetupStreamCount].IniSecName, value);
	    value=0;
	  }
	  SetupStreams[SetupStreamCount].TxNumaNode = value;
          DEBUG_PRINT("[%s] Tx numa node set to %d\n", SetupStreams[SetupStreamCount].IniSecName, value);
	}

	// Get Tx min host buffer size
        sprintf(key,"%s:TxMinHostBuffer", SetupStreams[SetupStreamCount].IniSecName);
        value = (uint32_t)iniparser_getint(pDict, key, -10000);
        if (value != -10000) {
	  if (value < 0) {
	    ERROR_PRINT("[%s] Invalid Tx min hostbuffer size (%i)\n", SetupStreams[SetupStreamCount].IniSecName, value);
	    value=0;
	  }
	  SetupStreams[SetupStreamCount].TxMinHostBuffer = value;
          DEBUG_PRINT("[%s] Tx min hostbuffer size set to %dMB\n", SetupStreams[SetupStreamCount].IniSecName, value);
	}

	// Get NanoSec
        sprintf(key,"%s:NanoSec", SetupStreams[SetupStreamCount].IniSecName);
        pStrVal = iniparser_getstring(pDict, key, "**");
        if (strcmp(pStrVal, "**")!=0) {
          if (strcmp(pStrVal,"true")==0) {
            value = 1;
          } else if (strcmp(pStrVal,"false")==0) {
            value = 0;
          } else {
            ERROR_PRINT("[%s] Invalid Tx min hostbuffer size (%i)\n", SetupStreams[SetupStreamCount].IniSecName, value);
            value=0;
          }
	  SetupStreams[SetupStreamCount].NanoSec = value;
          DEBUG_PRINT("[%s] NanoSec set to %d\n", SetupStreams[SetupStreamCount].IniSecName, value);
	}

	// Get NTInterface
        sprintf(key,"%s:NTInterface", SetupStreams[SetupStreamCount].IniSecName);
        pStrVal = iniparser_getstring(pDict, key, "**");
        if (strcmp(pStrVal, "**")!=0) {
          if (strcmp(pStrVal,"Packet")==0) {
            value = PACKET_IF;
          } else if (strcmp(pStrVal,"Segment")==0) {
            value = SEGMENT_IF;
          } else {
            ERROR_PRINT("[%s] Invalid NTInterface specified. \"%s\" is not recognized. Defaulting to Segment interface.\n", SetupStreams[SetupStreamCount].IniSecName, pStrVal);
            value = SEGMENT_IF;
          }
	  SetupStreams[SetupStreamCount].NTInterface = value;
          DEBUG_PRINT("[%s] NTInterface set to %d\n", SetupStreams[SetupStreamCount].IniSecName, value);
	}

	/* check if the streamId we want to use is already setup. If, then accept as created */
	if (StreamId != (uint32_t)-1 && StreamIdDefinedInDefStream(StreamId))
	{
	  SetupStreams[SetupStreamCount].NtplInfo[0].ntplId = (uint32_t)-1; // do not delete from here
	  SetupStreams[SetupStreamCount].StreamId = StreamId;   // use it
	}
	
	/* Check for valid Tx port */
	sprintf(key, "%s:Tx", SetupStreams[SetupStreamCount].IniSecName);
	TxPort = iniparser_getint(pDict, key, -1);
	if (TxPort < -1  || TxPort >= (int)SInfo.u.system.data.numPorts)
	{
	  ERROR_PRINT("[%s] Invalid Tx port specified\n", SetupStreams[SetupStreamCount].IniSecName);
	  ConfigOk = 0;
	}
	else
	{ 
         NtNetStreamTx_t hTxStream;

	  if (TxPort >= 0)  /* check if we may claim a tx host buffer */
	  {
	    char name[64];
	    sprintf(name,"libpcapTx%d", TxPort);

#if NTAPI_VERSION==0
	      res = NT_NetTxOpen(&hTxStream, name, 0x01<<TxPort, SetupStreams[SetupStreamCount].TxNumaNode);
#elif NTAPI_VERSION==1
  	      res = NT_NetTxOpen(&hTxStream, name, 0x01<<TxPort, SetupStreams[SetupStreamCount].TxNumaNode, SetupStreams[SetupStreamCount].TxMinHostBuffer);
#else
#error NTAPI_VERSION not supported
#endif

	    if (res == NT_SUCCESS)
	    {
	      DEBUG_PRINT("[%s] Tx stream setup and valid Port %d\n", SetupStreams[SetupStreamCount].IniSecName, TxPort);
	      SetupStreams[SetupStreamCount].TxPort = TxPort;
	      NT_NetTxClose(hTxStream);
	    }  
	    else
	    {
	      ERROR_PRINT("[%s] Could not open Tx port %i\n", SetupStreams[SetupStreamCount].IniSecName, TxPort);
	      ConfigOk = 0;
	    }  
	  }
	  else
	    ;// No Tx setup
	}
	
	if (!ConfigOk) 
	  continue;

        /* Do we have any Rx(n) specified and if, is the stream already created */
        if (StreamId != SetupStreams[SetupStreamCount].StreamId)
	{  
	  /* OK, now we have checked NTPL and streamIDs and also Tx port if specified with success */
	  /* Now try to set all Rx(x) NTPL strings to create receive stream */
	  rx = 1;
	  while ((rx <= MAX_FILTERS_PER_INI_SEC) && ConfigOk)
	  {  
	    sprintf(key, "%s:Rx%d", SetupStreams[SetupStreamCount].IniSecName, rx);
	    pNTPL = iniparser_getstring(pDict, key, NULL);
	    if (pNTPL == NULL)
	    {
	      SetupStreams[SetupStreamCount].NtplInfo[rx-1].ntplId = (uint32_t)-1;    // stop indicator
	      break; // no more keys
	    }
	    DEBUG_PRINT("[%s] Creating NTPL filter\n", SetupStreams[SetupStreamCount].IniSecName);
	    res = NT_NTPL(hStream, pNTPL, &SetupStreams[SetupStreamCount].NtplInfo[rx-1], 0);
	    if (res != NT_SUCCESS)
	    {
	      SetupStreams[SetupStreamCount].NtplInfo[rx-1].ntplId = (uint32_t)-1;
	      ConfigOk = 0;
	      ERROR_PRINT("[%s] Rx%d setup but NTPL failed\n", SetupStreams[SetupStreamCount].IniSecName, rx);
	      break;
	    }  
	    rx++;
	  }  
          SetupStreams[SetupStreamCount].StreamId = StreamId;
	}

        if (ConfigOk)  /* Yes! we are done and have created a new stream */
	  SetupStreamCount++;

      } /* for (i=0; i<numSec;i++) */
      NT_ConfigClose(hStream);
    }
    
    iniparser_freedict(pDict);
  }
  else
    DEBUG_PRINT("No ntpcap.ini file found\n");
  
  return NT_SUCCESS;
}




static int InitializeNT3G(void)
{
  int res;
  struct stat filestat;

  FastlockGet();

  if (NT3GInitialized) 
  {
    FastlockRelease();
    return NT_SUCCESS;
  }

  // If nt3gd driver is not loaded at all, we will exit silently
  if (stat(CHAR_DEVICE_3GD, &filestat) != 0) {
    FastlockRelease();
    NT3GInitialized = 1;
    return ERROR_3GD_NOT_LOADED; // char device not loaded. This library may be used to access other network devices
  }

  res = NT_Init(NTAPI_VERSION);

  if (res == NT_SUCCESS || res == NT_ERROR_NT_INIT_ALREADY_CALLED)
    res = UpdateStreamInfo();
  else
  {
    NT_Done();
    FastlockRelease();
    /* Probably not available or ntservice not running */
    return NT_SUCCESS;
  }
  
  if (res == NT_SUCCESS) {
    NtInfo_t SInfo;
    NtInfoStream_t hInfStream;

    /* get timestampType used by adapter */
    if (NT_InfoOpen(&hInfStream, "libpcap") != NT_SUCCESS)
    {
      ERROR_PRINT("ERROR Opening Info\n");
      return -1; 
    }
    SInfo.cmd = NT_INFO_CMD_READ_ADAPTER_V1;
    SInfo.u.adapter_v1.adapterNo = 0;
    if (NT_InfoRead(hInfStream, &SInfo) != NT_SUCCESS)
    {
      NT_InfoClose(hInfStream);
      ERROR_PRINT("ERROR reading Info from NT_Info_Stream\n");
      return -1; 
    }
    NT_InfoClose(hInfStream);
   
    timestampType = (int)SInfo.u.adapter_v1.data.timestampType;

    res = ReadAndSetupIniStreams();
  }
  
  if (res == NT_SUCCESS) 
  {  
    pthread_mutex_init(&LstMutex, NULL);
    atexit(atexit_handler);
    NT3GInitialized = 1;
    platform32 = (sizeof(long)-sizeof(int))?0:1;
  }  
  
  FastlockRelease();
  return res;
}




static int is_napatech3G_device( const char *deviceName)
{
  int res = 0;

  CALL_TRACE();
  res = InitializeNT3G();

  if (res != NT_SUCCESS) {
    if (res == ERROR_3GD_NOT_LOADED) {
      return 0;  // nt3gd not loaded
    }
   return res;  
  }
  /* Name from ini file ? */
  if (NameDefinedInSetupStream(deviceName)) return 1;

  /* Check to see if devName has the following format "nt3g<n>" and is thus a napatech device */
  if (strlen(deviceName) >= 5 &&
     ((deviceName[0] == 'n') || (deviceName[0] == 'N')) &&
     ((deviceName[1] == 't') || (deviceName[1] == 'T')) &&
      (deviceName[2] == '3') &&
     ((deviceName[3] == 'g') || (deviceName[3] == 'G')) &&
      (deviceName[4] >= '0') && (deviceName[4] <= '9' )) return 1;
  
 return 0;
}



pcap_t *napatech3G_create(const char *device, char *ebuf, int *is_ours)
{
 pcap_t *p;
 
  CALL_TRACE();
  *is_ours = is_napatech3G_device(device);
  if (!*is_ours) {
    return NULL;
  }
  p = pcap_create_common(device, ebuf);
  if (p == NULL)
    return NULL;

  p->activate_op = napatech_activate;
  return p;
}


/* napatech_pcap_stats
 *
 * Read the statistics of the pcap packet feed.
 *
 * Parameters
 * p  - pointer to pcap_t structure for device
 * ps - pointer to pcap_stat structure to put statistics in
 * Returns 0 on success or -1 on error
 */
static int napatech_pcap_stats( pcap_t *p, struct pcap_stat *ps )
{
  CALL_TRACE();
  /* sizeof(pcap_stat) varies according to whether REMOTE is defined.*/
  /* this can cause some crashes.  commenting this out fixes this.*/
  /*memset(ps, 0, sizeof(struct pcap_stat));*/
  
  ps->ps_recv = p->md.nt3G->pkts;
#if NTAPI_VERSION==0
  ps->ps_drop = (uint32_t)p->md.nt3G->dropped;
#elif NTAPI_VERSION==1
  ps->ps_drop = 0;
  if (p->md.nt3G->pRxStream) {
    NtNetRx_t RxRd;
    RxRd.cmd = NT_NETRX_READ_CMD_STREAM_DROP;
    if (NT_NetRxRead(p->md.nt3G->pRxStream, &RxRd) == NT_SUCCESS) {
      ps->ps_drop = (uint32_t)RxRd.u.streamDrop.pktsDropped;
    }
  }
  ps->ps_ifdrop = 0;
  if (p->md.nt3G->pStatStream) {
    NtStatistics_t stat;
    stat.cmd = NT_STATISTICS_READ_CMD_QUERY_V2;
    if (NT_StatRead(p->md.nt3G->pStatStream, &stat) == NT_SUCCESS) {
      ps->ps_ifdrop = stat.u.query_v2.data.stream.streamid[p->md.nt3G->StreamId].drop.pkts;
    }
  }
#else
#error NTAPI_VERSION not supported
#endif

return 0;
}





/* napatech_pcap_read_packet
 *
 * Read some packets and call a callback function on each of these
 * packets.
 *
 * Parameters
 * p   - pointer to pcap_t structure for device
 * cnt - number of packets to read, 0 means read as many as possible
 * callback - callback function to be called on each packet
 * user - user details
 * Returns number of packets read or -1 on error
 */

static int napatech_pcap_read_packet( pcap_t *p, int cnt, pcap_handler callback, u_char *user )
{
 int pkts, res = NT_SUCCESS;
 NtNetBuf_t pNetBuf;
 struct pcap_pkthdr hdr;
 uint64_t ts;
 struct timespec tp;
 char ErrorBuf[256];
 int breakCnt=0, tmo;

  CALL_TRACE();

  if (p->md.nt3G->StreamId == (uint32_t)-1)
  {
    DEBUG_PRINT("No read support on this device\n");
    return -1;
  }  
    
  /* Please note here that the Rx stream is not opened before the actual request for the first packet */
  if (p->md.nt3G->pRxStream == NULL) {
#if NTAPI_VERSION==0
      res = NT_NetRxOpen(&p->md.nt3G->pRxStream, "libpcap", NT_NET_INTERFACE_PACKET, p->md.nt3G->StreamId);
#elif NTAPI_VERSION==1
      res = NT_NetRxOpen(&p->md.nt3G->pRxStream, "libpcap", NT_NET_INTERFACE_PACKET, p->md.nt3G->StreamId, p->md.nt3G->HBAllow);
#else
#error NTAPI_VERSION not supported
#endif
  }

  if (res != NT_SUCCESS) 
  {
    NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
    ERROR_PRINT("NetRxOpen failed (%s)\n", ErrorBuf);
    return -1;
  }  

  pkts = 0;
  do 
  {  
    if (p->break_loop) break;

    // if cnt < 0 we want to do returns regularily. Apps like wireshark needs to return to update the screen
    if (cnt < 0) {
      breakCnt++;
      if ((pkts && breakCnt > 2) || breakCnt > 30) {
        return pkts;
      }
    }

    tmo = p->md.nt3G->nonblocking?0:100;

#ifdef USE_GET_ONLY_API
    res = NT_NetRxGetNextPacket(p->md.nt3G->pRxStream, &pNetBuf, tmo);
#else
    res = NT_NetRxGet(p->md.nt3G->pRxStream, &pNetBuf, tmo);
#endif
    if (res == NT_STATUS_TRYAGAIN)
      continue;
    if (res == NT_STATUS_TIMEOUT)
    {
      DEBUG_PRINT("Timeout\n");
      if (p->break_loop) break;
      return pkts;
    }  
    if (res != NT_SUCCESS) {
      NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
      ERROR_PRINT("NetRxGet failed (%s)\n", ErrorBuf);
      return -1;
    }

    DEBUG_PRINT("PACKET RECEIVED!!\n");
    p->md.nt3G->pkts++;

#if NTAPI_VERSION==0
      p->md.nt3G->dropped = NT_NET_GET_STREAM_DROPPED_PKTS(pNetBuf);
#elif NTAPI_VERSION==1
      // p->md.nt3G->dropped = 0; can be deferred to actual call of stat. Does not need netbuf
#else
#error NTAPI_VERSION not supported
#endif

    hdr.caplen = NT_NET_GET_PKT_CAP_LENGTH(pNetBuf) - NT_NET_GET_PKT_DESCR_LENGTH(pNetBuf);
    hdr.len = NT_NET_GET_PKT_WIRE_LENGTH(pNetBuf);
    /* stored length is 8 byte aligned, therefore we may get bigger caplen than wirelen */
    if (hdr.caplen > hdr.len) hdr.caplen = hdr.len;
    
#ifdef CAPTURE_WITHOUT_CRC
    // We need to decrement the lengths with 4 because it should be returned without CRC
    hdr.caplen -= 4;
    hdr.len -= 4;
#endif        
    
    if((p->fcode.bf_insns == NULL) || bpf_filter(p->fcode.bf_insns, (u_char *)NT_NET_GET_PKT_L2_PTR(pNetBuf),  hdr.len, hdr.caplen)) 
    {
      ts = NT_NET_GET_PKT_TIMESTAMP(pNetBuf);
      /* convert base */
      switch (NT_NET_GET_PKT_TIMESTAMP_TYPE(pNetBuf))
      {
	case NT_TIMESTAMP_TYPE_NATIVE:    // 10ns resolution, base of 0
	  if (p->md.nt3G->NativeTimeOffset == 0)
	  {  
	    clock_gettime(CLOCK_REALTIME, &tp);
	    p->md.nt3G->NativeTimeOffset = ((uint64_t)tp.tv_sec * 1000000000 + tp.tv_nsec)/10 - ts;
	  }
	  ts += p->md.nt3G->NativeTimeOffset;
	  break;
	case NT_TIMESTAMP_TYPE_NATIVE_NDIS:    // 10ns resolution, base 1 jan 1601
	  ts -= EPOCH_OFFSET;
	  break;
	case NT_TIMESTAMP_TYPE_NATIVE_UNIX:     // 10ns resolution, base 1 jan 1970
	  // Base is Ok
	  break;
	default:
	  ERROR_PRINT("Time stamp type not supported\n");
#ifndef USE_GET_ONLY_API
	  NT_NetRxRelease(p->md.nt3G->pRxStream, pNetBuf);
#endif
	  return -1;
      }  

      if ((ts-p->md.nt3G->lst_ts)<100000000) {
	p->md.nt3G->tlst.nsec += (long)(ts-p->md.nt3G->lst_ts)*10;
        if (p->md.nt3G->tlst.nsec>999999999) {
          p->md.nt3G->tlst.nsec-=1000000000;
          p->md.nt3G->tlst.sec++;
	}
      } else {
        p->md.nt3G->tlst.sec  = ts / 100000000;
        p->md.nt3G->tlst.nsec = (ts -((uint64_t)p->md.nt3G->tlst.sec*100000000))*10;
      }

      hdr.ts.tv_sec=p->md.nt3G->tlst.sec;
      if (p->md.nt3G->NanoSec) {
        hdr.ts.tv_usec=p->md.nt3G->tlst.nsec;
      } else {
        hdr.ts.tv_usec=p->md.nt3G->tlst.nsec/1000;
      }

      p->md.nt3G->lst_ts = ts;

      if (hdr.caplen > p->md.nt3G->snaplen) hdr.caplen = p->md.nt3G->snaplen;

      callback(user, &hdr, (u_char *)NT_NET_GET_PKT_L2_PTR(pNetBuf));
      pkts++;
    }
    
#ifndef USE_GET_ONLY_API
     NT_NetRxRelease(p->md.nt3G->pRxStream, pNetBuf);
#endif

    if (cnt > 0 && pkts >= cnt) {
      return pkts;
    }
    
  } while (1);
  
  if (p->break_loop) {
    p->break_loop = 0;
    return -2;
  }
  return 0;
}



/* napatech_pcap_read_segment
 *
 * Read some packets and call a callback function on each of these
 * packets.
 * 
 * This interface is used to optimized performance. It uses the 
 * Napatech 3GD Segment API.
 * 
 * Limitation: Cannot merge between adapters
 *
 * Parameters
 * p   - pointer to pcap_t structure for device
 * cnt - number of packets to read, 0 means read as many as possible
 * callback - callback function to be called on each packet
 * user - user details
 * Returns number of packets read or -1 on error
 * Only NTAPI Version 1+ supported
 */
static int napatech_pcap_read_segment( pcap_t *p, int cnt, pcap_handler callback, u_char *user )
{
 int pkts, res = NT_SUCCESS;
 struct pcap_pkthdr hdr;
 uint64_t ts;
 struct timespec tp;
 char ErrorBuf[256];
 int breakCnt=0, tmo;

  CALL_TRACE();

  if (p->md.nt3G->StreamId == (uint32_t)-1)
  {
    ERROR_PRINT("No read support on this device\n");
    return -1;
  }

#if NTAPI_VERSION < 1
  ERROR_PRINT("Segment API selected on device. NTAPI version %i not supported (must be version 1+).\n", NTAPI_VERSION);
  return -1;
#endif  

  /* Please note here that the Rx stream is not opened before the actual request for the first packet */
  if (p->md.nt3G->pRxStream == NULL) {
      res = NT_NetRxOpen(&p->md.nt3G->pRxStream, "libpcap", NT_NET_INTERFACE_SEGMENT, p->md.nt3G->StreamId, p->md.nt3G->HBAllow);
  }

  if (res != NT_SUCCESS) 
  {
    NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
    ERROR_PRINT("NetRxOpen failed (%s)\n", ErrorBuf);
    return -1;
  }

  pkts = 0;
  do {
    if (p->break_loop) break;

    // if cnt < 0 we want to do returns regularily. Apps like wireshark needs to return to update the screen
    if (cnt < 0) {
      breakCnt++;
      if ((pkts && breakCnt > 2) || breakCnt > 30) {
        return pkts;
      }
    }

    tmo = p->md.nt3G->nonblocking?0:100;

    /* Do we need a new segment ? */
    if (p->md.nt3G->segNetBuf == NULL) {
      res = NT_NetRxGet(p->md.nt3G->pRxStream, &p->md.nt3G->segNetBuf, tmo);

      if (res == NT_STATUS_TRYAGAIN)
        continue;
      if (res == NT_STATUS_TIMEOUT)
        {
          DEBUG_PRINT("Timeout\n");
          if (p->break_loop) break;
          return pkts;
        }  
      if (res != NT_SUCCESS) {
        NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
        ERROR_PRINT("NetRxGet failed (%s)\n", ErrorBuf);
        return -1;
      }
      /* if pcap timestamp, we also have a pcap header */
      if (NT_NET_GET_SEGMENT_TIMESTAMP_TYPE(p->md.nt3G->segNetBuf) == NT_TIMESTAMP_TYPE_PCAP ||
          NT_NET_GET_SEGMENT_TIMESTAMP_TYPE(p->md.nt3G->segNetBuf) == NT_TIMESTAMP_TYPE_PCAP_NANOTIME) {
        p->md.nt3G->pPCAPHdr = (struct ntpcap_hdr_s *)p->md.nt3G->segNetBuf->hHdr;
        p->md.nt3G->segLength = NT_NET_GET_SEGMENT_LENGTH(p->md.nt3G->segNetBuf);
      } else {
        _nt_net_build_pkt_netbuf(p->md.nt3G->segNetBuf, &p->md.nt3G->NetBuf);
      }
    }

    if (NT_NET_GET_SEGMENT_TIMESTAMP_TYPE(p->md.nt3G->segNetBuf) == NT_TIMESTAMP_TYPE_PCAP ||
        NT_NET_GET_SEGMENT_TIMESTAMP_TYPE(p->md.nt3G->segNetBuf) == NT_TIMESTAMP_TYPE_PCAP_NANOTIME) {

      /* Loop packets in current segment */
      while (p->md.nt3G->segLength > 0) {

        hdr.caplen = p->md.nt3G->pPCAPHdr->caplen;
        hdr.len = p->md.nt3G->pPCAPHdr->wirelen;

        if (hdr.len) {
          DEBUG_PRINT("PACKET RECEIVED!!\n");
          p->md.nt3G->pkts++;
        }

        /* stored length is 8 byte aligned, therefore we may get bigger caplen than wirelen */
        if (hdr.caplen > hdr.len) hdr.caplen = hdr.len;
    
#ifdef CAPTURE_WITHOUT_CRC
        if (hdr.len) {
          // We need to decrement the lengths with 4 because it should be returned without CRC
          hdr.caplen -= 4;
          hdr.len -= 4;
        }
#endif        
 
        if (hdr.len && 
            ((p->fcode.bf_insns == NULL) || 
             bpf_filter(p->fcode.bf_insns, (u_char *)p->md.nt3G->pPCAPHdr + sizeof(struct ntpcap_hdr_s),  hdr.len, hdr.caplen))) {
          hdr.ts.tv_sec = p->md.nt3G->pPCAPHdr->ts.sec;
          hdr.ts.tv_usec = p->md.nt3G->pPCAPHdr->ts.usec;
          if (hdr.caplen > p->md.nt3G->snaplen) hdr.caplen = p->md.nt3G->snaplen;
          callback(user, &hdr, (u_char *)p->md.nt3G->pPCAPHdr + sizeof(struct ntpcap_hdr_s));
          pkts++;
        }
        p->md.nt3G->segLength -= p->md.nt3G->pPCAPHdr->caplen + sizeof(struct ntpcap_hdr_s);
        if (p->md.nt3G->segLength <= 0) {
           NT_NetRxRelease(p->md.nt3G->pRxStream, p->md.nt3G->segNetBuf);
           p->md.nt3G->segNetBuf = NULL; 
        } else {
          p->md.nt3G->pPCAPHdr = (struct ntpcap_hdr_s *)((uint8_t *)p->md.nt3G->pPCAPHdr + p->md.nt3G->pPCAPHdr->caplen + sizeof(struct ntpcap_hdr_s));
        }

        if (cnt > 0 && pkts >= cnt) return pkts;
      }

      if (p->md.nt3G->segNetBuf) {
        NT_NetRxRelease(p->md.nt3G->pRxStream, p->md.nt3G->segNetBuf);
        p->md.nt3G->segNetBuf = NULL; 
      }

    } else {
      /* NT timestamp in packets */

      /* Loop packets in current segment */
      do {

        hdr.caplen = NT_NET_GET_PKT_CAP_LENGTH((&p->md.nt3G->NetBuf)) - NT_NET_GET_PKT_DESCR_LENGTH((&p->md.nt3G->NetBuf));
        hdr.len = NT_NET_GET_PKT_WIRE_LENGTH((&p->md.nt3G->NetBuf));
  
        if (hdr.len) {
          DEBUG_PRINT("PACKET RECEIVED\n");
          p->md.nt3G->pkts++;
        }

        /* stored length is 8 byte aligned, therefore we may get bigger caplen than wirelen */
        if (hdr.caplen > hdr.len) hdr.caplen = hdr.len;
    
#ifdef CAPTURE_WITHOUT_CRC
        if (hdr.len) {
          // We need to decrement the lengths with 4 because it should be returned without CRC
          hdr.caplen -= 4;
          hdr.len -= 4;
        }
#endif        
    
        if (hdr.len &&
            ((p->fcode.bf_insns == NULL) || bpf_filter(p->fcode.bf_insns, (u_char *)NT_NET_GET_PKT_L2_PTR((&p->md.nt3G->NetBuf)), hdr.len, hdr.caplen))) {
          ts = NT_NET_GET_PKT_TIMESTAMP((&p->md.nt3G->NetBuf));

          /* convert base */
          switch (NT_NET_GET_PKT_TIMESTAMP_TYPE((&p->md.nt3G->NetBuf)))
          {
            case NT_TIMESTAMP_TYPE_NATIVE:    // 10ns resolution, base of 0
              if (p->md.nt3G->NativeTimeOffset == 0)
                {  
                  clock_gettime(CLOCK_REALTIME, &tp);
                  p->md.nt3G->NativeTimeOffset = ((uint64_t)tp.tv_sec * 1000000000 + tp.tv_nsec)/10 - ts;
                }
              ts += p->md.nt3G->NativeTimeOffset;
              break;
            case NT_TIMESTAMP_TYPE_NATIVE_NDIS:    // 10ns resolution, base 1 jan 1601
              ts -= EPOCH_OFFSET;
              break;
            case NT_TIMESTAMP_TYPE_NATIVE_UNIX:     // 10ns resolution, base 1 jan 1970
              // Base is Ok
              break;
            default:
              ERROR_PRINT("Time stamp type not supported\n");
              if (p->md.nt3G->segNetBuf) {
                NT_NetRxRelease(p->md.nt3G->pRxStream, p->md.nt3G->segNetBuf);
                p->md.nt3G->pRxStream = NULL;
              }
              return -1;
          }  

          if ((ts-p->md.nt3G->lst_ts)<100000000) {
            p->md.nt3G->tlst.nsec += (long)(ts-p->md.nt3G->lst_ts)*10;
            if (p->md.nt3G->tlst.nsec>999999999) {
              p->md.nt3G->tlst.nsec-=1000000000;
              p->md.nt3G->tlst.sec++;
            }
          } else {
            p->md.nt3G->tlst.sec  = ts / 100000000;
            p->md.nt3G->tlst.nsec = (ts -((uint64_t)p->md.nt3G->tlst.sec*100000000))*10;
          }

          hdr.ts.tv_sec=p->md.nt3G->tlst.sec;
          if (p->md.nt3G->NanoSec) {
            hdr.ts.tv_usec=p->md.nt3G->tlst.nsec;
          } else {
            hdr.ts.tv_usec=p->md.nt3G->tlst.nsec/1000;
          }

          p->md.nt3G->lst_ts = ts;
          if (hdr.caplen > p->md.nt3G->snaplen) hdr.caplen = p->md.nt3G->snaplen;
          callback(user, &hdr, (u_char *)NT_NET_GET_PKT_L2_PTR((&p->md.nt3G->NetBuf)));
          pkts++;
        }

        if (!_nt_net_get_next_packet(p->md.nt3G->segNetBuf, NT_NET_GET_SEGMENT_LENGTH(p->md.nt3G->segNetBuf), &p->md.nt3G->NetBuf)) {
          NT_NetRxRelease(p->md.nt3G->pRxStream, p->md.nt3G->segNetBuf);
          /* Indicate that this segment is empty and we need a new one */
          p->md.nt3G->segNetBuf = NULL; 
        }
        if (cnt > 0 && pkts >= cnt) {
          DEBUG_PRINT("RETURN PACKET %i\n",pkts);
          return pkts;
        }
      } while (p->md.nt3G->segNetBuf);
    }
  } while (1);

  if (p->break_loop) {
    p->break_loop = 0;
    return -2;
  }
  return 0;
}



static int napatech_pcap_inject(pcap_t *p, const void *buf, size_t size)
{
 NtNetBuf_t pNetBuf; 
 int res, Retry;
 
 CALL_TRACE();

  do {
      Retry = 0;   
      res = NT_NetTxGet(p->md.nt3G->pTxStream, &pNetBuf, p->md.nt3G->TxPort, size + 4, NT_NETTX_PACKET_OPTION_DEFAULT, 1000);
      if (res == NT_SUCCESS)
      {  
	/**
	* It has been proven to be faster to only change what is needed. Writes are very costly.
	*/
	
	if((size&0x3)==0) 
	{ // Length must be multiple of 8
	  uint16_t i;
          uint64_t *pDest = (uint64_t *)NT_NET_GET_PKT_L2_PTR(pNetBuf);
	  uint64_t *pSrc = (uint64_t *)buf;
	  for(i=0; i<size>>3; i++) {
	    if(*(pDest+i) != *(pSrc+i)) {
	      *(pDest+i) = *(pSrc+i);
	    }
	  }
	} 
	else 
          memcpy((void *)NT_NET_GET_PKT_L2_PTR(pNetBuf), buf, size);
	
        NT_NET_SET_PKT_CALC_L2_CRC(pNetBuf,1);
	NT_NET_SET_PKT_TXNOW(pNetBuf,1);
	NT_NetTxRelease(p->md.nt3G->pTxStream, pNetBuf);
	return size;
      }
      if (res == NT_STATUS_TIMEOUT || res == NT_STATUS_TRYAGAIN) Retry = 1;
      else
      if (res == NT_ERROR_NTAPI_NET_TX_PKT_SIZE_INVALID)
	DEBUG_PRINT("ERROR: Packet size not supported");
      else	  
	DEBUG_PRINT("ERROR: while requesting Tx buffer.\n");
  }    
  while (Retry);  
  
 return 0;  
}


static int napatech_pcap_inject_nosupport(pcap_t *p, const void *buf, size_t size)
{
  fprintf(stderr, "TX is not supported by this device\n");
  return -1;
}
  
static int napatech_set_datalink( pcap_t *p, int dlt )
{
  DEBUG_PRINT("napatech_set_datalink\n");  
	/* Stub implementation keeps ethereal happy */
  return(0);
}

/* napatech_pcap_setfilter
 *
 * Set the BPF filter for a pcap packet stream.
 *
 * Parameters
 * p   - pointer to pcap_t structure for device
 * fp  - BPF program
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setfilter( pcap_t *p, struct bpf_program *fp )
{
  CALL_TRACE();
  /* Check for a 'NULL' filter - just a single return instruction filter */
  if(fp != NULL) {
    if(fp->bf_len == 1 && fp->bf_insns != NULL) {
      if (fp->bf_insns->code == BPF_RET)  {
        /* Looks like a 'NULL' filter */
        return 0;
      }
    }
  }
  
	/* user-mode filtering */
	/* check we have a pcap handle */
	if(!p) {
		return -1;
	}
	/* check for a filter */
	if(!fp) {
//		DEBUG_PRINT("setfilter: No filter specified");
		return -1;
	}

	/* make private copy of the filter */
	if(install_bpf_program(p, fp) < 0) {
		return -1;
	}
	
	/* set user-mode filter flag */
	p->md.use_bpf = 0;
  return 0;
}

/* napatech_pcap_cleanup
 *
 * Close an pcap packet stream on a Napatech device.
 *
 * Parameters
 * p - pointer to pcap_t structure for device to close
 */
static void napatech_pcap_cleanup( pcap_t *p )
{
  pcap_t *prev=NULL, *cur=pHead;
  
  CALL_TRACE();
  
  if (p->md.nt3G->pRxStream)
  {  
    if (p->md.nt3G->segNetBuf) {
      NT_NetRxRelease(p->md.nt3G->pRxStream, p->md.nt3G->segNetBuf);
    }
    NT_NetRxClose(p->md.nt3G->pRxStream);
    p->md.nt3G->pRxStream = NULL;
  }  

  if (p->md.nt3G->pTxStream)
  {  
    NT_NetTxClose(p->md.nt3G->pTxStream);
    p->md.nt3G->pTxStream = NULL;
  }  

  if (p->md.nt3G->pStatStream) {
    NT_StatClose(p->md.nt3G->pStatStream);
    p->md.nt3G->pStatStream = NULL;
  }


  pthread_mutex_lock(&LstMutex);
  // Remove the element from the list.
  while (cur && (cur != p)) 
  {
    prev=cur;
    cur=cur->md.nt3G->pNext;
  }
  
  if(cur==p) 
  {
    if(prev==NULL)  pHead = cur->md.nt3G->pNext;
    else 
      prev->md.nt3G->pNext = cur->md.nt3G->pNext;
  }
  
  pthread_mutex_unlock(&LstMutex);

  if (p->md.nt3G)
  {  
    free(p->md.nt3G);
    p->md.nt3G = NULL;
  }  
  
  pcap_cleanup_live_common(p);
}


/* napatech3G_pcap_findalldevs
 *
 * Populate device list with napatech devices.  
 *
 * Parameters
 * alldevsp - list to put devices in
 * errbuf   - error reporting buffer
 * Returns 0 on success, -1 on failure
 */

int napatech3G_platform_finddevs( pcap_if_t **devlistp, char *errbuf )
{
  char virtName[MAX_DEV_NAME_SIZE];
  char description[128];
  int cnt, res;

  CALL_TRACE();

  if ((res = InitializeNT3G()) != NT_SUCCESS)
  {  
    if (res == ERROR_3GD_NOT_LOADED) {
      return 0; // nt3gd not loaded
    }
    NT_ExplainError(res, errbuf, PCAP_ERRBUF_SIZE);
    return -1;
  }  

  for (cnt = 0; cnt < SetupStreamCount; cnt++)
  { 
    if (SetupStreams[cnt].TxPort > -1)
    {  
      if (SetupStreams[cnt].StreamId != (uint32_t)-1)
        snprintf(description, sizeof(description), "NT 3G configured device. Stream %d. Transmit %d", SetupStreams[cnt].StreamId, SetupStreams[cnt].TxPort);
      else
        snprintf(description, sizeof(description), "NT 3G configured device. Transmit only %d",  SetupStreams[cnt].TxPort);
    }  
    else
    if (SetupStreams[cnt].StreamId != (uint32_t)-1)
       snprintf(description, sizeof(description), "NT 3G configured device Stream %d. No transmit", SetupStreams[cnt].StreamId);
    else 
      continue;
    
    if(pcap_add_if(devlistp, SetupStreams[cnt].IniSecName, 0, description, errbuf) == -1) {
      /* Failure. */
      ERROR_PRINT("Failed adding virtual device [%s]\n", SetupStreams[cnt].IniSecName); 
      return -1;
    }
  }

  /* Read setup streams from ntservice and filter out the streamids setup by ini-file */
  
  for (cnt = 0; cnt < DefStreams.SInfo.u.stream.data.count; cnt++)
  {
    if (!StreamIdDefinedInSetupStream(DefStreams.SInfo.u.stream.data.streamIDList[cnt]))
    {  
      snprintf(description, sizeof(description), "NT 3G predefined stream. Stream %d (no transmit)", DefStreams.SInfo.u.stream.data.streamIDList[cnt]);
      snprintf(virtName, MAX_DEV_NAME_SIZE, "nt3g%d", DefStreams.SInfo.u.stream.data.streamIDList[cnt]);
      
      if(pcap_add_if(devlistp, virtName, 0, description, errbuf) == -1) {
	/* Failure. */
	ERROR_PRINT("Failed adding virtual device %s\n", virtName); 
	return -1;
      }
    }
  }
  return 0;
}

/* napatech_pcap_setbuff
 *
 * Set the size of the packet recieve buffer. Not necessary for Napatech
 * devices, but the stub function must be in the common code to aid
 * porting to different versions of LibPCap.
 *
 * Parameters
 * p    - pointer to pcap_t structure for device
 * size - size of buffer in bytes
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setbuff( pcap_t *p, int size )
{
  CALL_TRACE();
  DEBUG_PRINT("napatech_pcap_setbuff: Not required by Napatech devices");
  return 0;
}

/* napatech_pcap_setmintocopy
 *
 * Set the minimum amount of data to read to release a read call.
 * Not necessary for Napatech devices, but the stub function must be in the
 * common code to aid porting to different versions of LibPCap.
 *
 * Parameters
 * p    - pointer to pcap_t structure for device
 * size - size of minimum read in bytes
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setmintocopy( pcap_t *p, int size )
{
  CALL_TRACE();
  DEBUG_PRINT("napatech_pcap_setmintocopy: Not supported by Napatech devices");
  return 0;
}





static int napatech_pcap_setnonblock(pcap_t *p, int nonblock, char *errbuf)
{
  DEBUG_PRINT("set nonblock %i\n", nonblock);
  p->md.nt3G->nonblocking = nonblock;
  return 0;
}


static int napatech_pcap_getnonblock(pcap_t *p, char *errbuf)
{
  return p->md.nt3G->nonblocking;
}


/* napatech_activate
 *
 * Activate the pcap device
 */
static int napatech_activate(pcap_t *p)
{
 struct SetupStream_s *pS;
 const char *device = p->opt.source;
 char ErrorBuf[256];
 int res;
 int NTInterface;
  
  CALL_TRACE();

  if(is_napatech3G_device(device)) 
  {    
    if ((p->md.nt3G = malloc(sizeof(struct napatech3G_details))) == NULL) 
      return -1;
    
    memset(p->md.nt3G, 0, sizeof(struct napatech3G_details));
    
    if ((pS = NameDefinedInSetupStream(device)) != NULL)
    {  
      p->md.nt3G->StreamId = pS->StreamId;
      p->md.nt3G->TxPort = pS->TxPort;
      p->md.nt3G->HBAllow = pS->HBAllow;
      p->md.nt3G->TxNumaNode = pS->TxNumaNode;
      p->md.nt3G->TxMinHostBuffer = pS->TxMinHostBuffer;
      p->md.nt3G->NanoSec = pS->NanoSec;
      NTInterface = pS->NTInterface;
    }  
    else
    {  
      p->md.nt3G->StreamId = ExtractStreamIdFrom_nt3g_Device(device);
      p->md.nt3G->TxPort = -1;
      p->md.nt3G->HBAllow = -1;
      p->md.nt3G->TxNumaNode = 0;
      p->md.nt3G->TxMinHostBuffer = 0;
      p->md.nt3G->NanoSec = 0;
      NTInterface = SEGMENT_IF;
    }  

    if (timestampType == NT_TIMESTAMP_TYPE_PCAP_NANOTIME) {
      p->md.nt3G->NanoSec = 1; // Implicit nanotime in header
    }
    
    p->md.NanoSec = p->md.nt3G->NanoSec; // for use in sf-pcap.c

    p->md.nt3G->pkts = 0;
    p->md.nt3G->dropped = 0;
    p->md.nt3G->snaplen = p->snapshot;
    p->md.nt3G->promisc = p->opt.promisc;  // promisc is always on
    p->md.nt3G->to_ms   = p->md.timeout;
    p->md.nt3G->nonblocking = 0;


    DEBUG_PRINT("snaplen %i, promisc %i, timeout %i\n", p->snapshot, p->opt.promisc,p->md.timeout);     

    /*a Napatech adapter is assumed to be ethernet*/
    p->linktype = DLT_EN10MB;
    
    p->selectable_fd = -1;

    if (NTInterface == PACKET_IF) {
      p->read_op = napatech_pcap_read_packet;
    } else
    if (NTInterface == SEGMENT_IF) {
      p->read_op = napatech_pcap_read_segment;
    } else {
      ERROR_PRINT("Internal error! invalid NTInterface\n");
      return -1;
    }

    p->setfilter_op = napatech_pcap_setfilter;
    p->setdirection_op = NULL;  /* Cannot set direction */
    p->set_datalink_op = napatech_set_datalink;	/* set link type */
    p->getnonblock_op = napatech_pcap_getnonblock;
    p->setnonblock_op = napatech_pcap_setnonblock;


    p->stats_op = napatech_pcap_stats;
    p->cleanup_op = napatech_pcap_cleanup;
    p->inject_op = napatech_pcap_inject_nosupport;
    
    if (p->md.nt3G->TxPort >= 0)
    {
#if NTAPI_VERSION==0
	res = NT_NetTxOpen(&p->md.nt3G->pTxStream, "libpcap",0x01<<p->md.nt3G->TxPort, p->md.nt3G->TxNumaNode);
#elif NTAPI_VERSION==1
	res = NT_NetTxOpen(&p->md.nt3G->pTxStream, "libpcap",0x01<<p->md.nt3G->TxPort, p->md.nt3G->TxNumaNode, p->md.nt3G->TxMinHostBuffer); 
#else
#error NTAPI_VERSION not supported
#endif
      if (res == NT_SUCCESS) {
        p->inject_op = napatech_pcap_inject;
      } else {
        NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
	ERROR_PRINT("ERROR while opening stream for transmit! (%s)\n", ErrorBuf);
      }
    }

    res = NT_StatOpen(&p->md.nt3G->pStatStream, "libpcap");
    if (res != NT_SUCCESS) {
      NT_ExplainError(res, ErrorBuf, sizeof(ErrorBuf));
      ERROR_PRINT("ERROR while opening statistics stream! (%s)\n", ErrorBuf);
    }

    pthread_mutex_lock(&LstMutex);
    // Add entry to the list.
    p->md.nt3G->pNext = pHead;
    pHead = p;
    pthread_mutex_unlock(&LstMutex);

  }
  return 0;
}


#endif /*NAPATECH_SUPPORT*/
