#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_NAPATECH_API
#include "pcap-napatech.h"
#include "pcap-int.h"
#include "basetype.h"
#include "NTCommonInterface.h"
#include "NTCI_packetfeed.h"
#include "packetdescriptor.h"
#include "packetfeedstructure.h"
#include "NTCI_capabilities.h"
#include "capabilitiesstructure.h"
#include "NTCI_cardconfiguration.h"
#include "cardconfigurationstructure.h"
#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <sys/sem.h>
#include <syslog.h>
#include <pcap-napatech-intl.h>

// This code has been taken from the linux kernel
/*
 * i386 calling convention returns 64-bit value in edx:eax, while
 * x86_64 returns at rax. Also, the "A" constraint does not really
 * mean rdx:rax in x86_64, so we need specialized behaviour for each
 * architecture
 */
#if _NT_HOST_CPU_POINTER_SIZE==_NT_HOST_CPU_POINTER_SIZE_64
#define DECLARE_ARGS(val, low, high)	unsigned low, high
#define EAX_EDX_VAL(val, low, high)	((low) | ((uint64_t)(high) << 32))
#define EAX_EDX_ARGS(val, low, high)	"a" (low), "d" (high)
#define EAX_EDX_RET(val, low, high)	"=a" (low), "=d" (high)
#else
#define DECLARE_ARGS(val, low, high)	unsigned long long val
#define EAX_EDX_VAL(val, low, high)	(val)
#define EAX_EDX_ARGS(val, low, high)	"A" (val)
#define EAX_EDX_RET(val, low, high)	"=A" (val)
#endif

static inline unsigned long long __rdtscll(void)
{
	DECLARE_ARGS(val, low, high);

	asm volatile("rdtsc" : EAX_EDX_RET(val, low, high));

	return EAX_EDX_VAL(val, low, high);
}

static int napatech_activate(pcap_t *p);

/* struct napatech_details
 *
 * This is an additional part of the pcap_t structure, which we use to
 * hold the details of the packet feeds we create.
 */

struct napatech_details
{
  /* Virtual device name. */
  char virtName[MAX_DEV_NAME_SIZE];

  /* The card handle of the adapter containing the virtual adapter. */
  handle_t cardHandle;
  
  /* Feed handle associated with the packet feed */
  FeedHandle feedHandle;

  /* Buffer base address. */
  uint8_t *bufferBase;
  
  /* Open Live parameters. */
  int snaplen;
  int promisc;
  int to_ms;
  int nonblocking;

  /* Flag to indicate whether the feed is being closed */
  volatile int closing;

  struct {
    /* Current buffer section information */
    bool_t        holdingBuffer;
    BufferInformationSectionType8_t sectionInfo;
    uint32_t      framesAvailable;
    uint64_t      lastTime;
    
    /* Section details */
    uint32_t* currentPacket;
  }Rx;

  struct {
    pthread_t threadHandle;
    enum {
      THREAD_DESTROYED,
      THREAD_STOPPED,
      THREAD_QUIT,
      THREAD_START,
      THREAD_RUNNING,
      THREAD_DONE
    } threadRunFlag;
    unsigned long long pktDuration;       //!< The number of 10ns ticks it take to send the previous packet
    unsigned long long cyclesPerMicroSec; //!< CPU cycles per usec.
    unsigned long long firstCycles;       //!< The cycles count of first packet in a segment
    unsigned long long prevCycles;        //!< The previous packet cycles count
    uint32_t segmentSize;                 //!< The segment size
    uint64_t timestampLastPacket;         //!< The timestamp of the previous packet
    uint64_t timestamp;                   //!< The timestamp to patch into packets
    BufferInformationSectionType3_t  bufferSectionTx;
    BufferRequestSampledType2_t      request;
    PacketDescriptorNtSeriesType1_t *pCurDescriptor;
    uint32_t ifSpeed; //!< The interface speed.
    int txSupported;
    int pktTsLast; //!< Packets are timestamped when the last octet is received
  }Tx;

  /* Packet feed statistics */
  uint32_t packetsRecv;
  uint32_t packetsDrop;

  /* Mutex. */
  pthread_mutex_t mutex;

  /* List */
  pcap_t *pNext;
};

static void WriteToLog(bool_t boError, char *file, const char *function, uint32_t line, const char *format, ... )
{
  va_list ap;
  char *logStr = (char*)malloc(1024);
  va_start(ap,format);

  vsprintf(logStr, format, ap);

  if(file!=NULL) {
    syslog(boError?LOG_ERR:LOG_INFO, "NT: %s %d: %s: %s", file, line, function, logStr);
  } else {
    syslog(boError?LOG_ERR:LOG_INFO, "NT: %s", logStr);
  }
  if(boError) {
    if(file!=NULL) {
      fprintf(stderr, "NT: %s %d: %s: %s", file, line, function, logStr);
    } else {
      fprintf(stderr, "NT: %s", logStr);
    }
  }
  free(logStr);
}

static void WriteFailedApiToLog(handle_t cardHandle, char *file, uint32_t line, const char * CommandName)
{
  uint32_t HighLevelError;
  uint32_t LowLevelError;
  char errorBuffer[1000];
  HighLevelError = NTCI_GetManagerLastError();
  LowLevelError = NTCI_GetManagerLastLowLevelError();

  errorBuffer[0]='\0';
  if( HighLevelError == NTCI_ERRCODE_DRIVER_ERROR )
  {
    NTCI_ExplainLastCardError(cardHandle, errorBuffer, sizeof(errorBuffer));
    syslog(LOG_ERR, "NT: %s %d: Command %s failed: with driver errorcode 0x%08X: %s\n", file, line, CommandName, LowLevelError, errorBuffer);
    fprintf(stderr, "NT: %s %d: Command %s failed: with driver errorcode 0x%08X: %s\n", file, line, CommandName, LowLevelError, errorBuffer);
  } else {
    syslog(LOG_ERR, "NT: %s %d: Command %s failed: High level error %d (low level error 0x%08X)\n", file, line, CommandName, HighLevelError,LowLevelError);
    fprintf(stderr, "NT: %s %d: Command %s failed: High level error %d (low level error 0x%08X)\n", file, line, CommandName, HighLevelError,LowLevelError);
  }
}

#ifdef NAPATECH_CALL_TRACE
#define CALL_TRACE()                     WriteToLog(0, __FILE__, __FUNCTION__, __LINE__,"\n")
#else
#define CALL_TRACE() 
#endif

#ifdef NAPATECH_DEBUG
#define DEBUG_PRINT(string, args...)     WriteToLog(0, __FILE__, __FUNCTION__, __LINE__, string, ## args)
#else
#define DEBUG_PRINT(string, args...)
#endif

#define WARNING_PRINT(string,  args...)    WriteToLog(1, NULL, NULL, 0, string, ## args)
#define ERROR_PRINT(string,  args...)    WriteToLog(1, __FILE__, __FUNCTION__, __LINE__, string, ## args)
#define REPORT_NTCI_FAILURE(__cardHandle__, __CommandName__) WriteFailedApiToLog(__cardHandle__, __FILE__, __LINE__, __CommandName__)

static void napatech_pcap_cleanup( pcap_t *p );

// List of all pcap instances created. Use to destroy them on "atexit"
static int listMutexCreated = 0;
pthread_mutex_t mList; //!< List mutex
pcap_t *pHead=NULL;
int atexithandlerRegistered=0;
static void atexit_handler(void)
{
  pcap_t *cur;

  pthread_mutex_lock(&mList);
  cur=pHead;
  pthread_mutex_unlock(&mList);
  
  // Close all nt pcap instances
  while(cur) {
    // napatech_pcap_cleanup will take cur out of the list
    napatech_pcap_cleanup(cur);
    pthread_mutex_lock(&mList);
    cur=pHead;
    pthread_mutex_unlock(&mList);
  }
  // Destroy the mutex. We are done
  if(listMutexCreated) {
    pthread_mutex_destroy(&mList);
    listMutexCreated=0;
  }
}

#ifdef HAVE_NEW_CARDCAP_STRUCT     
static int MutexGet(pcap_t *p)
{
  while(p->md.ntDetails->closing==FALSE){
    if(pthread_mutex_lock(&p->md.ntDetails->mutex) == 0) {
      /* We got the mutex. */
      return 0;
    }
    /* The wait call was interrupted. This could be beacuse the system("clear") call
       is making an interrupt. Check if we are still running. . */
  }
  return -EINTR;
}

static int MutexPost(pcap_t *p)
{
  while(p->md.ntDetails->closing==FALSE) {
    if(pthread_mutex_unlock(&p->md.ntDetails->mutex) == 0) {
      return 0;
    }
    /* The wait call was interrupted. This could be beacuse the system("clear") call
       is making an interrupt. Check if we are still running. . */
  }
  return -EINTR;
}


static int InjectPacket(pcap_t *p, const void *buf, size_t size)
{
  uint32_t result;
  uint32_t storedLength;
  int releaseSegment=0;
  unsigned long long curCycles;
  uint64_t *pSrc, *pDest;
  unsigned long long dT;
  int firstpacket=0;
//  CALL_TRACE();

  // Read current "timestamp".
  curCycles=__rdtscll();

  storedLength = ((sizeof(PacketDescriptorNtSeriesType1_t)+size+4+7)&~7);
  // Release segment if this is a timeout or there is not enough space left for the packet
  if(p->md.ntDetails->Tx.pCurDescriptor != NULL) {
    if((buf==NULL) && (size==0) &&
       ((((curCycles - p->md.ntDetails->Tx.firstCycles)/p->md.ntDetails->Tx.cyclesPerMicroSec) >= TX_SEGMENT_TIMEOUT_MICROSEC))) {
      releaseSegment = 1;
    } else if((p->md.ntDetails->Tx.bufferSectionTx.numBytesAvailable+storedLength) > p->md.ntDetails->Tx.segmentSize) {
      releaseSegment = 1;
    } else if((buf==NULL) && (size==-1)) {
      releaseSegment = 1;
    }
    
    if(releaseSegment) {
      p->md.ntDetails->Tx.bufferSectionTx.transmitFlag = 1;
      p->md.ntDetails->Tx.pCurDescriptor = NULL;
      result = NTCI_ReleaseBuffer(p->md.ntDetails->cardHandle,
                                  p->md.ntDetails->feedHandle,
                                  STID_BUFFER_INFORMATION_SECTION_TYPE3,
                                  &p->md.ntDetails->Tx.bufferSectionTx);      
    }
  }

  if((buf==NULL) || (size < 1)) {
    p->md.ntDetails->Tx.prevCycles=curCycles;
    return 0;
  }
  
  // Check if we already have a segment to use
  if(p->md.ntDetails->Tx.pCurDescriptor == NULL) {
    // Force wait for new segment
    p->md.ntDetails->Tx.request.common.waitTimeoutMS = GET_NEXT_BUFFER_BLOCKING;

    result = NTCI_GetNextBuffer(p->md.ntDetails->cardHandle,
                                p->md.ntDetails->feedHandle,
                                STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                &p->md.ntDetails->Tx.request,
                                STID_BUFFER_INFORMATION_SECTION_TYPE3,
                                &p->md.ntDetails->Tx.bufferSectionTx);

    if(result == NTCI_ERRCODE_SUCCESS) {
      /* Insert packet into the transmit-only buffer */
      p->md.ntDetails->Tx.pCurDescriptor =
        (PacketDescriptorNtSeriesType1_t *)(p->md.ntDetails->bufferBase + p->md.ntDetails->Tx.bufferSectionTx.sectionOffset);
      p->md.ntDetails->Tx.segmentSize = p->md.ntDetails->Tx.bufferSectionTx.numBytesAvailable;
      // Segment now contains no data.
      p->md.ntDetails->Tx.bufferSectionTx.numBytesAvailable = 0;
      // Store cycles of the first packet
      p->md.ntDetails->Tx.firstCycles = curCycles;
      p->md.ntDetails->Tx.prevCycles = curCycles;
      firstpacket=1;
    } else if(result == NTCI_STATUSCODE_NO_PACKETS_AVAILABLE) {
      // We cannot send. Return 0 bytes sent.
      p->md.ntDetails->Tx.prevCycles=curCycles;
      return 0;
    } else {
      // Error.
      REPORT_NTCI_FAILURE(p->md.ntDetails->cardHandle, "NTCI_GetNextBuffer");
      p->md.ntDetails->Tx.prevCycles=curCycles;
      return -1;
    }
  }

  //Clear the header
  *((uint64_t*)p->md.ntDetails->Tx.pCurDescriptor)=0;
  *((uint64_t*)p->md.ntDetails->Tx.pCurDescriptor+1)=0;

  if (firstpacket)
    p->md.ntDetails->Tx.pCurDescriptor->txNow=1;

  // Inject the packet.
  //+20 because of default 12 IPG and 8 PREEAMPLE
  // Timestamps are sampled on the last packet leaving the adapter
  if(p->md.ntDetails->Tx.ifSpeed==10000) {
    p->md.ntDetails->Tx.pktDuration=(((uint64_t)size+4+20)*328)>>12; //  multiply size with 0.08 = 10ns ticks
  } else if(p->md.ntDetails->Tx.ifSpeed==1000) {
    p->md.ntDetails->Tx.pktDuration=(((uint64_t)size+4+20)*3277)>>12;//  multiply size with 0.8  = 10ns ticks        
  } else if(p->md.ntDetails->Tx.ifSpeed==100) {
    p->md.ntDetails->Tx.pktDuration=(((uint64_t)size+4+20)*8);       //  multiply size with 8    = 10ns ticks    
  } else {
    p->md.ntDetails->Tx.pktDuration=(((uint64_t)size+4+20)*80);      //  multiply size with 80   = 10ns ticks
  }

  if(p->md.ntDetails->Tx.pktTsLast) {
    // Update the timestamp with the function call delta and packet duration
    if(buf != NULL) {
      // dT in 10ns ticks.
      dT=(((curCycles-p->md.ntDetails->Tx.prevCycles)*100)/p->md.ntDetails->Tx.cyclesPerMicroSec);
      if(dT >= p->md.ntDetails->Tx.pktDuration) {
        // dT is larger or equal to the packet duration so we need to use the delta instead of packet duration
        p->md.ntDetails->Tx.timestamp+=dT;      
      } else {
        // The caller is calling faster then the packet duration. This must be equal to line rate tx request
        p->md.ntDetails->Tx.timestamp+=p->md.ntDetails->Tx.pktDuration;	
        p->md.ntDetails->Tx.pCurDescriptor->txNow=1;
      }
    }
    // Update the packet timestamp
    p->md.ntDetails->Tx.pCurDescriptor->timestamp = p->md.ntDetails->Tx.timestamp;
  } else {
    // Update the packet timestamp
    p->md.ntDetails->Tx.pCurDescriptor->timestamp = p->md.ntDetails->Tx.timestamp;

    // Update the timestamp with the function call delta and packet duration
    if(buf != NULL) {
      // dT in 10ns ticks.
      dT=(((curCycles-p->md.ntDetails->Tx.prevCycles)*100)/p->md.ntDetails->Tx.cyclesPerMicroSec);
      if(dT >= p->md.ntDetails->Tx.pktDuration) {
        // dT is larger or equal to the packet duration so we need to use the delta instead of packet duration
        p->md.ntDetails->Tx.timestamp+=dT;      
      } else {
        // The caller is calling faster then the packet duration. This must be equal to line rate tx request
        p->md.ntDetails->Tx.timestamp+=p->md.ntDetails->Tx.pktDuration;	
        p->md.ntDetails->Tx.pCurDescriptor->txNow=1;
      }
    }
  }
  
//  printf("%d %lld #%lld\n", size+4, (unsigned long long)p->md.ntDetails->Tx.pCurDescriptor->timestamp,
//         (unsigned long long)(p->md.ntDetails->Tx.timestamp-p->md.ntDetails->Tx.timestampLastPacket));
//  printf("#%d %d\n", size+4, ((((uint64_t)size+4+20)*8*100)/p->md.ntDetails->Tx.ifSpeed));
  p->md.ntDetails->Tx.pCurDescriptor->storedLength = storedLength;
  p->md.ntDetails->Tx.pCurDescriptor->wireLength = size+4; // Incl. crc.
  p->md.ntDetails->Tx.pCurDescriptor->descriptorType=1;
  // Calc crc.
  p->md.ntDetails->Tx.pCurDescriptor->txCrcOverride=1;
  p->md.ntDetails->Tx.pCurDescriptor->txChannel=atoi(&p->md.ntDetails->virtName[6]);

#ifndef USE_FUNCTION_CALL_IPG
  // Force the use of txNow.
  p->md.ntDetails->Tx.pCurDescriptor->txNow=1;
#endif  
  
  // Copy payload
  /**
   * Is has proven to be faster to only change what is needed. Writes are very costly.
   */
  // Skip descriptors
  pSrc=(uint64_t*)buf;
  pDest=(uint64_t*)((uint8_t*)p->md.ntDetails->Tx.pCurDescriptor+sizeof(PacketDescriptorNtSeriesType1_t));
  if((size&0x3)==0) { // Length must be multiple of 8
    uint16_t i;
    for(i=0; i<size>>3; i++) {
      if(*(pDest+i) != *(pSrc+i)) {
        *(pDest+i) = *(pSrc+i);
      }
    }
  } else {
    memcpy(pDest, pSrc, size);
  }


  // Check if we need to txNow and resync hw.
  if(firstpacket || 
     ((p->md.ntDetails->Tx.timestamp-p->md.ntDetails->Tx.timestampLastPacket) < 1000000)) {
    // Move the descriptor    
    p->md.ntDetails->Tx.pCurDescriptor = (PacketDescriptorNtSeriesType1_t*)
      ((uint8_t*)p->md.ntDetails->Tx.pCurDescriptor+storedLength);
    p->md.ntDetails->Tx.bufferSectionTx.numBytesAvailable += storedLength;    
  } else {
    // if we have more than a 1msec between packets we use TX now.
    p->md.ntDetails->Tx.pCurDescriptor->txNow=1;
    p->md.ntDetails->Tx.bufferSectionTx.transmitFlag = 1;
    p->md.ntDetails->Tx.pCurDescriptor = NULL;
    result = NTCI_ReleaseBuffer(p->md.ntDetails->cardHandle,
                                p->md.ntDetails->feedHandle,
                                STID_BUFFER_INFORMATION_SECTION_TYPE3,
                                &p->md.ntDetails->Tx.bufferSectionTx);    
  }
  
  // Store last timestamp
  p->md.ntDetails->Tx.timestampLastPacket=p->md.ntDetails->Tx.timestamp;

  // Read cycles on exit, so we have no processing penalty. The caller will control the delay
  p->md.ntDetails->Tx.prevCycles=curCycles;

  return size;
}


static int StartPacketFeed(pcap_t *p)
{
  uint32_t result=0, segment=0;
  handle_t cardHandle=NULL;
  FeedDescriptionType1_t description;
  FeedStatusSampledType4_t feedStatus;
  BufferInformationSectionType8_t bufInfo;
  BufferRequestSampledType2_t  bufferRequest;
  FeedBufferInformationType1_t  bufferInfo;
  struct deviceEntry *myDevice=NULL;

  CALL_TRACE();

  bufferRequest.common.waitTimeoutMS = 0;
  
  /* Open the adapter. */
  if((cardHandle = NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, atoi(&p->md.ntDetails->virtName[4]))) == NULL) {
    DEBUG_PRINT("Failed to open device: %s", p->md.ntDetails->virtName);
    return -1;
  }
  
  /* Check that the feed is not already in use. */
  if(NTCI_GetPacketFeedDescription(cardHandle, 
                                   atoi(&p->md.ntDetails->virtName[6]),
                                   STID_FEED_DESCRIPTION_TYPE1,
                                   &description) != NTCI_ERRCODE_SUCCESS) {
    REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetPacketFeedDescription");
    NTCI_CloseCard(cardHandle);
    return -1;    
  }
    
  if(NTCI_GetPacketFeedStatus(cardHandle, description.handle,
                              STID_FEED_STATUS_SAMPLED_TYPE4, &feedStatus) != NTCI_ERRCODE_SUCCESS) {
    REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetPacketFeedStatus");
    NTCI_CloseCard(cardHandle);    
    return -1;
  }
  if(feedStatus.mode == SAMPLED_FEED_STOPPED) {
    /* Check that the feed has all segments "busy". This means that the application has everything. */
    if((feedStatus.numSegmentsReady != 0) ||
       (feedStatus.numSegmentsBusy == 0) ||
       (feedStatus.numSegmentsIdle != 0)) {
      DEBUG_PRINT("%s: Feed not ready %d:%d:%d\n", p->md.ntDetails->virtName,
                  feedStatus.numSegmentsReady, feedStatus.numSegmentsBusy, feedStatus.numSegmentsIdle);
      NTCI_CloseCard(cardHandle);
      return -1;
    }
  } else {
    DEBUG_PRINT("%s: Feed already running\n", p->md.ntDetails->virtName);
    NTCI_CloseCard(cardHandle);
    return -1;    
  }    

  /* All segments allocated with success. Start the packetfeed. */
  if(NTCI_StartPacketFeed(cardHandle, description.handle) != NTCI_ERRCODE_SUCCESS) {
    REPORT_NTCI_FAILURE(cardHandle, "NTCI_StartPacketFeed");
    NTCI_CloseCard(cardHandle);
    return -1;    
  }

  // Find base address.
  if((result = NTCI_GetPacketFeedBuffers(cardHandle, 
                                         description.handle,
                                         STID_FEED_BUFFER_INFORMATION_TYPE1,
                                         &bufferInfo)) != NTCI_ERRCODE_SUCCESS) {
    REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetNextBuffer");
    NTCI_CloseCard(cardHandle);
    return -1;    
  }

  p->md.ntDetails->bufferBase = bufferInfo.bufferBaseAddress;
  
  /* Everything went ok. Save the feed parameters. */
  p->md.ntDetails->cardHandle = cardHandle;
  p->md.ntDetails->feedHandle = description.handle;
  return 0;
}


static THREAD_API NapatechPcapTxHandler(void *data)
{
  pcap_t *p = (pcap_t*)data;
  struct timeval  tWait;
  CALL_TRACE();

  p->md.ntDetails->Tx.threadRunFlag = THREAD_START;

  p->md.ntDetails->Tx.threadRunFlag = THREAD_RUNNING;

  while(p->md.ntDetails->Tx.threadRunFlag != THREAD_QUIT) {
    tWait.tv_usec=TX_SEGMENT_TIMEOUT_MICROSEC;
    tWait.tv_sec=0;
    // Wait then issue a timeout request.
    select(0, NULL, NULL, NULL, &tWait);
    if(MutexGet(p)!=0) {
      p->md.ntDetails->Tx.threadRunFlag = THREAD_STOPPED;  
      break;
    }
    InjectPacket(p, NULL, 0);
    MutexPost(p);
  }
  p->md.ntDetails->Tx.threadRunFlag = THREAD_STOPPED;  
}
#endif

int is_napatech_device( const char *deviceName )
{
  int result = 0;
  handle_t cardHandle=NULL;
  char devName[MAX_DEV_NAME_SIZE];

  CALL_TRACE();
  strncpy(devName, deviceName, MAX_DEV_NAME_SIZE);

  /* Check to see if devName has the following format "ntxc?:?" and is thus a napatech device */
  if(((devName[0] == 'n') || (devName[0] == 'N')) &&
     ((devName[1] == 't') || (devName[1] == 'T')) &&
     ((devName[2] == 'x') || (devName[2] == 'X')) &&
     ((devName[3] == 'c') || (devName[3] == 'C')) &&
     ((devName[5] == ':'))) {
    /* Check that the adapter is found. */
    if((cardHandle = NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, atoi(&devName[4]))) != NULL) {
      /* Ok the card was there, what about the feed?. */
      FeedDescriptionType2_t description;

      DEBUG_PRINT("%s: Checking if feed %d exists.\n", deviceName, atoi(&devName[6]));
      /* Get the description */
      if(NTCI_GetPacketFeedDescription(cardHandle, 
                                       atoi(&devName[6]),
                                       STID_FEED_DESCRIPTION_TYPE2,
                                       &description) == NTCI_ERRCODE_SUCCESS) {
        /* The feed exitst so the device is a Napatech device. */
	result = 1;
        /* Check that the feed is running with PCAP descriptor and PCAP timestamp. */
        if((description.timestampType != TIMESTAMP_PCAP) && (description.timestampType != TIMESTAMP_PCAP_NANOTIME)){
          DEBUG_PRINT("%s is running timestamp type %d. PCAP type is expected. Ignoring device.\n",
                      deviceName, description.timestampType);
          result = 0;
        } else if(description.descriptorType != PCAP_DESCRIPTOR) {
          DEBUG_PRINT("%s is running descriptor type %d. PCAP type is expected. Ignoring device.\n",
                      deviceName, description.descriptorType);
          result = 0;
        }
      }
      NTCI_CloseCard(cardHandle);
    }
  } 
  return result;
}

pcap_t *napatech_create(const char *device, char *ebuf, int *is_ours)
{
  pcap_t *p;

  CALL_TRACE();

  *is_ours = is_napatech_device(device);
  if (!*is_ours) {
    return NULL;
  }
  p = pcap_create_common(device, ebuf);
  if (p == NULL)
    return NULL;

  p->activate_op = napatech_activate;
  return p;
}


/* napatech_pcap_stats
 *
 * Read the statistics of the pcap packet feed.
 *
 * Parameters
 * p  - pointer to pcap_t structure for device
 * ps - pointer to pcap_stat structure to put statistics in
 * Returns 0 on success or -1 on error
 */
static int napatech_pcap_stats( pcap_t *p, struct pcap_stat *ps )
{
  CALL_TRACE();
  /* sizeof(pcap_stat) varies according to whether REMOTE is defined.*/
  /* this can cause some crashes.  commenting this out fixes this.*/
  
  /*memset(ps, 0, sizeof(struct pcap_stat));*/
  
  ps->ps_recv = p->md.ntDetails->packetsRecv; /* number of packets received */
  ps->ps_drop = p->md.ntDetails->packetsDrop; /* number of packets dropped */
  ps->ps_ifdrop = 0; /* drops by interface XXX not yet supported */
  return 0;
}


/* napatech_pcap_read
 *
 * Read some packets and call a callback function on each of these
 * packets.
 *
 * Parameters
 * p   - pointer to pcap_t structure for device
 * cnt - number of packets to read, 0 means read as many as possible
 * callback - callback function to be called on each packet
 * user - user details
 * Returns number of packets read or -1 on error
 */
static int napatech_pcap_read( pcap_t *p, int cnt, pcap_handler callback,
                               u_char *user )
{
  int n = 0;
  int i;
  int result;
//  CALL_TRACE();
  
  if( p->md.ntDetails->closing == TRUE ) {
    ERROR_PRINT("napatech_pcap_read: Packet feed is being closed: feed %d\n", p->md.ntDetails->feedHandle&0xFF);
    snprintf(p->errbuf, PCAP_ERRBUF_SIZE,"napatech_pcap_read: Packet feed is being closed");
    return -2;
  }
  
  if(p->md.ntDetails->feedHandle==0) {
    if(MutexGet(p)!=0) {
      return -2;
    }
    /* The creation start of packetfeeds is a critical section. */
    if((result = StartPacketFeed(p)) != 0) {
      MutexPost(p);
      return result;
    }
    /* Release the mutex again. */
    MutexPost(p);
  }
  
  /* Check if there are any frames ready to be processed */
  if(p->md.ntDetails->Rx.holdingBuffer == FALSE ) {
    uint32_t status;
    BufferRequestSampledType2_t bufferRequest;
    
    /* Configure the maximum wait to be the same as the read timeout */
    if (p->md.ntDetails->nonblocking) {
      bufferRequest.common.waitTimeoutMS = GET_NEXT_BUFFER_POLL;
    } else {
      bufferRequest.common.waitTimeoutMS = GET_NEXT_BUFFER_BLOCKING;
    }    
    /* Get the next section */
    status = NTCI_GetNextBuffer(p->md.ntDetails->cardHandle,
                                p->md.ntDetails->feedHandle,
                                STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                &bufferRequest,
                                STID_BUFFER_INFORMATION_SECTION_TYPE8,
                                &p->md.ntDetails->Rx.sectionInfo);

    if(status == NTCI_STATUSCODE_NO_PACKETS_AVAILABLE ) {
      if (p->break_loop) {
        p->break_loop = 0;
        return -2;
      }
      return 0;     
    } else if(status != NTCI_ERRCODE_SUCCESS) {
      REPORT_NTCI_FAILURE(p->md.ntDetails->cardHandle, "NTCI_GetNextBuffer");
      return(-1);
    }

    if(p->md.ntDetails->Rx.lastTime==0) {
      p->md.ntDetails->Rx.lastTime = p->md.ntDetails->Rx.sectionInfo.segmentTimestamp;
    }
    
    p->md.ntDetails->Rx.holdingBuffer = TRUE;
        
    /* Update buffer details */      
    p->md.ntDetails->Rx.framesAvailable = p->md.ntDetails->Rx.sectionInfo.numDescriptorsAvailable;
    
    /* Start processing from the beginning of the list */
    p->md.ntDetails->Rx.currentPacket = (uint32_t *) (p->md.ntDetails->Rx.sectionInfo.externalBuffer.user.u.pVirt);
    
    /* Update statistics */             
    p->md.ntDetails->packetsDrop += p->md.ntDetails->Rx.sectionInfo.numDroppedFrames;
  }

  if(p->break_loop) {
    p->md.ntDetails->closing = TRUE;
    /* user requested break */
    n = -2;
    p->break_loop = 0;
    return n;
  }  

  /* Go through the section and process each packet.
   */
//  DEBUG_PRINT("Entering read loop %d frames", p->md.ntDetails->framesAvailable);
  while( p->md.ntDetails->Rx.framesAvailable > 0 )
  {
    struct pcap_pkthdr hdrDetails;
    PacketDescriptorNtSeriesType1_t *pDescriptor;
    uint32_t storedLength;

    if(p->break_loop) {
      p->md.ntDetails->closing = TRUE;
      /* user requested break */
      n = -2;
      break;
    }  

    /* Get entry in descriptor list */
    /* Get location of descriptor */
    pDescriptor = (PacketDescriptorNtSeriesType1_t *)p->md.ntDetails->Rx.currentPacket;

    storedLength = pDescriptor->storedLength;
    
#ifdef CAPTURE_WITHOUT_CRC
    // We need to decrement the lengths with 4 because it should be returned without CRC
    // StoredLength is used to find the next packet so we need to store the original
    pDescriptor->wireLength-=4;
    pDescriptor->storedLength-=4;
#endif  

    /* Update the statistics */
    p->md.ntDetails->packetsRecv++;
    /* Run the packet filter if not using kernel filter */
    
    if((p->fcode.bf_insns == NULL) || bpf_filter(p->fcode.bf_insns, (u_char *)(pDescriptor+1), pDescriptor->wireLength, pDescriptor->storedLength)) {
      /*
        Check if we are running on a system that uses 64bit ulong for tv_sec in
        struct timeval.
      */

      if (hdrDetails.caplen > p->md.ntDetails->snaplen) hdrDetails.caplen = p->md.ntDetails->snaplen;

      hdrDetails.ts.tv_sec  = *((uint32_t*)pDescriptor);
      hdrDetails.ts.tv_usec = *(((uint32_t*)pDescriptor)+1);        
      hdrDetails.caplen = pDescriptor->storedLength;
      hdrDetails.len = pDescriptor->wireLength;
      /* stored length is 8 byte aligned, therefore we may get bigger caplen than wirelen */
      if (hdrDetails.caplen > hdrDetails.len) hdrDetails.caplen = hdrDetails.len;
      /* Process the current packet. */
      (*callback)(user, &hdrDetails, (u_char *)(pDescriptor+1));
      n++;
    }
      
    p->md.ntDetails->Rx.framesAvailable--;
    p->md.ntDetails->Rx.currentPacket = (uint32_t*)((uint8_t*)(p->md.ntDetails->Rx.currentPacket) + (storedLength + sizeof(PacketDescriptorNtSeriesType1_t)));

    /* Check whether we have read the maximum number of packets
     * requested */
    if( n >= cnt && cnt > 0 ) {
      break;
    }    
  }

  if(p->md.ntDetails->Rx.framesAvailable == 0 || p->break_loop ) {
    /* When all packets in current buffer have been read,
     * release the buffer back to the driver */
    NTCI_ReleaseBuffer(p->md.ntDetails->cardHandle,
                       p->md.ntDetails->feedHandle,
                       STID_BUFFER_INFORMATION_SECTION_TYPE8,
                       &p->md.ntDetails->Rx.sectionInfo);    
    p->md.ntDetails->Rx.holdingBuffer = FALSE;
  }
  
  if (p->break_loop && n == -2) p->break_loop = 0;
  return n;
}


#ifdef HAVE_NEW_CARDCAP_STRUCT  
static int napatech_pcap_inject(pcap_t *p, const void *buf, size_t size)
{
  int result;

  // Take the mutex.
  if(MutexGet(p)!=0) {
    return -1;
  }  
  if(p->md.ntDetails->feedHandle==0) {
    /* The creation start of packetfeeds is a critical section. */
    if((result = StartPacketFeed(p)) != 0) {
      MutexPost(p);
      return result;
    }

  }
  if(p->md.ntDetails->Tx.threadHandle==0) {
    int maxWait=0;
    uint64_t        tStart, tStop;
    struct timeval  tWait;
    ChannelStatusType1_t chanStatus;
    // We need to create the thread that will handle TX of segments.
    memset(&p->md.ntDetails->Tx, 0x00, sizeof(p->md.ntDetails->Tx));
    
    // Determine the link speed
    if(NTCI_GetChannelStatus(p->md.ntDetails->cardHandle, 
			     1<<atoi(&p->md.ntDetails->virtName[6]),
			     STID_CHANNEL_STATUS_TYPE1,
			     &chanStatus)!=NTCI_ERRCODE_SUCCESS) {
      REPORT_NTCI_FAILURE(p->md.ntDetails->cardHandle, "NTCI_GetChannelStatus");
      MutexPost(p);
      return -2;        
    }
    switch(chanStatus.actualSpeed) {
    case LINK_SPEED_10M:  p->md.ntDetails->Tx.ifSpeed=   10; break;
    case LINK_SPEED_100M: p->md.ntDetails->Tx.ifSpeed=  100; break;
    case LINK_SPEED_1G:   p->md.ntDetails->Tx.ifSpeed= 1000; break;
    case LINK_SPEED_10G:  p->md.ntDetails->Tx.ifSpeed=10000; break;
    default:
      p->md.ntDetails->Tx.ifSpeed=10;
    }
    
    // Get the first timestamp.
    p->md.ntDetails->Tx.prevCycles=__rdtscll();
    
    /**
     * Calculate speed of CPU timer.
     */    
    tWait.tv_usec=0;
    tWait.tv_sec=1;
    tStart=__rdtscll();
    select(0, NULL, NULL, NULL, &tWait);
    tStop=__rdtscll();
    p->md.ntDetails->Tx.cyclesPerMicroSec = (tStop-tStart)/1000000;
    DEBUG_PRINT("CPU speed: %lld Mhz (%lld cycles per sec)\n", (tStop-tStart)/1000000, 
		(unsigned long long)(tStop-tStart));
    
    p->md.ntDetails->Tx.threadRunFlag = THREAD_DONE;
    if(pthread_create(&p->md.ntDetails->Tx.threadHandle, NULL,
		      NapatechPcapTxHandler, p) != 0) {
      ERROR_PRINT("%s: Could not create TX feed%d thread\n", p->md.ntDetails->feedHandle&0xFF);
      snprintf(p->errbuf, PCAP_ERRBUF_SIZE,"napatech_pcap_read: Packet feed is being closed");
      MutexPost(p);
      return -1;
    }
    // Wait until the thread has been created to avoid close race conditions
    while(p->md.ntDetails->Tx.threadRunFlag != THREAD_RUNNING) {
      tWait.tv_usec=1000;
      tWait.tv_sec=0;
      // Wait then issue a timeout request.
      select(0, NULL, NULL, NULL, &tWait);
      maxWait++;
      if(maxWait == 10) {
        // Only wait one second
        ERROR_PRINT("%s: Failed to start TX thread for feed %d\n", p->md.ntDetails->feedHandle&0xFF);
        break;
      }
    }
  }
  // Inject the packet
  result = InjectPacket(p, buf, size);
  /* Release the mutex again. */
  MutexPost(p);
  return result;
}
#endif


static int napatech_pcap_inject_nosupport(pcap_t *p, const void *buf, size_t size)
{
  fprintf(stderr, "TX is not supported by this feed. Have you run the driver with \"PcapTx=1\" or are you injecting on a feed higher than port count?\n");
  return -1;
}
  
static int napatech_set_datalink( pcap_t *p, int dlt )
{
	/* Stub implementation keeps ethereal happy */
	return(0);
}

/* napatech_pcap_setfilter
 *
 * Set the BPF filter for a pcap packet stream.
 *
 * Parameters
 * p   - pointer to pcap_t structure for device
 * fp  - BPF program
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setfilter( pcap_t *p, struct bpf_program *fp )
{
  CALL_TRACE();
  /* Check for a 'NULL' filter - just a single return instruction filter */
  if(fp != NULL) {
    if(fp->bf_len == 1 && fp->bf_insns != NULL) {
      if (fp->bf_insns->code == BPF_RET)  {
        /* Looks like a 'NULL' filter */
        return 0;
      }
    }
  }
  
	/* user-mode filtering */
	/* check we have a pcap handle */
	if(!p) {
		return -1;
	}
	/* check for a filter */
	if(!fp) {
		DEBUG_PRINT("setfilter: No filter specified");
		return -1;
	}

	/* make private copy of the filter */
	if(install_bpf_program(p, fp) < 0) {
		return -1;
	}
	
	/* set user-mode filter flag */
	p->md.use_bpf = 0;
  return 0;
}

/* napatech_pcap_cleanup
 *
 * Close an pcap packet stream on a Napatech device.
 *
 * Parameters
 * p - pointer to pcap_t structure for device to close
 */
static void napatech_pcap_cleanup( pcap_t *p )
{
  uint32_t result=0, segment=0;
  BufferRequestSampledType2_t bufferRequest;
  BufferInformationSectionType8_t bufInfo;
  FeedStatusSampledType4_t feedStatus;
  pcap_t *prev=NULL, *cur;
  
  CALL_TRACE();

  pthread_mutex_lock(&mList);
  cur=pHead;
  
  // Remove the element from the list.
  while(cur && (cur != p)) {
    prev=cur;
    cur=cur->md.ntDetails->pNext;
  }
  
  if(cur==p) {
    if(prev==NULL) {
      pHead = cur->md.ntDetails->pNext;
    } else {
      prev->md.ntDetails->pNext = cur->md.ntDetails->pNext;
    }
  }
  pthread_mutex_unlock(&mList);
  
  if(p->md.ntDetails) {
    // Destroy TX thread
    if(p->md.ntDetails->Tx.threadHandle) {      
      if(p->md.ntDetails->Tx.threadRunFlag!=THREAD_STOPPED) {
        p->md.ntDetails->Tx.threadRunFlag = THREAD_QUIT;
      }
      while(p->md.ntDetails->Tx.threadRunFlag != THREAD_STOPPED) {  
        sleep(0);
      }
      if(pthread_detach(p->md.ntDetails->Tx.threadHandle) != 0) {
        ERROR_PRINT("Failed to detach thread: %s\n", p->md.ntDetails->virtName); 
      }
      p->md.ntDetails->Tx.threadRunFlag = THREAD_DESTROYED;
      // Release TX segments     
      if(MutexGet(p)==0) {
        InjectPacket(p, NULL, -1);
        MutexPost(p);
      }     
    }
    
    /* Indicate that the stream is closing, so that we don't try to get more
     * packets */
    p->md.ntDetails->closing = TRUE;    
    
    /* Make sure that any buffers that are held, are released back to
     * the driver */
    if( p->md.ntDetails->feedHandle != 0 ) {

      if(p->md.ntDetails->Rx.holdingBuffer) {
        NTCI_ReleaseBuffer(p->md.ntDetails->cardHandle,
                           p->md.ntDetails->feedHandle,
                           STID_BUFFER_INFORMATION_SECTION_TYPE8,
                           &p->md.ntDetails->Rx.sectionInfo);    
        p->md.ntDetails->Rx.holdingBuffer = FALSE;
      }

      /* Stop the packetfeed. */
      NTCI_StopPacketFeed(p->md.ntDetails->cardHandle,
                          p->md.ntDetails->feedHandle);

      p->md.ntDetails->Rx.holdingBuffer = FALSE;
    }
    
    /* Record that this device is no longer running a feed */
    p->md.ntDetails->feedHandle = 0;

    /* Destroy mutex. */
    pthread_mutex_destroy(&p->md.ntDetails->mutex);

    free(p->md.ntDetails);
    p->md.ntDetails = NULL;
    pcap_cleanup_live_common(p);
  }
}

/* napatech_pcap_findalldevs
 *
 * Populate device list with napatech devices. Create the list from the
 * device names returned by NTCI_Get....Device. 
 *
 * Parameters
 * alldevsp - list to put devices in
 * errbuf   - error reporting buffer
 * Returns 0 on success, -1 on failure
 */
int napatech_platform_finddevs( pcap_if_t **devlistp, char *errbuf )
{
  char virtName[MAX_DEV_NAME_SIZE];
  char description[100];
	int ret = 0;
	int c,v;
  uint32_t numFeeds=0, result=0;
  handle_t cardHandle=NULL;
  CardHardwareVersionCapabilitiesType3_t data3;
  CardServicesCapabilitiesType1_t services;
  CALL_TRACE();
  
  for(c=0;c<MAX_SUPPORTED_CARDS; c++) {    
    if((cardHandle = NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, c)) != NULL) {
      DEBUG_PRINT("Found adapter %d\n", c);

      /* Check if we need to use NTPL to set up feeds. */
      if(NTCI_GetCapabilities(cardHandle,
                              CAPID_VERSION,
                              NOT_SUB_CAPABILITIES,
                              STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                              &data3) != NTCI_ERRCODE_SUCCESS) {
        REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetPacketFeedCount");
        exit(-1);
      } else {
        if(data3.AdapterCard.dwInfoType==1) {
          if((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) < 33) {
            ERROR_PRINT("Unsupported FPGA on adapter %d\n", c);
            NTCI_CloseCard(cardHandle);
            continue;
          }
        }else if(data3.AdapterCard.dwInfoType==2) {
          if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) < 33) {
            ERROR_PRINT("Unsupported FPGA on adapter %d\n", c);
            NTCI_CloseCard(cardHandle);
            continue;
          }
#ifdef VERSION_INFO_TYPE3          
        }else if(data3.AdapterCard.dwInfoType==3) {
          if((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) < 33) {
            ERROR_PRINT("Unsupported FPGA on adapter %d\n", c);
            NTCI_CloseCard(cardHandle);
            continue;
          }
#endif          
        } else {
          ERROR_PRINT("Unsupported FPGA on adapter %d\n", c);
          NTCI_CloseCard(cardHandle);
          continue;
        }
      }

      /* FPGA is supported. */

      /* Check if packetfeed is supported. */
      if(NTCI_GetCapabilities(cardHandle, CAPID_SERVICES, NOT_SUB_CAPABILITIES,
                              STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                              &services) != NTCI_ERRCODE_SUCCESS) {
        NTCI_CloseCard(cardHandle);
        continue;
      }
      if(!services.packetFeedSupported) {
        /* Skip card it has no packetfeed support. */
        NTCI_CloseCard(cardHandle);
        continue;
      }        
      
      /* Find number of feeds. */
      if((result = NTCI_GetPacketFeedCount(cardHandle, &numFeeds)) != NTCI_ERRCODE_SUCCESS) {
        REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetPacketFeedCount");
        exit(-1);
      }
      DEBUG_PRINT("Adapter %d has %d feeds.\n", c, numFeeds);      
      for(v=0; v<numFeeds; v++) {
        snprintf(virtName, sizeof(virtName)-1, "ntxc%d:%d", c, v);
        /* Check that the feed exist. */
        if(is_napatech_device(virtName)) {
          DEBUG_PRINT("Adding virtual device: %s.\n", virtName);
          sprintf(description, "NT adapter %d feed %d", c, v);
          if(pcap_add_if(devlistp, virtName, 0, description, errbuf) == -1) {
            /* Failure. */
            ERROR_PRINT("Failed adding virtual device %s\n", virtName); 
            return -1;
          }
        }
      }
      NTCI_CloseCard(cardHandle);
    }
  }
  return 0;
}

/* napatech_pcap_setbuff
 *
 * Set the size of the packet recieve buffer. Not necessary for Napatech
 * devices, but the stub function must be in the common code to aid
 * porting to different versions of LibPCap.
 *
 * Parameters
 * p    - pointer to pcap_t structure for device
 * size - size of buffer in bytes
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setbuff( pcap_t *p, int size )
{
  CALL_TRACE();
  DEBUG_PRINT("napatech_pcap_setbuff: Not required by Napatech devices");
  return 0;
}

/* napatech_pcap_setmintocopy
 *
 * Set the minimum amount of data to read to release a read call.
 * Not necessary for Napatech devices, but the stub function must be in the
 * common code to aid porting to different versions of LibPCap.
 *
 * Parameters
 * p    - pointer to pcap_t structure for device
 * size - size of minimum read in bytes
 * Returns 0 on success, -1 on failure
 */
static int napatech_pcap_setmintocopy( pcap_t *p, int size )
{
  CALL_TRACE();
  DEBUG_PRINT("napatech_pcap_setmintocopy: Not supported by Napatech devices");
  return 0;
}





static int napatech_pcap_setnonblock(pcap_t *p, int nonblock, char *errbuf)
{
  p->md.ntDetails->nonblocking = nonblock;
  return 0;
}


static int napatech_pcap_getnonblock(pcap_t *p, char *errbuf)
{
  return p->md.ntDetails->nonblocking;
}




/* napatech_activate
 *
 * Activate the pcap device
 */
static int napatech_activate(pcap_t *p)
{
  handle_t cardHandle=NULL;
  CardPacketFeedCapabilitiesType2_t packetfeed;
  CardHardwareCapabilitiesType1_t cardCapabilities;
  uint8_t numChannels=0, i=0;
  char *device = p->opt.source;
  
  CALL_TRACE();

  WriteToLog(FALSE, NULL, NULL, 0, "%s : %s", PCAP_PACKAGE_NAME, PCAP_PACKAGE_TIMESTAMP);

  if(is_napatech_device(device)) {
    /* Allocate the private structure. */
    if((p->md.ntDetails = malloc(sizeof(struct napatech_details))) == NULL) {
      return -1;
    }
    memset(p->md.ntDetails, 0, sizeof(struct napatech_details));

    strcpy(p->md.ntDetails->virtName, device);
    p->md.ntDetails->closing = FALSE;    
    p->md.ntDetails->snaplen = p->snapshot;
    p->md.ntDetails->promisc = p->opt.promisc;
    p->md.ntDetails->to_ms   = p->md.timeout;
    p->md.ntDetails->nonblocking = 0;
   
    /*a Napatech adapter is assumed to be ethernet*/
    p->linktype = DLT_EN10MB;
    
    p->selectable_fd = -1;
    
    p->read_op = napatech_pcap_read;
    p->setfilter_op = napatech_pcap_setfilter;
    p->setdirection_op = NULL;  /* Cannot set direction */
    p->set_datalink_op = napatech_set_datalink;	/* set link type */
    p->getnonblock_op = napatech_pcap_getnonblock;
    p->setnonblock_op = napatech_pcap_setnonblock;
    p->stats_op = napatech_pcap_stats;
    p->cleanup_op = napatech_pcap_cleanup;

    /* Create the mutex. */
    pthread_mutex_init(&p->md.ntDetails->mutex, NULL);

    if((cardHandle = NTCI_OpenCard(NTCI_ANY_NAPATECH_CARD_TYPE, atoi(&device[4]))) != NULL) {
      /* Check if we have TX support. */
#ifdef HAVE_NEW_CARDCAP_STRUCT      
      if(NTCI_GetCapabilities(cardHandle, CAPID_PACKET_FEED, NOT_SUB_CAPABILITIES,
                              STID_CARD_PACKET_FEED_CAPABILITIES_TYPE2,
                              &packetfeed) != NTCI_ERRCODE_SUCCESS) {
        REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetCapabilities");
        NTCI_CloseCard(cardHandle);
        p->inject_op = napatech_pcap_inject_nosupport;
        p->md.ntDetails->Tx.txSupported=0;
      } else {
#ifdef HAVE_CARD_CAPABILITIES4
        CardHardwareCapabilitiesType4_t cardCapabilities4;
        // Default we timestamp when the last octet is received
        p->md.ntDetails->Tx.pktTsLast =1;
        if(NTCI_GetCapabilities(cardHandle, 
                                CAPID_HARDWARE,
                                NOT_SUB_CAPABILITIES,
                                STID_CARD_HARDWARE_CAPABILITIES_TYPE4,
                                &cardCapabilities4) == NTCI_ERRCODE_SUCCESS) {
          if(cardCapabilities4.pktTsMethod == PACKET_TIMESTAMP_FIRST) {
            p->md.ntDetails->Tx.pktTsLast=0;
          }
          if((cardCapabilities4.inlineCard) && (!packetfeed.type1.txBoundToPort)) {
            WARNING_PRINT("%s: Inline adapter not loaded with PcapTx=1\n", device);
            NTCI_CloseCard(cardHandle);
            p->inject_op = napatech_pcap_inject_nosupport;
            p->md.ntDetails->Tx.txSupported=0;
            return -1;
          }          
        }
#else
        p->md.ntDetails->Tx.pktTsLast =1;
#endif          
        if((packetfeed.type1.externalTxSupported) && (packetfeed.type1.txBoundToPort==1)) {
          // We only allow TX to happen on feeds 0 .. numChannels-1
          if(NTCI_GetCapabilities(cardHandle, 
                                  CAPID_HARDWARE,
                                  NOT_SUB_CAPABILITIES,
                                  STID_CARD_HARDWARE_CAPABILITIES_TYPE1,
                                  &cardCapabilities) != NTCI_ERRCODE_SUCCESS) {
            REPORT_NTCI_FAILURE(cardHandle, "NTCI_GetCapabilities");
            NTCI_CloseCard(cardHandle);
            p->inject_op = napatech_pcap_inject_nosupport;
            p->md.ntDetails->Tx.txSupported=0;
          } else {         
            numChannels = 0;
            i = cardCapabilities.totalChannelsBitmask;
            while (i != 0) {
              if (i & 0x1)
                ++numChannels;    
              i >>= 1;
            }
            if(atoi(&device[6]) < numChannels) {
              p->inject_op = napatech_pcap_inject;
              p->md.ntDetails->Tx.txSupported=1;
            } else {
              // Allow capture on this interface, nothing more.
              p->inject_op = napatech_pcap_inject_nosupport;
              p->md.ntDetails->Tx.txSupported=1;
            }
          }
        } else if(packetfeed.type1.externalTxSupported) {
          // Old drivers has a bug. They report TX even though they don't support it.
          NtVersionInfoType1_t versionInfo;
          if(NTCI_GetCardInfo(
               cardHandle,
               STID_CARD_INFO_VERSION_INFO_TYPE1, &versionInfo) != NTCI_ERRCODE_SUCCESS) {
            p->inject_op = NULL;
            p->md.ntDetails->Tx.txSupported=0;            
          } else {
            if(versionInfo.DriverMajorVersion<=4) {
              if(versionInfo.DriverMinorVersion<17) {
                p->inject_op = NULL;
                p->md.ntDetails->Tx.txSupported=0;
              } else {                
                p->inject_op = napatech_pcap_inject_nosupport;
                p->md.ntDetails->Tx.txSupported=0;          
              }
            } else {
              p->inject_op = napatech_pcap_inject_nosupport;
              p->md.ntDetails->Tx.txSupported=0;
            }
          }
        } else {
          p->inject_op = NULL;
          p->md.ntDetails->Tx.txSupported=0;
        }
      }
#else
      p->inject_op = NULL;
      p->md.ntDetails->Tx.txSupported=0;
#endif      
      NTCI_CloseCard(cardHandle);      
    }
    if(listMutexCreated == 0) {
      pthread_mutex_init(&mList, NULL);
      listMutexCreated=1;
    }
    pthread_mutex_lock(&mList);
    // Add entry to the list.
    p->md.ntDetails->pNext = pHead;
    pHead = p;
    pthread_mutex_unlock(&mList);
    
    //ATEXIT
    if(!atexithandlerRegistered) {
      atexit(atexit_handler);
      atexithandlerRegistered=1;
    }
  }
  return 0;
}


#endif /*NAPATECH_SUPPORT*/
