// $Id: Diagnostics.c 24943 2013-08-16 06:50:59Z am $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * This application runs sanity checks on an adapter.
 */

#ifdef WIN32
#include <windows.h>
#include "platformwin.h"
#include <stdbool.h>
#else
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <stdbool.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#endif
#include <toolslib.h>
#include "tcppacket.c"

/* These headers are needed when monitoring sensors. */
#include <hwmonstructure.h>
#include <NTCI_hardwaremonitor.h>

/* These headers are needed when getting capabilities. */
#include <capabilitiesstructure.h>
#include <NTCI_capabilities.h>

/* These headers are neede when getting timesync info. */
#include <cardconfigurationstructure.h>
#include <NTCI_cardconfiguration.h>

/* These headers are needed when handling packetfeeds. */
#include "NTCI_packetfeed.h"
#include "packetdescriptor.h"
#include "packetfeedstructure.h"

/* These headers are needed for the packet classification API */
#include <NTCI_packetclassification.h>
#include <packetclassificationstructure.h>

#include <NTCI_registeraccess.h>
#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>

enum { CONFIG, PCI, NAPA, FILTER_HW_SHADOW, NAPA_NMB, NAPA_DBG, I2C, NAPA_FLASH_C=100, NAPA_FLASH_NC} gl_region=0;
#define NT_REG_P9020_DDR3_ADDR_BASE                          0x28000000
#define NT_REG_DDR3_BUFFER_ADDR_ADDR_REL                     0x00000008
#define NT_REG_DDR3_BUFFER_DATA_ADDR_REL                     0x0000000c
#define NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR                   0x28000008
#define NT_REG_P9020_DDR3_BUFFER_DATA_ADDR                   0x2800000c
#define NT_REG_P9020_DDR3_COMMANDS_ADDR                      0x28000004

#define NT_REG_P9020_RXAUI0_ADDR_BASE                        0x10100000
#define NT_REG_P9020_RXAUI1_ADDR_BASE                        0x10101000

extern NTCOMMONLIB uint32_t NTCOMMONCALL NTCI_SendPacket(  IN  handle_t hCard,
                                                           IN int      channel,
                                                           IN void *   packetData,
                                                           IN uint32_t packetDataLength);

/* Test functions. */
static uint32_t TemperaturSensorTest(uint8_t cardIndex);
static uint32_t SDRAMTest(uint8_t cardIndex);
static uint32_t Ddr3RamTest(uint8_t cardIndex, uint32_t ramSize);
static uint32_t TimesyncTest(uint8_t cardIndex);
static uint32_t TrafficTest(uint8_t cardIndex);

static int stopApplication=0;

static char *verbose=NULL;
static char *rxaui_loopback=NULL;
static char *fullramtest=NULL;
static char *sensoronly=NULL;

static uint32_t writePos=0;
static char screenBuffer[65536]; /* A 64k screenbuffer should be anough. */

static CommandLineOptions_t tools_options[] = {
  {"v", "Verbose.", NULL, 2, &verbose},
  {"l", "Loopback.", NULL, 2, &rxaui_loopback},
  {"f", "Full memory (SDRAM) test. Time-out is seconds between writing\n"
   "\t       to SDRAM and readback.", "time-out", 0, &fullramtest},
  {"s", "Sensor only.", NULL, 2, &sensoronly},
  {NULL, NULL, NULL, 0, NULL}
};

typedef struct {
  const char *testname;
  const char *testdescription;
  uint32_t (*pfTest)(uint8_t);
} TestCase_t;

TestCase_t Tests[] =
{
  {"Sensor", "Verify that temperature sensors are within limits", TemperaturSensorTest},
  {"SDRAM", "Verify that the SDRAM is working", SDRAMTest},
  {"Timesync", "Verify that a timesync link is connected and working", TimesyncTest},
  {"Traffic", "Send and receive packets on all interfaces.", TrafficTest},
  {NULL, NULL, NULL}
};
#ifdef WIN32
#define PRINT(a, ...)   writePos+=WriteToBuffer(&screenBuffer[writePos], sizeof(screenBuffer)-writePos, a, __VA_ARGS__);
#else
#define PRINT(args...)   writePos+=WriteToBuffer(&screenBuffer[writePos], sizeof(screenBuffer)-writePos, ## args);
#endif
#define NTCI_ERRCODE_DIAGNOSTICS_ERROR 0x12345678
#define NTCI_ERRCODE_DIAGNOSTICS_NA    0x87654321
#define NTCI_ERRCODE_DIAGNOSTICS_IGNORED  0x11223344

#define NUM_PACKETS_TO_SEND  10000

static char tool_description[] = {
  "Run diagnostics on an adapter.\n"
  "The following is tested:\n"
  " 1) Temperature sensors - Verify that the onboard sensors can be read and\n"
  "                          are within limits.\n"
  " 2) SDRAM               - Verify that the SDRAM is mounted.\n"
  "                          (Not applicable for X adapters.)\n"
  " 3) Timesync            - Verify that the timesync port is working.\n"
  "                          (Not applicable for X adapters).\n"
  " 4) Traffic (and LED)   - Verify that packets can be sent and received\n"
  "                          on all interfaces. The LEDs on the adapter flash\n"
  "                          during this operation and can be verified manually.\n"
  "NOTE:\n"
  "1) An NT adapter must be a slave adapter and must be connected to a Napatech\n"
  "   timesync unit, or connected or daisy-chained to a master adapter for the\n"
  "   \"Timesync\" test to pass.\n"
  "2) The traffic test only passes if the adapter interfaces are connected\n"
  "   to each other.\n" };
static char output_description[] = {"\n"};


/* CRC Table */
static const uint32_t crc32_table[256] =
{
  0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419,
  0x706af48f, 0xe963a535, 0x9e6495a3, 0x0edb8832, 0x79dcb8a4,
  0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07,
  0x90bf1d91, 0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
  0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7, 0x136c9856,
  0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9,
  0xfa0f3d63, 0x8d080df5, 0x3b6e20c8, 0x4c69105e, 0xd56041e4,
  0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
  0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3,
  0x45df5c75, 0xdcd60dcf, 0xabd13d59, 0x26d930ac, 0x51de003a,
  0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599,
  0xb8bda50f, 0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
  0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d, 0x76dc4190,
  0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f,
  0x9fbfe4a5, 0xe8b8d433, 0x7807c9a2, 0x0f00f934, 0x9609a88e,
  0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
  0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed,
  0x1b01a57b, 0x8208f4c1, 0xf50fc457, 0x65b0d9c6, 0x12b7e950,
  0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3,
  0xfbd44c65, 0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
  0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb, 0x4369e96a,
  0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5,
  0xaa0a4c5f, 0xdd0d7cc9, 0x5005713c, 0x270241aa, 0xbe0b1010,
  0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
  0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17,
  0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad, 0xedb88320, 0x9abfb3b6,
  0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615,
  0x73dc1683, 0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
  0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1, 0xf00f9344,
  0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb,
  0x196c3671, 0x6e6b06e7, 0xfed41b76, 0x89d32be0, 0x10da7a5a,
  0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
  0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1,
  0xa6bc5767, 0x3fb506dd, 0x48b2364b, 0xd80d2bda, 0xaf0a1b4c,
  0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef,
  0x4669be79, 0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
  0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f, 0xc5ba3bbe,
  0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31,
  0x2cd99e8b, 0x5bdeae1d, 0x9b64c2b0, 0xec63f226, 0x756aa39c,
  0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
  0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b,
  0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21, 0x86d3d2d4, 0xf1d4e242,
  0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1,
  0x18b74777, 0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
  0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45, 0xa00ae278,
  0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7,
  0x4969474d, 0x3e6e77db, 0xaed16a4a, 0xd9d65adc, 0x40df0b66,
  0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
  0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605,
  0xcdd70693, 0x54de5729, 0x23d967bf, 0xb3667a2e, 0xc4614ab8,
  0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b,
  0x2d02ef8d
};



#ifdef WIN32
  static void PrintError(char *str)
{
  CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
  HANDLE handle=GetStdHandle(STD_OUTPUT_HANDLE);
  GetConsoleScreenBufferInfo(handle, &ConsoleScreenBufferInfo);
  SetConsoleTextAttribute(handle, FOREGROUND_RED|FOREGROUND_INTENSITY);
  printf("%s", str);
  SetConsoleTextAttribute(handle, ConsoleScreenBufferInfo.wAttributes);
}
static void PrintPass(char *str)
{
  CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
  HANDLE handle=GetStdHandle(STD_OUTPUT_HANDLE);
  GetConsoleScreenBufferInfo(handle, &ConsoleScreenBufferInfo);
  SetConsoleTextAttribute(handle, FOREGROUND_GREEN|FOREGROUND_INTENSITY);
  printf("%s", str);
  SetConsoleTextAttribute(handle, ConsoleScreenBufferInfo.wAttributes);
}
static void PrintNotTested(char *str)
{
  CONSOLE_SCREEN_BUFFER_INFO ConsoleScreenBufferInfo;
  HANDLE handle=GetStdHandle(STD_OUTPUT_HANDLE);
  GetConsoleScreenBufferInfo(handle, &ConsoleScreenBufferInfo);
  SetConsoleTextAttribute(handle, FOREGROUND_BLUE|FOREGROUND_INTENSITY);
  printf("%s", str);
  SetConsoleTextAttribute(handle, ConsoleScreenBufferInfo.wAttributes);
}
#else
static void PrintError(char *str)
{
  printf("%c[%d;%dm%s%c[%d;%dm", 0x1B,1,31, str, 0x1B,0,39);
}
static void PrintPass(char *str)
{
  printf("%c[%d;%dm%s%c[%d;%dm", 0x1B,1,32, str, 0x1B,0,39);
}
static void PrintNotTested(char *str)
{
  printf("%c[%d;%dm%s%c[%d;%dm", 0x1B,1,34, str, 0x1B,0,39);
}
#endif

#ifdef WIN32
static BOOL CatchSignalCTRL_C(DWORD CtrlType)
{
  stopApplication=1;
  return TRUE;
}
#else
static void CatchSignalCTRL_C (int sig __attribute__((unused)))
{
  stopApplication=1;
}
#endif

static uint32_t TemperaturSensorTest(uint8_t cardIndex)
{
  handle_t cardHandle;
  uint32_t result, i;
  uint32_t errorDetected=0;
  HardwareMonitorType2_t* sensor;
  HardwareMonitorType3_t* sensor3;
  HardwareMonitorType2_t SensorList2[50]; /* Max sensors */
  HardwareMonitorType3_t SensorList3[50]; /* Max sensors */
  uint32_t blockSize;
  uint32_t SensorListLength;
  bool     DataAvailable;

  if ((fullramtest) && (!sensoronly)) {
    // Sensor Test ignored when running full SDRAM test
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  GetCardHandle(cardIndex, &cardHandle);

  result = NTCI_GetPartSensorsBlockSize(cardHandle,
                                        STID_HARDWARE_MONITOR_TYPE2,
                                        &blockSize);

  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT(">>> Error: Failed getting sensor block size.\nIs the driver running with the sensor monitoring service?\n");
    PRINT("Please enable the sensor monitoring service in default.cfg (SensorAlerts=1)\n"
           "and reload the driver.\n");
    return result;
  }

  if(blockSize > sizeof(SensorList2)) {
    PRINT(">>> Error: Too many sensors at Adapter. Not enough memory.\n");
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  result = NTCI_GetPartSensors(cardHandle,
                               STID_HARDWARE_MONITOR_TYPE2,
                               &SensorList2,
                               blockSize);

  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT(">>> Error: Failed getting sensors.\n");
    return result;
  }

  /* One entry for each sensor, plus the "all sensors" entry */
  SensorListLength = blockSize/sizeof(HardwareMonitorType2_t);

  printf("%12s ", "");
  printf(" %-8s %-8s %-8s %-8s %-8s %-8s %-8s\n", "Sensor", "Low", "High", "Current", "Alarm", "Lowest", "Highest");
  printf("%12s ", "");
  printf(" %-8s %-8s %-8s %-8s %-8s %-8s %-8s\n", "type", "limit", "limit", "value", "state", "value", "value");

  for (i=0; i<SensorListLength; i++) {
    sensor = (HardwareMonitorType2_t*)&SensorList2[i];
    if ((SensorList2[i].SensorType == TEMPERATURE)
     || (SensorList2[i].SensorType == FANSPEED )
     || (SensorList2[i].SensorType == POWER)) {
      printf("%-14s", sensor->SensorName);
    }

    switch(sensor->SensorType) {
    case TEMPERATURE: printf("%-8s", "Temp."); break;
    case FANSPEED: printf("%-8s", "Fan."); break;
    default: continue;
    }

    if (sensor->SensorType == FANSPEED ){
      printf("%8d ", sensor->lowAlarmLimit);
      printf("%8d ", sensor->highAlarmLimit);
      printf("%8d ", sensor->sensorMeasurement);
      printf("%8s ", sensor->SensorAlarm==TRUE?"Alarm!":"");
      if(((HardwareMonitorType1_t*)sensor)->StructureVersion == 1) {
        printf("%8d ", sensor->lowestValue);
        printf("%8d ", sensor->highestValue);
      }
    } else if (sensor->SensorType == TEMPERATURE) {
      printf("%8.2f ", (float)sensor->lowAlarmLimit/10);
      printf("%8.2f ", (float)sensor->highAlarmLimit/10);

      DataAvailable = (sensor->lowestValue != 2550);

      if (DataAvailable)
        printf("%8.2f ", (float)sensor->sensorMeasurement/10);
      else
        printf("%8s", "NA");

      printf("%8s ", sensor->SensorAlarm==TRUE?"Alarm!":"");

      if(((HardwareMonitorType1_t*)sensor)->StructureVersion == 1) {
        if (DataAvailable)
          printf("%8.2f ", (float)sensor->lowestValue/10);
        else
          printf("%8s", "NA");

        if (DataAvailable)
          printf("%8.2f ", (float)sensor->highestValue/10);
        else
          printf("%8s", "NA");
      }
    }
    if ((SensorList2[i].SensorType == TEMPERATURE)
     || (SensorList2[i].SensorType == FANSPEED )) {
      printf("\n");
    }
  }

  result = NTCI_GetPartSensorsBlockSize(cardHandle,
                                        STID_HARDWARE_MONITOR_TYPE3,
                                        &blockSize);
  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT(">>> Error: Failed getting sensor block size.\nIs the driver running with the sensor monitoring service?\n");
    PRINT("Please enable the sensor monitoring service in default.cfg (SensorAlerts=1)\n"
           "and reload the driver.\n");
    return result;
  }

  if(blockSize > sizeof(SensorList3)) {
    PRINT(">>> Error: Too many sensors at Adapter. Not enough memory.\n");
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  result = NTCI_GetPartSensors(cardHandle,
                               STID_HARDWARE_MONITOR_TYPE3,
                               &SensorList3,
                               blockSize);

  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT(">>> Error: Failed getting sensors.\n");
    return result;
  }

  /* One entry for each sensor, plus the "all sensors" entry */
  SensorListLength = blockSize/sizeof(HardwareMonitorType3_t);
  if (SensorListLength > 0) {
    /* Print Power headline */
    printf("\n%12s ", "");
    printf(" %-8s %-8s\n", "Sensor", "Current");
    printf("%12s ", "");
    printf(" %-8s %-8s\n", "type", "value");
  }

  for (i=0; i<SensorListLength; i++) {
    sensor3 = (HardwareMonitorType3_t*)&SensorList3[i];
    if (SensorList3[i].SensorType == POWER) {
      printf("%-14s", sensor3->SensorName);
    }

    switch(sensor3->SensorType) {
    case POWER: printf("%-8s", "Power."); break;
    default: continue;
    }

    if (sensor3->SensorType == POWER) {
      printf("%8.2f ", (float)sensor3->sensorMeasurement/1000);
      printf("\n");
    }
  }

  if(errorDetected) {
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }
  return NTCI_ERRCODE_SUCCESS;
}

static uint32_t crc32 (uint32_t value)
{
  uint32_t crc=0;

  crc = crc32_table[(crc ^  (value&0xff)) & 0xff] ^ (crc >> 8);
  crc = crc32_table[(crc ^  (value&0xff00)>>8) & 0xff] ^ (crc >> 8);
  crc = crc32_table[(crc ^  (value&0xff0000)>>16) & 0xff] ^ (crc >> 8);
  crc = crc32_table[(crc ^  (value&0xff000000)>>24) & 0xff] ^ (crc >> 8);
  return ~crc;
}


static uint32_t SDRAMTest(uint8_t cardIndex)
{
  handle_t cardHandle;
  uint32_t result;
  CardHardwareCardCapabilitiesType2_t capability;
  CardHardwareVersionCapabilitiesType3_t data3;
  uint64_t fpgaId=0;

  if (sensoronly) {
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  GetCardHandle(cardIndex, &cardHandle);

  {
    CardServicesCapabilitiesType1_t services;

    /* A card has been found. Check if it has packetfeeds support. */
    if((result = NTCI_GetCapabilities(cardHandle, CAPID_SERVICES, NOT_SUB_CAPABILITIES,
                                      STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                                      &services)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_INFO, "Failed to read card services capability.\n");
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
    if(!services.packetFeedSupported) {
      /* Skip card it has no packetfeed support. */
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
  }

  result = NTCI_GetCapabilities(CardHandle,
                                CAPID_CARD_INDEXED,
                                0,
                                STID_CARD_HARDWARE_CARD_CAPABILITIES_TYPE2,
                                &capability);
  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed getting capability\n");
    return result;
  }
  if(!capability.SDRAMController) {
    PRINT("No SDRAM controller support\n");
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }
  if(!capability.SDRAMPresent) {
    PRINT("No SDRAM detected\n");
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }
  if(capability.SDRAMSize & 0x3FF) {
    /* Special encoding. 64bit. */
    capability.SDRAMSize*=4096;
    PRINT("SDRAM detected: Size %d MB (a)\n", capability.SDRAMSize);
  } else {
    PRINT("SDRAM detected: Size %d MB (b)\n", capability.SDRAMSize/(1024*1024));
    if((result = NTCI_GetCapabilities(cardHandle,
      CAPID_VERSION,
      NOT_SUB_CAPABILITIES,
      STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
      &data3)) != NTCI_ERRCODE_SUCCESS) {
      return CARD_UNKNOWN;
    }
    if(data3.AdapterCard.dwInfoType==1) {
      fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
     }else if(data3.AdapterCard.dwInfoType==2) {
      fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
    }else if(data3.AdapterCard.dwInfoType==3) {
      fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
     }

    // Should only be done for NT20E2 and NT40E2 cards
    if ((((fpgaId&0xFFFF0000) >> 16) == 9020) || (((fpgaId&0xFFFF0000) >> 16) == 9220) ||
        (((fpgaId&0xFFFF0000) >> 16) == 9221) || (((fpgaId&0xFFFF0000) >> 16) == 9222)) {
      // Check if full RAM test should be performed
      if (fullramtest) {
        Ddr3RamTest(cardIndex, (capability.SDRAMSize/(1024*1024)));
      }
    }
  }
  return NTCI_ERRCODE_SUCCESS;
}


/* Write BL8 read or write command into the command buffer */
int Ddr3WriteCmd(handle_t cardHandle, uint32_t nr, uint32_t rd, uint32_t addr) {
  uint32_t value;

  NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, &value);
  value = value & 0xfffffc00;
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, value | nr);

  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_DATA_ADDR, ( addr << 1 ) | ( rd & 0x1 ));
  //printf("Write cmd %x data %x\n", value | nr,  ( addr << 1 ) | ( rd & 0x1 ));
  return 0;
}

/* Write BL8 read or write command into the command buffer */
int Ddr3WriteCmd_v44_1G(handle_t cardHandle, uint32_t nr, uint32_t rd, uint32_t addr) {
  uint32_t value;
  uint32_t row;
  NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, &value);
  value = value & 0xfffffc00;
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, value | nr);

  row=0x0E00000&addr;  //New row
  addr=(addr&0x01FFFFF)|(row<<1);
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_DATA_ADDR, ( addr << 1 ) | ( rd & 0x1 ));
  //printf("Write cmd %x data %x\n", value | nr,  ( addr << 1 ) | ( rd & 0x1 ));
  return 0;
}
/* Write data into the data buffer */
int Ddr3WriteData(handle_t cardHandle, uint32_t nr, uint32_t data) {
  uint32_t value;

  NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, &value);
  value = value & 0xfffffc00;
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, value | 0x200 | nr);
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_DATA_ADDR, data);
  return 0;
}


/* Execute DDR3 BL8 read or write commands written into the command and
   data buffers */
int Ddr3ExecuteCmds(handle_t cardHandle, uint32_t nr) {
  uint32_t timeout, data;
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_COMMANDS_ADDR, nr);

  timeout = 1000000;
  data = 0;
  while ((timeout > 0) && (data < nr)) {
    NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_COMMANDS_ADDR, &data);
    timeout = timeout - 1;
  }

  if ( timeout == 0 ) {
    printf(" ddr3 commands not executed\n");
    return -1;
  }
  NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_COMMANDS_ADDR, 0);
  return 0;
}


static uint32_t Ddr3RamTest(uint8_t cardIndex, uint32_t ramSize)
{
  handle_t cardHandle;
  uint32_t i, res, addr=0;
  uint32_t start, end, num_cmds, crc;
  uint32_t value, result;
  uint32_t percent=0;
  uint16_t fpgaVer=0;
  CardHardwareVersionCapabilitiesType3_t data3;

  crc=0x12345678;
  start = 0;
  end = 1024*1024*(ramSize/(16*4));
  end = end;
  GetCardHandle(cardIndex, &cardHandle);

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_VERSION,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                    &data3)) != NTCI_ERRCODE_SUCCESS) {
    return CARD_UNKNOWN;
  }

  if(data3.AdapterCard.dwInfoType==1) {
    fpgaVer =(data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF;
  }else if(data3.AdapterCard.dwInfoType==2) {
    fpgaVer =(data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF;
  }else if(data3.AdapterCard.dwInfoType==3) {
    fpgaVer =(data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF;
  }
  printf("FULL DDR3 RAM test\n");
  printf("Fill complete RAM\n");
  printf("Start %x end %x \n", start, end);
  while (start < end) {
    num_cmds = ( end - start );
    if ( num_cmds > 15 ) {
      num_cmds = 15;
    }

    /* Exit if Ctrl-C has been pushed */
    if (stopApplication==1) {
      /* shutdown card */
      CardShutdown(0);
      DisplayProgramFooter();

      exit(-1);
    }

    // Check progress
    if ((100 - (((end-start)*100)/end)) > percent) {
      percent = 100 - (((end-start)*100)/end);
      printf("%3d%% of RAM filled\n", (percent-1));
    }

    /* Write data to DDR3 RAM */

    for (i=0;i<num_cmds;i++) {
      //printf("Write CMD to nr %x addr %x \n", i, start + i);
      if((fpgaVer>=44) && (ramSize == 1024)) {
        Ddr3WriteCmd_v44_1G(cardHandle, i, 0, start + i);
      } else {
        Ddr3WriteCmd(cardHandle, i, 0, start + i);
      }
    }
    for (i=0;i<(16 * num_cmds);i+=2) {
      if(start+ num_cmds > end) {
        break;
      }

      crc=crc32(crc);

      Ddr3WriteData(cardHandle, i, crc);
      Ddr3WriteData(cardHandle, i+1, ~crc);
      addr++;
    }

    /* execute */
    if (Ddr3ExecuteCmds(cardHandle, num_cmds) != 0) {
      printf("ERROR in executing DDR3 commands!\n");
      return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
    }
    start += num_cmds;
  }
  if (atoi(fullramtest)) {
    printf("Checking refresh controller by waiting %d seconds\n", atoi(fullramtest) );
    printf("before reading back and verifying RAM\n\n");
    sleep(atoi(fullramtest));
  }
  start = 0;
  end = 1024*1024*(ramSize/(16*4));
  percent=0;
  crc=0x12345678;
  printf("Read back and verify complete RAM\n");
  while (start < end) {
    num_cmds = ( end - start );
    if ( num_cmds > 15 ) {
      num_cmds = 15;
    }

    /* Exit if Ctrl-C has been pushed */
    if (stopApplication==1) {
      /* shutdown card */
      CardShutdown(0);
      DisplayProgramFooter();

      exit(-1);
    }

    /* Read data from DDR3 RAM */
    // Check progress
    if ((100 - (((end-start)*100)/end)) > percent) {
      percent = 100 - (((end-start)*100)/end);
      printf("%3d%% of the RAM verified\n", (percent-1));
    }

    for (i=0;i<num_cmds;i++) {
      if((fpgaVer>=44) && (ramSize == 1024)) {
        Ddr3WriteCmd_v44_1G(cardHandle, i, 1, start + i);
      } else {
        Ddr3WriteCmd(cardHandle, i, 1, start + i);
      }
    }

    /* execute */
    if (Ddr3ExecuteCmds(cardHandle, num_cmds) != 0) {
      printf("ERROR in executing DDR3 commands!\n");
      return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
    }
    if((fpgaVer>=44) && (ramSize == 1024)) {
      Ddr3WriteCmd_v44_1G(cardHandle, 0, 0, 0);
    } else {
      Ddr3WriteCmd(cardHandle, 0, 0, 0);
    }
    for (i=0;i<(16 * num_cmds);i+=2) {
      crc=crc32(crc);
      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, &value);
      value = value & 0xfffffc00;
      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, value | 0x300 | i);
      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_DATA_ADDR, &res);
      if ( res != crc ) {
        printf("ERROR\n");
        printf(" Wrong read data on addr=%08x \n", 16*start + i);
        printf(" Wrong read data=%08x \n", res);
        printf(" expected data=%08x \n", crc);
        printf("\nNOTE: Did you remember to reload the driver before running this test?\n");
        printf("      The SDRAM will fail if the card has been configured to run traffic before the test!\n");
        return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
      }

      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_ADDR_ADDR, value | 0x300 | (i+1));
      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_DDR3_BUFFER_DATA_ADDR, &res);

      if ( res != ~crc ) {
        printf(" Wrong read data on addr=%08x \n", 16*start + i+1);
        printf(" Wrong read data=%08x \n", res);
        printf(" expected data=%08x \n", ~crc);
        printf("\nNOTE: Did you remember to reload the driver before running this test?\n");
        printf("      The SDRAM will fail if the card has been configured to run traffic before the test!\n");
        return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
      }

    }
    start += num_cmds;
  }
  printf("RAM verify check test completed\n\n");

  return NTCI_ERRCODE_SUCCESS;
}

static uint32_t TimesyncTest(uint8_t cardIndex)
{
  handle_t cardHandle;
  uint32_t result, errorDetected=0;
  TimeSyncConfigurationType2_t config2;
  TimeSyncConfigurationType3_t config3;
  CardHardwareCardCapabilitiesType2_t capability;

  if (sensoronly) {
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  if (fullramtest) {
    // Timesync Test ignored when running full SDRAM test
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  GetCardHandle(cardIndex, &cardHandle);

  {
    CardServicesCapabilitiesType1_t services;

    /* A card has been found. Check if it has packetfeeds support. */
    if((result = NTCI_GetCapabilities(cardHandle, CAPID_SERVICES, NOT_SUB_CAPABILITIES,
                                      STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                                      &services)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_INFO, "Failed to read card services capability.\n");
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
    if(!services.packetFeedSupported) {
      /* Skip card it has no packetfeed support. */
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
  }

  result = NTCI_GetCapabilities(CardHandle,
                                CAPID_CARD_INDEXED,
                                0,
                                STID_CARD_HARDWARE_CARD_CAPABILITIES_TYPE2,
                                &capability);
  if(result != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed getting capability\n");
    return result;
  }
  if(!capability.TimeSyncSupported) {
    PRINT("No TimeSync support - TimeSync test skipped\n");
    return NTCI_ERRCODE_DIAGNOSTICS_NA;
  }

  result = NTCI_GetCardConfigurationDetails(cardHandle,
                                            0 /* Not needed. */,
                                            STID_TIMESYNC_CONFIGURATION_TYPE3,
                                            &config3);
  if(result != NTCI_ERRCODE_SUCCESS) {
    result = NTCI_GetCardConfigurationDetails(cardHandle,
                                              0 /* Not needed. */,
                                              STID_TIMESYNC_CONFIGURATION_TYPE2,
                                              &config2);

    if(result != NTCI_ERRCODE_SUCCESS) {
      PRINT("Failed reading time synchronization status. %i\n", result);
      return result;
    }
    PRINT("TimeSync Status:\n");
    if (config2.timesyncConfigMaster) {
      PRINT("Adapter is time synchronization Master.\n");
    }
    else {
      PRINT("Adapter is time synchronization Slave.\n");
      PRINT("Master present: %s\nMaster in synchronization: %s\n",
            (config2.link_active_status ? "Yes" : "No"),
            (config2.slave_in_sync ? "Yes" : "No"));
      if (config2.GPS_Box_is_Master) {
        PRINT("GPS Box is master\n");
        PRINT("GPS Box PPS is %s \n",
              (config2.GPS_Box_PPS_Ok ? "OK" : "not OK"));
        PRINT("GPS Box is %s synchronization\n",
              (config2.GPS_Box_In_sync ? "in" : "not in"));
        PRINT("GPS Box serial stream is %s\n",
              (config2.GPS_Box_Serial_Stream_OK ? "OK" : "not OK"));
      }
      if (config2.PTA_is_Master) {
        PRINT("PTA is master\n");
      }

      if (config2.OS_Mode_Sync_enabled) {
        PRINT("OS mode %s enabled\n",
              (config2.OS_Mode_Sync_enabled ? "is" : "is not"));
        PRINT("OS mode is %s synchronization\n",
              (config2.OS_Mode_Sync_enabled ? "in" : "not in"));
      }
      PRINT("Master Id code: %d\n",
            config2.Master_ID_Code);
    }
    if(config2.timesyncConfigMaster) {
      PRINT(">>> Master mode detected. Slave mode is required for this test\n");
      errorDetected=1;
    }
    if(!config2.link_active_status) {
      PRINT(">>> No NT timesync link detected\n");
      errorDetected=1;
    }
  }
  else {
    PRINT("TimeSync Status:\n");
    if (config3.timesyncConfigMaster) {
      PRINT("Adapter is time synchronization Master.\n");
    }
    else {
      PRINT("Adapter is time synchronization Slave.\n");
      PRINT("Master present: %s\nMaster in synchronization: %s\n",
            (config3.link_active_status ? "Yes" : "No"),
            (config3.slave_in_sync ? "Yes" : "No"));
      if (config3.GPS_Box_is_Master) {
        PRINT("GPS Box is master\n");
        PRINT("GPS Box PPS is %s \n",
              (config3.GPS_Box_PPS_Ok ? "OK" : "not OK"));
        PRINT("GPS Box is %s synchronization\n",
              (config3.GPS_Box_In_sync ? "in" : "not in"));
        PRINT("GPS Box serial stream is %s\n",
              (config3.GPS_Box_Serial_Stream_OK ? "OK" : "not OK"));
      }
      if (config3.PTA_is_Master) {
        PRINT("PTA is master\n");
      }
      if (config3.OS_Mode_Sync_enabled) {
        PRINT("OS mode %s enabled\n",
              (config3.OS_Mode_Sync_enabled ? "is" : "is not"));
        PRINT("OS mode is %s synchronization\n",
              (config3.OS_Mode_Sync_enabled ? "in" : "not in"));
      }
      PRINT("Master Id code: %d\n",
            config3.Master_ID_Code);
    }
    PRINT("External time: %llX\n",
          config3.External_Time);
    PRINT("Adapter time:  %llX\n",
          config3.Adapter_Time);
    if(config3.timesyncConfigMaster) {
      PRINT(">>> Master mode detected. Slave mode is required for this test\n");
      errorDetected=1;
    }
    if(!config3.link_active_status) {
      PRINT(">>> No NT timesync link detected\n");
      errorDetected=1;
    }
  }
  if(errorDetected) {
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }
  return NTCI_ERRCODE_SUCCESS;
}

//-----------------------------------------------------------------------------
//Check the status for the port indicated and return true if the port is ready.
//-----------------------------------------------------------------------------
bool CheckPort(handle_t hAdapter, uint PortNo, char* ErrorMsg)
{
  uint32_t             ErrorCode;
  ChannelStatusType1_t ChanStatus;

  strcpy(ErrorMsg, "");

  ErrorCode = NTCI_GetChannelStatus(hAdapter,
                                    1 << PortNo,
                                    STID_CHANNEL_STATUS_TYPE1,
                                    &ChanStatus);

  if (ErrorCode != NTCI_ERRCODE_SUCCESS)
  {
    sprintf(ErrorMsg, "Error reading channel status: %i\n", ErrorCode);
    return false;
  }

  if (ChanStatus.portType == PORT_TYPE_NOT_AVAILABLE)
  {
    sprintf(ErrorMsg, "No physical port found");
    return false;
  }

  if (ChanStatus.portType == PORT_TYPE_SFP_NOT_PRESENT)
  {
    sprintf(ErrorMsg, "No NIM module found");
    return false;
  }

  if (!ChanStatus.linkUp)
  {
    sprintf(ErrorMsg, "Link down");
    return false;
  }

  return true;
}

static uint32_t TrafficTest(uint8_t cardIndex)
{
  handle_t cardHandle;
  uint32_t result, errorDetected=0;
  uint32_t packetFeedHandle=0;
  uint32_t channel=0, count=0, pktLoop=0, done=0;
  uint32_t feedPriorities[] = { 1 };
  uint32_t value;
  FeedConfigSampledType3_t feedConfig;
  BufferInformationSectionType1_t bufferInformation;
  FeedBufferInformationType1_t bufferInfo;
  BufferRequestSampledType2_t request;
  CardHardwareCapabilitiesType1_t cardCapabilities;
  uint32_t channelsBitmask=0;
  CardHardwareVersionCapabilitiesType3_t data3;
  uint32_t sendPackets[8]={0,0,0,0,0,0,0,0};
  uint32_t receivedPackets[8]={0,0,0,0,0,0,0,0};
  uint64_t fpgaId=0;
  bool     AllLinksOk;
  char     ErrorMsg[128];
  TimestampConfigurationType1_t ts_conf;

  if (sensoronly) {
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  if (fullramtest) {
    // Traffictest ignored when running full SDRAM test
    return NTCI_ERRCODE_DIAGNOSTICS_IGNORED;
  }

  GetCardHandle(cardIndex, &cardHandle);

  {
    CardServicesCapabilitiesType1_t services;

    /* A card has been found. Check if it has packetfeeds support. */
    if((result = NTCI_GetCapabilities(cardHandle, CAPID_SERVICES, NOT_SUB_CAPABILITIES,
                                      STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                                      &services)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_INFO, "Failed to read card services capability.\n");
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
    if(!services.packetFeedSupported) {
      /* Skip card it has no packetfeed support. */
      return NTCI_ERRCODE_DIAGNOSTICS_NA;
    }
  }

  /* Shut down all packet feeds before starting the new one. */
  ShutdownAllPacketFeeds(cardHandle);

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_HARDWARE,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_CAPABILITIES_TYPE1,
                                    &cardCapabilities)) != NTCI_ERRCODE_SUCCESS) {

    PRINT("Error finding adapter capabilities: %i\n", result);
    return result;
  }
  channelsBitmask = cardCapabilities.totalChannelsBitmask;

  /* Check if we need to use NTPL to set up feeds. */
  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_VERSION,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                    &data3)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Error finding HW capabilities: %i\n", result);
    return result;
  }

  AllLinksOk = true;

  for (channel = 0; cardCapabilities.totalChannelsBitmask & (1 << channel); channel++)
  {
    if (!CheckPort(cardHandle, channel, ErrorMsg))
    {
      PRINT("Port %d:  %s\n", channel, ErrorMsg);
      AllLinksOk = false;
    }
  }

  if (!AllLinksOk)
  {
    PRINT(">>> All links are not ready - test aborted.\n");
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }

#ifdef WIN32
    switch (data3.AdapterCard.dwInfoType) {
    case 1:
      if((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 32) {
        HKEY  hKey;
        DWORD value;
        DWORD length;
        LONG  lRet;

        lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, MAIN_REG_KEY, 0, KEY_QUERY_VALUE, &hKey);
        if (lRet == ERROR_SUCCESS)
        {
          length = sizeof(DWORD);
          lRet = RegQueryValueEx(hKey, RESERVED_DMA_KEY, NULL, NULL, (LPBYTE)&value, &length);
          RegCloseKey(hKey);
          if (lRet == ERROR_SUCCESS) {
            if (value == 0) {
              channelsBitmask = 0;
            }
          } else {
            PrintError("Registry keys not found. Driver not installed correctly\n");
            exit(1);
          }
        } else {
          PrintError("Registry keys not found. Driver not installed correctly\n");
          exit(1);
        }
      }
      break;
    case 2:
      if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 32) {
        HKEY  hKey;
        DWORD value;
        DWORD length;
        LONG  lRet;

        lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, MAIN_REG_KEY, 0, KEY_QUERY_VALUE, &hKey);
        if (lRet == ERROR_SUCCESS) {
          length = sizeof(DWORD);
          lRet = RegQueryValueEx(hKey, RESERVED_DMA_KEY, NULL, NULL, (LPBYTE)&value, &length);
          RegCloseKey(hKey);
          if (lRet == ERROR_SUCCESS) {
            if (value == 0) {
              channelsBitmask = 0;
            }
          } else {
            PrintError("Registry keys not found. Driver not installed correctly\n");
            exit(1);
          }
        } else {
          PrintError("Registry keys not found. Driver not installed correctly\n");
          exit(1);
        }
      }
      break;
    case 3:
      if((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 32) {
        HKEY  hKey;
        DWORD value;
        DWORD length;
        LONG  lRet;

        lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, MAIN_REG_KEY, 0, KEY_QUERY_VALUE, &hKey);
        if (lRet == ERROR_SUCCESS) {
          length = sizeof(DWORD);
          lRet = RegQueryValueEx(hKey, RESERVED_DMA_KEY, NULL, NULL, (LPBYTE)&value, &length);
          RegCloseKey(hKey);
          if (lRet == ERROR_SUCCESS) {
            if (value == 0) {
              channelsBitmask = 0;
            }
          } else {
            PrintError("Registry keys not found. Driver not installed correctly\n");
            exit(1);
          }
        } else {
          PrintError("Registry keys not found. Driver not installed correctly\n");
          exit(1);
        }
      }
      break;
    default:
      fprintf(stderr, "Unsupported FPGA!\n");
      return -1;
    }
#else
    switch (data3.AdapterCard.dwInfoType) {
    case 1:
      if((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 32) {
        if(system("grep \"#disableMultiFeed\" /opt/napatech/config/default.cfg > /dev/null") != 0) {
          if(system("grep disableMultiFeed /opt/napatech/config/default.cfg | grep 0 > /dev/null") == 0) {
            channelsBitmask = 0;
          }
        } else {
          channelsBitmask = 0;
        }
      }
      break;
    case 2:
      if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 32) {
        if(system("grep \"#disableMultiFeed\" /opt/napatech/config/default.cfg > /dev/null") != 0) {
          if(system("grep disableMultiFeed /opt/napatech/config/default.cfg | grep 0 > /dev/null") == 0) {
            channelsBitmask = 0;
          }
        } else {
          channelsBitmask = 0;
        }
      }
      break;
    case 3:
      if((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 32) {
        if(system("grep \"#disableMultiFeed\" /opt/napatech/config/default.cfg > /dev/null") != 0) {
          if(system("grep disableMultiFeed /opt/napatech/config/default.cfg | grep 0 > /dev/null") == 0) {
            channelsBitmask = 0;
          }
        } else {
          channelsBitmask = 0;
        }
      }
      break;
    default:
      PRINT("Unsupported FPGA!\n");
      return -1;
    }
#endif

  /* Set up the packet feed engine for a single feed.
   *
   * The feedPriorities array is used to assign the
   * priorities of different feeds when creating multiple
   * simultaneous feeds; in this example it is redundant
   * as only one entry is required */
  if((result = NTCI_SetupPacketFeedEngine(cardHandle, 1, feedPriorities)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed to initialise packet feed engine (%i)\n", result);
    return result;
  }


  /* Configure packet feed */

  /* Read time stamp type */
  result = NTCI_GetTimestampConfiguration(cardHandle, 0, STID_TIMESTAMP_CONFIGURATION_TYPE1, &ts_conf);
  if (result != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed to retrieve time-stamp configuration (%u)\n", result);
    return result;
  }

  /* set timestamp type */
  feedConfig.timestampType = ts_conf.timestampType;

  /* slicing is disabled */
  feedConfig.sliceLength = 0;

  /* Select standard descriptor type */
  feedConfig.descriptorType = STANDARD_DESCRIPTOR;
  feedConfig.extendedDescriptorFormat = 0;
  feedConfig.extendedDescriptorLength = 0;

  /* Set latency to 1000us */
  feedConfig.maxLatency = 1000;

  /* Create packet feed */
  if((result = NTCI_CreatePacketFeed(cardHandle,
                                     0,
                                     channelsBitmask,
                                     0,  // Reserved0 (unused parameter)
                                     0,  // Reserved1 (unused parameter)
                                     STID_FEED_CONFIG_SAMPLED_TYPE3,
                                     &feedConfig,
                                     &packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed to create a packet feed (%i)\n", result);
    return result;
  }

  if(channelsBitmask==0) {
    PassFilter_t command;
    command.WriteHW = TRUE;
    sprintf(command.achFilterString, "Capture[Priority=0;Feed=0]=ALL");

    if((result = NTCI_PacketClassification(cardHandle,
                                           STID_PASS_FILTER,
                                           &command)) != NTCI_ERRCODE_SUCCESS) {
      FilterError_t error;
      result = NTCI_PacketClassification(cardHandle,
                                         STID_GET_FILTER_ERROR,
                                         &error);
      if (result != NTCI_ERRCODE_SUCCESS) {
        PRINT("Failed to classify packet (%u)\n", result);
        return result;
      }

      PRINT("%s\n",error.achFilterError1);
      PRINT("%s\n",error.achFilterError2);
      PRINT("%s\n",error.achFilterError3);
      NTCI_CloseCard(cardHandle);
    }
  }

  if((result = NTCI_StartPacketFeed(cardHandle, packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Error while starting packet feed (%i)\n",
            result);
    return result;
  }

  /* find the location of the feed buffer where new
   * data will be stored */
  if((result = NTCI_GetPacketFeedBuffers(cardHandle,
                                         packetFeedHandle,
                                         STID_FEED_BUFFER_INFORMATION_TYPE1,
                                         &bufferInfo)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Unable to retrieve the location of the feed "
            "buffer.\n");
    return result;
  }

  if(data3.AdapterCard.dwInfoType==1) {
    fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==2) {
    fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==3) {
    fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
  }

  if (rxaui_loopback) {
    // Should only be done for NT20E2 cards
    if (((fpgaId&0xFFFF0000) >> 16) == 9020) {
      printf("Enabling RXAUI loopback\n");
      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI0_ADDR_BASE, &value);
      value = value | 0x00000001;
      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI0_ADDR_BASE, value);

      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI1_ADDR_BASE, &value);
      value = value | 0x00000001;
      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI1_ADDR_BASE, value);
    }
  }

  pktLoop = 0;
  /* We need to send NUM_PACKETS_TO_SEND packets on all interfaces. */
  while((pktLoop++ < NUM_PACKETS_TO_SEND) &&
        (stopApplication==0)){
    /* Send a packet on all interfaces. */
    channel=0;
    printf("Sending packet: %05d of %05d\r", pktLoop, NUM_PACKETS_TO_SEND);
    fflush(stdout);
    for(;;) {
      if(cardCapabilities.totalChannelsBitmask&(1<<channel)) {
        if((result = NTCI_SendPacket(cardHandle, channel,
                                     tcp_packet,
                                     sizeof(tcp_packet))) != NTCI_ERRCODE_SUCCESS) {
          PRINT("Failed to send packet on channel %d\n", channel);
          errorDetected=1;
          break;
        }
        sendPackets[channel]++;
      }else {
        break;
      }
      channel++;
    }
    if(errorDetected) {
      break;
    }
    {
      uint32_t badCells=0;

      if(data3.AdapterCard.dwInfoType==2) {
        if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) <= 34) {
          NTCI_ReadRegister32(cardHandle, 2, 0xA404E600, &badCells);
        } else if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) <= 38){
          NTCI_ReadRegister32(cardHandle, 4, 0x0240E600, &badCells);
        } else if((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) <= 40){
          NTCI_ReadRegister32(cardHandle, 4, 0x02405800, &badCells);
        } else {
          if(((data3.AdapterCard.u.InfoType2.qwFpgaId&0xFFFF0000)>>16) == 9220) {
            NTCI_ReadRegister32(cardHandle, 4, 0x20005800, &badCells);
          } else {
            NTCI_ReadRegister32(cardHandle, 4, 0x02405800, &badCells);
          }
        }
      } else if(data3.AdapterCard.dwInfoType==3) {
        if(((data3.AdapterCard.u.InfoType2.qwFpgaId&0xFFFF0000)>>16) == 9220) {
          NTCI_ReadRegister32(cardHandle, 4, 0x20005800, &badCells);
        } else {
          NTCI_ReadRegister32(cardHandle, 4, 0x02405800, &badCells);
        }
      }

      if(badCells!=0) {
        PRINT("######### Feed %d has %d bad SDRAM cells. #########.\n", 0, badCells);
        errorDetected=1;
        break;
      }
    }
#ifdef WIN32
          count = 10000; /* Poll 10000 times. */
#else
    count = 1000;  /* Poll 1000 times. */
#endif
    while((count--) && (stopApplication==0)) {
      request.common.waitTimeoutMS=GET_NEXT_BUFFER_BLOCKING;
      result = NTCI_GetNextBuffer(cardHandle,
                                  packetFeedHandle,
                                  STID_BUFFER_REQUEST_SAMPLED_TYPE2,
                                  &request,
                                  STID_BUFFER_INFORMATION_SECTION_TYPE1,
                                  &bufferInformation);

      if(result == NTCI_ERRCODE_SUCCESS) {
        PacketDescriptorNtSeriesType1_t *descriptor=NULL;
        uint32_t offset=0, i=0;
        for (i=0; i<bufferInformation.numDescriptorsAvailable; i++) {
          /* Add the new data into the packet buffer */
          descriptor = (PacketDescriptorNtSeriesType1_t *)(((uint8_t*)bufferInfo.bufferBaseAddress)+bufferInformation.sectionOffset+offset);

          if(descriptor->wireLength != (sizeof(tcp_packet)+4)) {
            PRINT(">>> Packet not received with correct wireLength (%d). Expected: %d\n",
                  descriptor->wireLength,
                  sizeof(tcp_packet));
            errorDetected=1;
            break;
          }
          receivedPackets[descriptor->channel]++;

          offset += descriptor->storedLength;
        }
        NTCI_ReleaseBuffer(cardHandle,
                           packetFeedHandle,
                           STID_BUFFER_INFORMATION_SECTION_TYPE1,
                           &bufferInformation);

        /* Check if everything has been received. */
        channel=0;
        done=1;
        for(;;) {
          if(cardCapabilities.totalChannelsBitmask&(1<<channel)) {
            if(sendPackets[channel] > receivedPackets[channel]) {
              done=0;
            } else if(sendPackets[channel] < receivedPackets[channel]) {
              errorDetected=1;
              PRINT(">>> Channel %d has sent %d packets but received %d. This is an error. Are you connected to a switch?\n",
                    channel, sendPackets[channel], receivedPackets[channel]);
            }
          } else {
            break;
          }
          channel++;
        }
        if(done) {
          break;
        }
        if(errorDetected) {
          break;
        }
      } else if(result != NTCI_STATUSCODE_NO_SEGMENT_AVAILABLE) {
        break;
      }
    }
  }
  if(stopApplication) {
    PRINT(">>> Application stopped by user.\n");
    errorDetected=1;
  }

  if (rxaui_loopback) {
    // Should only be done for NT20E2 cards
    if (((fpgaId&0xFFFF0000) >> 16) == 9020) {
      printf("Disabling RXAUI loopback\n");
      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI0_ADDR_BASE, &value);
      value = value & 0xFFFFFFFE;
      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI0_ADDR_BASE, value);

      NTCI_ReadRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI1_ADDR_BASE, &value);
      value = value & 0xFFFFFFFE;
      NTCI_WriteRegister32(cardHandle, NAPA_NMB, NT_REG_P9020_RXAUI1_ADDR_BASE, value);
    }
  }

  printf("                                                                    \r");
  NTCI_StopPacketFeed(cardHandle, packetFeedHandle);
  if((result = NTCI_DestroyPacketFeed(cardHandle, packetFeedHandle)) != NTCI_ERRCODE_SUCCESS) {
    PRINT("Failed to destroy packet feed! (%i)\n", result);
    return result;
  }

  if(channelsBitmask==0) {
    PassFilter_t command;
    command.WriteHW = TRUE;
    sprintf(command.achFilterString, "DeleteFilter=All");

    if((result = NTCI_PacketClassification(cardHandle,
                                           STID_PASS_FILTER,
                                           &command)) != NTCI_ERRCODE_SUCCESS) {
      FilterError_t error;
      result = NTCI_PacketClassification(cardHandle,
                                         STID_GET_FILTER_ERROR,
                                         &error);

      PRINT("%s\n",error.achFilterError1);
      PRINT("%s\n",error.achFilterError2);
      PRINT("%s\n",error.achFilterError3);
      return result;
    }
  }

  channel=0;
  for(;;) {
    if(cardCapabilities.totalChannelsBitmask&(1<<channel)) {
      if(sendPackets[channel] != receivedPackets[channel]) {
        errorDetected=1;
        PRINT(">>> Channel %d send %d packets. Received %d\n", channel,
              sendPackets[channel], receivedPackets[channel]);
      }
      PRINT("Channel %d send %d packets. Received %d\n", channel,
            sendPackets[channel], receivedPackets[channel]);
    } else {
      break;
    }
    channel++;
  }

  if(errorDetected) {
    return NTCI_ERRCODE_DIAGNOSTICS_ERROR;
  }
  return NTCI_ERRCODE_SUCCESS;
}


int main(int argc, char *argv[])
{
  uint32_t  cardIndex=0, result=0, test=0;
  handle_t  cardHandle;
  char      *ProgramName = argv[0];
  char      line[80];
  bool_t    showOutput=FALSE;

  /* Parse the command line and initialize the needed variables. */
  if(InitApplication(&argc, argv, FALSE /* any card. */, &cardIndex,
                     tools_options, tool_description,
                     output_description) == -1) {
    exit(-1);
  }

  if((result = GetCardHandle(cardIndex, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, ">>> Error: Failed to get card handle.\n");
    goto errorHandling;
  }

  DisplayProgramHeader(ProgramName, PRG_TOOL);

  if(verbose) {
    showOutput=1;
  }

#ifdef WIN32
  SetConsoleCtrlHandler((PHANDLER_ROUTINE)CatchSignalCTRL_C, TRUE);
#else
  signal(SIGINT, CatchSignalCTRL_C);
#endif

  /* Run the tests. */
  while((Tests[test].pfTest != NULL) && (stopApplication==0)) {
    writePos=0;
    PRINT("--------------------------------------------------------------------------------\n");
    PRINT("Testing: %s - %s\n", Tests[test].testname, Tests[test].testdescription);
    result = Tests[test].pfTest(cardIndex);
    PRINT("--------------------------------------------------------------------------------\n");
    if(result != NTCI_ERRCODE_SUCCESS) {
      if(result == NTCI_ERRCODE_DIAGNOSTICS_NA) {
        snprintf(line, sizeof(line), ">>> Test: %s cannot be run on the adapter.\n", Tests[test].testname);
        PrintNotTested(line);
      } else {
        if(result == NTCI_ERRCODE_DIAGNOSTICS_IGNORED) {
          snprintf(line, sizeof(line), ">>> Test: %s ignored\n", Tests[test].testname);
          if(showOutput) {
            WriteToLog(LOG_INFO, screenBuffer);
          }
          PrintNotTested(line);
        } else {
          WriteToLog(LOG_INFO, screenBuffer);
          if(result != NTCI_ERRCODE_DIAGNOSTICS_ERROR) {
            fprintf(stderr, ">> NTCI function call error.\n");
          }
          snprintf(line, sizeof(line), ">>> Test: %s failed\n", Tests[test].testname);
          PrintError(line);
        }
      }
    } else {
      snprintf(line, sizeof(line), ">>> Test: %s passed\n", Tests[test].testname);
      if(showOutput) {
        WriteToLog(LOG_INFO, screenBuffer);
      }
      PrintPass(line);
    }
    test++;

  }
  printf("--------------------------------------------------------------------------------\n");
  printf("Done with the test suite.\n");
  printf("--------------------------------------------------------------------------------\n");

errorHandling:
  /* shutdown card */
  CardShutdown(0);
  DisplayProgramFooter();

  return 0;
}
