// $Id: LinkTool.c 24595 2013-07-09 08:34:11Z am $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * This tools demonstrates link query/configuration options
 */

#ifdef _WIN32
#include <windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <signal.h>
#include <time.h>
#include <termios.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#endif

#include <toolslib.h>

/* The NTCommonInterface.h header provides the interface definitions
 * for the Napatech NTCI library */

#include <NTCommonInterface.h>
#include <cardstatusstructure.h>
#include <cardconfigurationstructure.h>
#include <NTCI_cardstatus.h>
#include <NTCI_capabilities.h>

static char *req_channelBitmask=NULL;
static char *req_cmd=NULL;
static char *req_LinkOptionsBitmask=NULL;
static char *req_MdiMode=NULL;

static CommandLineOptions_t tools_options[] = {
  {"cb",  "Channel bitmask. Set one bit only", "0x01..0x80",  0, &req_channelBitmask},
  {"cmd", "Channel link command", "SetAuto|SetManual|Get|EnablePortID|DisablePortID|GetPortIDStatus",  1, &req_cmd},
  {"link", "Link options bitmask:\n"
   "                 Set default        : 0x00 (Not allowed for manual speed and duplex setup)\n"
   "                 1G full duplex     : 0x01\n"
   "                 100M full duplex   : 0x04\n"
   "                 100M half duplex   : 0x08\n"
   "                 10M full duplex    : 0x10\n"
   "                 10M half duplex    : 0x20",
   "0..0x3F", 0, &req_LinkOptionsBitmask},
  {"mdi", "MDI mode:\n"
   "                 Auto               : 0\n"
   "                 MDI                : 1\n"
   "                 MDIX               : 2\n",
   "0|1|2", 0, &req_MdiMode},
  {NULL, NULL, NULL, 0, NULL}};

static char tool_description[] = {"Display link status and capabilities and setup link options\n"};
static char output_description[] = {"\n"};

static uint32_t MenuRunning;
struct MenuEntry
{
    char Key;
    char const *Description;
    uint32_t (*Command)(uint8_t cardIndex, uint32_t channelsBitmask);
};


uint32_t Quit (uint8_t cardIndex __attribute__((unused)),
               uint32_t channelsBitmask __attribute__((unused)))
{
    MenuRunning = 0;
    return NTCI_ERRCODE_SUCCESS;
}

uint32_t GetPortIDStatus(uint8_t cardIndex, uint32_t channelMask)
{
  return GetPortIdentificationStatus(cardIndex, channelMask);
}

uint32_t SetPortIDStatus(uint8_t cardIndex, uint32_t channelMask, uint32_t enable)
{
  return SetPortIdentificationStatus(cardIndex, channelMask, enable);
}

uint32_t GetLinkSetup(uint8_t cardIndex, uint32_t channelMask)
{
  uint32_t result;

  result = GetChannelStatus(cardIndex, channelMask);
  if (result != NTCI_ERRCODE_SUCCESS) {
    return result;
  }
  result = GetAutonegotiationMode(cardIndex, channelMask);
  return result;
}

uint32_t SetAuto(uint8_t cardIndex, uint32_t channelsBitmask, uint32_t linkOptions, uint8_t mdiMode)
{
  uint32_t result;
  uint8_t enable = 1;
  uint8_t required = 0;
  uint8_t restartAutoNeg = 1;
  uint8_t advertise10000FullDuplex = 0;
  uint8_t       advertise10000HalfDuplex = 0;
  uint8_t       advertise1000FullDuplex = 0;
  uint8_t       advertise1000HalfDuplex = 0;
  uint8_t       advertise100FullDuplex = 0;
  uint8_t       advertise100HalfDuplex = 0;
  uint8_t       advertise10FullDuplex = 0;
  uint8_t       advertise10HalfDuplex = 0;
  uint8_t pauseControlMode = 0;
  uint8_t       manualFullDuplex = 0;
  uint8_t       manualSpeed = 0;
  uint8_t setdefault = 0;

  if (linkOptions == 0) {
    setdefault = 1;
  } else {
    if (linkOptions & 0x1) {
      advertise1000FullDuplex = 1;
    }
    if (linkOptions & 0x2) {
      /* 1G Half duplex is not supported */
      fprintf(stderr, ">>> Error: \"%s\" command has no support for link option 0x%02X (bit 0x%02X is enabled)\n", req_cmd, linkOptions, (linkOptions & 0x2));
      return -1;
    }
    if (linkOptions & 0x4) {
      advertise100FullDuplex = 1;
    }
    if (linkOptions & 0x8) {
      advertise100HalfDuplex = 1;
    }
    if (linkOptions & 0x10) {
      advertise10FullDuplex = 1;
    }
    if (linkOptions & 0x20) {
      advertise10HalfDuplex = 1;
    }
  }

  result = SetAutonegotiationMode(cardIndex, channelsBitmask,
                                  enable, required, restartAutoNeg,
                                  advertise10000FullDuplex, advertise10000HalfDuplex,
                                  advertise1000FullDuplex, advertise1000HalfDuplex,
                                  advertise100FullDuplex, advertise100HalfDuplex,
                                  advertise10FullDuplex, advertise10HalfDuplex,
                                  pauseControlMode, manualFullDuplex,
                                  manualSpeed, mdiMode,
                                  setdefault);
  if (result == NTCI_ERRCODE_SUCCESS) {
    printf("Autonegotiation settings changed succesfully \n");
  } else {
    printf("Autonegotiation settings could not be changed - Check if PktTsFirst has been set in default.cfg\n");
  }

  return result;
}


uint32_t SetManual(uint8_t cardIndex, uint32_t channelsBitmask, uint32_t linkOptions, uint8_t mdiMode)
{
  uint32_t result;
  uint8_t enable = 0;
  uint8_t required = 0;
  uint8_t restartAutoNeg = 0;
  uint8_t advertise10000FullDuplex = 0;
  uint8_t       advertise10000HalfDuplex = 0;
  uint8_t       advertise1000FullDuplex = 0;
  uint8_t       advertise1000HalfDuplex = 0;
  uint8_t       advertise100FullDuplex = 0;
  uint8_t       advertise100HalfDuplex = 0;
  uint8_t       advertise10FullDuplex = 0;
  uint8_t       advertise10HalfDuplex = 0;
  uint8_t pauseControlMode = 0;
  uint8_t       manualFullDuplex = 0;
  uint8_t       manualSpeed = 0;
  uint8_t setdefault = 0;


  if (linkOptions & 0x1) {
    /* 1G full duplex */
    manualSpeed = MANUAL_SPEED_1000;
    manualFullDuplex = MANUAL_DUPLEX_FULL;
  } else  if (linkOptions & 0x4) {
    /* 100 Mb full duplex */
    manualSpeed = MANUAL_SPEED_100;
    manualFullDuplex = MANUAL_DUPLEX_FULL;
  } else if (linkOptions & 0x8) {
    /* 100 Mb half duplex */
    manualSpeed = MANUAL_SPEED_100;
    manualFullDuplex = MANUAL_DUPLEX_HALF;
  } else if (linkOptions & 0x10) {
    /* 10 Mb full duplex */
    manualSpeed = MANUAL_SPEED_10;
    manualFullDuplex = MANUAL_DUPLEX_FULL;
  } else if (linkOptions & 0x20) {
    /* 10 Mb half duplex */
    manualSpeed = MANUAL_SPEED_10;
    manualFullDuplex = MANUAL_DUPLEX_HALF;
  } else {
    fprintf(stderr, ">>> Error: \"%s\" link option cannot be 0x%02X.\n", req_cmd, (linkOptions & 0x2));
    return NTCI_ERRCODE_SUCCESS;
  }



  result = SetAutonegotiationMode(cardIndex, channelsBitmask,
                         enable, required, restartAutoNeg,
                         advertise10000FullDuplex, advertise10000HalfDuplex,
                         advertise1000FullDuplex, advertise1000HalfDuplex,
                         advertise100FullDuplex, advertise100HalfDuplex,
                         advertise10FullDuplex, advertise10HalfDuplex,
                         pauseControlMode, manualFullDuplex,
                         manualSpeed, mdiMode,
                         setdefault);

  if (result == NTCI_ERRCODE_SUCCESS) {
    printf("Manual link settings changed succesfully \n");
  } else {
    printf("Cannot change manual link settings on fiber port\n");
  }

  return result;
}


struct MenuEntry Menu[] = {
    {'q', "Quit",                    Quit} /* Must be last */
};

void PrintSeparator()
{
    int i;

    for (i=0; i<78; ++i)
        putchar('-');
    putchar('\n');
}

void PaddedPrintf(int columns, const char *s, ...)
{
    va_list args;
    int i;

    va_start(args, s);

    i = vprintf(s, args);

    va_end(args);

    /* pad out to reach the required number of columns */

    for (; i<columns; ++i)
        putchar(' ');
}

void DisplayKeyOptions()
{
    int32_t i;

    printf("\n");
    PrintSeparator();

    /* Display list of keyboard shortcuts */

    for (i=0; i < (int32_t)((sizeof(Menu) / sizeof(struct MenuEntry)) - 1) ; i++)
    {
        PaddedPrintf(30, "%c - %s", toupper(Menu[i].Key), Menu[i].Description);
        /* Two menu options per line */
        if ((i+1) % 2 == 0)
            printf("\n");
    }

    /* Print the last menu option (Q - Quit) on separate line */
    if ( ((sizeof(Menu) / sizeof(struct MenuEntry)) - 1) % 2 != 0 ) {
      printf("\n");
    }
    printf("%c - %s\n", toupper(Menu[i].Key), Menu[i].Description);
}


int main(int argc, char *argv[])
{
  uint32_t         channel=1;
  uint32_t         cardIndex;
  uint32_t         linkoptions=0;
  uint32_t         mdiMode=0;
  int32_t          result;
  handle_t         cardHandle;
  bool_t           setlinksupport=1;
  CardHardwareVersionCapabilitiesType3_t data3;
  uint64_t fpgaId=0;


  /* Parse the command line and initialize the needed variables. */
  if(InitApplication(&argc, argv, TRUE /* capture capable cards only. */, &cardIndex,
                     tools_options, tool_description,
                     output_description) == -1) {
    exit(-1);
  }

  if((result = GetCardHandle(cardIndex, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, ">>> Error: Failed to get card handle.\n");
    goto errorHandling;
  }

  DisplayProgramHeader(argv[0], PRG_TOOL);

  if(req_channelBitmask) {
    channel = AsciiToInt(req_channelBitmask);
  }

  if(channel==0) {
    fprintf(stderr, ">>> Error: The Channel bitmask (-cb) cannot be zero.\n");
    goto errorHandling;
  }

  if(req_LinkOptionsBitmask) {
    linkoptions = AsciiToInt(req_LinkOptionsBitmask);
  }

  if(req_MdiMode) {
    mdiMode = AsciiToInt(req_MdiMode);
  }

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_VERSION,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                    &data3)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "NTCI_GetCapabilities returned error: %i\n", result);
    goto errorHandling;
  }

  if(data3.AdapterCard.dwInfoType==1) {
    fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==2) {
    fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==3) {
    fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
  }

  switch((fpgaId&0xFFFF0000) >> 16) {
    case 9008:
    case 9014:
    case 9114:
    case 9020:
    case 9220:
    case 9221:
    case 9222:
    {
      setlinksupport = 0;
      break;
    }

    case 9015:
    case 9115:
    case 9017:
    case 9117:
       break;
    default:
    {
      fprintf(stderr, "Unknown nic type\n");
      goto errorHandling;
    }
  }

  //   // check channel bitmask
  if ( ValidateAdapterChannelBitmask(cardHandle, channel) != 0) {
    goto errorHandling;
  }

  if(req_cmd != NULL) {
    if(!strcasecmp("EnablePortID", req_cmd)) {
      result = SetPortIDStatus(cardIndex, channel, 1);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, ">>> Error: Port Identification feature is only supported by the NT20 family adapters\n");
        goto errorHandling;
      }
    } else if(!strcasecmp("DisablePortID", req_cmd)) {
      result = SetPortIDStatus(cardIndex, channel, 0);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, ">>> Error: Port Identification feature is only supported by the NT20 family adapters\n");
        goto errorHandling;
      }
    } else if(!strcasecmp("GetPortIDStatus", req_cmd)) {
      result = GetPortIDStatus(cardIndex, channel);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, ">>> Error: Port Identification feature is only supported by the NT20 family adapters\n");
        goto errorHandling;
      }
    } else if(!strcasecmp("Get", req_cmd)) {
      result = GetLinkSetup(cardIndex, channel);
      if (result != NTCI_ERRCODE_SUCCESS) {
        fprintf(stderr, ">>> Error: Cannot retrieve link set-up\n");
        goto errorHandling;
      }
    } else if(!strcasecmp("SetAuto", req_cmd)) {
      if(setlinksupport) {
        if (SetAuto(cardIndex, channel, linkoptions, mdiMode) !=  NTCI_ERRCODE_SUCCESS) {
          goto errorHandling;
        }
      } else {
        fprintf(stderr,">>> Error: Changing link settings is NOT supported on NT20e2, NT40e2, NT20e and NT20X\n");
        goto errorHandling;
      }
    } else if(!strcasecmp("SetManual", req_cmd)) {
      if(setlinksupport) {
        if (SetManual(cardIndex, channel, linkoptions, mdiMode) != NTCI_ERRCODE_SUCCESS) {
          goto errorHandling;
        }
      } else {
        fprintf(stderr,">>> Error: Changing link settings is NOT supported on NT20e2, NT40e2, NT20e and NT20X\n");
        goto errorHandling;
      }
    } else {
      fprintf(stderr, ">>> Error: \"%s\" is an unknown \"-cmd\".\n", req_cmd);
      goto errorHandling;
    }
  }



  errorHandling:
  /* shutdown card */
  CardShutdown(0);

  DisplayProgramFooter();

  return 0;
}
