// $Id: Statistics.c 25600 2013-09-19 07:40:19Z mh $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * This tool demonstrates a RMON/HW statistics dump.
 */

#ifdef _WIN32
#include <windows.h>
#include "platformwin.h"
#else 
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <signal.h>
#include <time.h>
#include <termios.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>

#endif

#include "toolslib.h"
#include "NTCI_capabilities.h"
#include "capabilitiesstructure.h"

/* The NTCommonInterface.h header provides the interface definitions
 * for the Napatech NTCI library */

#include "NTCommonInterface.h"

bool_t hex=TRUE;
bool_t boBlockStatSupported=FALSE;

static char *req_channelBitmask=NULL;
static char *req_num=NULL;
static char *req_time=NULL;
static char *req_cmd=NULL;
static char *req_bcmd=NULL;

static char *interactive_mode=NULL;
static char *req_blockChannelsBitmask=NULL;
static char *req_blockFilterBitmask=NULL;
static char *req_blockFeedsBitmask=NULL;
static char *summary=NULL;
static char *dec=NULL;
static char *tx=NULL;
static char *legacy=NULL;

static CommandLineOptions_t tools_options[] = {
  {"cb",  "Channel bitmask.", "0x01..0xFF",  0, &req_channelBitmask},
  {"cmd", "Channel stat command.", "Start|Stop|Reset",  0, &req_cmd},
  {"bch", "Block statistics : \"Channels bitmask\" (overwrites \"cb\").", "0..0xFF", 0, &req_blockChannelsBitmask},
  {"bfi", "Block statistics : \"Filter bitmask\"   (overwrites \"cb\").", "0..0xFFFFFFFFFFFFFFFF", 0, &req_blockFilterBitmask},
  {"bfe", "Block statistics : \"Feeds bitmask\"    (overwrites \"cb\").", "0..0xFFFFFFFF", 0, &req_blockFeedsBitmask},
  {"bcmd","Block statistics : \"Command\"          (overwrites \"cmd\"). ", "Start|Stop|Reset",  0, &req_bcmd},
  {"n",   "Number of consecutive statistics readouts (default is 1).", "1..9999",  0, &req_num},
  {"t",   "Time in milliseconds between readouts (default is 1000).", "1..999999",  0, &req_time},
  {"interactive", "Interactive mode (continuous readout - overrides option \"n\").", NULL,  2, &interactive_mode},
  {"summary", "Summarise the counters within each group.", NULL, 2, &summary},
  {"dec", "Show counters as decimal.", NULL, 2, &dec},
  {"legacy", "Show the legacy statistics using STID_ETHER_STATISTICS_TYPE2.", NULL, 2, &legacy},  
  {"tx", "Show TX statistics. (Requires the \"cb\" option too.)", NULL, 2, &tx},
  {NULL, NULL, NULL, 0, NULL}};

static char tool_description[] = {"Display adapter statistics.\n"};
static char output_description[] = {"\n"};

static uint32_t MenuRunning;
static bool_t   DisplayMode=0;
struct MenuEntry 
{
  char Key;
  const char *Description;
  uint32_t (*Command)(uint8_t cardIndex, uint32_t channelsBitmask);
};

uint32_t Noop (uint8_t cardIndex __attribute__((unused)),
               uint32_t channelsBitmask __attribute__((unused)))
{
  return NTCI_ERRCODE_SUCCESS;
}


uint32_t Quit (uint8_t cardIndex __attribute__((unused)),
               uint32_t channelsBitmask __attribute__((unused)))
{
  MenuRunning = 0;
  return NTCI_ERRCODE_SUCCESS;
}

uint32_t Toggle (uint8_t cardIndex __attribute__((unused)),
                 uint32_t channelsBitmask __attribute__((unused)))
{
  DisplayMode = !DisplayMode;
  return NTCI_ERRCODE_SUCCESS;
}

uint32_t Reset (uint8_t cardIndex, uint32_t channelsBitmask)
{
  StatisticsReset(cardIndex, channelsBitmask);
  return NTCI_ERRCODE_SUCCESS;
}

uint32_t Start(uint8_t cardIndex, uint32_t channelsBitmask)
{
  StatisticsStart(cardIndex, channelsBitmask);
  return NTCI_ERRCODE_SUCCESS;
}

uint32_t Stop(uint8_t cardIndex, uint32_t channelsBitmask)
{
  StatisticsStop(cardIndex, channelsBitmask);
  return NTCI_ERRCODE_SUCCESS;
}


struct MenuEntry Menu[] = {
    {'t', "Toggle RMON/HW counters",  Toggle},
    {'r', "Reset counters",           Reset},
    {'s', "Start statistics",         Start},
    {'o', "Stop statistics",          Stop},
    {'q', "Quit",                     Quit} /* Must be last */
};

void PrintSeparator()
{
    int i;

    for (i=0; i<78; ++i)
        putchar('-');
    putchar('\n');
}

void PaddedPrintf(int columns, const char *s, ...)
{
    va_list args;
    int i;

    va_start(args, s);

    i = vprintf(s, args);

    va_end(args);

    /* pad out to reach the required number of columns */

    for (; i<columns; ++i)
        putchar(' ');
}

void DisplayKeyOptions()
{
    uint32_t i;

    printf("\n");
    PrintSeparator();        

    /* Display list of keyboard shortcuts */

    for (i=0; i < ((sizeof(Menu) / sizeof(struct MenuEntry)) - 1) ; i++)
    {
      if(Menu[i].Key!='\0' && Menu[i].Description!=NULL) {
        if ((legacy != NULL) && (i == 0)) {
          // Do nothing
        } else {
          PaddedPrintf(30, "%c - %s", toupper(Menu[i].Key), Menu[i].Description);
          /* Two menu options per line */
          if ((i+1) % 2 == 0)
          printf("\n");
        }
      }
    }    
    
    /* Print the last menu option (Q - Quit) on separate line */
    if ( ((sizeof(Menu) / sizeof(struct MenuEntry)) - 1) % 2 != 0 ) {
      printf("\n");
    }
    printf("%c - %s\n", toupper(Menu[i].Key), Menu[i].Description);
}

#ifdef _WIN32

#include <windows.h>

uint32_t RunMenu (uint32_t cardIndex, uint32_t channel, handle_t cardHandle, uint32_t t, char *prgName, ePrgType prgType)
{
    uint32_t i;

    for (MenuRunning = 1; MenuRunning; )
    {
        DWORD eventsRead=0;

        ClearScreen();
        DisplayProgramHeader(prgName, prgType);
        printf("%s statistics:\n", tx?"TX":"RX");
        if (legacy != NULL) {
          StatisticsGetSummary(cardIndex, channel, tx?0:1);          
          StatisticsGetType2Counters(cardIndex, channel, hex);          
        } else {

          if (DisplayMode == 0) {
            /* Summary and RMON Counters */
            StatisticsGetSummary((uint8_t)cardIndex, channel, tx?0:1);
            StatisticsGetRMONCounters((uint8_t)cardIndex, channel, hex, tx?0:1);
          } else {
            /* Display HW counters only  */
            StatisticsGetSummary((uint8_t)cardIndex, channel, tx?0:1);              
            StatisticsGetHwCounters((uint8_t)cardIndex, channel, hex, tx?0:1); 
          }
        }
        
        DisplayKeyOptions();

        /* Wair for key press */
        WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), t);

        /* Read console input */
        GetNumberOfConsoleInputEvents(GetStdHandle(STD_INPUT_HANDLE), &eventsRead);

        if (eventsRead > 0)
        {
            INPUT_RECORD event;
            uint32_t key;

            ReadConsoleInput(GetStdHandle(STD_INPUT_HANDLE),
                &event,
                1,
                &eventsRead);

            if (eventsRead > 0 && event.EventType == KEY_EVENT && event.Event.KeyEvent.bKeyDown)
            {
                key = event.Event.KeyEvent.uChar.UnicodeChar;

                for (i=0; i<sizeof(Menu) / sizeof(struct MenuEntry); ++i)
                {
                    if (toupper(key) == toupper(Menu[i].Key))
                        Menu[i].Command((uint8_t)cardIndex, channel);
                }
            }
        }
    }
    return NTCI_ERRCODE_SUCCESS;
}


#else 

uint32_t RunMenu (uint32_t cardIndex, uint32_t channel, NOUSE handle_t cardHandle, uint32_t t, char *prgName, ePrgType prgType)
{
    struct termios original_io, current_io;
    uint8_t        i, num_channels=0;

    for(i=0;i<8;i++) {
      if (channel & 1<<i) {
        num_channels++;
      }
    }
    
    /* save the state of the terminal */

    tcgetattr(fileno(stdin), &original_io);
    
    /* set the terminal into unbuffered mode */

    current_io = original_io;
    current_io.c_cc[VTIME] = 0;        /* dont wait for enter */
    current_io.c_cc[VMIN] = 1;         /* min 1 character to trigger input */
    current_io.c_lflag &= ~ECHO;       /* terminal echo off */
    current_io.c_lflag &= ~ICANON;     /* unbuffered */
    tcsetattr(fileno(stdin), TCSANOW, &current_io);

    for (MenuRunning = 1; MenuRunning; )
    {
        fd_set fd_list;
        int32_t result;
        struct timeval tv;

        ClearScreen();
        DisplayProgramHeader(prgName, prgType);
        printf("%s statistics:\n", tx?"TX":"RX");

        if (legacy != NULL) {
          StatisticsGetSummary(cardIndex, channel, tx?0:1);          
          StatisticsGetType2Counters(cardIndex, channel, hex);          
        } else {
          if (DisplayMode == 0) {
            /* Summary and RMON Counters */
            StatisticsGetSummary(cardIndex, channel, tx?0:1);
            StatisticsGetRMONCounters(cardIndex, channel, hex, tx?0:1);
          } else {
            /* Display HW counters only  */
            StatisticsGetSummary(cardIndex, channel, tx?0:1);              
            StatisticsGetHwCounters(cardIndex, channel, hex, tx?0:1); 
          }
        }
        
        DisplayKeyOptions();
        
        /* Listen to standard input for keypresses */
        FD_ZERO(&fd_list);
        FD_SET(fileno(stdin), &fd_list);

        /* Convert the t value (ms) to seconds and microseconds */
        tv.tv_sec = t/1000;
        tv.tv_usec = (t % 1000 ) * 1000;
        result = select(fileno(stdin) + 1, &fd_list, NULL, NULL, &tv);

        if (result < 0)
        {
            fprintf(stderr, ">>> Error: Error on select: %s\n", strerror(errno));
            return result;
        }

        /* Check for a keypress */
        
        if (FD_ISSET(fileno(stdin), &fd_list))
        {
            uint32_t key;
            uint32_t j;

            key = getchar();

            for (j=0; j<sizeof(Menu) / sizeof(struct MenuEntry); ++j)
            {
                if (toupper(key) == toupper(Menu[j].Key))
                    Menu[j].Command(cardIndex, channel);
            }
        }
    }

    /* Restore the terminal */
    tcsetattr(fileno(stdin), TCSANOW, &original_io);

    return NTCI_ERRCODE_SUCCESS;
}
#endif

int main(int argc, char *argv[])
{
  uint32_t         channel=1;
  uint32_t         channels=1;
  uint32_t         cardIndex;
  uint32_t         i, n=1, t=1000;
  int32_t         result;
  handle_t         cardHandle;
#ifndef WIN32
  fd_set           fd_list;
  struct timeval   tv;
#endif

  /* Parse the command line and initialize the needed variables. */    
  if(InitApplication(&argc, argv, TRUE /* capture capable cards only. */, &cardIndex,
                     tools_options, tool_description,
                     output_description) == -1) {
    exit(-1);
  }

  if (legacy != NULL) {
    /* Check whether the user has specified invalid/unsupported options
       when in legacy mode (-legacy option). */
    if (req_cmd || req_blockChannelsBitmask || req_blockFilterBitmask ||
	req_blockFeedsBitmask || req_bcmd || req_num || req_time ||
	summary || tx) {
      fprintf(stderr,
	      ">>> Error: Unsupported option(s) detected together with the "
	      "-legacy option;\nuse only -adapter, -interactive, -cb, and -dec "
	      "options together with -legacy.\n");
      goto errorHandling;
    }
  }

  if((result = GetCardHandle(cardIndex, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, ">>> Error: Failed to get card handle.\n");
    goto errorHandling;
  }

  // Check if we are running in DMA block statistics mode or port statistics
  if(1) {
    uint32_t s=0;
    CardInterfaceSupportedSTIDType1_t supportedStids;
        
    /* Check that BLOCK statistics is supported. */      
    if((result = NTCI_GetCapabilities(cardHandle, CAPID_STATISTICS_IF, NOT_SUB_CAPABILITIES,
                                      STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, ">>> Error: Failed to get statistics interface capabilities.\n");
      goto errorHandling;
    }
    s=0;
    while (strlen(supportedStids.InterfaceSTID[s]) != 0) {    
      if((strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[s],
                  sizeof("STID_STATISTICS_BLOCK_TYPE1")) == 0) ||
         (tx!=NULL)) {

        int j;

        // Dont print TX if driver is loaded with PerPortStatistics=1
        if((strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[s],
                     sizeof("STID_STATISTICS_BLOCK_TYPE1")) != 0) && (tx)) {
          fprintf(stderr, "\nTX Statistics: Not supported when driver is loaded with \"PerPortStatistics=1\"");
          goto errorHandling;
        }
        // Modify the menu for interactive mode
        boBlockStatSupported=TRUE;
        for(j=0; j<(int)(sizeof(Menu)/sizeof(struct MenuEntry)); j++) {
          switch(Menu[j].Key) {
          case 't':
          case 'q':
            break;
          default:
            Menu[j].Key='\0';
            Menu[j].Description=NULL;
            Menu[j].Command = Noop;
            break;
          }
        }
      }
      s++;
    }
  }

  if (req_num != NULL) {
    n = AsciiToInt(req_num);
  }

  if (req_time != NULL) {
    t = AsciiToInt(req_time);
    if(t==0) {
      t=1;
    }
  }

  hex = dec?0:1;

  DisplayProgramHeader(argv[0], PRG_TOOL);

  if(legacy != NULL) {
    if(boBlockStatSupported==FALSE) {   

      // read and check channel bitmask
      if (req_channelBitmask != NULL) {
        channels = AsciiToInt(req_channelBitmask);
        if (channels) {
          if ( ValidateAdapterChannelBitmask(cardHandle, channels) != 0) {
            goto errorHandling;
          }
        }
      } else {
        // get defaults to show stats for all channels on adaptercard
        if ( GetAdapterChannelBitmask(cardHandle, &channels) != 0) {
          goto errorHandling;
        }
      }
      
      if (interactive_mode == NULL) {      
        StatisticsGetSummary(cardIndex, channels, tx?0:1);                
        StatisticsGetType2Counters(cardIndex, channels, hex);                  
      } else {
        RunMenu(cardIndex, channels, cardHandle, t, argv[0], PRG_TOOL);
      }      
    } else {
      fprintf(stderr, ">>> Error: Legacy mode statistics requires the driver to be loaded with \"PerPortStatistics=1\"");
      goto errorHandling;
    }
    goto errorHandling; /* Nothing more to be done when in legacy mode. */
  } 

  if(req_bcmd != NULL) {
    if(boBlockStatSupported==FALSE) {
      fprintf(stderr, ">>> Error: Driver only supports per port statistics.");
      goto errorHandling;
    }
    if(!strcasecmp("Reset", req_bcmd)) {
      StatisticsReset(cardIndex, 0);
    } else if(!strcasecmp("Start", req_bcmd)) {
      StatisticsStart(cardIndex, 0);
    } else if(!strcasecmp("Stop", req_bcmd)) {
      StatisticsStop(cardIndex, 0);
    } else {
      fprintf(stderr, ">>> Error: \"%s\" is an unknown \"-cmd\".\n", req_bcmd);
      goto errorHandling;
    }
  }
  if((req_blockChannelsBitmask!=NULL) ||
     (req_blockFilterBitmask!=NULL) ||
     (req_blockFeedsBitmask!=NULL)) {
    if(boBlockStatSupported==FALSE) {
      fprintf(stderr, ">>> Error: Driver only supports per port statistics.");
      goto errorHandling;
    }
    if(interactive_mode!=NULL) {
      fprintf(stderr, ">>> Error: The block statistics does not support interactive mode");
      goto errorHandling;
    }    
    
    /* Block statistics. */
    for (i=0; i < n; i++) {        
      StatisticsGetBlockCounters(cardIndex,
                                 (uint32_t)AsciiToInt(req_blockChannelsBitmask),
                                 AsciiToInt(req_blockFilterBitmask),
                                 (uint32_t)AsciiToInt(req_blockFeedsBitmask),
                                 summary?1:0, hex);
    }
  } else {

    // read and check channel bitmask
    if (req_channelBitmask != NULL) {
      channels = AsciiToInt(req_channelBitmask);
      if (channels) {
        if ( ValidateAdapterChannelBitmask(cardHandle, channels) != 0) {
          goto errorHandling;
        }
      }
    } else {
      // get defaults to show stats for all channels on adaptercard
      if ( GetAdapterChannelBitmask(cardHandle, &channels) != 0) {
        goto errorHandling;
      }
    }

    if(req_cmd != NULL) {
      if(boBlockStatSupported==TRUE) {
        fprintf(stderr, ">>> Error: Driver only supports block statistics.");
        goto errorHandling;
      }
      if(!strcasecmp("Reset", req_cmd)) {
        StatisticsReset(cardIndex, channels);
      } else if(!strcasecmp("Start", req_cmd)) {
        StatisticsStart(cardIndex, channels);
      } else if(!strcasecmp("Stop", req_cmd)) {
        StatisticsStop(cardIndex, channels);
      } else {
        fprintf(stderr, ">>> Error: \"%s\" is an unknown \"-cmd\".\n", req_cmd);
        goto errorHandling;
      }
    }
    if (interactive_mode == NULL) {
      /* Non-interactive mode */  
      for (i=0; i < n; i++) {        
        printf("                         -- Statistics %s snapshot %d of %d --\n", tx?"TX":"RX", i+1, n);
        for(channel=0;channel<8;channel++) {          
          if(!summary) {
            if(!(channels & (1 << channel))) {
              continue;
            }
          } else {
            channel=channels;
          }
          StatisticsGetSummary(cardIndex, (summary ? channels : (uint32_t)(1<<channel)), tx?0:1);
          StatisticsGetRMONCounters(cardIndex, (summary ? channels : (uint32_t)(1<<channel)), hex, tx?0:1);
          printf("\n");
          StatisticsGetHwCounters(cardIndex, (summary ? channels : (uint32_t)(1<<channel)), hex, tx?0:1);
          
          /* wait the specified amount of time before next sample */
          /* Convert the t value (ms) to seconds and microseconds */
          if (i+1 < n) { /* Wait only if this is not the last loop iteration */
            printf("\n");

#ifdef _WIN32
            WaitForSingleObject(GetStdHandle(STD_INPUT_HANDLE), t);
#else
            tv.tv_sec = t/1000;
            tv.tv_usec = (t % 1000 ) * 1000;
            result = select(fileno(stdin) + 1, &fd_list, NULL, NULL, &tv);
            if (result < 0)
            {
              fprintf(stderr, ">>> Error: Error on select: %s\n", strerror(errno));
              goto errorHandling;
            }
#endif
          }
          printf("\n");
          if(summary) {
            break;
          }
        }
      }
    } else {
      /* Interactive mode */
      RunMenu(cardIndex, channels, cardHandle, t, argv[0], PRG_TOOL);
    }
  }

  errorHandling:
  /* shutdown card */
  CardShutdown(0);
  
  DisplayProgramFooter();

  return 0;
}


