// $Id: SupportInfo.c 25600 2013-09-19 07:40:19Z mh $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * SupportInfo: Dump info to help solving support issues.
 */

#ifdef WIN32
#include <Windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>
#endif

#include <toolslib.h>
#include <ImageAdmin.h>
#include <DriverLog.h>

#include <NTCommonInterface.h>

/* These headers are needed for accessing the capabilities interface: */

#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>

/* This header is needed for accessing the configurations interface: */
#include <NTCI_cardconfiguration.h>
#include <cardconfigurationstructure.h>

/* This header is needed for enumerating the Filter HW shadow register RAMs */
#include <packetclassificationstructure.h>

#include <NTCI_registeraccess.h>

#define OUTPUT_TARFILE "supportinfo.tar.gz"

static CommandLineOptions_t tools_options[] = {
  {NULL, NULL, NULL, 0, NULL}
};

static char tool_description[] = {"Collect status information for debugging purposes.\n"
                                  "The output is stored in file \"" OUTPUT_TARFILE "\"\n"};
static char output_description[] = {"\n"};

void DumpCards(uint8_t cardIndex)
{
  int i;

  for(i=0; i<4; i++) {
    if (i != cardIndex) {
      if(OpenAnyCard(NTCI_CARD_TYPE_ANY, i, i) == NTCI_ERRCODE_SUCCESS) {
        GetAdapterInfo(i);
        WriteToLog(LOG_INFO, "\n");
        CloseCard(i);
      }
    }
  }
}

static int _dump_tsm_v5(FILE *fHandle, handle_t cardHandle)
{
  const uint32_t address = 0x10010000;
  uint32_t value32;

  (void)fprintf(fHandle, "Time Stamp Module (TSM) V5 Registers:\n");

  (void)fprintf(fHandle, "TSM_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0, value32);

  (void)fprintf(fHandle, "TSM_STAT\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);

  (void)fprintf(fHandle, "TSM_INT_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8, value32);

  (void)fprintf(fHandle, "TSM_INT_STAT\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xC, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xC, value32);

  (void)fprintf(fHandle, "TSM_LED\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x10, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x10, value32);

  (void)fprintf(fHandle, "TSM_TS_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x14, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);

  (void)fprintf(fHandle, "TSM_TS_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x18, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);

  (void)fprintf(fHandle, "TSM_TS_OFFSET\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x1C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);

  (void)fprintf(fHandle, "TSM_TIME_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x20, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);

  (void)fprintf(fHandle, "TSM_TIME_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x24, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x24, value32);

  (void)fprintf(fHandle, "TSM_TIME_RATE_ADJ\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x28, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x28, value32);

  (void)fprintf(fHandle, "TSM_TIME_HARDSET_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x2C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x2C, value32);

  (void)fprintf(fHandle, "TSM_TIME_HARDSET_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x30, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x30, value32);

  (void)fprintf(fHandle, "TSM_NTTS_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x34, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x34, value32);

  (void)fprintf(fHandle, "TSM_NTTS_STAT\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x38, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x38, value32);

  (void)fprintf(fHandle, "TSM_NTTS_EXT_STAT\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x3C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x3C, value32);

  (void)fprintf(fHandle, "TSM_NTTS_TS_T0_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x40, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x40, value32);

  (void)fprintf(fHandle, "TSM_NTTS_TS_T0_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x44, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x44, value32);

  (void)fprintf(fHandle, "TSM_NTTS_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x48, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x48, value32);

  (void)fprintf(fHandle, "TSM_NTTS_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x4C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4C, value32);

  (void)fprintf(fHandle, "TSM_NTTS_TS_T0_OFFSET\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x50, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x50, value32);

  (void)fprintf(fHandle, "TSM_NTTS_OFFSET\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x54, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x54, value32);

  (void)fprintf(fHandle, "TSM_NTTS_LIMIT_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x58, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x58, value32);

  (void)fprintf(fHandle, "TSM_NTTS_LIMIT_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x5C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x5C, value32);

  (void)fprintf(fHandle, "TSM_CON0_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x60, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x60, value32);

  (void)fprintf(fHandle, "TSM_CON0_INTERFACE\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x64, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x64, value32);

  (void)fprintf(fHandle, "TSM_CON0_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x68, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x68, value32);

  (void)fprintf(fHandle, "TSM_CON0_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x6C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x6C, value32);

  (void)fprintf(fHandle, "TSM_CON1_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x70, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x70, value32);

  (void)fprintf(fHandle, "TSM_CON1_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x74, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x74, value32);

  (void)fprintf(fHandle, "TSM_CON1_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x78, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x78, value32);

  (void)fprintf(fHandle, "TSM_CON2_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x7C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x7C, value32);

  (void)fprintf(fHandle, "TSM_CON2_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x80, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x80, value32);

  (void)fprintf(fHandle, "TSM_CON2_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x84, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x84, value32);

  (void)fprintf(fHandle, "TSM_CON3_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x88, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x88, value32);

  (void)fprintf(fHandle, "TSM_CON3_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x8C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8C, value32);

  (void)fprintf(fHandle, "TSM_CON3_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x90, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x90, value32);

  (void)fprintf(fHandle, "TSM_CON4_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x94, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x94, value32);

  (void)fprintf(fHandle, "TSM_CON4_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x98, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x98, value32);

  (void)fprintf(fHandle, "TSM_CON4_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x9C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x9C, value32);

  (void)fprintf(fHandle, "TSM_CON5_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xA0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xA0, value32);

  (void)fprintf(fHandle, "TSM_CON5_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xA4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xA4, value32);

  (void)fprintf(fHandle, "TSM_CON5_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xA8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xA8, value32);

  (void)fprintf(fHandle, "TSM_CON6_CONFIG\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xAC, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xAC, value32);

  (void)fprintf(fHandle, "TSM_CON6_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xB0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xB0, value32);

  (void)fprintf(fHandle, "TSM_CON6_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xB4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xB4, value32);

  (void)fprintf(fHandle, "TSM_CON7_HOST_SAMPLE_LO\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xB8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xB8, value32);

  (void)fprintf(fHandle, "TSM_CON7_HOST_SAMPLE_HI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xBC, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xBC, value32);

  (void)fprintf(fHandle, "TSM_TIMER_CTRL\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xC0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xC0, value32);

  (void)fprintf(fHandle, "TSM_TIMER_T0\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xC4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xC4, value32);

  (void)fprintf(fHandle, "TSM_TIMER_T1\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xC8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xC8, value32);

  (void)fprintf(fHandle, "TSM_PB_CTRL\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xCC, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xCC, value32);

  (void)fprintf(fHandle, "TSM_PB_INSTMEM\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xD0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xD0, value32);

  (void)fprintf(fHandle, "TSM_PI_CTRL_KP\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xD4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xD4, value32);

  (void)fprintf(fHandle, "TSM_PI_CTRL_KI\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0xD8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0xD8, value32);

  return 0;
}

static int _dump_tsm_v4(FILE *fHandle, handle_t cardHandle)
{
  const uint32_t address = 0x10010000;
  uint32_t value32;

  (void)fprintf(fHandle, "Time Stamp Module (TSM) V4 Registers:\n");

  (void)fprintf(fHandle, "TimeSync Control Register\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x0, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0, value32);

  (void)fprintf(fHandle, "TimeSync Status Register\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x4, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);

  (void)fprintf(fHandle, "TimeSync Status - External source\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x8, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8, value32);

  /* Skip current time-stamp at offset 0xC */

  (void)fprintf(fHandle, "TimeSync Basic parameters for algorithm (Bresenham))\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x14, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x18, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);

  (void)fprintf(fHandle, "TimeSync Slowdown parameters for algorithm (Bresenham)\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x1C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);

  (void)fprintf(fHandle, "TimeSync Speedup parameters for algorithm (Bresenham)\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x20, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);

  (void)fprintf(fHandle, "TimeSync Jump limit\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x24, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x24, value32);
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x28, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x28, value32);

  (void)fprintf(fHandle, "TimeSync Offset\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x2C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x2C, value32);
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x30, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x30, value32);

  (void)fprintf(fHandle, "TimeSync External sampled time\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x34, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x34, value32);
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x38, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x38, value32);

  (void)fprintf(fHandle, "TimeSync Internal sampled time\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x3C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x3C, value32);
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x40, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x40, value32);

  (void)fprintf(fHandle, "TimeSync Event timer\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x44, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x44, value32);

  (void)fprintf(fHandle, "TimeSync Interface\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x48, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x48, value32);

  (void)fprintf(fHandle, "TimeSync Timer Control\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x4C, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4C, value32);

  (void)fprintf(fHandle, "TimeSync Timer T1\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x50, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x50, value32);

  (void)fprintf(fHandle, "TimeSync Timer D T0\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x54, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x54, value32);

  (void)fprintf(fHandle, "TimeSync RSync Count\n");
  (void)NTCI_ReadRegister32(cardHandle, 4, address+0x58, &value32);
  (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x58, value32);

  return 0;
}

uint32_t CollectInfo(uint8_t cardIndex, handle_t cardHandle)
{
  uint32_t value32;
  uint64_t value64;
  uint64_t address64;
  uint64_t address32;
  uint32_t address;
  uint32_t loopcount, loopmax, i;
  FILE *fHandle=NULL;
  uint32_t shadow_offset;
  NOUSE int res;

  /* Find out which FPGA version we are working with */
  bool_t   fpga34 = FALSE;
  bool_t   fpga39 = FALSE;
  bool_t   fpga41 = FALSE;
  bool_t   fpga42 = FALSE;
  bool_t   fpga44 = FALSE;
  bool_t   fpga46 = FALSE;
  bool_t   fpgaInline = FALSE;
  bool_t   nt4std = FALSE;
  bool_t   ntbp   = FALSE;
  uint32_t result;
  uint16_t fpgaType;
  CardHardwareVersionCapabilitiesType3_t data3;
  CardHardwareCapabilitiesType1_t data1;
  CardServicesCapabilitiesType1_t services;

  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_HARDWARE,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_CAPABILITIES_TYPE1,
                                    &data1)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error finding HW capabilities: %i\n", result);
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }
  if (data1.CardType == NT4STD) {
    nt4std = TRUE;
  } else if (data1.CardType == NTBPE) {
    ntbp = TRUE; /* Bypass Adapter */
  }

  if (!ntbp) {
    if((result = NTCI_GetCapabilities(cardHandle,
                                      CAPID_VERSION,
                                      NOT_SUB_CAPABILITIES,
                                      STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                      &data3)) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error finding HW capabilities: %i\n", result);
      return NTCI_ERRCODE_INTERNAL_ERROR;
    }
    if(data3.AdapterCard.dwInfoType==1) {
      fpgaType = ((uint16_t) ((data3.AdapterCard.u.InfoType1.qwFpgaId&0xFFFF0000)>>16));
      if ((9100 <= fpgaType) && (fpgaType <=9199)) {
        fpgaInline = TRUE;
      }
      if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 34) {
        fpga34 = TRUE;
        if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 39) {
          fpga39 = TRUE;
          if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 41) {
            fpga41 = TRUE;
            if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 42) {
              fpga42 = TRUE;
              if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 44) {
                fpga44 = TRUE;
                if ((uint16_t)((data3.AdapterCard.u.InfoType1.qwFpgaId>>8)&0xFF) >= 46) {
                  fpga46 = TRUE;
                }
              }
            }
          }
        }
      }
    }else if (data3.AdapterCard.dwInfoType==2) {
      fpgaType = ((uint16_t) ((data3.AdapterCard.u.InfoType2.qwFpgaId&0xFFFF0000)>>16));
      if ((9100 <= fpgaType) && (fpgaType <=9199)) {
        fpgaInline = TRUE;
      }
      if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 34) {
        fpga34 = TRUE;
        if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 39) {
          fpga39 = TRUE;
          if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 41) {
            fpga41 = TRUE;
            if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 42) {
              fpga42 = TRUE;
              if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 44) {
                fpga44 = TRUE;
                if ((uint16_t)((data3.AdapterCard.u.InfoType2.qwFpgaId>>8)&0xFF) >= 46) {
                  fpga46 = TRUE;
                }
              }
            }
          }
        }
      }
    }else if (data3.AdapterCard.dwInfoType==3) {
      fpgaType = ((uint16_t) ((data3.AdapterCard.u.InfoType3.qwFpgaId&0xFFFF0000)>>16));
      if ((9100 <= fpgaType) && (fpgaType <=9199)) {
        fpgaInline = TRUE;
      }
      if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 34) {
        fpga34 = TRUE;
        if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 39) {
          fpga39 = TRUE;
          if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 41) {
            fpga41 = TRUE;
            if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 42) {
              fpga42 = TRUE;
              if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 44) {
                fpga44 = TRUE;
                if ((uint16_t)((data3.AdapterCard.u.InfoType3.qwFpgaId>>8)&0xFF) >= 46) {
                  fpga46 = TRUE;
                }
              }
            }
          }
        }
      }
    } else {
      fprintf(stderr, "Unsupported FPGA!\n");
      return NTCI_ERRCODE_INTERNAL_ERROR;
    }
  }

  /* Dump card information */
  fHandle = fopen("debuginfo.log", "w+");
  if(fHandle == NULL) {
    fprintf(stderr, ">>> Error: Failed to open file \"debuginfo.log\" for updating\n");
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  if (!ntbp) {
    fprintf(fHandle, "---- FPGA Info start -----\n");
    if (DisplayFpgaStatus(cardHandle, fHandle) == NTCI_ERRCODE_INTERNAL_ERROR) {
      fprintf(stderr, ">>> Error: Reading FPGA information failed.\n");
    }
    fprintf(fHandle, "----- FPGA Info end ------\n\n");
  }
  fclose(fHandle);
  if (StartLog((char *)"debuginfo.log", TRUE, TRUE) == NTCI_ERRCODE_SUCCESS)
    {
      WriteToLog(LOG_INFO, "---- Other cards start ----\n");
      DumpCards(cardIndex);
      WriteToLog(LOG_INFO, "----- Other cards end -----\n\n");

      WriteToLog(LOG_INFO, "---- Driver info start ----\n");
      GetDriverInfo(cardIndex);
      WriteToLog(LOG_INFO, "----- Driver info end -----\n\n");

      WriteToLog(LOG_INFO, "----- Driver log start ----\n");
      GetDriverLog(cardIndex, 0x3F, 0, 0);
      WriteToLog(LOG_INFO, "----- Driver log end -----\n\n");

      WriteToLog(LOG_INFO, "----- Sensor info start ----\n");
      GetSensorInfo(cardIndex);
      WriteToLog(LOG_INFO, "----- Sensor info end -----\n\n");

      if (ntbp) {
        FailoverConfigurationType1_t config;
        const char  *modeStr[] = {"Normal", "Bypass", "Undefined"};

        WriteToLog(LOG_INFO, "------- Failover Configuration --------------\n");
        if (NTCI_GetCardConfigurationDetails(cardHandle,
                                             0 /* Not needed. */,
                                             STID_FAILOVER_CONFIGURATION_TYPE1,
                                             &config) == NTCI_ERRCODE_SUCCESS) {

          WriteToLog(LOG_INFO, "Failover Current Mode           : %s\n",  modeStr[config.currentMode]);
          WriteToLog(LOG_INFO, "Failover PwrOn Mode             : %s\n",  modeStr[config.powerOnMode]);
          WriteToLog(LOG_INFO, "Failover Heartbeat value        : %d\n",  config.heartBeatTimeout);
          WriteToLog(LOG_INFO, "Failover PowerLoss detect       : %s\n",  config.powerLossDetect == 0 ? "Disabled" : "Enabled");
          WriteToLog(LOG_INFO, "Failover PowerLoss switch delay : %d\n",  config.powerLossSwitchDelay);
          WriteToLog(LOG_INFO, "Failover Autonomous Mode        : %s\n",  config.autonomousMode == 0 ? "Off" : "On");
          WriteToLog(LOG_INFO, "Failover Firmware Version ID    : 0x%x\n",  config.fwVersion);
        } else {
          WriteToLog(LOG_INFO, "Failed reading Failover configuration.\n");
        }
        WriteToLog(LOG_INFO, "------- Failover Configuration end ----------\n\n");
      }
    }
  StopLog((char *)"debuginfo.log");

#ifdef WIN32
  DumpInfo("debuginfo.log");
#else
  /* Getting DMESG info. */
  res = system("echo \"----- System log start -----\" >> debuginfo.log");
  res = system("dmesg >> debuginfo.log");
  res = system("echo \"----- System log end -----\" >> debuginfo.log");
#endif

  if (ntbp) {
    goto close_and_exit;
  }

  if ((result = NTCI_GetCapabilities(cardHandle,
                                     CAPID_SERVICES,
                                     NOT_SUB_CAPABILITIES,
                                     STID_CARD_SERVICES_CAPABILITIES_TYPE1,
                                     &services)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "Error finding HW capabilities: %i\n", result);
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }
  if ((!services.packetClassificationSupported) && (!nt4std)) {
    goto close_and_exit;
  }

  /* Read HW filters */
  fHandle = fopen("filterinfo.log", "w+");
  if(fHandle == NULL) {
    fprintf(stderr, ">>> Error: Failed to create file \"filterinfo.log\"\n");
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  if (nt4std) {
    address32 = 0x02200000;
    loopcount = 0;
    fprintf(fHandle, "Slice ram\n");
    while (loopcount < 4) {
      NTCI_ReadRegister32(cardHandle, 4, address32+loopcount*4, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)(address32+loopcount), value32);
      loopcount++;
    }
    if (fpga41) {
      address = 0x02060000;
    } else {
      address = 0x02050000;
    }
    loopcount = 0;
    fprintf(fHandle, "Packet Feed LUT\n");
    while (loopcount < 4) {
      NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
      loopcount++;
    }

    address = 0x01010000;
    fprintf(fHandle, "TimeSync Control Register\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x0, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0, value32);
    fprintf(fHandle, "TimeSync Status Register\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x4, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);
    fprintf(fHandle, "TimeSync Status - External source\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x8, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8, value32);
    fprintf(fHandle, "TimeSync Basic parameters for algorithm\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x14, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);
    NTCI_ReadRegister32(cardHandle, 4, address+0x18, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);
    fprintf(fHandle, "TimeSync Slowdown parameters for algorithm\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x1C, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);
    fprintf(fHandle, "TimeSync Speedup parameters for algorithm\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x20, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);
    fprintf(fHandle, "TimeSync Jump limit\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x24, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x24, value32);
    NTCI_ReadRegister32(cardHandle, 4, address+0x28, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x28, value32);
    fprintf(fHandle, "TimeSync Offset\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x2C, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x2C, value32);
    NTCI_ReadRegister32(cardHandle, 4, address+0x30, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x30, value32);
    fprintf(fHandle, "TimeSync External sampled time\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x34, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x34, value32);
    NTCI_ReadRegister32(cardHandle, 4, address+0x38, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x38, value32);
    fprintf(fHandle, "TimeSync Internal sampled time\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x3C, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x3C, value32);
    NTCI_ReadRegister32(cardHandle, 4, address+0x40, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x40, value32);
    fprintf(fHandle, "TimeSync Event timer\n");
    NTCI_ReadRegister32(cardHandle, 4, address+0x44, &value32);
    fprintf(fHandle, "0x%08X 0x%08X\n", address+0x44, value32);
  }
  else { /* Non STD adapters */
    if (fpga34) { /* FPGA >= 34 */
      if (fpga41) {
        address = 0x02040000;
      } else {
        address = 0x02000000;
      }
      loopcount = 0;
      fprintf(fHandle, "Grouping 4Plus function lookup\n");
      while (loopcount < 0x800) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02042000;
      } else {
        address = 0x02004000;
      }
      loopcount = 0;
      fprintf(fHandle, "Grouping 8 function lookup\n");
      while (loopcount < 0x800) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga46) {
        address = 0x02054800;
      } else if (fpga41) {
        address = 0x02052400;
      } else {
        address = 0x02008000;
      }
      loopcount = 0;
      fprintf(fHandle, "Frame size compare values\n");
      while (loopcount < 1) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga46) {
        address = 0x02054000;
        loopmax = 0x200;
      } else if (fpga41) {
        address = 0x02052000;
        loopmax = 0x100;
      } else {
        address = 0x0200A000;
        loopmax = 0x100;
      }
      loopcount = 0;
      fprintf(fHandle, "Error function lookup\n");
      while (loopcount < loopmax) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga46) {
        address = 0x02050000;
        loopmax = 0x1000;
      } else if (fpga41) {
        address = 0x02050000;
        loopmax = 0x800;
      } else {
        address = 0x0200C000;
        loopmax = 0x800;
      }
      loopcount = 0;
      fprintf(fHandle, "Protocol function lookup\n");
      while (loopcount < loopmax) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02000000;
      } else {
        address = 0x02010000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 1\n");
      while (loopcount < 0x2000) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02010000;
      } else {
        address = 0x02020000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 2\n");
      while (loopcount < 0x2000) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02020000;
      } else {
        address = 0x02030000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 3\n");
      while (loopcount < 0x2000) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02030000;
      } else {
        address = 0x02040000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 4\n");
      while (loopcount < 0x2000) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02008000;
      } else {
        address = 0x02018000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern static offset group 1\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02018000;
      } else {
        address = 0x02028000;
      }
      loopcount = 0;
      fprintf(fHandle, "Pattern static offset group 2\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02028000;
      } else {
        address = 0x02038000;
      }
      loopcount = 0;
      fprintf(fHandle, "Patterne static offset group 3\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02038000;
      } else {
        address = 0x02048000;
      }
      loopcount = 0;
      fprintf(fHandle, "Patterne static offset group 4\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02008080;
      } else {
        address = 0x02019000;
      }
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 1\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02018080;
      } else {
        address = 0x02029000;
      }
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 2\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02028080;
      } else {
        address = 0x02039000;
      }
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 3\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02038080;
      } else {
        address = 0x02049000;
      }
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 4\n");
      while (loopcount < 0x20) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02008100;
      } else {
        address = 0x0201C000;
      }
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 1\n");
      while (loopcount < 1) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }

      if (fpga41) {
        address = 0x02018100;
      } else {
        address = 0x0202C000;
      }
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 2\n");
      while (loopcount < 1) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }
      if ((fpgaInline) || (fpga41)) {
        if (fpga41) {
          address = 0x03000000;
        } else {
          address = 0x03000000;
        }
        loopcount = 0;
        fprintf(fHandle, "Tx Control\n");
        while (loopcount < 2) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
      }
      if ((!fpgaInline) || (fpga41)) {
        if (fpga41) {
          address = 0x02028100;
        } else {
          address = 0x0203C000;
        }
        loopcount = 0;
        fprintf(fHandle, "Dynamic offsets used for pattern compares group 3\n");
        while (loopcount < 1) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }

        if (fpga41) {
          address = 0x02038100;
        } else {
          address = 0x0204C000;
        }
        loopcount = 0;
        fprintf(fHandle, "Dynamic offsets used for pattern compares group 4\n");
        while (loopcount < 1) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }

        if (fpga41) {
          address32 = 0x02200000;
        } else {
          address32 = 0x02200000;
        }
        loopcount = 0;
        fprintf(fHandle, "Slice ram\n");
        while (loopcount < 64) {
          NTCI_ReadRegister32(cardHandle, 4, address32+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)(address32+loopcount), value32);
          loopcount++;
        }
      }
      if (fpga46) {
        address = 0x02068000;
      } else if (fpga41) {
        address = 0x02060000;
      } else {
        address = 0x02050000;
      }
      loopcount = 0;
      fprintf(fHandle, "Packet Feed LUT\n");
      while (loopcount < 0x800) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }
      if(fpga39) {
        if (fpga46) {
          address = 0x02060000;
          loopmax = 0x2000;
        } else if (fpga41) {
          address = 0x02062000;
          loopmax = 0x200;
        } else {
          address = 0x02052000;
          loopmax = 0x200;
        }
        loopcount = 0;
        fprintf(fHandle, "Hash ram\n");
        while (loopcount < loopmax) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
        if (fpga46) {
          address = 0x0206a400;
        } else if (fpga41) {
          address = 0x02062b00;
        } else {
          address = 0x02054000;
        }
        loopcount = 0;
        fprintf(fHandle, "Hash CRC/Seed Init value\n");
        while (loopcount < 1) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
        if (fpga42) {
          if(fpga46) {
            address = 0x0206a404;
          } else {
            address = 0x02062b04;
          }
          loopcount = 0;
          fprintf(fHandle, "Hash split legacy mode\n");
          while (loopcount < 1) {
            NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
            fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
            loopcount++;
          }
          if(fpga46) {
            address = 0x0206a300;
          } else {
            address = 0x02062c00;
          }
          loopcount = 0;
          fprintf(fHandle, "Hash split lut\n");
          while (loopcount < 64) {
            NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
            fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
            loopcount++;
          }
          if(fpga46) {
            address = 0x0206a408;
          } else {
            address = 0x02062b08;
          }
          fprintf(fHandle, "Hash mode masks\n");
          for (loopcount=0; loopcount<16; loopcount++) {
            fprintf(fHandle, "0x%02d : ", loopcount);
            for (i=0; i<10; i++) {
              NTCI_WriteRegister32(cardHandle, 4, address, (loopcount << 4) + i);
              NTCI_ReadRegister32(cardHandle, 4, address+4, &value32);
              fprintf(fHandle, "0x%08X ", value32);
            }
            fprintf(fHandle, "\n");
          }
        }
      } else {
        address = 0x02052000;
        loopcount = 0;
        fprintf(fHandle, "Hash Algorithm\n");
        while (loopcount < 1) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
        address = 0x02052004;
        loopcount = 0;
        fprintf(fHandle, "Hash CRC Init value\n");
        while (loopcount < 1) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
      }
      if (fpga46) {
        address = 0x0206a000;
      } else if (fpga41) {
        address = 0x02062800;
      } else {
        address = 0x02060000;
      }
      loopcount = 0;
      fprintf(fHandle, "Group Priority group\n");
      while (loopcount < 0x40) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }
      if ((!fpgaInline) || (fpga41)) {
        if (fpga41) {
          address = 0x02070000;
        } else {
          address = 0x02061000;
        }
        loopcount = 0;
        fprintf(fHandle, "Deduplication DataSet\n");
        while (loopcount < 0x8) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
        if (fpga46) {
          address = 0x0206a100;
        } else if (fpga41) {
          address = 0x02062900;
        } else {
          address = 0x02060100;
        }
        loopcount = 0;
        fprintf(fHandle, "Retransmit Priority group\n");
        while (loopcount < 0x40) {
          NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
          fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
          loopcount++;
        }
      }
      if (fpga46) {
        address = 0x0206A200;
      } else if (fpga41) {
        address = 0x02062A00;
      } else {
        address = 0x02060200;
      }
      loopcount = 0;
      fprintf(fHandle, "Port LUT group\n");
      while (loopcount < 0x40) {
        NTCI_ReadRegister32(cardHandle, 4, address+loopcount*4, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+loopcount*4, value32);
        loopcount++;
      }
      if (fpga46) {
        (void)_dump_tsm_v5(fHandle, cardHandle);
      } else if (fpga44) {
        (void)_dump_tsm_v4(fHandle, cardHandle);
      } else {
        address = 0x01010000;
        (void)fprintf(fHandle, "TimeSync Control Register\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x0, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0, value32);
        (void)fprintf(fHandle, "TimeSync Status Register\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x4, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);
        (void)fprintf(fHandle, "TimeSync Status - External source\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x8, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8, value32);
        (void)fprintf(fHandle, "TimeSync Basic parameters for algorithm\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x14, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x18, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);
        (void)fprintf(fHandle, "TimeSync Slowdown parameters for algorithm\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x1C, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);
        (void)fprintf(fHandle, "TimeSync Speedup parameters for algorithm\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x20, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);
        (void)fprintf(fHandle, "TimeSync Jump limit\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x24, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x24, value32);
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x28, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x28, value32);
        (void)fprintf(fHandle, "TimeSync Offset\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x2C, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x2C, value32);
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x30, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x30, value32);
        (void)fprintf(fHandle, "TimeSync External sampled time\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x34, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x34, value32);
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x38, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x38, value32);
        (void)fprintf(fHandle, "TimeSync Internal sampled time\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x3C, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x3C, value32);
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x40, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x40, value32);
        (void)fprintf(fHandle, "TimeSync Event timer\n");
        (void)NTCI_ReadRegister32(cardHandle, 4, address+0x44, &value32);
        (void)fprintf(fHandle, "0x%08X 0x%08X\n", address+0x44, value32);
      }

      address = 0x02300008;
      fprintf(fHandle, "Alternative TPID\n");
      NTCI_ReadRegister32(cardHandle, 4, address, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address, value32);

      if (fpga44) {
        address = 0x02100000;
        fprintf(fHandle, "IPF Control register\n");
        NTCI_ReadRegister32(cardHandle, 4, address, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address, value32);
        fprintf(fHandle, "IPF Unmatched feed\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x04, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x04, value32);
        fprintf(fHandle, "IPF Timeout\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x08, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x08, value32);
        fprintf(fHandle, "IPF FTF\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x0C, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0C, value32);
        fprintf(fHandle, "IPF Size\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x10, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x10, value32);
        fprintf(fHandle, "IPF Stat Max 1\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x14, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);
        fprintf(fHandle, "IPF Stat Max 2\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x18, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);
        fprintf(fHandle, "IPF Stat Max 3\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x1C, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);
        fprintf(fHandle, "IPF Stat Max 4\n");
        NTCI_ReadRegister32(cardHandle, 4, address+0x20, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);
      }
    }
    else { /* FPGA < 34 */
      address64 = 0x0;
      loopcount = 0;
      fprintf(fHandle, "Exclude filters\n");
      while (loopcount < 4) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x1000;
      loopcount = 0;
      fprintf(fHandle, "Include filters\n");
      while (loopcount < 4) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x4000;
      loopcount = 0;
      fprintf(fHandle, "Grouping function lookup\n");
      while (loopcount < 1024) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x8000;
      loopcount = 0;
      fprintf(fHandle, "Frame size compare values\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0xA000;
      loopcount = 0;
      fprintf(fHandle, "Error function lookup\n");
      while (loopcount < 128) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0xC000;
      loopcount = 0;
      fprintf(fHandle, "Protocol function lookup\n");
      while (loopcount < 2048) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x10000;
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 1\n");
      while (loopcount < 4096) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x20000;
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 2\n");
      while (loopcount < 4096) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x30000;
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 3\n");
      while (loopcount < 4096) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x40000;
      loopcount = 0;
      fprintf(fHandle, "Pattern compare match bits group 4\n");
      while (loopcount < 4096) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x18000;
      loopcount = 0;
      fprintf(fHandle, "Pattern static offset group 1\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x28000;
      loopcount = 0;
      fprintf(fHandle, "Pattern static offset group 2\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x38000;
      loopcount = 0;
      fprintf(fHandle, "Patterne static offset group 3\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x48000;
      loopcount = 0;
      fprintf(fHandle, "Patterne static offset group 4\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x19000;
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 1\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x29000;
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 2\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x39000;
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 3\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x49000;
      loopcount = 0;
      fprintf(fHandle, "Packet end check LUT group 4\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x1C000;
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 1\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x2C000;
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 2\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x3C000;
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 3\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address64 = 0x4C000;
      loopcount = 0;
      fprintf(fHandle, "Dynamic offsets used for pattern compares group 4\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }

      address32 = 0xA4046000;
      loopcount = 0;
      fprintf(fHandle, "Slice ram\n");
      while (loopcount < 1024) {
        NTCI_ReadRegister32(cardHandle, 2, address32+loopcount, &value32);
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)(address32+loopcount), value32);
        loopcount++;
      }
      address64 = 0x50000;
      loopcount = 0;
      fprintf(fHandle, "Packet Feed LUT\n");
      while (loopcount < 2048) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }
      address64 = 0x54000;
      loopcount = 0;
      fprintf(fHandle, "Hash Algorithm\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }
      address64 = 0x54008;
      loopcount = 0;
      fprintf(fHandle, "Hash CRC Init value\n");
      while (loopcount < 1) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }
      address64 = 0x60000;
      loopcount = 0;
      fprintf(fHandle, "Group Priority\n");
      while (loopcount < 32) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }
      address64 = 0x61000;
      loopcount = 0;
      fprintf(fHandle, "Deduplication DataSet\n");
      while (loopcount < 8) {
        NTCI_WriteRegister64(cardHandle, 2, 0xA404C000, address64+loopcount*8);
        NTCI_ReadRegister64(cardHandle, 2, 0xA404C008, &value64);
        fprintf(fHandle, "0x%016llX 0x%016llX\n", (ULL)(address64+loopcount*8), (ULL)value64);
        loopcount++;
      }
      address = 0xA4042000;
      fprintf(fHandle, "TimeSync Control Register\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x0, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x0, value32);
      fprintf(fHandle, "TimeSync Status Register\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x4, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);
      fprintf(fHandle, "TimeSync Status - External source\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x8, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x8, value32);
      fprintf(fHandle, "TimeSync Basic parameters for algorithm\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x14, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x14, value32);
      NTCI_ReadRegister32(cardHandle, 2, address+0x18, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x18, value32);
      fprintf(fHandle, "TimeSync Slowdown parameters for algorithm\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x1C, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x1C, value32);
      fprintf(fHandle, "TimeSync Speedup parameters for algorithm\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x20, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x20, value32);
      fprintf(fHandle, "TimeSync Jump limit\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x24, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x24, value32);
      NTCI_ReadRegister32(cardHandle, 2, address+0x28, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x28, value32);
      fprintf(fHandle, "TimeSync Offset\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x2C, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x2C, value32);
      NTCI_ReadRegister32(cardHandle, 2, address+0x30, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x30, value32);
      fprintf(fHandle, "TimeSync External sampled time\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x34, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x34, value32);
      NTCI_ReadRegister32(cardHandle, 2, address+0x38, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x38, value32);
      fprintf(fHandle, "TimeSync Internal sampled time\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x3C, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x3C, value32);
      NTCI_ReadRegister32(cardHandle, 2, address+0x40, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x40, value32);
      fprintf(fHandle, "TimeSync Event timer\n");
      NTCI_ReadRegister32(cardHandle, 2, address+0x44, &value32);
      fprintf(fHandle, "0x%08X 0x%08X\n", address+0x4, value32);
    }
  }

  /* Dump shadow RAMs */
  fprintf(fHandle, "Pattern Compare HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowPatternCompare << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  fprintf(fHandle, "Pattern Offset HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowPatternOffset << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  fprintf(fHandle, "Pattern Endcheck HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowPatternEndcheck << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  fprintf(fHandle, "Size HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowSize << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  fprintf(fHandle, "Protocol HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowProtocol << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  fprintf(fHandle, "Error HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowError << 24);
  while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
    fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
    shadow_offset += sizeof(uint64_t);
  }

  if (0) {
    fprintf(fHandle, "Retransmit Include HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowRetransmitInclude << 24);
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if (0) {
    fprintf(fHandle, "Retransmit Exclude HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowRetransmitExclude << 24);
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if (0) {
    fprintf(fHandle, "Retransmit Config HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowRetransmitConfig << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if (0) {
    fprintf(fHandle, "Search LUT HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowSearchLUT << 24);

    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if (0) {
    int x;
    int count = 0;
    uint32_t retval=NTCI_ERRCODE_SUCCESS;
    uint32_t shadow_offset_value = 0x0 | (eShadowSearch << 24);
    uint32_t shadow_offset_mask = 0x80000 | (eShadowSearch << 24);

    fprintf(fHandle, "Search HW shadow registers:\nOffset     Value         8-0         |        17-9       |        26-18      |        27-35\n");

    while(retval == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X ", (uint32_t)((shadow_offset_value&0xFFFFFF)/64));
      fprintf(fHandle, "value:");
      for (x=0; x<4; x++) {
        for (i=0; i<2; i++) {
          retval = NTCI_ReadRegister64(cardHandle, 3, shadow_offset_value, &value64);
          if (retval != NTCI_ERRCODE_SUCCESS){
            break;
          }
          if (i==0) {
            fprintf(fHandle, "|%016llX", (ULL)value64);
          } else {
            fprintf(fHandle, "|%02X", (uint8_t)value64);
          }
          shadow_offset_value += sizeof(uint64_t);
        }
      }
      fprintf(fHandle,"\n");

      fprintf(fHandle, "           mask :");
      for (x=0; x<4; x++) {
        for (i=0; i<2; i++) {
          retval = NTCI_ReadRegister64(cardHandle, 3, shadow_offset_mask, &value64);
          if (retval != NTCI_ERRCODE_SUCCESS){
            break;
          }
          if (i==0) {
            fprintf(fHandle, "|%016llX", (ULL)value64);
          } else {
            fprintf(fHandle, "|%02X", (uint8_t)value64);
          }
          shadow_offset_mask += sizeof(uint64_t);
        }
      }
      fprintf(fHandle,"\n");

      count++;
    }
  }

  if (0) {
    int x;
    int count = 0;
    uint32_t retval=NTCI_ERRCODE_SUCCESS;
    uint32_t hw_offset_value = 0x0;
    uint32_t hw_offset_mask = 0x100000;
    CardPacketClassificationCapabilitiesType1_t CardPacketClassificationCapabilities;

    if((result = NTCI_GetCapabilities(cardHandle, CAPID_PACKET_CLASSIFICATION, NOT_SUB_CAPABILITIES,
                                      STID_CARD_PACKET_CLASSIFICATION_CAPABILITIES_TYPE1,
                                      &CardPacketClassificationCapabilities)) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "Error finding HW capabilities: %i\n", result);
      return NTCI_ERRCODE_INTERNAL_ERROR;
    } else {
      /* Dump the Search HW registers only if maxSignatureMatchPatternLength > 0 (i.e. if CAM is present) */
      if (CardPacketClassificationCapabilities.maxSignatureMatchPatternLength > 0) {

        fprintf(fHandle, "Search HW registers:\nOffset     Value         8-0         |        17-9       |        26-18      |        27-35\n");

        while(retval == NTCI_ERRCODE_SUCCESS) {
          fprintf(fHandle, "0x%08X ", hw_offset_value/8);
          fprintf(fHandle, "value:");
          for (x=0; x<4; x++) {
            for (i=0; i<2; i++) {
              retval = NTCI_ReadRegister64(cardHandle, 4, hw_offset_value, &value64);
              if (retval != NTCI_ERRCODE_SUCCESS){
                break;
              }
              if (i==0) {
                fprintf(fHandle, "|%016llX", (ULL)value64);
              } else {
                fprintf(fHandle, "|%02X", (uint8_t)value64);
              }
              hw_offset_value++;
            }
          }
          fprintf(fHandle,"\n");

          fprintf(fHandle, "           mask :");
          for (x=0; x<4; x++) {
            for (i=0; i<2; i++) {
              retval = NTCI_ReadRegister64(cardHandle, 4, hw_offset_mask, &value64);
              if (retval != NTCI_ERRCODE_SUCCESS){
                break;
              }
              if (i==0) {
                fprintf(fHandle, "|%016llX", (ULL)value64);
              } else {
                fprintf(fHandle, "|%02X", (uint8_t)value64);
              }
              hw_offset_mask++;
            }
          }
          fprintf(fHandle,"\n");

          count++;
        }
      }
    }
  }

  if ((0) || (fpga34)) {
    fprintf(fHandle, "Group4Plus HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowGroup4Plus << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if ((0) || (fpga34)) {
    fprintf(fHandle, "Group8 HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowGroup8 << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if ((1) && (!fpga34)) {
    fprintf(fHandle, "Group HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowGroup << 24);
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint64_t);
    }
  }

  if ((0) || (!fpga34)) {
    fprintf(fHandle, "Include HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowInclude << 24);
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint64_t);
    }

    fprintf(fHandle, "Exclude HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowExclude << 24);
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint64_t);
    }
  }

  if (!fpgaInline) {
    if (1) {
      fprintf(fHandle, "Slice HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowSlice << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }
    }

    if (0) {
      fprintf(fHandle, "Slice Lookup HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowSliceLookup << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }

      fprintf(fHandle, "Slice Length HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowSliceLength << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }
    }
  }

  fprintf(fHandle, "Dynoffset HW shadow registers:\nOffset     Value\n");
  shadow_offset = 0x0 | (eShadowDynoffset << 24);
  if (1) {
    while (NTCI_ReadRegister64(cardHandle, 3, shadow_offset, &value64) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%016llX\n", (uint32_t)shadow_offset&0xFFFFFF, (ULL)value64);
      shadow_offset += sizeof(uint64_t);
    }
  } else {
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }
  }

  if (1) {
    fprintf(fHandle, "Feed LUT HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowFeedLUT << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }

    fprintf(fHandle, "Port LUT HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowPortLUT << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }

    fprintf(fHandle, "Group Priority HW shadow registers:\nOffset     Value\n");
    shadow_offset = 0x0 | (eShadowPriority << 24);
    while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
      fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
      shadow_offset += sizeof(uint32_t);
    }

    if (!fpgaInline) {
      fprintf(fHandle, "Retransmit Priority HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowRetransmitPriority << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }

      fprintf(fHandle, "Deduplication HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowDeduplication << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }
    }
    if (fpga39) {
      fprintf(fHandle, "Hashmode HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowHashmode << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }
    }
    if (fpga42) {
      fprintf(fHandle, "Hashmask HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowHashmask << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }

      fprintf(fHandle, "Hash SplitNum HW shadow registers:\nOffset     Value\n");
      shadow_offset = 0x0 | (eShadowHashSplitNum << 24);
      while (NTCI_ReadRegister32(cardHandle, 3, shadow_offset, &value32) == NTCI_ERRCODE_SUCCESS) {
        fprintf(fHandle, "0x%08X 0x%08X\n", (uint32_t)shadow_offset&0xFFFFFF, value32);
        shadow_offset += sizeof(uint32_t);
      }
    }
  }
  fclose(fHandle);

 close_and_exit:

#ifdef WIN32
  /* tar is not a part of windows */
#else
  res = system("which dmidecode >/dev/null && dmidecode > dmidecode.log || "
         "echo \"No dmidecode tool located on the system\" > dmidecode.log");
  /* Make debuginfo tar file. */
  if ((ntbp) || (!services.packetClassificationSupported)) {
    res = system("tar czf " OUTPUT_TARFILE
           " debuginfo.log dmidecode.log");
  } else {
    res = system("tar czf " OUTPUT_TARFILE
           " debuginfo.log dmidecode.log filterinfo.log");
  }
  res = system("rm -f debuginfo.log dmidecode.log filterinfo.log");
#endif

  return NTCI_ERRCODE_SUCCESS;
}

int main(int argc, char *argv[])
{
  uint32_t cardIndex, result;
  char     *ProgramName = argv[0];
  handle_t cardHandle;

  /* Parse the command line and initialize the needed variables. */
  if(InitApplication(&argc, argv, FALSE /* all adapter types */, &cardIndex,
                     tools_options, tool_description,
                     output_description) == -1) {
    exit(-1);
  }

  DisplayProgramHeader(ProgramName, PRG_TOOL);

  if((result = GetCardHandle(cardIndex, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, ">>> Error: Failed to get card handle.\n");
    goto errorHandling;
  }

  if (CollectInfo(cardIndex, cardHandle) == NTCI_ERRCODE_SUCCESS) {
#ifdef WIN32
    printf(">>> Info: Data successfully collected. Output in \"debuginfo.log and filterinfo.log\"\n");
#else
    printf(">>> Info: Data successfully collected. Output in \"%s\"\n", OUTPUT_TARFILE);
#endif
  }

 errorHandling:
  /* shutdown card */
  CardShutdown(0);

  DisplayProgramFooter();

  return 0;
}
