// $Id: DriverLog.c 14228 2011-09-09 13:04:51Z ps $
/*
* Copyright 2013 Napatech A/S. All Rights Reserved.
* 
*   The source code contained or described herein and all documents
*   related to the source code ("Material") are owned by Napatech
*   A/S or its suppliers or licensors.  Title to the Material
*   remains with Napatech Inc. or its suppliers and licensors. The
*   software is licensed under the Napatech A/S SCLA License.
* 
*   Unless otherwise agreed by Napatech A/S in writing, you may
*   not remove or alter this notice or any other notice embedded
*   in Materials by Napatech A/S or Napatech's suppliers or
*   licensors in any way.
*/

/**
* @file
* DriverLog.c : This file contain functions related to R&D debugging
*/

#ifdef WIN32
#include <Windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <time.h>

#include <sys/types.h>
#include <sys/stat.h>
#endif

#include "toolslib.h"

#include <NTCommonInterface.h>

/* These headers are needed for accessing the capabilities interface: */
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>

/* These headers are needed for CardInfo functions */
#include <cardinfo.h>
#include <NTCI_memoryaccess.h>

int toolRunning=1;

static const char* LogTypeToString(uint64_t logType)
{
  switch((eLogType)logType) {
  case NT_LOG_ERROR:
    return "ERR";
  case NT_LOG_WARNING:
    return "WARN";
  case NT_LOG_INFO:
    return "INFO";
  case NT_LOG_DEBUG:
    return "DBG";
  case NT_LOG_UPDATEIMAGE:
    return "UIMG";
  case NT_LOG_DIAGNOSTICS:
    return "DIAG";
  case NT_LOG_AVR:
    return "AVR";
  default:
    return "???";
  }
}

uint32_t GetDriverLog(uint8_t card, uint64_t logLevelMask, uint8_t waitForNewEntry, uint8_t clearAfterRead)
{
  CardInterfaceSupportedSTIDType1_t supportedStids;
  uint32_t result=0, i=0;
  handle_t cardHandle;
  MemoryDMAType1_t memType1;
  CardInfoLogType1_t cardInfoLogType1;
  CardInfoLogEntryType1_t *pEntry=NULL;
  bool_t boDriverLogSupported=FALSE;

  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  /* Check that BLOCK statistics is supported. */      
  if((result = NTCI_GetCapabilities(cardHandle, CAPID_CARD_INFO_IF, NOT_SUB_CAPABILITIES,
    STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_ERROR, "Failed to get card info interface capabilities.\n");    
      return result;
  }
  i=0;
  while (strlen(supportedStids.InterfaceSTID[i]) != 0) {    
    if(strncmp("STID_CARD_INFO_LOG_TYPE1", supportedStids.InterfaceSTID[i],
      sizeof("STID_CARD_INFO_LOG_TYPE1")) == 0) {
        boDriverLogSupported=TRUE;
        break;
    }
    i++;
  }

  if(boDriverLogSupported) {
    memset(&cardInfoLogType1, 0x00, sizeof(cardInfoLogType1));    
    /* Get the buffer size required for the log. */
    if((result =
      NTCI_GetCardInfo(cardHandle, STID_CARD_INFO_LOG_TYPE1 , &cardInfoLogType1)) != NTCI_ERRCODE_SUCCESS) { 
        WriteToLog(LOG_ERROR, "Failed to get required log size.\n");    
        return result;
    }
    memcpy(&memType1.dmaBuffer, &cardInfoLogType1.log, sizeof(ExternalBufferInfoType1_t));    
    if((result =
      NTCI_Memory(cardHandle, STID_MEMORY_ALLOC_DMA_TYPE1, &memType1)) != NTCI_ERRCODE_SUCCESS) {
        WriteToLog(LOG_ERROR, "Failed to allocate %d bytes shared memory.\n",
          memType1.dmaBuffer.bufferSize);
        return result;
    }
    memcpy(&cardInfoLogType1.log, &memType1.dmaBuffer, sizeof(ExternalBufferInfoType1_t));    

    WriteToLog(LOG_INFO, "DriverLog: LogLevel mask: 0x%X\n", logLevelMask);
    WriteToLog(LOG_INFO, "------------------+---------+------------------------------------------------------------------------\n");
    WriteToLog(LOG_INFO, "Timestamp         | LogType | Log entry\n");
    WriteToLog(LOG_INFO, "------------------+---------+------------------------------------------------------------------------\n");

    if(waitForNewEntry==0) {
      /* Only run once. */
      toolRunning=0;
    } else {
      /* If we need to wait for new entries it only makes sense to clear the log. */
      clearAfterRead=1;
    }

    /* Get the log. */
    cardInfoLogType1.waitForNewEntry = 0;
    do {
      cardInfoLogType1.clearAfterRead = clearAfterRead;       /* Clear the log after the read. */

      if((result = NTCI_GetCardInfo(cardHandle, STID_CARD_INFO_LOG_TYPE1, &cardInfoLogType1)) != NTCI_ERRCODE_SUCCESS) {
        WriteToLog(LOG_ERROR, "Failed to get log.\n");
      }

      /* Dump the entries. */
      if(result==NTCI_ERRCODE_SUCCESS) {
        pEntry = (CardInfoLogEntryType1_t*)cardInfoLogType1.log.user.u.pVirt;
        for(i=0; i<cardInfoLogType1.numEntries; i++) {
          if(pEntry->logType & logLevelMask) {
            WriteToLog(LOG_INFO, "%08X.%08X | #%-6s | %s",
              (uint32_t)pEntry->timestamp&0xFFFFFFFF,
              (uint32_t)(pEntry->timestamp>>32),
              LogTypeToString(pEntry->logType), pEntry->logStr);
          } 
          pEntry++;
        }
      }
      cardInfoLogType1.waitForNewEntry = waitForNewEntry;  /* We want to wait for new data on the next run. */
    } while(toolRunning);
    WriteToLog(LOG_INFO, "------------------+---------+------------------------------------------------------------------------\n");

    NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
    return result;
  } else {
    WriteToLog(LOG_ERROR, "Driver doesn't support \"DriverLog\".\n");
    return NTCI_ERRCODE_DRIVER_ERROR;
  }
}
