// $Id: ImageAdmin.c 14206 2011-09-08 16:47:22Z ps $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * ImageAdmin.c : Contains image upload and management functions
 */

#ifdef WIN32
#include <Windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#include <strings.h>
#endif

/* The NTCommonInterface.h header provides the interface definitions
 * for the NTCI library */

#include <NTCommonInterface.h>
#include <NTCI_registeraccess.h>

/* These headers are needed for accessing the capabilities interface: */
#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>

/* These headers are needed for accessing the devicecontrol interface: */
#include <NTCI_devicecontrol.h>
#include <devicecontrolstructure.h>

/* These headers are needed for accessing the fpga information: */
#include <fpgainfostructure.h>

#include "toolslib.h"
#include "ImageAdmin.h"


/*  ----------------------------------------------------------------------------
**
**  Name           :DestroyMemoryBlock
**
**  Description    :a type-safe NULL-safe wrapper for free()
**                  Call this function when finished with the data returned
**                  by AllocateAndLoadImage (above)
**
**  Parameters     :uint8_t * pData
**
**  Returns        :void
**
**
**  --------------------------------------------------------------------------*/
void DestroyMemoryBlock(uint8_t * pData)
{
    if (pData != NULL)
    {
        free(pData);
    }
}

/*  ----------------------------------------------------------------------------
**
**  Name           :CreateImageMemoryBlock
**
**  Description    :a type-safe wrapper for malloc()
**
**  Parameters     :uint32_t Length
**
**  Returns        :uint8_t *
**
**
**  --------------------------------------------------------------------------*/
uint8_t * CreateImageMemoryBlock(uint32_t Length)
{
    void * pUntypedMemory = malloc(Length);
    uint8_t * pTypedMemory = (uint8_t *)pUntypedMemory;
    return pTypedMemory;
}


/*  ----------------------------------------------------------------------------
**
**  Name           :AllocateAndLoadImage
**
**  Description    :function to load a whole image file from disk to a new
**                  memory block. *pLoadedLength is set to length of the file
**                  if file has been read successfully.
**
**  Parameters     :const char * ImageName
**                  int * pLoadedLength
**
**  Returns        :uint8_t *   - Set to NULL if any error occurs
**
**
**  --------------------------------------------------------------------------*/
uint8_t * AllocateAndLoadImage(const char * ImageName,uint32_t * pLoadedLength)
{
    uint8_t * pFileData = NULL;
    FILE * fHandle = fopen(ImageName, "rb");
    *pLoadedLength = 0;

    if (fHandle != NULL)
    {
        /* Image opened - try to read its length */
        int FileLength = 0;
        int numread = 0;
        fseek(fHandle, 0, SEEK_END);
        FileLength = ftell(fHandle);
        fseek(fHandle, 0, SEEK_SET);

        /* Try to create enough memory to hold the whole image */
        pFileData = CreateImageMemoryBlock(FileLength);
        if (pFileData != NULL)
        {
            /* try to read the file up to the length specified */
            numread = fread( pFileData, sizeof(char), FileLength, fHandle );
            if (numread ==FileLength)
            {
                /* Success */
                *pLoadedLength = FileLength;
            }
            else
            {
                /* User shall not be given a partial file*/
                WriteToLog(LOG_ERROR, "Failed to read entire length (%d), only read:%d bytes (ptr:%p)\n",
                                                                                  FileLength, numread, pFileData);
                DestroyMemoryBlock(pFileData);
                pFileData = NULL;
            }
        }
        else
        {
            WriteToLog(LOG_ERROR, "Failed to create a buffer for the image\n");
        }
        /* Always close the file even after a read failure */
        fclose(fHandle);
    }
    else
    {
        WriteToLog(LOG_ERROR, "Failed to open file '%s' for reading\n",ImageName);
    }
    return pFileData;
}

uint32_t UpdateImage (uint32_t card, const char *filename, char *initials, char *option, char type, uint8_t *buffer, uint32_t length)
{
  uint32_t             FileContentLength = 0;
  char*                pTempImageFileName = NULL;    
  ProgramDeviceType1_t programDevice;
  uint8_t*             pFileContent = NULL;
  uint32_t             result;
  handle_t             cardHandle;
  CardHardwareCapabilitiesType1_t data1;
  
#define VERIFY  0
#define PROGRAM 1
#define AVR_FRONT_IMAGE       0xFED0
#define AVR_MAIN_IMAGE        0xFED1
#define AVR_MAIN_PSU          0xFED3
#define AVR_FRONT_PSU         0xFED5
#define AVR_FRONT_FUS         0xFED6
#define AVR_MAIN_FUS          0xFED7
#define AVR_FRONT_EEPROM      0xFED8
#define AVR_MAIN_EEPROM       0xFED9
  if((result = GetCardHandle((uint8_t)card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  if (!strcasecmp(option, "Verify")) {
    programDevice.programFlags = VERIFY;
  } else if(!strcasecmp(option, "Program")) {
    programDevice.programFlags = PROGRAM;
  } else if(!strcasecmp(option, "AvrFrontFus")) {
    programDevice.programFlags = AVR_FRONT_FUS;
  } else if(!strcasecmp(option, "AvrMainFus")) {
    programDevice.programFlags = AVR_MAIN_FUS;
  } else if(!strcasecmp(option, "AvrFrontImg")) {
    programDevice.programFlags = AVR_FRONT_IMAGE;
  } else if(!strcasecmp(option, "AvrMainImg")) {
    programDevice.programFlags = AVR_MAIN_IMAGE;
  } else if(!strcasecmp(option, "AvrMainPsu")) {
    programDevice.programFlags = AVR_MAIN_PSU;
  } else if(!strcasecmp(option, "AvrFrontPsu")) {
    programDevice.programFlags = AVR_FRONT_PSU;
  } else if(!strcasecmp(option, "AvrMainEeprom")) {
    programDevice.programFlags = AVR_MAIN_EEPROM;
  } else if(!strcasecmp(option, "AvrFrontEeprom")) {
    programDevice.programFlags = AVR_FRONT_EEPROM;
  }

  else {
    return NTCI_ERRCODE_INVALID_PARAMETER;
  }
      
  if (type == UPDATE_IMAGE_TYPE_FILE) {
    pFileContent = AllocateAndLoadImage(filename, &FileContentLength);
  }
  else {
    pFileContent = buffer;
    FileContentLength = length;
  }
      
  if (pFileContent == NULL) {
    return NTCI_ERRCODE_INSUFFICIENT_MEMORY;
  }

#ifdef WIN32
  pTempImageFileName = (strrchr(filename, '\\'));
#else
  pTempImageFileName = (strrchr(filename, '/'));
#endif
      
  if (pTempImageFileName != NULL) {
    filename = pTempImageFileName+1;
  }

  programDevice.deviceImageSize = FileContentLength;
  
  memcpy(&programDevice.achImageFileName, filename, IMAGE_FILE_NAME_LENGTH);
  if (initials == NULL) {
    strcpy(programDevice.achLoadedBy, "noname");
  } else {
    memcpy(&programDevice.achLoadedBy, initials, LOADED_BY_LENGTH);
  }
  programDevice.pDeviceImage = pFileContent;

  if((result = NTCI_GetCapabilities(cardHandle,
                                        CAPID_HARDWARE,
                                        NOT_SUB_CAPABILITIES,
                                        STID_CARD_HARDWARE_CAPABILITIES_TYPE1,
                                        &data1)) != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Failed to get HW capabilities. Result = 0x%x\n", result);
    return result;
  }

  if (programDevice.programFlags != VERIFY) {
    if (data1.CardType == NTBPE) {
      /* Bypass Adapter */
      WriteToLog(LOG_INFO, ">>> Info: Adapter firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: This will take a few seconds...\n");
    } else if(!strcasecmp(option, "AvrFrontFus")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Front FUSE module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    } else if(!strcasecmp(option, "AvrMainFus")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Main FUSE module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    } else if(!strcasecmp(option, "AvrFrontImg")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Front module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    } else if(!strcasecmp(option, "AvrMainImg")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Main module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    } else if(!strcasecmp(option, "AvrMainPsu")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Main PSU module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    } else if(!strcasecmp(option, "AvrFrontPsu")) {
      WriteToLog(LOG_INFO, ">>> Info: AVR Front PSU module  firmware image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
      WriteToLog(LOG_INFO, ">>> Info: While firmware is updated, no sensor monitoring is performed\n");
    }else {
      WriteToLog(LOG_INFO, ">>> Info: FPGA image is being updated - DO NOT INTERRUPT\n");
      WriteToLog(LOG_INFO, ">>> Info: Image name: %s\n", filename);
      WriteToLog(LOG_INFO, ">>> Info: Check your console/dmesg for feedback\n");
    }
  } else {
    WriteToLog(LOG_INFO, ">>> Info: FPGA image is being verified\n");
  }     

  result = NTCI_ImageUpload(cardHandle,
                            STID_PROGRAM_DEVICE_TYPE1,
                            &programDevice);
  if ( result == NTCI_ERRCODE_SUCCESS) {    
    if (programDevice.programFlags != VERIFY) {          
      WriteToLog(LOG_INFO, "\n>>> Info: Image update succesfully\n");
    } else {
      WriteToLog(LOG_INFO, "\n>>> Info: Image verified succesfully\n");
    }
  } else {
      WriteToLog(LOG_ERROR, "Image update/verification FAILED\n");
      if (data1.CardType == NTBPE) {
        WriteToLog(LOG_ERROR, "Remembered to engage the Adapter FW Prg. Switch?\n");
      }
  }

  return result;
}

uint32_t ChangePriImage (uint32_t card)
{
  handle_t               cardHandle;
  uint32_t               result;
  ChangeBootImageType1_t changeBootImage;

  if((result = GetCardHandle((uint8_t)card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  result = NTCI_DeviceControl(cardHandle,
                              STID_CHANGE_BOOT_IMAGE_TYPE1,
                              &changeBootImage );    
  return result;
}

int DisplayFpgaStatus(handle_t cardHandle, FILE *stream)
{
  FpgaImageBootStatusType1_t ImageBootStatus;
  FpgaImageStatusType_t ImageStatus;
  FpgaImageHeaderType2_t flashHeaderType2;
  CardHardwareVersionCapabilitiesType3_t data3;

  char achImageState[128];

  uint64_t fpgaId=0;
  uint32_t result;
  int i = 0;

  typedef enum
  {
    NT_PRIM_IMAGE_OK                    =1,
    NT_PRIM_IMAGE_LOAD_ERROR_SEC_LOADED =2,
    NT_SEC_IMAGE_OK                     =3,
    NT_SEC_IMAGE_LOAD_OK_TEST_FAILED    =4,
    NT_SEC_IMAGE_LOAD_ERROR_PRIM_LOADED =5,
    NT_IMAGE_LOAD_FROM_PROBE            =6,
  } NT_IMAGE_BOOT_STATUS;

  typedef enum
  {
    NT_IMAGE_NOT_UPLOADED_VIA_SAFEMODE               =0,
    NT_IMAGE_UPLOADED_NOT_VERIFIED                   =1,
    NT_IMAGE_UPLOADED_VERIFIED_NOT_TESTED            =2,
    NT_IMAGE_UPLOADED_VERIFICATION_FAILED            =3,
    NT_IMAGE_NOT_LOADED_OK_FROM_FLASH                =4,
    NT_IMAGE_LOADED_OK_FROM_FLASH_IMAGE_TEST_FAILED  =5,
    NT_IMAGE_LOADED_VERIFIED_AND_TESTED_OK           =6,
  } NT_IMAGE_STATUS;

  if ( NTCI_DeviceControl(cardHandle,
    STID_FPGA_IMAGE_BOOT_STATUS_TYPE1,
    &ImageBootStatus ) != NTCI_ERRCODE_SUCCESS) {        
      return NTCI_ERRCODE_INTERNAL_ERROR;    
  } else {

    fprintf(stream, "NIC image boot status\n");
    fprintf(stream, "---------------------------------------------------------\n");
    switch (ImageBootStatus.imageState)
    {
    case NT_PRIM_IMAGE_OK:
      {
        fprintf(stream, "Image %d (Primary) loaded, no special events\n", ImageBootStatus.primaryImageNumber);
        fprintf(stream, "Primary image = %d\n", ImageBootStatus.primaryImageNumber);
      }
      break;
    case NT_PRIM_IMAGE_LOAD_ERROR_SEC_LOADED:
      {
        fprintf(stream, "Secondary image loaded, due to failure when loading Primary image\n");
        if (ImageBootStatus.primaryImageChanged == 1) {
          fprintf(stream, "Primary image has been changed to image number %d\n", ImageBootStatus.primaryImageNumber);          
        } else {        
          fprintf(stream, "Primary image = %d\n", ImageBootStatus.primaryImageNumber);
        }
      }
      break;
    case NT_SEC_IMAGE_OK:
      {
        fprintf(stream, "Image %d (Secondary) loaded and tested OK\n", ImageBootStatus.primaryImageNumber);
        if (ImageBootStatus.primaryImageChanged == 1) {
          fprintf(stream, "Primary image has been changed to image number %d\n", ImageBootStatus.primaryImageNumber);          
        }
      }
      break;
    case NT_SEC_IMAGE_LOAD_OK_TEST_FAILED:
      {
        fprintf(stream, "Image %d (Secondary) loaded OK, but image test failed\n", ~ImageBootStatus.primaryImageNumber & 0x1);
        fprintf(stream, "Primary image = %d\n", ImageBootStatus.primaryImageNumber);
      }
      break;
    case NT_SEC_IMAGE_LOAD_ERROR_PRIM_LOADED:
      {
        fprintf(stream, "Image %d (Primary) loaded, after failed attempt to load image %d (Secondary)\n", ImageBootStatus.primaryImageNumber, (~ImageBootStatus.primaryImageNumber & 0x1));
        fprintf(stream, "Primary image = %d\n", ImageBootStatus.primaryImageNumber);
      }
      break;
    case NT_IMAGE_LOAD_FROM_PROBE:
      {
        fprintf(stream, "Image loaded from probe\n");
        fprintf(stream, "Primary image = %d\n", ImageBootStatus.primaryImageNumber);
      }
      break;
    default:
      {
        fprintf(stream, "Unknown Fpga image status\n");
      }
      break;    
    }
    fprintf(stream, "\n");
  }

  fprintf(stream, "Fpga image status in flash\n");
  fprintf(stream, "---------------------------------------------------------\n");
  for (i=0; i<2; i++) {
    result = NTCI_ReadVitalProductData(cardHandle,
      i,
      STID_FPGA_IMAGE_HEADER_TYPE2,
      &flashHeaderType2);

    if (result != NTCI_ERRCODE_SUCCESS)
    {
      fprintf(stderr, "Unable to read flash header for slot '%i', probably does not contain a valid FPGA image\n", i);
      /*exit(-1);*/
    } else {
      fprintf(stream, "Image Number    : %d\n", flashHeaderType2.imageNumber);
      fprintf(stream, "Image type      : %d\n",flashHeaderType2.imageType);
      fprintf(stream, "version         : %02d.%02d\n",
        ((flashHeaderType2.imageVersion>>8)&0xFF),
        (flashHeaderType2.imageVersion    &0xFF));
      fprintf(stream, "Created by      : %s\n", flashHeaderType2.imageCreatedBy);
      fprintf(stream, "Build by        : %s\n",flashHeaderType2.imageBuildBy);
      fprintf(stream, "Source filename : %s\n", flashHeaderType2.imageSourceFileName);
      fprintf(stream, "Source date     : %s\n", flashHeaderType2.imageSourceFileDateAndTime);
      fprintf(stream, "Note            : %s\n", flashHeaderType2.imageNote);
      fprintf(stream, "Downloaded by   : %s\n", flashHeaderType2.imageLoadedBy);
      fprintf(stream, "Filename        : %s\n", flashHeaderType2.imageFileName);
      fprintf(stream, "Download date   : %s\n", ctime((time_t*)(void*)&flashHeaderType2.imageLoadDate));
    }    
    if ( NTCI_ReadVitalProductData(cardHandle,
                                   i,
                                   STID_FPGA_IMAGE_STATUS_TYPE,
                                   &ImageStatus ) != NTCI_ERRCODE_SUCCESS) {
      fprintf(stderr, "\n");      
      if (i != 0) {
        return NTCI_ERRCODE_INTERNAL_ERROR;
      }
    } else {      
      switch(ImageStatus.State) {
      case NT_IMAGE_NOT_UPLOADED_VIA_SAFEMODE:
      {
        sprintf(achImageState, "Image has not been uploaded with the Safemode functionality. Hence no image state information\n");
      }
      break;
      
      case NT_IMAGE_UPLOADED_NOT_VERIFIED:
      {
        sprintf(achImageState, "Image has been uploaded but not verified. Image could be corrupted please upload image again\n");          
      }
      break;
        
      case NT_IMAGE_UPLOADED_VERIFIED_NOT_TESTED:
      {
        sprintf(achImageState, "Image has been uploaded and verified but not yet tested\n");                    
      }
      break;
      
      case NT_IMAGE_UPLOADED_VERIFICATION_FAILED:
      {
        sprintf(achImageState, "Image has been uploaded but verification during image upload FAILED. Image is corrupted\n");                    
      }
      break;
      
      case NT_IMAGE_NOT_LOADED_OK_FROM_FLASH:
      {
        sprintf(achImageState, "Image FAILED during load from flash. Image is corrupted \n");                              
      }
      break;
      
      case NT_IMAGE_LOADED_OK_FROM_FLASH_IMAGE_TEST_FAILED:
      {
        sprintf(achImageState, "Image FAILED during test. Image does not work correctly\n");                                        
      }
      break;
      
      case NT_IMAGE_LOADED_VERIFIED_AND_TESTED_OK:
      {
        sprintf(achImageState, "Image is OK\n");          
      }
      break;
      
      default:
      {
        sprintf(achImageState, "Unknown image status\n");
      }
      break;    
      
      }
      fprintf(stream, "Image state     : %s\n", achImageState);
    }
  }
  
  if((result = NTCI_GetCapabilities(cardHandle,
                                    CAPID_VERSION,
                                    NOT_SUB_CAPABILITIES,
                                    STID_CARD_HARDWARE_VERSION_CAPABILITIES_TYPE3,
                                    &data3)) != NTCI_ERRCODE_SUCCESS) {
    fprintf(stderr, "NTCI_GetCapabilities returned error: %i\n", result);      
    return NTCI_ERRCODE_INTERNAL_ERROR;
  }

  if(data3.AdapterCard.dwInfoType==1) {
    fpgaId = data3.AdapterCard.u.InfoType1.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==2) {
    fpgaId = data3.AdapterCard.u.InfoType2.qwFpgaId;
  }else if(data3.AdapterCard.dwInfoType==3) {
    fpgaId = data3.AdapterCard.u.InfoType3.qwFpgaId;
  }
  if ((((fpgaId&0xFFFF0000) >> 16) == 9020) || 
      (((fpgaId&0xFFFF0000) >> 16) == 9220) ||
      (((fpgaId&0xFFFF0000) >> 16) == 9221) ||
      (((fpgaId&0xFFFF0000) >> 16) == 9222) ) {

  
    fprintf(stream, "Main board AVR image status\n");
    fprintf(stream, "---------------------------------------------------------\n");
  
    result = NTCI_ReadVitalProductData(cardHandle,
                                       2,
                                       STID_FPGA_IMAGE_HEADER_TYPE2,
                                       &flashHeaderType2);
    
    if (result != NTCI_ERRCODE_SUCCESS)
    {
      fprintf(stderr, "Unable to read flash header for AVR main board\n");
      fprintf(stderr, "Main board AVR has probably not been updated since production\n");
      /*exit(-1);*/
    } else {
      fprintf(stream, "Image Number    : %d\n", flashHeaderType2.imageNumber);
      fprintf(stream, "Image type      : AVR Main Image\n");
      fprintf(stream, "version         : %c.%c%c\n",
              ((flashHeaderType2.imageVersion>>16)&0xFF),
              ((flashHeaderType2.imageVersion>>8)&0xFF),
              (flashHeaderType2.imageVersion    &0xFF));
      fprintf(stream, "Source filename : %s\n", flashHeaderType2.imageSourceFileName);   
      fprintf(stream, "Downloaded by   : %s\n", flashHeaderType2.imageLoadedBy);  
      fprintf(stream, "Download date   : %s\n", ctime((time_t*)(void*)&flashHeaderType2.imageLoadDate));
    }    
    
    fprintf(stream, "Front board AVR image status\n");
    fprintf(stream, "---------------------------------------------------------\n");
    
    result = NTCI_ReadVitalProductData(cardHandle,
                                       3,
                                       STID_FPGA_IMAGE_HEADER_TYPE2,
                                       &flashHeaderType2);

    if (result != NTCI_ERRCODE_SUCCESS)
    {
      fprintf(stderr, "Unable to read flash header for AVR front board\n");
      fprintf(stderr, "Front board AVR has probably not been updated since production\n");
      /*exit(-1);*/
    } else {
      fprintf(stream, "Image Number    : %d\n", flashHeaderType2.imageNumber);
      fprintf(stream, "Image type      : AVR Front Image\n");
      fprintf(stream, "version         : %c.%c%c\n",
              ((flashHeaderType2.imageVersion>>16)&0xFF),
              ((flashHeaderType2.imageVersion>>8)&0xFF),
              (flashHeaderType2.imageVersion    &0xFF));
       fprintf(stream, "Source filename : %s\n", flashHeaderType2.imageSourceFileName);
      fprintf(stream, "Downloaded by   : %s\n", flashHeaderType2.imageLoadedBy);
      fprintf(stream, "Download date   : %s\n", ctime((time_t*)(void*)&flashHeaderType2.imageLoadDate));
    }    
  }
  return 0;
}
