// $Id: Statistics.c 25153 2013-08-30 11:14:55Z ph $
/*
 * Copyright 2013 Napatech A/S. All Rights Reserved.
 * 
 *   The source code contained or described herein and all documents
 *   related to the source code ("Material") are owned by Napatech
 *   A/S or its suppliers or licensors.  Title to the Material
 *   remains with Napatech Inc. or its suppliers and licensors. The
 *   software is licensed under the Napatech A/S SCLA License.
 * 
 *   Unless otherwise agreed by Napatech A/S in writing, you may
 *   not remove or alter this notice or any other notice embedded
 *   in Materials by Napatech A/S or Napatech's suppliers or
 *   licensors in any way.
 */

/**
 * @file
 * Statistics.c : Contain functions related to statistics
 */

#ifdef WIN32
#include <Windows.h>
#include "platformwin.h"
#else
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>
#endif

/* The NTCommonInterface.h header provides the interface definitions
 * for the NTCI library */

#include <NTCommonInterface.h>

/* These headers are needed for accessing the capabilities interface: */

#include <NTCI_capabilities.h>
#include <capabilitiesstructure.h>

/* These headers are needed for the statistics API */

#include <NTCI_statistics.h>
#define _STAT_CONVERT_FUNCTIONS_ 
#include <statisticsstructure.h>

#include <NTCI_cardstatus.h>
#include <cardstatusstructure.h>

#include <NTCI_memoryaccess.h>
#include <memoryaccess.h>
#include "toolslib.h"
#include "Statistics.h"

#ifdef _WIN32
#define PRINTF_64BIT "%I64d"
#else
#define PRINTF_64BIT "%Ld"
#define PRINTF_64BITHEX "%016llX"
#endif

/* arrays for storing stats history */
static uint64_t prevOctCount[MAX_CARDS][MAX_CHANNELS];
static uint64_t prevTimeCount[MAX_CARDS][MAX_CHANNELS];
static uint64_t prevFrameCount[MAX_CARDS][MAX_CHANNELS];

static void StatisticsDumpEthType7(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex);
static void StatisticsDumpEthType8(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex);
static void StatisticsDumpMisc(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex);
static void StatisticsDumpIpFragment(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex);
static void StatisticsDumpSimpleType1(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex);
static void StatisticsSummaryMisc(StatBlockDescriptorType1_t *src,
                                  StatBlockDescriptorType1_t *dest);
static void StatisticsSummaryIpFragment(StatBlockDescriptorType1_t *src,
                                        StatBlockDescriptorType1_t *dest);
static void StatisticsSummaryEthType7(StatBlockDescriptorType1_t *src,
                                      StatBlockDescriptorType1_t *dest);
static void StatisticsSummaryEthType8(StatBlockDescriptorType1_t *src,
                                      StatBlockDescriptorType1_t *dest);
static void StatisticsSummarySimpleType1(StatBlockDescriptorType1_t *src,
                                         StatBlockDescriptorType1_t *dest);

static char *NumAsString(uint64_t number, bool_t hex)
{
  static char str[20];
  if(hex) {
    snprintf(str, 20, "0x%014llX", (ULL)number);
  } else {
    snprintf(str, 20, "%16lld", (ULL)number);
  }
  return str;
}


static char writeBuffer[4096];
static uint32_t writePos;
#define STATS_LOG(...) writePos += WriteToBuffer(&writeBuffer[writePos], sizeof(writeBuffer)-writePos,  __VA_ARGS__);

/* Start recording statistics on a particular channel or set of channels */
uint32_t StatisticsStart(uint8_t card, uint32_t channelsBitmask)
{
  StatisticsOptionsType3_t options;
  uint32_t result;
  handle_t cardHandle;

  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  options.startNow = 1;
  options.stopNow = 0;
  options.resetNow = 0;

  result = NTCI_SetStatisticsOptions(cardHandle,
                                     channelsBitmask,
                                     STID_STATISTICS_OPTIONS_TYPE3,
                                     &options);

  if (result != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Statistics: Error starting statistics: %i\n", result);    
  }
  return result;
}

/* Stop recording statistics */
uint32_t StatisticsStop(uint8_t card, uint32_t channelsBitmask)
{
  StatisticsOptionsType3_t options;
  uint32_t result;
  handle_t cardHandle;

  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }
  
  options.startNow = 0;
  options.stopNow = 1;
  options.resetNow = 0;
  
  result = NTCI_SetStatisticsOptions(cardHandle,
                                     channelsBitmask,
                                     STID_STATISTICS_OPTIONS_TYPE3,
                                     &options);
  
  if (result != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Statistics: Error stopping statistics: %i\n", result);
  }
  return result;
}

/* Reset statistics counter */
uint32_t StatisticsReset(uint8_t card, uint32_t channelsBitmask)
{
  StatisticsOptionsType3_t options;
  uint32_t result;
  handle_t cardHandle;
    
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  options.startNow = 0;
  options.stopNow = 0;
  options.resetNow = 1;
  
  result = NTCI_SetStatisticsOptions(cardHandle,
                                     channelsBitmask,
                                     STID_STATISTICS_OPTIONS_TYPE3,
                                     &options);
  
  if (result != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Statistics: Error resetting statistics: %i\n", result);
  }
  return result;
}


/* Print simple statistics of a particular channel, for
 * use in the menu */
uint32_t StatisticsGetSummary(uint8_t card, uint32_t channelBitmask, bool_t RxStat)
{
  StatisticsEtherType2_t stats;
  uint32_t result=0;
  int64_t deltaOctCount=0;
  int64_t deltaTimeCount=0;
  int64_t deltaFrameCount=0;
  double utilization=0;
  double bytesum=0;
  double rate=0;
  double timeInSecs=0;
  double bitsPerSec=0;
  uint32_t channels=0, i=0, channel=0;
  ChannelStatusType1_t chanStatus;
  uint64_t lastValidAt=0;
  TimestampType_t timestampType=0;
  uint64_t dwlBytesReceived=0, dwlFramesReceived=0;
  handle_t cardHandle;
  uint32_t s=0;
  CardInterfaceSupportedSTIDType1_t supportedStids;
  bool_t boBlockStatSupported=FALSE;    
  
  writePos = 0;

  memset(&stats, 0x00, sizeof(stats));
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  /* Check that BLOCK statistics is supported. */      
  if((result = NTCI_GetCapabilities(cardHandle, CAPID_STATISTICS_IF, NOT_SUB_CAPABILITIES,
                                    STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Failed to get statistics if capabilities.\n");    
    return result;
  }
  s=0;
  while (strlen(supportedStids.InterfaceSTID[s]) != 0) {    
    if(strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[s],
               sizeof("STID_STATISTICS_BLOCK_TYPE1")) == 0) {      
      boBlockStatSupported=TRUE;
      break;
    }
    s++;
  }       

  channels=0;
  for(i=0; i<MAX_CHANNELS; i++) {
    if((1<<i) & channelBitmask) {
      bitsPerSec=0;
      timeInSecs=0;
      utilization=0;
      channels++;
      channel=i;
      
      /*calculate the change since the last sample */
      /* Get statistics */
      if (!boBlockStatSupported) {
        if(RxStat) {
          result = NTCI_GetStatistics(cardHandle, 
                                      1<<channel,
                                      STID_ETHER_STATISTICS_TYPE2,
                                      &stats);
        } else {
          // TX is default not supported.
          result = NTCI_ERRCODE_INTERFACE_NOT_AVAILABLE;
        }

        if (result == NTCI_ERRCODE_SUCCESS) {
          dwlBytesReceived = stats.RMON1Statistics.dwlOctetsReceivedOK;
          dwlFramesReceived = stats.RMON1Statistics.dwlFramesReceivedOK;
          lastValidAt = stats.dwlTimeValidAt;
          timestampType = stats.timestampType;
        }                   
      } else {
        /* Try the block statistics instead. */
        StatisticsBlockType1_t statBlockType1;
        MemoryDMAType1_t memType1;
        StatBlockDescriptorType1_t *descriptor;
        /* Summary descriptors. */
        StatisticsEtherType7_t *type7;
        

        memset(&statBlockType1, 0x00, sizeof(statBlockType1));    
        /* Get the dma buffer size required for the stat block. */
        if((result =
            NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
          WriteToLog(LOG_ERROR, "Failed to get required stat block size.\n");    
          return result;
        }
        
        memcpy(&memType1.dmaBuffer, &statBlockType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
        if((result =
            NTCI_Memory(cardHandle, STID_MEMORY_ALLOC_DMA_TYPE1, &memType1)) != NTCI_ERRCODE_SUCCESS) {
          WriteToLog(LOG_ERROR, "Failed to allocate %d bytes DMA memory.\n",
                     memType1.dmaBuffer.bufferSize);
        }
        memcpy(&statBlockType1.dmaBuffer, &memType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
        
        /* Get the next statistics block. */
        /* We want to wait for new data. */
        statBlockType1.waitForNewData = 0;
        if((result =
            NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
          WriteToLog(LOG_ERROR, "Failed to get next stat block.\n");    
          return result;
        }
        
        descriptor = (StatBlockDescriptorType1_t*)statBlockType1.dmaBuffer.user.u.pVirt;
        dwlBytesReceived=0;
        dwlFramesReceived=0;
        memset(&stats, 0x00, sizeof(stats));
        for(s=0; s<statBlockType1.numDescriptors; s++) {
          if(descriptor->headerVersion==1) {
            if(RxStat) {
              switch(descriptor->type) {
              case STAT_TYPE_PORT:
              case STAT_TYPE_PORT_FILTERED:
                if(channel == descriptor->index) {                  
                  type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
                  ConvertEtherType7ToEtherType2(&stats, type7);
                  lastValidAt = stats.dwlTimeValidAt;
                  timestampType = stats.timestampType;
                }
                break;
              }
            } else {
              switch(descriptor->type) {
              case STAT_TYPE_PORT_TX:
                if(channel == descriptor->index) {
                  // Type7 and type8 are alike. Only element naming differ.
                  type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
                  ConvertEtherType7ToEtherType2(&stats, type7);
                  lastValidAt = stats.dwlTimeValidAt;
                  timestampType = stats.timestampType;
                }
                break;
              }                
            }
          } else {
            WriteToLog(LOG_ERROR, "Statblock unsupported header version %d.\n",
                       descriptor->headerVersion);
            NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
            return result;
          }
          descriptor = (StatBlockDescriptorType1_t*)((uint8_t*)descriptor+descriptor->storedLength);
        }
        dwlBytesReceived += stats.RMON1Statistics.dwlOctetsReceivedOK;
        dwlFramesReceived += stats.RMON1Statistics.dwlFramesReceivedOK;
        NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
      }
      deltaOctCount =  dwlBytesReceived - prevOctCount[card][channel];
      /* Convert last valid at to NDIS native */
      if(timestampType!=TIMESTAMP_NATIVE) {
        ConvertSystemTickTime(&lastValidAt, timestampType, TIMESTAMP_NATIVE_NDIS);
      }
      deltaTimeCount = lastValidAt - prevTimeCount[card][channel];
      deltaFrameCount = dwlFramesReceived - prevFrameCount[card][channel];
      
      /* we don't want division by zero from deltaTimeCount when we first run this*/
      if(deltaTimeCount) {
        bytesum = (double) ((deltaFrameCount * 20) + deltaOctCount);
        rate = (double)1250000; /* rate adjusted for link speed (10Mbps)*/
        
        result = NTCI_GetChannelStatus(cardHandle, 
                                       1<<channel,
                                       STID_CHANNEL_STATUS_TYPE1,
                                       &chanStatus);
        if(result == NTCI_ERRCODE_SUCCESS) {
          switch(chanStatus.actualSpeed) {
          case LINK_SPEED_10M:  break;
          case LINK_SPEED_100M: rate*=10; break;
          case LINK_SPEED_1G:   rate*=100; break;
          case LINK_SPEED_10G:  rate*=1000; break;
          case LINK_SPEED_40G:  rate*=4000; break;            
          }
        }
        if(bytesum) {
          timeInSecs = (double)deltaTimeCount/100000000; /* bytes per sec, not 10nsec */        
          utilization = ((bytesum *100)/ (rate * timeInSecs));
          if (utilization > 100) {
            utilization = 100;
          }        
          bitsPerSec = ((bytesum/1000/1000)*8)/timeInSecs;
        }
      }
      if (prevTimeCount[card][channel]) {
	STATS_LOG("\nCh%d: Packets=0x%08llX : Bytes=0x%014llX : "
		  "Mbps=%7.02f : Util=%6.02f",
		  channel,
		  (ULL)dwlFramesReceived,
		  (ULL)dwlBytesReceived,
		  bitsPerSec, utilization);
      } else {
	/* prevTimeCount[card][channel] is zero on the first invocation.
	   The code simply states "pending" for both load and utilisation.
	   The next invocation will report the correct numbers. */
	STATS_LOG("\nCh%d: Packets=0x%08llX : Bytes=0x%014llX : "
		  "Mbps=%7s : Util=%6s",
		  channel,
		  (ULL)dwlFramesReceived,
		  (ULL)dwlBytesReceived,
		  "pending", "pending");
      }

      /* save previous values so we can calculate utilization next time */
      prevOctCount[card][channel]  = dwlBytesReceived;
      prevTimeCount[card][channel]  = lastValidAt;
      prevFrameCount[card][channel] = dwlFramesReceived;
    }
  }
  STATS_LOG("\n");
  WriteToLog(LOG_INFO, writeBuffer);
  return NTCI_ERRCODE_SUCCESS;
}


/* Print detailed channel statistics for the selected channel on a NT card using StatisticsEtherType2_t */
uint32_t StatisticsGetType2Counters(uint8_t card, uint32_t channelsBitmask, bool_t hex)
{
  StatisticsEtherType2_t stats2;
  uint32_t result=0;
  handle_t cardHandle=NULL;

  writePos = 0;

  memset(&stats2, 0x00, sizeof(stats2));
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }  

  /* Get statistics */
  result = NTCI_GetStatistics(cardHandle, 
                              channelsBitmask,
                              STID_ETHER_STATISTICS_TYPE2,
                              &stats2);

  if (result != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Statistics: Failed to retrieve statistics for mask: %X (%i)\n",
               channelsBitmask, result);
    return result;
  }
  
  STATS_LOG("\nRMON1Statistics structure of the STID_ETHER_STATISTICS_TYPE2 for channelmask 0x%08lX:\n", channelsBitmask);  
  /* Display RMON statistics */  
  STATS_LOG("dwlFramesReceivedOK               : %s ",  NumAsString(stats2.RMON1Statistics.dwlFramesReceivedOK, hex));
  STATS_LOG("dwlOctetsReceivedOK               : %s\n", NumAsString(stats2.RMON1Statistics.dwlOctetsReceivedOK, hex));
  STATS_LOG("dwlGroupFramesReceivedOk          : %s ",  NumAsString(stats2.RMON1Statistics.dwlGroupFramesReceivedOk, hex));
  STATS_LOG("dwlGroupOctetsReceivedOk          : %s\n", NumAsString(stats2.RMON1Statistics.dwlGroupOctetsReceivedOk, hex));
  STATS_LOG("dwlFragmentsReceivedOk            : %s ",  NumAsString(stats2.RMON1Statistics.dwlFragmentsReceivedOk, hex));    
  STATS_LOG("dwlOversizeFramesReceivedOk       : %s\n", NumAsString(stats2.RMON1Statistics.dwlOversizeFramesReceivedOk, hex));
  STATS_LOG("dwlUnderSizedFramesReceivedOk     : %s ",  NumAsString(stats2.RMON1Statistics.dwlUnderSizedFramesReceivedOk, hex));  
  STATS_LOG("dwlJabberFramesReceivedOk         : %s\n", NumAsString(stats2.RMON1Statistics.dwlJabberFramesReceivedOk, hex));
  STATS_LOG("dwlBroadcastFramesReceivedOK      : %s ",  NumAsString(stats2.RMON1Statistics.dwlBroadcastFramesReceivedOK, hex));
  STATS_LOG("dwlUnicastFramesReceivedOK        : %s\n", NumAsString(stats2.RMON1Statistics.dwlUnicastFramesReceivedOK, hex));    
  STATS_LOG("dwlMulticastFramesReceivedOK      : %s ",  NumAsString(stats2.RMON1Statistics.dwlMulticastFramesReceivedOK, hex));  
  STATS_LOG("dwlCRCErrors                      : %s\n", NumAsString(stats2.RMON1Statistics.dwlCRCErrors, hex));
  STATS_LOG("dwlCharacterCVErrors              : %s ",  NumAsString(stats2.RMON1Statistics.dwlCharacterCVErrors, hex));
  STATS_LOG("dwlReceiveFrames64Octets          : %s\n", NumAsString(stats2.RMON1Statistics.dwlReceiveFrames64Octets, hex));
  STATS_LOG("dwlReceiveFrames65To127Octets     : %s ",  NumAsString(stats2.RMON1Statistics.dwlReceiveFrames65To127Octets, hex));
  STATS_LOG("dwlReceiveFrames128To255Octets    : %s\n", NumAsString(stats2.RMON1Statistics.dwlReceiveFrames128To255Octets, hex));
  STATS_LOG("dwlReceiveFrames256To511Octets    : %s ",  NumAsString(stats2.RMON1Statistics.dwlReceiveFrames256To511Octets, hex));    
  STATS_LOG("dwlReceiveFrames512To1023Octets   : %s\n", NumAsString(stats2.RMON1Statistics.dwlReceiveFrames512To1023Octets, hex));
  STATS_LOG("dwlReceiveFrames1024To1518Octets  : %s ",  NumAsString(stats2.RMON1Statistics.dwlReceiveFrames1024To1518Octets, hex));
  STATS_LOG("dwlJumboFrames1519To2047Octets    : %s\n", NumAsString(stats2.RMON1Statistics.dwlJumboFrames1519To2047Octets, hex));    
  STATS_LOG("dwlJumboFrames2048To4095Octets    : %s ",  NumAsString(stats2.RMON1Statistics.dwlJumboFrames2048To4095Octets, hex));
  STATS_LOG("dwlJumboFrames4096To8191Octets    : %s\n", NumAsString(stats2.RMON1Statistics.dwlJumboFrames4096To8191Octets, hex)); 
  STATS_LOG("dwlJumboFrames8192To9018Octets    : %s ",  NumAsString(stats2.RMON1Statistics.dwlJumboFrames8192To9018Octets, hex));
  STATS_LOG("dwlVLANJumboFrames9019To9022Octets: %s\n", NumAsString(stats2.RMON1Statistics.dwlVLANJumboFrames9019To9022Octets, hex));
  STATS_LOG("dwlBabyGiantVLANFramesReceivedOK  : %s ",  NumAsString(stats2.RMON1Statistics.dwlBabyGiantVLANFramesReceivedOK, hex));
  STATS_LOG("dwlBabyGiantISLFramesReceivedOK   : %s\n", NumAsString(stats2.RMON1Statistics.dwlBabyGiantISLFramesReceivedOK, hex));
  STATS_LOG("dwlBabyGiantJumboFramesReceivedOK : %s\n", NumAsString(stats2.RMON1Statistics.dwlBabyGiantJumboFramesReceivedOK, hex));  

  STATS_LOG("\nExtendedStatistics structure of the STID_ETHER_STATISTICS_TYPE2 for channelmask 0x%08lX:\n", channelsBitmask);  
  /* Display Extended Statistics */
  STATS_LOG("dwlLayer3ChecksumErrors           : %s ",  NumAsString(stats2.ExtendedStatistics.dwlLayer3ChecksumErrors, hex));
  STATS_LOG("dwlLayer4ChecksumErrors           : %s\n", NumAsString(stats2.ExtendedStatistics.dwlLayer4ChecksumErrors, hex));  
  STATS_LOG("dwlTCPChecksumErrors              : %s ",  NumAsString(stats2.ExtendedStatistics.dwlTCPChecksumErrors, hex));
  STATS_LOG("dwlUDPChecksumErrors              : %s\n", NumAsString(stats2.ExtendedStatistics.dwlUDPChecksumErrors, hex));
  
  STATS_LOG("\nPacketFeedStatistics structure of the STID_ETHER_STATISTICS_TYPE2 for channelmask 0x%08lX:\n", channelsBitmask);  
  /* Display  Packetfeed Statistics */
  STATS_LOG("dwlSmallFramesDropped             : %s ",  NumAsString(stats2.PacketFeedStatistics.dwlSmallFramesDropped, hex));
  STATS_LOG("dwlBandwidthDropped               : %s\n", NumAsString(stats2.PacketFeedStatistics.dwlBandwidthDropped, hex));
  STATS_LOG("dwlPacketFeedFilteredDropped      : %s ",  NumAsString(stats2.PacketFeedStatistics.dwlPacketFeedFilteredDropped, hex));  
  STATS_LOG("dwlPacketFeedBandwidthDropped     : %s\n", NumAsString(stats2.PacketFeedStatistics.dwlPacketFeedBandwidthDropped, hex));
  STATS_LOG("dwlPacketFeedFrames               : %s ",  NumAsString(stats2.PacketFeedStatistics.dwlPacketFeedFrames, hex));
  STATS_LOG("dwlMaxFrameSizeDropped            : %s\n", NumAsString(stats2.PacketFeedStatistics.dwlMaxFrameSizeDropped, hex));
  STATS_LOG("dwlHardSlicedFrames               : %s ",  NumAsString(stats2.PacketFeedStatistics.dwlHardSlicedFrames, hex));
  STATS_LOG("overflow                          : %s\n", NumAsString(stats2.PacketFeedStatistics.overflow, hex));
  STATS_LOG("timestampType                     : %s\n", NumAsString(stats2.timestampType, hex));
  STATS_LOG("dwlTimeValidAt                    : %s\n", NumAsString(stats2.dwlTimeValidAt, hex));    
  
  WriteToLog(LOG_INFO, writeBuffer);  
  return NTCI_ERRCODE_SUCCESS;  
}


/* Print detailed channel statistics for the selected channel */
uint32_t StatisticsGetRMONCounters(uint8_t card, uint32_t channelsBitmask, bool_t hex, bool_t RxStat)
{
  StatisticsEtherType2_t stats2;
  StatisticsEtherType6_t stats6;
  uint32_t result=0;
  handle_t cardHandle=NULL;
  bool_t boBlockStatSupported=FALSE;
  uint32_t s=0;
  CardInterfaceSupportedSTIDType1_t supportedStids;  
        
  writePos = 0;

  memset(&stats2, 0x00, sizeof(stats2));
  memset(&stats6, 0x00, sizeof(stats6));  
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  /* Check that BLOCK statistics is supported. */      
  if((result = NTCI_GetCapabilities(cardHandle, CAPID_STATISTICS_IF, NOT_SUB_CAPABILITIES,
                                    STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Failed to get statistics if capabilities.\n");    
    return result;
  }
  s=0;
  while (strlen(supportedStids.InterfaceSTID[s]) != 0) {    
    if(strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[s],
               sizeof("STID_STATISTICS_BLOCK_TYPE1")) == 0) {      
      boBlockStatSupported=TRUE;
      break;
    }
    s++;
  }       
  
  /* Get statistics */
  if (!boBlockStatSupported) {
    if(RxStat) {
      result = NTCI_GetStatistics(cardHandle, 
                                  channelsBitmask,
                                  STID_ETHER_STATISTICS_TYPE6,
                                  &stats6);
    } else {
      // TX is default not supported.
      result = NTCI_ERRCODE_INTERFACE_NOT_AVAILABLE;
    }
    
  } else {

   /* Try the block statistics instead. */
    StatisticsBlockType1_t statBlockType1;
    MemoryDMAType1_t memType1;
    StatBlockDescriptorType1_t *descriptor;
    /* Summary descriptors. */
    StatisticsEtherType7_t *type7;        
    

    memset(&statBlockType1, 0x00, sizeof(statBlockType1));    
    /* Get the dma buffer size required for the stat block. */
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get required stat block size.\n");    
      return result;
    }
    
    memcpy(&memType1.dmaBuffer, &statBlockType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
    if((result =
        NTCI_Memory(cardHandle, STID_MEMORY_ALLOC_DMA_TYPE1, &memType1)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_ERROR, "Failed to allocate %d bytes DMA memory.\n",
                 memType1.dmaBuffer.bufferSize);
    }
    memcpy(&statBlockType1.dmaBuffer, &memType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
    
    /* Get the next statistics block. */
    /* We want to wait for new data. */
    statBlockType1.waitForNewData = 0;
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get next stat block.\n");    
      return result;
    }
    
    descriptor = (StatBlockDescriptorType1_t*)statBlockType1.dmaBuffer.user.u.pVirt;
    memset(&stats6, 0x00, sizeof(stats6));
    for(s=0; s<statBlockType1.numDescriptors; s++) {
      if(descriptor->headerVersion==1) {
        if(RxStat) {
          switch(descriptor->type) {
          case STAT_TYPE_PORT:
          case STAT_TYPE_PORT_FILTERED:
            if(channelsBitmask & (1<<descriptor->index)) {
              type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
              ConvertEtherType7ToEtherType6(&stats6, type7);
            }
            break;
          }
        } else {
          switch(descriptor->type) {
          case STAT_TYPE_PORT_TX:
            if(channelsBitmask & (1<<descriptor->index)) {
              // Type 7 and 8 are equal. It is just the naming that is different.
              type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
              ConvertEtherType7ToEtherType6(&stats6, type7);
            }
            break;
          }
        }
      } else {
        WriteToLog(LOG_ERROR, "Statblock unsupported header version %d.\n",
                   descriptor->headerVersion);
        NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
        return result;
      }
      descriptor = (StatBlockDescriptorType1_t*)((uint8_t*)descriptor+descriptor->storedLength);
    }
    NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
  }
  
  if(boBlockStatSupported) {
    STATS_LOG("\nDMA block statistics is used and has been converted to show the counters below:\n");
  }
  
  STATS_LOG("\nRMON statistics for channelmask 0x%08lX:\n", channelsBitmask);

  /* Display RMON statistics */
  STATS_LOG("Total packets        : %s ",  NumAsString(stats6.RMONStatistics.dwlFramesReceived, hex));
  STATS_LOG("Total bytes          : %s\n", NumAsString(stats6.RMONStatistics.dwlOctetsReceived, hex));
  STATS_LOG("Undersized packets   : %s ",  NumAsString(stats6.RMONStatistics.dwlUnderSizedFramesReceived, hex));
  STATS_LOG("Oversized packets    : %s\n", NumAsString(stats6.RMONStatistics.dwlOversizeFramesReceived, hex));
  STATS_LOG("Fragments            : %s ",  NumAsString(stats6.RMONStatistics.dwlFragmentsReceived, hex));
  STATS_LOG("Jabber packets       : %s\n", NumAsString(stats6.RMONStatistics.dwlJabberFramesReceived, hex));
  STATS_LOG("Multicast packets    : %s ",  NumAsString(stats6.RMONStatistics.dwlMulticastFramesReceived, hex));
  STATS_LOG("Broadcast packets    : %s\n", NumAsString(stats6.RMONStatistics.dwlBroadcastFramesReceived, hex));
  STATS_LOG("CRC Align errors     : %s ",  NumAsString(stats6.RMONStatistics.dwlCRCAlignErrors, hex));
  STATS_LOG("Drop Events          : %s\n", NumAsString(stats6.RMONStatistics.dwlDropEvents, hex));    
  STATS_LOG("Frames size 64       : %s ",  NumAsString(stats6.RMONStatistics.dwlReceiveFrames64Octets, hex));
  STATS_LOG("Frames size 65-127   : %s\n", NumAsString(stats6.RMONStatistics.dwlReceiveFrames65To127Octets, hex));    
  STATS_LOG("Frames size 128-255  : %s ",  NumAsString(stats6.RMONStatistics.dwlReceiveFrames128To255Octets, hex));
  STATS_LOG("Frames size 256-511  : %s\n", NumAsString(stats6.RMONStatistics.dwlReceiveFrames256To511Octets, hex));    
  STATS_LOG("Frames size 512-1023 : %s ",  NumAsString(stats6.RMONStatistics.dwlReceiveFrames512To1023Octets, hex));
  STATS_LOG("Frames size 1024-1518: %s\n", NumAsString(stats6.RMONStatistics.dwlReceiveFrames1024To1518Octets, hex));

  WriteToLog(LOG_INFO, writeBuffer);
  return NTCI_ERRCODE_SUCCESS;
}

/* Print detailed channel statistics for the selected channel */
uint32_t StatisticsGetHwCounters(uint8_t card, uint32_t channelsBitmask, bool_t hex, bool_t RxStat)
{
  StatisticsEtherType6_t stats;
  uint32_t result;
  handle_t cardHandle;
  bool_t boBlockStatSupported=FALSE;
  uint32_t s=0;
  CardInterfaceSupportedSTIDType1_t supportedStids;

  writePos = 0;

  memset(&stats, 0x00, sizeof(stats));  
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  /* Check that BLOCK statistics is supported. */      
  if((result = NTCI_GetCapabilities(cardHandle, CAPID_STATISTICS_IF, NOT_SUB_CAPABILITIES,
                                    STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Failed to get statistics if capabilities.\n");    
    return result;
  }
  s=0;
  while (strlen(supportedStids.InterfaceSTID[s]) != 0) {    
    if(strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[s],
               sizeof("STID_STATISTICS_BLOCK_TYPE1")) == 0) {      
      boBlockStatSupported=TRUE;
      break;
    }
    s++;
  }       
  

  /* Get statistics */
  if (!boBlockStatSupported) {
    if(RxStat) {
      result = NTCI_GetStatistics(cardHandle, 
                                  channelsBitmask,
                                  STID_ETHER_STATISTICS_TYPE6,
                                  &stats);
    } else {
      // TX is default not supported.
      result = NTCI_ERRCODE_INTERFACE_NOT_AVAILABLE;    
    }
  } else {
    
    /* Try the block statistics instead. */
    StatisticsBlockType1_t statBlockType1;
    MemoryDMAType1_t memType1;
    StatBlockDescriptorType1_t *descriptor;
    /* Summary descriptors. */
    StatisticsEtherType7_t *type7;
    
    memset(&statBlockType1, 0x00, sizeof(statBlockType1));    
    /* Get the dma buffer size required for the stat block. */
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get required stat block size.\n");    
      return result;
    }
    
    memcpy(&memType1.dmaBuffer, &statBlockType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
    if((result =
        NTCI_Memory(cardHandle, STID_MEMORY_ALLOC_DMA_TYPE1, &memType1)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_ERROR, "Failed to allocate %d bytes DMA memory.\n",
                 memType1.dmaBuffer.bufferSize);
    }
    memcpy(&statBlockType1.dmaBuffer, &memType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
    
    /* Get the next statistics block. */
    /* We want to wait for new data. */
    statBlockType1.waitForNewData = 0;
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get next stat block.\n");    
      return result;
    }
    
    descriptor = (StatBlockDescriptorType1_t*)statBlockType1.dmaBuffer.user.u.pVirt;
    memset(&stats, 0x00, sizeof(stats));
    for(s=0; s<statBlockType1.numDescriptors; s++) {
      if(descriptor->headerVersion==1) {
        if(RxStat) {
          switch(descriptor->type) {
          case STAT_TYPE_MISC:
            if(channelsBitmask & (1<<descriptor->index)) {
              StatisticsMiscType1_t *misc = (StatisticsMiscType1_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
              stats.HardwareStatistics.dwlRxXGMIISmallFrameDrop+=misc->dwlRxXGMIISmallFrameDrop;
              stats.HardwareStatistics.dwlRxPacketMergeDrop+=misc->dwlRxPacketMergeDrop;
            }
            break;
          case STAT_TYPE_PORT:
          case STAT_TYPE_PORT_FILTERED:
            if(channelsBitmask & (1<<descriptor->index)) {
              type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
              ConvertEtherType7ToEtherType6(&stats, type7);
            }
            break;
          }
        } else {
          switch(descriptor->type) {
          case STAT_TYPE_PORT_TX:
            if(channelsBitmask & (1<<descriptor->index)) {
              // Type 7 and 8 are equal. It is just the naming that is different.
              type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));  
              ConvertEtherType7ToEtherType6(&stats, type7);
            }
            break;
          }
        }
      }else {
        WriteToLog(LOG_ERROR, "Statblock unsupported header version %d.\n",
                   descriptor->headerVersion);
        NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
        return result;
      }
      descriptor = (StatBlockDescriptorType1_t*)((uint8_t*)descriptor+descriptor->storedLength);
    }
    NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
  }

  if(boBlockStatSupported) {
    STATS_LOG("\nDMA block statistics is used and has been converted to show the counters below:\n");
  }
  /* Display Hardware counters */    
  STATS_LOG("\nHardware counters for channelmask 0x%08lX:\n", channelsBitmask);
  STATS_LOG("Frames size 64       : %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize64, hex));
  STATS_LOG("Frames size 65-127   : %s\n", NumAsString(stats.HardwareStatistics.dwlRxFrameSize65To127, hex));
  STATS_LOG("Frames size 128-255  : %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize128To255, hex));
  STATS_LOG("Frames size 256-511  : %s\n", NumAsString(stats.HardwareStatistics.dwlRxFrameSize256To511, hex));    
  STATS_LOG("Frames size 512-1023 : %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize512To1023, hex));
  STATS_LOG("Frames size 1024-1518: %s\n", NumAsString(stats.HardwareStatistics.dwlRxFrameSize1024To1518, hex));    
  STATS_LOG("Frames size 1519-2047: %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize1519To2047, hex));
  STATS_LOG("Frames size 2048-4095: %s\n", NumAsString(stats.HardwareStatistics.dwlRxFrameSize2048To4095, hex));    
  STATS_LOG("Frames size 4096-8191: %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize4096To8191, hex));
  if(boBlockStatSupported) {
    STATS_LOG("Frames size 8192-Max : %s\n",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize8192To9018, hex));    
    STATS_LOG("                                        ");
  } else {
    STATS_LOG("Frames size 8192-9018: %s\n", NumAsString(stats.HardwareStatistics.dwlRxFrameSize8192To9018, hex));    
    STATS_LOG("Frames size 9019-Max : %s ",  NumAsString(stats.HardwareStatistics.dwlRxFrameSize9019ToMax, hex));
  } 
  STATS_LOG("Total frames         : %s\n", NumAsString(stats.HardwareStatistics.dwlRxTotalFrames, hex));    
  STATS_LOG("Oversize frames      : %s ",  NumAsString(stats.HardwareStatistics.dwlRxOversizeFrames, hex));
  STATS_LOG("Oversize Hard sliced : %s\n", NumAsString(stats.HardwareStatistics.dwlRxOversizeHardSlicedFrames, hex));    
  STATS_LOG("Jabber frames        : %s ",  NumAsString(stats.HardwareStatistics.dwlRxJabberFrames, hex));
  STATS_LOG("Jabber Hard sliced   : %s\n", NumAsString(stats.HardwareStatistics.dwlRxJabberHardSlicedFrames, hex));
  STATS_LOG("Undersized frames    : %s ",  NumAsString(stats.HardwareStatistics.dwlRxUndersizeFrames, hex));
  STATS_LOG("Ok pause bad MAC ctrl: %s\n", NumAsString(stats.HardwareStatistics.dwlRxGoodPauseUnsupportedMACControl, hex));    
  STATS_LOG("Good bytes           : %s ",  NumAsString(stats.HardwareStatistics.dwlRxGoodFrameBytes, hex));
  STATS_LOG("Bad bytes            : %s\n", NumAsString(stats.HardwareStatistics.dwlRxBadFrameBytes, hex));        
  STATS_LOG("Unicast frames       : %s ",  NumAsString(stats.HardwareStatistics.dwlRxUnicastFrames, hex));
  STATS_LOG("Multicast frames     : %s\n", NumAsString(stats.HardwareStatistics.dwlRxMulticastFrames, hex));        
  STATS_LOG("Broadcast frames     : %s ",  NumAsString(stats.HardwareStatistics.dwlRxBroadcastFrames, hex));
  STATS_LOG("Pause frames         : %s\n", NumAsString(stats.HardwareStatistics.dwlRxPauseFrames, hex));         
  STATS_LOG("Unsupported MAC Ctrl : %s ",  NumAsString(stats.HardwareStatistics.dwlRxUnsupportedMACControlFrames, hex));
  STATS_LOG("Crc Errors           : %s\n", NumAsString(stats.HardwareStatistics.dwlRxCrcErrors, hex));          
  STATS_LOG("Fragments            : %s ",  NumAsString(stats.HardwareStatistics.dwlRxFragments, hex));        
  STATS_LOG("Code Violation Errors: %s\n", NumAsString(stats.HardwareStatistics.dwlRxCodeViolationErrors, hex));          
  if(RxStat) {
    STATS_LOG("Alignment Errors     : %s ",  NumAsString(stats.HardwareStatistics.dwlRxAlignmentErrors, hex));
    STATS_LOG("XGMII Small Drop     : %s\n", NumAsString(stats.HardwareStatistics.dwlRxXGMIISmallFrameDrop, hex));         
    STATS_LOG("Packet Merge Drop    : %s ",  NumAsString(stats.HardwareStatistics.dwlRxPacketMergeDrop, hex));
    if(!boBlockStatSupported) {
      STATS_LOG("Smoother Drop        : %s\n", NumAsString(stats.HardwareStatistics.dwlRxSmootherDrop, hex));
    } else {
      STATS_LOG("\nSmoother Drop: Counter not supported.\nLoad driver with \"PerPortStatistics=1\"");
    }
  }
  WriteToLog(LOG_INFO, writeBuffer);
  return NTCI_ERRCODE_SUCCESS;
}


/* Print the selected statistics from the block statistics. */
uint32_t StatisticsGetBlockCounters(uint8_t card,
                                    uint32_t channelsBitmask,
                                    uint64_t filtersBitmask,
                                    uint32_t feedsBitmask,
                                    uint8_t   summary,
                                    bool_t hex)
{
  StatisticsBlockType1_t statBlockType1;
  uint32_t result=0, i=0;
  handle_t cardHandle;
  CardInterfaceSupportedSTIDType1_t supportedStids;
  bool_t boBlockStatSupported=FALSE;
  MemoryDMAType1_t memType1;
  StatBlockDescriptorType1_t *descriptor;
  /* Summary descriptors. */
  uint8_t chanDisc[sizeof(StatisticsEtherType7_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t chanFwrd[sizeof(StatisticsEtherType7_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t filterFwrd[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t filterDisc[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t feedFwrd[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t feedInvl[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t feedBnwd[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t chanDedup[sizeof(StatisticsSimpleType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t misc[sizeof(StatisticsMiscType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t chanTx[sizeof(StatisticsEtherType8_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t ipFragFeed[sizeof(StatisticsIpFragType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t ipFragPort[sizeof(StatisticsIpFragType1_t)+sizeof(StatBlockDescriptorType1_t)];
  uint8_t ipFragPortDrop[sizeof(StatisticsIpFragType1_t)+sizeof(StatBlockDescriptorType1_t)];

  StatisticsRetransmit_t *retransmit = NULL;          
  
  memset(chanDisc, 0x00, sizeof(chanDisc));
  memset(chanFwrd, 0x00, sizeof(chanFwrd));
  memset(filterFwrd, 0x00, sizeof(filterFwrd));
  memset(filterDisc, 0x00, sizeof(filterDisc));
  memset(feedFwrd, 0x00, sizeof(feedFwrd));
  memset(feedInvl, 0x00, sizeof(feedInvl));
  memset(feedBnwd, 0x00, sizeof(feedBnwd));
  memset(chanDedup, 0x00, sizeof(chanDedup));
  memset(chanTx, 0x00, sizeof(chanTx));
  memset(misc, 0x00, sizeof(misc));
  memset(ipFragFeed, 0x00, sizeof(ipFragFeed));
  memset(ipFragPort, 0x00, sizeof(ipFragPort));
  memset(ipFragPortDrop, 0x00, sizeof(ipFragPortDrop));
  
  writePos = 0;
  
  if((result = GetCardHandle(card, &cardHandle)) != NTCI_ERRCODE_SUCCESS) {
    return result;
  }

  /* Check that BLOCK statistics is supported. */      
  if((result = NTCI_GetCapabilities(cardHandle, CAPID_STATISTICS_IF, NOT_SUB_CAPABILITIES,
                                    STID_CARD_INTERFACE_SUPPORTED_STID_TYPE1, &supportedStids)) != NTCI_ERRCODE_SUCCESS) {
    WriteToLog(LOG_ERROR, "Failed to get statistics if capabilities.\n");    
    return result;
  }
  i=0;
  while (strlen(supportedStids.InterfaceSTID[i]) != 0) {    
    if(strncmp("STID_STATISTICS_BLOCK_TYPE1", supportedStids.InterfaceSTID[i],
               sizeof("STID_STATISTICS_BLOCK_TYPE1")) == 0) {
      boBlockStatSupported=TRUE;
      break;
    }
    i++;
  }

  if(boBlockStatSupported) {
    memset(&statBlockType1, 0x00, sizeof(statBlockType1));    
    /* Get the dma buffer size required for the stat block. */
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get required stat block size.\n");    
      return result;
    }

    memcpy(&memType1.dmaBuffer, &statBlockType1.dmaBuffer, sizeof(MemoryDMAType1_t));    
    if((result =
        NTCI_Memory(cardHandle, STID_MEMORY_ALLOC_DMA_TYPE1, &memType1)) != NTCI_ERRCODE_SUCCESS) {
      WriteToLog(LOG_ERROR, "Failed to allocate %d bytes DMA memory.\n",
                 memType1.dmaBuffer.bufferSize);
    }
    memcpy(&statBlockType1.dmaBuffer, &memType1.dmaBuffer, sizeof(MemoryDMAType1_t));    

    /* Get the next statistics block. */
    /* We want to wait for new data. */
    statBlockType1.waitForNewData = 1;
    if((result =
        NTCI_GetStatistics(cardHandle, 0, STID_STATISTICS_BLOCK_TYPE1, &statBlockType1)) != NTCI_ERRCODE_SUCCESS) { 
      WriteToLog(LOG_ERROR, "Failed to get next stat block.\n");    
      return result;
    }

    descriptor = (StatBlockDescriptorType1_t*)statBlockType1.dmaBuffer.user.u.pVirt;
    for(i=0; i<statBlockType1.numDescriptors; i++) {
      if(descriptor->headerVersion==1) {
        switch((eStatSubType)descriptor->type) {
        case STAT_TYPE_PORT:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE7) {
                StatisticsSummaryEthType7(descriptor, (StatBlockDescriptorType1_t*)chanFwrd);
              }
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsSummaryEthType8(descriptor, (StatBlockDescriptorType1_t*)chanFwrd);
              }
            } else {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE7) {
                StatisticsDumpEthType7(descriptor, summary, hex);
              }
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsDumpEthType8(descriptor, summary, hex);                
              }
            }
          }
          break;
        case STAT_TYPE_PORT_FILTERED:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE7) {
                StatisticsSummaryEthType7(descriptor, (StatBlockDescriptorType1_t*)chanDisc);
              }
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsSummaryEthType8(descriptor, (StatBlockDescriptorType1_t*)chanDisc);
              }
            } else {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE7) {
                StatisticsDumpEthType7(descriptor, summary, hex);
              }
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsDumpEthType8(descriptor, summary, hex);
              }              
            }
          }
          break;
        case STAT_TYPE_PORT_TX:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsSummaryEthType8(descriptor, (StatBlockDescriptorType1_t*)chanTx);
              }
            }
            else {
              if(descriptor->structureId==STID_ETHER_STATISTICS_TYPE8) {
                StatisticsDumpEthType8(descriptor, summary, hex);
              }
            }
          }
          break;          
        case STAT_TYPE_FILTER_FORWARD:
          if(filtersBitmask&(1LL<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)filterFwrd);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_FILTER_DISCARD:
          if(filtersBitmask) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)filterDisc);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_FEED_FORWARD:
          if(feedsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)feedFwrd);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_FEED_INVALID_DROP:
          if(feedsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)feedInvl);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_FEED_BANDWIDTH_DROP:
          if(feedsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)feedBnwd);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_DEDUP:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)chanDedup);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_MISC:        
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummaryMisc(descriptor, (StatBlockDescriptorType1_t*)misc);
            } else {
              StatisticsDumpMisc(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_FEED_IP_FRAGMENT:
          if(feedsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummaryIpFragment(descriptor, (StatBlockDescriptorType1_t*)ipFragFeed);
            } else {
              StatisticsDumpIpFragment(descriptor, summary, hex);
            }
          }
          break;

        case STAT_TYPE_PORT_IP_FRAGMENT:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummaryIpFragment(descriptor, (StatBlockDescriptorType1_t*)ipFragPort);
            } else {
              StatisticsDumpIpFragment(descriptor, summary, hex);
            }
          }
          break;
        case STAT_TYPE_DROP_IP_FRAGMENT:
          if(channelsBitmask&(1<<descriptor->index)) {
            if(summary) {
              StatisticsSummarySimpleType1(descriptor, (StatBlockDescriptorType1_t*)ipFragPortDrop);
            } else {
              StatisticsDumpSimpleType1(descriptor, summary, hex);
            }
          }
          break;

          /* Subject to change in long beach */
        case STAT_TYPE_RETRANSMIT:
          retransmit = (StatisticsRetransmit_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
          if(!summary) {            
            writePos = 0;            
            if(channelsBitmask&0x1) {
              STATS_LOG("Hardware counters for channel %d: - Retransmit -\n", 0);        
              STATS_LOG("Packets discarded        : %s ",  NumAsString( retransmit->Port0.packetsDiscard, hex));
              STATS_LOG("Packets sent             : %s\n", NumAsString(retransmit->Port0.packetsSent, hex));
              STATS_LOG("---------------------------------------------------------------------------------------\n");
            }
            if(channelsBitmask&0x2) {
              STATS_LOG("Hardware counters for channel %d: - Retransmit -\n", 1);
              STATS_LOG("Packets discarded        : %s ",  NumAsString( retransmit->Port1.packetsDiscard, hex));
              STATS_LOG("Packets sent             : %s\n", NumAsString(retransmit->Port1.packetsSent, hex));
              STATS_LOG("---------------------------------------------------------------------------------------\n");
            }
            if(channelsBitmask) {
              WriteToLog(LOG_INFO, writeBuffer);
            }
          }
          break;
        default:
          WriteToLog(LOG_ERROR, "Statblock unsupported type %d.\n",
                     descriptor->type);
          NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
          return NTCI_ERRCODE_INTERNAL_ERROR;
        }
      } else {
        WriteToLog(LOG_ERROR, "Statblock unsupported header version %d.\n",
                   descriptor->headerVersion);
        NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
        return result;
      }
      descriptor = (StatBlockDescriptorType1_t*)((uint8_t*)descriptor+descriptor->storedLength);
    }
    if(summary) {
      if(channelsBitmask) {
        StatisticsDumpEthType7((StatBlockDescriptorType1_t*)chanFwrd, summary, hex);
        StatisticsDumpEthType7((StatBlockDescriptorType1_t*)chanDisc, summary, hex);
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)chanDedup, summary, hex);
        StatisticsDumpMisc((StatBlockDescriptorType1_t*)misc, summary, hex);   
        StatisticsDumpIpFragment((StatBlockDescriptorType1_t*)ipFragPort, summary, hex);
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)ipFragPortDrop, summary, hex);
        if(retransmit) {
          uint64_t packetsDiscard=0;
          uint64_t packetsSent=0;
          writePos = 0;          
          if(channelsBitmask&0x1) {
            packetsDiscard += retransmit->Port0.packetsDiscard;
            packetsSent += retransmit->Port0.packetsSent;
          }
          if(channelsBitmask&0x2) {
            packetsDiscard += retransmit->Port1.packetsDiscard;
            packetsSent += retransmit->Port1.packetsSent;
          }            
          STATS_LOG("Hardware counters for channelmask %X: - Retransmit -\n", 1);
          STATS_LOG("Packets discarded        : %s ",  NumAsString(packetsDiscard, hex));
          STATS_LOG("Packets sent             : %s\n", NumAsString(packetsSent, hex));
          STATS_LOG("---------------------------------------------------------------------------------------\n");
          WriteToLog(LOG_INFO, writeBuffer);
        }
      }
      if(filtersBitmask) {
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)filterFwrd, summary, hex);
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)filterDisc, summary, hex);
      }
      if(feedsBitmask) {
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)feedFwrd, summary, hex);
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)feedInvl, summary, hex);
        StatisticsDumpSimpleType1((StatBlockDescriptorType1_t*)feedBnwd, summary, hex);
        StatisticsDumpIpFragment((StatBlockDescriptorType1_t*)ipFragFeed, summary, hex);
      }
    }    
    NTCI_Memory(cardHandle, STID_MEMORY_FREE_DMA_TYPE1, &memType1);
  } else {
    WriteToLog(LOG_ERROR, "Block statistics is not supported by this adapter.\n");
  }
  return NTCI_ERRCODE_SUCCESS;
}


static void StatisticsSummaryEthType7(StatBlockDescriptorType1_t *src,
                                      StatBlockDescriptorType1_t *dest)
{
  unsigned int i;
  uint64_t *pSrc, *pDest;

  dest->index |= (1<<src->index);
  dest->type = src->type;
  dest->headerVersion = src->headerVersion;
  dest->structureId = src->structureId;  
  ((StatisticsEtherType7_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->timestampType =
    ((StatisticsEtherType7_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->timestampType;
  ((StatisticsEtherType7_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt =
    ((StatisticsEtherType7_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt;
      
  pSrc=(uint64_t*)&((StatisticsEtherType7_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->Set1;
  pDest=(uint64_t*)&((StatisticsEtherType7_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->Set1;
  for(i=0; i<(sizeof(StatSetType1_t)/sizeof(uint64_t)); i++) {
    *pDest+=*pSrc;
    pSrc++; pDest++;
  }  
  pSrc=(uint64_t*)&((StatisticsEtherType7_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->Set2;
  pDest=(uint64_t*)&((StatisticsEtherType7_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->Set2;
  for(i=0; i<(sizeof(StatSetType1_t)/sizeof(uint64_t)); i++) {
    *pDest+=*pSrc;
    pSrc++; pDest++;
  }    
}

static void StatisticsDumpEthType7(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex)
{
  StatisticsEtherType7_t *type7 = NULL;

  type7 = (StatisticsEtherType7_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
  
  writePos = 0;

  if(summary) {
    STATS_LOG("Hardware counters for channelmask 0x%X: - %s -\n",
              descriptor->index, descriptor->type==STAT_TYPE_PORT?"Forwarded by filter":descriptor->type==STAT_TYPE_PORT_TX?"Transmitted":"Discarded by filter");
  } else {
    STATS_LOG("Hardware counters for channel %d: - %s -\n",
              descriptor->index, descriptor->type==STAT_TYPE_PORT?"Forwarded by filter":descriptor->type==STAT_TYPE_PORT_TX?"Transmitted":"Discarded by filter");
  }
  /* Set 1. */
  STATS_LOG("Frames size 64           : %s ",  NumAsString( type7->Set1.dwlRxFrameSize64, hex));
  STATS_LOG("Frames size 65-127       : %s\n", NumAsString(type7->Set1.dwlRxFrameSize65To127, hex));    
  STATS_LOG("Frames size 128-255      : %s ",  NumAsString( type7->Set1.dwlRxFrameSize128To255, hex));
  STATS_LOG("Frames size 256-511      : %s\n", NumAsString(type7->Set1.dwlRxFrameSize256To511, hex));    
  STATS_LOG("Frames size 512-1023     : %s ",  NumAsString( type7->Set1.dwlRxFrameSize512To1023, hex));
  STATS_LOG("Frames size 1024-1518    : %s\n", NumAsString(type7->Set1.dwlRxFrameSize1024To1518, hex));
  STATS_LOG("Frames size 1519-2047    : %s ",  NumAsString( type7->Set1.dwlRxFrameSize1519To2047, hex));
  STATS_LOG("Frames size 2048-4095    : %s\n", NumAsString(type7->Set1.dwlRxFrameSize2048To4095, hex));    
  STATS_LOG("Frames size 4096-8191    : %s ",  NumAsString( type7->Set1.dwlRxFrameSize4096To8191, hex));
  STATS_LOG("Frames size 8192-Max     : %s\n", NumAsString(type7->Set1.dwlRxFrameSize8192ToMax, hex));
  STATS_LOG("Undersized frames        : %s ",  NumAsString( type7->Set1.dwlRxUndersizeFrames, hex));
  STATS_LOG("Fragments                : %s\n", NumAsString(type7->Set1.dwlRxFragments, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Undersized wo encap      : %s ",  NumAsString( type7->Set1.dwlRxUndersizeWithoutEncapsulation, hex));
  }
  STATS_LOG("Oversize frames          : %s\n", NumAsString(type7->Set1.dwlRxOversizeFrames, hex));
  STATS_LOG("Jabber frames            : %s ",  NumAsString( type7->Set1.dwlRxJabberFrames, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Baby giant frames        : %s\n", NumAsString(type7->Set1.dwlRxBabyGiantFrames, hex));
  } else {
    STATS_LOG("\n");
  }
  STATS_LOG("Oversize Hard sliced     : %s ",  NumAsString( type7->Set1.dwlRxOversizeHardSlicedFrames, hex));    
  STATS_LOG("Jabber Hard sliced       : %s\n", NumAsString(type7->Set1.dwlRxJabberHardSlicedFrames, hex));
  STATS_LOG("Good bytes               : %s ",  NumAsString( type7->Set1.dwlRxGoodFrameBytes, hex));
  STATS_LOG("Bad bytes                : %s\n", NumAsString(type7->Set1.dwlRxBadFrameBytes, hex));        
  STATS_LOG("Good pause/mac ctrl bytes: %s\n", NumAsString(type7->Set1.dwlRxGoodPauseMacFramesBytes, hex));
  /* Set 2. */
  STATS_LOG("Unicast frames           : %s ",  NumAsString( type7->Set2.dwlRxUnicastFrames, hex));
  STATS_LOG("Multicast frames         : %s\n", NumAsString(type7->Set2.dwlRxMulticastFrames, hex));        
  STATS_LOG("Broadcast frames         : %s ",  NumAsString( type7->Set2.dwlRxBroadcastFrames, hex));
  STATS_LOG("Unicast pause frames     : %s\n", NumAsString(type7->Set2.dwlRxUnicastPauseFrames, hex));         
  STATS_LOG("Multicast pause frames   : %s ",  NumAsString( type7->Set2.dwlRxMulticastPauseFrames, hex));         
  STATS_LOG("Unicast MAC frames       : %s\n", NumAsString(type7->Set2.dwlRxMacUnicastFrames, hex));
  STATS_LOG("Multicast MAC frames     : %s ",  NumAsString( type7->Set2.dwlRxMacMulticastFrames, hex));
  STATS_LOG("Broadcast MAC frames     : %s\n", NumAsString(type7->Set2.dwlRxMacBroadcastFrames, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Frames with no encap     : %s\n", NumAsString(type7->Set2.dwlRxNotIslVlanMplsFrames, hex));  
    STATS_LOG("Frames with ISL          : %s ",  NumAsString( type7->Set2.dwlRxIslFrames, hex));
    STATS_LOG("Frames with VLAN         : %s\n", NumAsString(type7->Set2.dwlRxVlanFrames, hex));
    STATS_LOG("Frames with ISL,VLAN     : %s ",  NumAsString( type7->Set2.dwlRxIslVlanFrames, hex));
    STATS_LOG("Frames with MPLS         : %s\n", NumAsString(type7->Set2.dwlRxMplsFrames, hex));
    STATS_LOG("Frames with ISL,MPLS     : %s ",  NumAsString( type7->Set2.dwlRxIslMplsFrames, hex));
    STATS_LOG("Frames with VLAN,MPLS    : %s\n", NumAsString(type7->Set2.dwlRxVlanMplsFrames, hex));
    STATS_LOG("Frames with ISL,VLAN,MPLS: %s ",  NumAsString( type7->Set2.dwlRxIslVlanMplsFrames, hex));
  }
  STATS_LOG("CRC errors               : %s\n", NumAsString(type7->Set2.dwlRxCrcErrors, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Alignment errors         : %s ",  NumAsString( type7->Set2.dwlRxAlignmentErrors, hex));
  }
  STATS_LOG("Code violation errors    : %s\n", NumAsString(type7->Set2.dwlRxCodeViolationErrors, hex));
  STATS_LOG("Small frames             : %s ",  NumAsString( type7->Set2.dwlRxSmallFrames, hex));
  STATS_LOG("Large frames             : %s\n", NumAsString(type7->Set2.dwlRxLargePackets, hex));
  STATS_LOG("Time stamp               : 0x%016llX\n", type7->dwlTimeValidAt);
  STATS_LOG("---------------------------------------------------------------------------------------\n");
  WriteToLog(LOG_INFO, writeBuffer);
}


static void StatisticsSummaryEthType8(StatBlockDescriptorType1_t *src,
                                      StatBlockDescriptorType1_t *dest)
{
  unsigned int i;
  uint64_t *pSrc, *pDest;

  dest->index |= (1<<src->index);
  dest->type = src->type;
  dest->headerVersion = src->headerVersion;
  dest->structureId = src->structureId;  
  ((StatisticsEtherType8_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->timestampType =
    ((StatisticsEtherType8_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->timestampType;
  ((StatisticsEtherType8_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt =
    ((StatisticsEtherType8_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt;
      
  pSrc=(uint64_t*)&((StatisticsEtherType8_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->Set1;
  pDest=(uint64_t*)&((StatisticsEtherType8_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->Set1;
  for(i=0; i<(sizeof(StatSetType1_t)/sizeof(uint64_t)); i++) {
    *pDest+=*pSrc;
    pSrc++; pDest++;
  }  
  pSrc=(uint64_t*)&((StatisticsEtherType8_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->Set2;
  pDest=(uint64_t*)&((StatisticsEtherType8_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->Set2;
  for(i=0; i<(sizeof(StatSetType1_t)/sizeof(uint64_t)); i++) {
    *pDest+=*pSrc;
    pSrc++; pDest++;
  }    
}

static void StatisticsDumpEthType8(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex)
{
  StatisticsEtherType8_t *type8 = NULL;

  type8 = (StatisticsEtherType8_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
  
  writePos = 0;

  if(summary) {
    STATS_LOG("Hardware counters for channelmask 0x%X: - %s -\n",
              descriptor->index, descriptor->type==STAT_TYPE_PORT?"Forwarded by filter":descriptor->type==STAT_TYPE_PORT_TX?"Transmitted":"Discarded by filter");
  } else {
    STATS_LOG("Hardware counters for channel %d: - %s -\n",
              descriptor->index, descriptor->type==STAT_TYPE_PORT?"Forwarded by filter":descriptor->type==STAT_TYPE_PORT_TX?"Transmitted":"Discarded by filter");
  }
  /* Set 1. */
  STATS_LOG("Frames size 64           : %s ",  NumAsString(type8->Set1.dwlFrameSize64, hex));
  STATS_LOG("Frames size 65-127       : %s\n", NumAsString(type8->Set1.dwlFrameSize65To127, hex));    
  STATS_LOG("Frames size 128-255      : %s ",  NumAsString(type8->Set1.dwlFrameSize128To255, hex));
  STATS_LOG("Frames size 256-511      : %s\n", NumAsString(type8->Set1.dwlFrameSize256To511, hex));    
  STATS_LOG("Frames size 512-1023     : %s ",  NumAsString(type8->Set1.dwlFrameSize512To1023, hex));
  STATS_LOG("Frames size 1024-1518    : %s\n", NumAsString(type8->Set1.dwlFrameSize1024To1518, hex));
  STATS_LOG("Frames size 1519-2047    : %s ",  NumAsString(type8->Set1.dwlFrameSize1519To2047, hex));
  STATS_LOG("Frames size 2048-4095    : %s\n", NumAsString(type8->Set1.dwlFrameSize2048To4095, hex));    
  STATS_LOG("Frames size 4096-8191    : %s ",  NumAsString(type8->Set1.dwlFrameSize4096To8191, hex));
  STATS_LOG("Frames size 8192-Max     : %s\n", NumAsString(type8->Set1.dwlFrameSize8192ToMax, hex));
  STATS_LOG("Undersized frames        : %s ",  NumAsString(type8->Set1.dwlUndersizeFrames, hex));
  STATS_LOG("Fragments                : %s\n", NumAsString(type8->Set1.dwlFragments, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Undersized wo encap      : %s ",  NumAsString(type8->Set1.dwlUndersizeWithoutEncapsulation, hex));
  }
  STATS_LOG("Oversize frames          : %s\n", NumAsString(type8->Set1.dwlOversizeFrames, hex));
  STATS_LOG("Jabber frames            : %s ",  NumAsString(type8->Set1.dwlJabberFrames, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Baby giant frames        : %s\n", NumAsString(type8->Set1.dwlBabyGiantFrames, hex));
  } else {
    STATS_LOG("\n");
  }
  STATS_LOG("Oversize Hard sliced     : %s ",  NumAsString(type8->Set1.dwlOversizeHardSlicedFrames, hex));    
  STATS_LOG("Jabber Hard sliced       : %s\n", NumAsString(type8->Set1.dwlJabberHardSlicedFrames, hex));
  STATS_LOG("Good bytes               : %s ",  NumAsString(type8->Set1.dwlGoodFrameBytes, hex));
  STATS_LOG("Bad bytes                : %s\n", NumAsString(type8->Set1.dwlBadFrameBytes, hex));        
  STATS_LOG("Good pause/mac ctrl bytes: %s\n", NumAsString(type8->Set1.dwlGoodPauseMacFramesBytes, hex));
  /* Set 2. */
  STATS_LOG("Unicast frames           : %s ",  NumAsString(type8->Set2.dwlUnicastFrames, hex));
  STATS_LOG("Multicast frames         : %s\n", NumAsString(type8->Set2.dwlMulticastFrames, hex));        
  STATS_LOG("Broadcast frames         : %s ",  NumAsString(type8->Set2.dwlBroadcastFrames, hex));
  STATS_LOG("Unicast pause frames     : %s\n", NumAsString(type8->Set2.dwlUnicastPauseFrames, hex));         
  STATS_LOG("Multicast pause frames   : %s ",  NumAsString(type8->Set2.dwlMulticastPauseFrames, hex));         
  STATS_LOG("Unicast MAC frames       : %s\n", NumAsString(type8->Set2.dwlMacUnicastFrames, hex));
  STATS_LOG("Multicast MAC frames     : %s ",  NumAsString(type8->Set2.dwlMacMulticastFrames, hex));
  STATS_LOG("Broadcast MAC frames     : %s\n", NumAsString(type8->Set2.dwlMacBroadcastFrames, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Frames with no encap     : %s\n", NumAsString(type8->Set2.dwlNotIslVlanMplsFrames, hex));  
    STATS_LOG("Frames with ISL          : %s ",  NumAsString(type8->Set2.dwlIslFrames, hex));
    STATS_LOG("Frames with VLAN         : %s\n", NumAsString(type8->Set2.dwlVlanFrames, hex));
    STATS_LOG("Frames with ISL,VLAN     : %s ",  NumAsString(type8->Set2.dwlIslVlanFrames, hex));
    STATS_LOG("Frames with MPLS         : %s\n", NumAsString(type8->Set2.dwlMplsFrames, hex));
    STATS_LOG("Frames with ISL,MPLS     : %s ",  NumAsString(type8->Set2.dwlIslMplsFrames, hex));
    STATS_LOG("Frames with VLAN,MPLS    : %s\n", NumAsString(type8->Set2.dwlVlanMplsFrames, hex));
    STATS_LOG("Frames with ISL,VLAN,MPLS: %s ",  NumAsString(type8->Set2.dwlIslVlanMplsFrames, hex));
  }
  STATS_LOG("CRC errors               : %s\n", NumAsString(type8->Set2.dwlCrcErrors, hex));
  if(descriptor->type!=STAT_TYPE_PORT_TX) {
    STATS_LOG("Alignment errors         : %s ",  NumAsString( type8->Set2.dwlAlignmentErrors, hex));
  }
  STATS_LOG("Code violation errors    : %s\n", NumAsString(type8->Set2.dwlCodeViolationErrors, hex));
  STATS_LOG("Small frames             : %s ",  NumAsString(type8->Set2.dwlSmallFrames, hex));
  STATS_LOG("Large frames             : %s\n", NumAsString(type8->Set2.dwlLargePackets, hex));
  STATS_LOG("IP checksum errors       : %s ",  NumAsString(type8->Set2.dwlIpChecksumErrors, hex));
  STATS_LOG("TCP checksum errors      : %s\n", NumAsString(type8->Set2.dwlTcpChecksumErrors, hex));
  STATS_LOG("UDP checksum errors      : %s\n", NumAsString(type8->Set2.dwlUdpChecksumErrors, hex));
  STATS_LOG("Time stamp               : 0x%016llX\n", type8->dwlTimeValidAt);
  STATS_LOG("---------------------------------------------------------------------------------------\n");
  WriteToLog(LOG_INFO, writeBuffer);
}

static void StatisticsSummarySimpleType1(StatBlockDescriptorType1_t *src,
                                         StatBlockDescriptorType1_t *dest)
{
  unsigned int i;
  uint64_t *pSrc, *pDest;
  
  dest->index |= (1<<src->index);
  dest->type = src->type;
  dest->headerVersion = src->headerVersion;
  dest->structureId = src->structureId;
  
  ((StatisticsSimpleType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->timestampType =
    ((StatisticsSimpleType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->timestampType;
  ((StatisticsSimpleType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt =
    ((StatisticsSimpleType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt;
  
  pSrc=(uint64_t*)&((StatisticsSimpleType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->simple;
  pDest=(uint64_t*)&((StatisticsSimpleType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->simple;
  for(i=0; i<(sizeof(StatSetSimpleType1_t)/sizeof(uint64_t)); i++) {
    *pDest+=*pSrc;
    pSrc++; pDest++;
  }  
}

static void StatisticsDumpSimpleType1(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex)
{
  StatisticsSimpleType1_t *type1 = NULL;
  
  type1 = (StatisticsSimpleType1_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
  
  writePos = 0;
  
  switch(descriptor->type) {
  case STAT_TYPE_FILTER_FORWARD:
    if(summary) {
      STATS_LOG("Hardware counters for filtermask 0x%X: - Forwarded by filter -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for filter %d: - Forwarded by filter -\n", descriptor->index);
    }
    break;
  case STAT_TYPE_FILTER_DISCARD:
    if(summary) {
      STATS_LOG("Hardware counters for filtermask 0x%X: - Discarded by filter -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for filters : - Discarded by filter -\n");
    }
    break;
  case STAT_TYPE_FEED_FORWARD:
    if(summary) {
      STATS_LOG("Hardware counters for feedmask 0x%X: - Forwarded by filter -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for feed %d: - Forwarded by filter -\n", descriptor->index);
    }
    break;
  case STAT_TYPE_FEED_INVALID_DROP:
    if(summary) {
      STATS_LOG("Hardware counters for feedmask 0x%X: - Inactive feed drop -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for feed %d: - Inactive feed drop -\n", descriptor->index);
    }
    break;
  case STAT_TYPE_FEED_BANDWIDTH_DROP:
    if(summary) {
      STATS_LOG("Hardware counters for feedmask 0x%X: - On-board buffer overflow drop -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for feed %d: - On-board buffer overflow drop -\n", descriptor->index);
    }
    break;
  case STAT_TYPE_DEDUP:
    if(summary) {
      STATS_LOG("Hardware counters for channelmask 0x%X: - Deduplication drop -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for channel %d: - Deduplication drop -\n", descriptor->index);
    }
    break;
  
  case STAT_TYPE_DROP_IP_FRAGMENT:
    if(summary) {
      STATS_LOG("Hardware counters for channelmask 0x%X: - IP fragment drop -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for channel %d: - IP fragment drop -\n", descriptor->index);
    }
    break;
  }
  STATS_LOG("Good frames              : %s ",  NumAsString(type1->simple.dwlGoodFrames, hex)); 
  STATS_LOG("Bad frames               : %s\n", NumAsString(type1->simple.dwlBadFrames, hex)); 
  STATS_LOG("Good bytes               : %s ",  NumAsString(type1->simple.dwlGoodBytes, hex)); 
  STATS_LOG("Bad bytes                : %s\n", NumAsString(type1->simple.dwlBadBytes, hex)); 
  STATS_LOG("Time stamp               : 0x%016llX\n", type1->dwlTimeValidAt);
  STATS_LOG("---------------------------------------------------------------------------------------\n");
  
  WriteToLog(LOG_INFO, writeBuffer);
}

static void StatisticsSummaryMisc(StatBlockDescriptorType1_t *src,
                                  StatBlockDescriptorType1_t *dest)
{
  StatisticsMiscType1_t *pSrc, *pDest;
  
  dest->index |= (1<<src->index);
  dest->type = src->type;
  dest->headerVersion = src->headerVersion;
  dest->structureId = src->structureId;

  ((StatisticsMiscType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->timestampType =
    ((StatisticsMiscType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->timestampType;
  ((StatisticsMiscType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt =
    ((StatisticsMiscType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt;
  
  pSrc=((StatisticsMiscType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)));
  pDest=((StatisticsMiscType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)));
  pDest->dwlRxXGMIISmallFrameDrop+=pSrc->dwlRxXGMIISmallFrameDrop;
  pDest->dwlRxPacketMergeDrop+=pSrc->dwlRxPacketMergeDrop;
}

static void StatisticsDumpMisc(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex)
{
  StatisticsMiscType1_t *misc= NULL;

  misc = (StatisticsMiscType1_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
  
  writePos = 0;

  if(summary) {
    STATS_LOG("Hardware counters for channelmask 0x%X: - Misc -\n", descriptor->index);
  } else {
    STATS_LOG("Hardware counters for channel %d: - Misc -\n", descriptor->index);
  }
  STATS_LOG("XGMIISmallFrameDrop      : %s ",  NumAsString(misc->dwlRxXGMIISmallFrameDrop, hex)); 
  STATS_LOG("PacketMergeDrop          : %s\n", NumAsString(misc->dwlRxPacketMergeDrop, hex)); 
  STATS_LOG("Time stamp               : 0x%016llX\n", misc->dwlTimeValidAt);
  STATS_LOG("---------------------------------------------------------------------------------------\n");

  WriteToLog(LOG_INFO, writeBuffer);
}

static void StatisticsSummaryIpFragment(StatBlockDescriptorType1_t *src,
                                        StatBlockDescriptorType1_t *dest)
{
  StatisticsIpFragType1_t *pSrc, *pDest;
  
  dest->index |= (1<<src->index);
  dest->type = src->type;
  dest->headerVersion = src->headerVersion;
  dest->structureId = src->structureId;

  ((StatisticsIpFragType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->timestampType =
    ((StatisticsIpFragType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->timestampType;
  ((StatisticsIpFragType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt =
    ((StatisticsIpFragType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)))->dwlTimeValidAt;
  
  pSrc=((StatisticsIpFragType1_t*)((uint8_t*)src+sizeof(StatBlockDescriptorType1_t)));
  pDest=((StatisticsIpFragType1_t*)((uint8_t*)dest+sizeof(StatBlockDescriptorType1_t)));

  pDest->dwlIpFragFirstHit += pSrc->dwlIpFragFirstHit;
  pDest->dwlIpFragFirstNoHit+=pSrc->dwlIpFragFirstNoHit;
  pDest->dwlIpFragMidHit += pSrc->dwlIpFragMidHit;
  pDest->dwlIpFragMidNoHit += pSrc->dwlIpFragMidNoHit;
  pDest->dwlIpFragLastHit += pSrc->dwlIpFragLastHit;
  pDest->dwlIpFragLastNoHit += pSrc->dwlIpFragLastNoHit;
  pDest->dwlIpFragFirstLast += pSrc->dwlIpFragFirstLast;
  pDest->dwlIpFragNotConsidered += pSrc->dwlIpFragNotConsidered;
}

static void StatisticsDumpIpFragment(StatBlockDescriptorType1_t *descriptor, bool_t summary, bool_t hex)
{
  StatisticsIpFragType1_t *frag_stats= NULL;

  frag_stats = (StatisticsIpFragType1_t*)((uint8_t*)descriptor+sizeof(StatBlockDescriptorType1_t));
  
  writePos = 0;
  if (descriptor->type == STAT_TYPE_FEED_IP_FRAGMENT) {
    if(summary) {
      STATS_LOG("Hardware counters for feedmask 0x%X: - Ip Fragments -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for feed %d: - Ip Fragments -\n", descriptor->index);
    }
  } else {
    if(summary) {
      STATS_LOG("Hardware counters for channelmask 0x%X: - Ip Fragments -\n", descriptor->index);
    } else {
      STATS_LOG("Hardware counters for channel %d: - Ip Fragments -\n", descriptor->index);
    }
  }
  STATS_LOG("IpFragFirstHit           : %s ",    NumAsString(frag_stats->dwlIpFragFirstHit, hex)); 
  STATS_LOG("IpFragFirstNoHit         : %s \n",  NumAsString(frag_stats->dwlIpFragFirstNoHit, hex)); 
  STATS_LOG("IpFragMidHit             : %s ",    NumAsString(frag_stats->dwlIpFragMidHit, hex)); 
  STATS_LOG("IpFragMidNoHit           : %s \n",  NumAsString(frag_stats->dwlIpFragMidNoHit, hex)); 
  STATS_LOG("IpFragLastHit            : %s ",    NumAsString(frag_stats->dwlIpFragLastHit, hex)); 
  STATS_LOG("IpFragLastNoHit          : %s \n",  NumAsString(frag_stats->dwlIpFragLastNoHit, hex)); 
  STATS_LOG("IpFragFirstLastHit       : %s ",    NumAsString(frag_stats->dwlIpFragFirstLast, hex)); 
  STATS_LOG("IpFragNotConsidered      : %s \n",  NumAsString(frag_stats->dwlIpFragNotConsidered, hex)); 
  STATS_LOG("Time stamp               : 0x%016llX\n", frag_stats->dwlTimeValidAt);
  
  STATS_LOG("---------------------------------------------------------------------------------------\n");

  WriteToLog(LOG_INFO, writeBuffer);
}


uint32_t _StatisticsStart(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsStart(card, channelsBitmask);
}

uint32_t _StatisticsStop(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsStop(card, channelsBitmask);
}

uint32_t _StatisticsReset(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsReset(card, channelsBitmask);
}

uint32_t _StatisticsGetSummary(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetSummary(card, channelsBitmask, TRUE);
}

uint32_t _StatisticsGetTxSummary(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetSummary(card, channelsBitmask, FALSE);
}

uint32_t _StatisticsGetRMONCounters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetRMONCounters(card, channelsBitmask, TRUE, TRUE);
}

uint32_t _StatisticsGetTxRMONCounters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetRMONCounters(card, channelsBitmask, TRUE, FALSE);
}

uint32_t _StatisticsGetHwCounters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetHwCounters(card, channelsBitmask, TRUE, TRUE);
}

uint32_t _StatisticsGetTxHwCounters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetHwCounters(card, channelsBitmask, TRUE, FALSE);
}

uint32_t _StatisticsGetBlockCounters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;
  uint64_t filtersBitmask;
  uint32_t feedsBitmask;
  uint8_t  summary;

  card            = get8incr(&argptr);  
  channelsBitmask = get32incr(&argptr);
  filtersBitmask  = get64incr(&argptr);
  feedsBitmask    = get32incr(&argptr);
  summary         = get8incr(&argptr);
  return StatisticsGetBlockCounters(card,
                                    channelsBitmask,
                                    filtersBitmask,
                                    feedsBitmask,
                                    summary, TRUE);
}


uint32_t _StatisticsGetType2Counters(void *arglist)
{
  uint8_t  *argptr = arglist;
  uint8_t card;
  uint32_t channelsBitmask;

  card            = get8incr(&argptr);
  channelsBitmask = get32incr(&argptr);   
  return StatisticsGetType2Counters(card, channelsBitmask, TRUE);
}
